/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//Useful tasks:
//
//      $ gulp Run                          - build and run TestCafe
//      ---------------------------------------------------------------------------------------
//      Params:
//          --release       - builds release version
//          --check-style   - runs jshint on TestCafe sources before run
//
//
//      $ gulp Distributive                 - creates TestCafe distribution in directory _dist_
//      ---------------------------------------------------------------------------------------
//      Params:
//          --public-fix    - create public fix distribution
//
//
//      $ gulp Nodeunit                     - run nodeunit tests
//      $ gulp QUnit                        - run qunit test page
//      $ gulp FunctionalTests:TestCafe     - run test runner and recorder functional tests
//      $ gulp FunctionalTests:ControlPanel - run ControlPanel functional tests
//      $ gulp HammerheadPlayground         - run Hammerhead playground page
//      $ gulp RecorderPlayground           - run Recorder playground page
//      $ gulp HammerhedHealthMonitor       - run Hammerhead Health Monitor
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


var fs = require('fs'),
    path = require('path'),
    gulp = require('gulp'),
    env = require('gulp-util').env,
    concat = require('gulp-concat'),
    cssmin = require('gulp-minify-css'),
    less = require('gulp-less'),
    uglify = require('gulp-uglify'),
    jshint = require('gulp-jshint'),
    nodeunit = require('gulp-nodeunit'),
    filter = require('gulp-filter'),
    rename = require('gulp-rename'),
    editJson = require('gulp-json-editor'),
    gzip = require('gulp-gzip'),
    tar = require('gulp-tar'),
    stylish = require('jshint-stylish'),
    del = require('del'),
    mergeTasks = require('merge-stream'),
    HgPrecommitHook = require('./hg_precommit');


//Promises
//---------------------------------------------------------------------------
function makePromise(fn) {
    return {then: fn};
}

var HangPromise = makePromise(function () {
    //Never resolves =)
});

//============================================================================================================

(function SETUP_GULP() {
    function isFarmTask() {
        for (var i = 0; i < env._.length; i++) {
            if (/^Farm-/g.test(env._[i]))
                return true;
        }

        return false;
    }

    //Determine if we are running on farm and build everything in release configuration
    if (isFarmTask() || env._.indexOf('Distributive') > -1)
        env.release = true;

    //Setup precommit hook otherwise
    else
        HgPrecommitHook.setup();


    //Force Gulp to exit process when all tasks are done
    //---------------------------------------------------------------------------
    gulp.on('stop', function () {
        setTimeout(function () {
            process.exit();
        });
    });

    gulp.on('err', function () {
        setTimeout(function () {
            process.exit(1);
        });
    });
})();

//============================================================================================================

(function HAMMERHEAD() {
    var OUTPUT_DIR = '_compiled_/hammerhead_client';


    //Internal tasks
    //---------------------------------------------------------------------------
    gulp.task('Hammerhead-Clean', function () {
        return makePromise(function (done) {
            del([OUTPUT_DIR], done);
        });
    });

    gulp.task('Hammerhead-JsTemplates', ['Hammerhead-Clean'], function () {
        return gulp
            .src('hammerhead/client/templates/*.jstmpl', {silent: false})
            .pipe(gulp.dest(OUTPUT_DIR));
    });

    gulp.task('Hammerhead-Scripts', ['Hammerhead-Clean'], function () {
        var task = gulp
            .src([
                'vendor/mods.js',
                'hammerhead/client/pre.jspart',
                'vendor/jquery-1.7.2.js',
                'vendor/async.js',
                'vendor/bowser.js',
                'hammerhead/client/**/*.js',
                'hammerhead/shared/**/*.js',
                'hammerhead/client/post.jspart'
            ])
            .pipe(concat('hammerhead.js'));

        if (env.release) {
            task.pipe(uglify({
                preserveComments: 'some' //NOTE: don't delete comments that starts with '!'
            }));
        }

        return task.pipe(gulp.dest(OUTPUT_DIR));
    });


    //Public tasks
    //---------------------------------------------------------------------------
    gulp.task('Hammerhead-Build', [
        'Hammerhead-JsTemplates',
        'Hammerhead-Scripts'
    ]);
})();

//============================================================================================================

(function CONTROL_PANEL() {
    var OUTPUT_DIR = '_compiled_/control_panel',
        IMAGES_OUTPUT_DIR = path.join(OUTPUT_DIR, '/images'),
        CODE_EDITOR_OUTPUT_DIR = path.join(OUTPUT_DIR, '/ace'),
        DX_GRID_VIEW_OUTPUT_DIR = path.join(OUTPUT_DIR, '/dx');


    //Internal tasks
    //---------------------------------------------------------------------------
    gulp.task('ControlPanel-Clean', function () {
        return makePromise(function (done) {
            del([OUTPUT_DIR], done);
        });
    });

    gulp.task('ControlPanel-Images', ['ControlPanel-Clean'], function () {
        return gulp
            .src('control_panel/client/images/**')
            .pipe(gulp.dest(IMAGES_OUTPUT_DIR));
    });

    gulp.task('ControlPanel-CodeEditor', ['ControlPanel-Clean'], function () {
        return gulp
            .src('vendor/ace/**')
            .pipe(gulp.dest(CODE_EDITOR_OUTPUT_DIR));
    });

    gulp.task('ControlPanel-Styles', ['ControlPanel-Clean'], function () {
        var task = gulp
            .src('control_panel/client/less/styles.less')
            .pipe(less());

        if (env.release)
            task.pipe(cssmin({showLog: true}));

        return task.pipe(gulp.dest(OUTPUT_DIR));
    });

    gulp.task('ControlPanel-Scripts', ['ControlPanel-Clean'], function () {
        var task = gulp
            .src([
                'vendor/mods.js',
                'control_panel/client/init.js',
                'vendor/json2.js',
                'vendor/jquery-2.1.3.js',
                'vendor/jquery.history.js',
                'vendor/jquery.cookie.js',
                'vendor/jquery-ui-1.11.2.effects.js',
                'vendor/qrcode.js',
                'control_panel/client/js/**/*.js'
            ])
            .pipe(concat('control_panel.js'));

        if (env.release) {
            task.pipe(uglify({
                preserveComments: 'some' //NOTE: don't delete comments that starts with '!'
            }));
        }

        return task.pipe(gulp.dest(OUTPUT_DIR));
    });

    gulp.task('ControlPanel-GridView', ['ControlPanel-Clean'], function () {
        return gulp
            .src('vendor/dx/**')
            .pipe(gulp.dest(DX_GRID_VIEW_OUTPUT_DIR));
    });

    //Public tasks
    //---------------------------------------------------------------------------
    gulp.task('ControlPanel-Build', [
        'ControlPanel-Images',
        'ControlPanel-CodeEditor',
        'ControlPanel-GridView',
        'ControlPanel-Styles',
        'ControlPanel-Scripts'
    ]);
})();

//============================================================================================================

(function TEST_CAFE_CLIENT() {
    var OUTPUT_DIR = '_compiled_/testcafe_client',
        STYLES_OUTPUT = path.join(OUTPUT_DIR, '/styles.css');


    //Internal tasks
    //---------------------------------------------------------------------------
    gulp.task('TestCafeClient-Clean', function () {
        return makePromise(function (done) {
            del([OUTPUT_DIR], done);
        });
    });

    gulp.task('TestCafeClient-Styles:TempCopy', ['TestCafeClient-Clean'], function () {
        return gulp
            .src('client/styles.less')
            .pipe(gulp.dest(OUTPUT_DIR));
    });

    gulp.task('TestCafeClient-Styles:Compile', ['TestCafeClient-Styles:TempCopy'], function () {
        var fs = require('fs'),
            path = require('path'),
            lessCompiler = require('less'),
            SharedConst = require('./hammerhead/shared/const');

        //Const
        var TEMP_SRC_FILE = path.join(process.cwd(), OUTPUT_DIR, 'styles.less');


        //NOTE: add unique UI postfix to each LESS class and id selector
        function addUIClassPostfix(lessRules) {
            lessRules.forEach(function (rule) {
                if (rule.selectors) {
                    rule.selectors.forEach(function (selector) {
                        if (selector.elements) {
                            selector.elements.forEach(function (element) {
                                var isPureMixin = typeof element.index === 'undefined',
                                    isTargetSelector = element.value && (element.value.indexOf('.') === 0 ||
                                        element.value.indexOf('#') === 0);

                                if (isTargetSelector && !isPureMixin)
                                    element.value += SharedConst.TEST_CAFE_UI_CLASSNAME_POSTFIX;
                            });
                        }
                    });
                }

                if (rule.rules)
                    addUIClassPostfix(rule.rules);
            });
        }

        return makePromise(function (done, fail) {
            fs.readFile(TEMP_SRC_FILE, function (readErr, src) {
                var parser = new lessCompiler.Parser();

                src = src.toString();

                parser.parse(src, function (parseErr, tree) {
                    if (parseErr) {
                        fail('Failed to parse client runtime LESS: ' + parseErr.message);
                        return;
                    }

                    addUIClassPostfix([tree]);

                    fs.writeFile(STYLES_OUTPUT, tree.toCSS(), function (writeErr) {
                        if (writeErr) {
                            fail('Failed to write dest file: ' + writeErr.message);
                            return;
                        }

                        fs.unlink(TEMP_SRC_FILE, function (unlinkErr) {
                            if (unlinkErr) {
                                fail('Failed to delete temp src file: ' + unlinkErr.message);
                                return;
                            }
                            done();
                        });
                    });
                });
            });
        });
    });

    gulp.task('TestCafeClient-Styles', ['TestCafeClient-Styles:Compile'], function () {
        var task = gulp.src(STYLES_OUTPUT);

        if (env.release)
            task.pipe(cssmin());

        return task.pipe(gulp.dest(OUTPUT_DIR));
    });

    gulp.task('TestCafeClient-Sprite', ['TestCafeClient-Clean'], function () {
        return gulp
            .src('client/sprite.png')
            .pipe(rename('uisprite.png'))
            .pipe(gulp.dest(OUTPUT_DIR));
    });

    function createScriptsTask(src, outFileName) {
        var task = gulp
            .src(src)
            .pipe(concat(outFileName));

        if (env.release) {
            task.pipe(uglify({
                preserveComments: 'some' //NOTE: don't delete comments that starts with '!'
            }));
        }

        return task.pipe(gulp.dest(OUTPUT_DIR));
    }

    gulp.task('TestCafeClient-Scripts', ['TestCafeClient-Clean'], function () {
        var core = createScriptsTask(
                [
                    'client/jsparts/core_pre.jspart',

                    'client/core/**/*.js',
                    '!client/Gruntfile.js',

                    'lib/shared/**/*.js',
                    'client/jsparts/core_post.jspart'
                ],

                'testcafe_core.js'
            ),

            recorder = createScriptsTask(
                [
                    'client/jsparts/recorder_pre.jspart',
                    'client/recorder/**/*.js',
                    'client/jsparts/recorder_post.jspart'
                ],

                'testcafe_recorder.js'
            ),

            uiCore = createScriptsTask(
                [
                    'client/jsparts/ui_pre.jspart',
                    'client/ui/core/**/*.js',
                    'client/jsparts/ui_post.jspart'
                ],

                'testcafe_ui_core.js'
            ),

            uiRecorder = createScriptsTask(
                [
                    'client/jsparts/recorder_ui_pre.jspart',
                    'client/ui/recorder/**/*.js',
                    'client/jsparts/recorder_ui_post.jspart'
                ],

                'testcafe_ui_recorder.js'
            );

        return mergeTasks(core, recorder, uiCore, uiRecorder);
    });

    gulp.task('TestCafeClient-JsTemplates', ['TestCafeClient-Clean'], function () {
        var task = gulp
            .src('client/task_templates/*.jstmpl');

        if (env.release) {
            task.pipe(uglify({
                preserveComments: 'all',
                compress: false,
                mangle: false,
                output: {
                    beautify: true
                }
            }));
        }

        return task.pipe(gulp.dest(OUTPUT_DIR));
    });


    //Public tasks
    //---------------------------------------------------------------------------
    gulp.task('TestCafeClient-Build', [
        'TestCafeClient-Styles',
        'TestCafeClient-Sprite',
        'TestCafeClient-Scripts',
        'TestCafeClient-JsTemplates'
    ]);
})();

//============================================================================================================

(function CHECK_CODE_STYLE() {
    var TEST_CAFE_TARGETS = [
            'control_panel/**/*.js',
            'client/**/*.js',
            'hammerhead/**/*.js',
            'lib/**/*.js',

            '!client/**/*.jspart',
            '!hammerhead/**/*.jspart',
            '!hammerhead/client/json.js',
            '!hammerhead/client/sandboxed_jquery.js',
            '!hammerhead/shared/js_parsing_tools.js'
        ],
        HAMERHEAD_HEALTH_MONITOR_TARGETS = [
            'test/hammerhead_health_monitor/**/*.js',

            '!test/hammerhead_health_monitor/vendor/*.js',
            '!test/hammerhead_health_monitor/node_modules/**/*.js'
        ],
        ALL_TARGETS = TEST_CAFE_TARGETS.concat(HAMERHEAD_HEALTH_MONITOR_TARGETS),
        OPTIONS = {
            sub: true,          // http://www.jshint.com/docs/options/#sub
            evil: true,         // http://www.jshint.com/docs/options/#evil
            eqeqeq: true,       // http://www.jshint.com/docs/options/#eqeqeq
            bitwise: true,      // http://www.jshint.com/docs/options/#bitwise
            forin: true,        // http://www.jshint.com/docs/options/#forin
            immed: true,        // http://www.jshint.com/docs/options/#immed
            undef: true,        // http://www.jshint.com/docs/options/#undef
            unused: true,       // http://www.jshint.com/docs/options/#unused
            debug: true,        // http://www.jshint.com/docs/options/#debug
            devel: true,        // http://www.jshint.com/docs/options/#devel
            browser: true,
            node: false,

            globals: {
                //NOTE: Option node:true defines global console object. It isn't suitable for checking client code on console.log
                //We define global objects of NodeJS without console
                __filename: false, __dirname: false, GLOBAL: false,
                global: false, module: false, require: false,
                Buffer: true, exports: true, process: true,
                setTimeout: true, clearTimeout: true, setInterval: true,
                clearInterval: true, setImmediate: true, clearImmediate: true,

                'TestCafeClient': true,
                'HammerheadClient': true,
                'NodeList': true,
                'HTMLCollection': true,
                'Proxy': true,
                'ControlPanel': true
            }
        };


    function checkCodeStyle(targets) {
        OPTIONS.debug = !env.release;
        OPTIONS.devel = !env.release;

        return gulp
            .src(targets)
            .pipe(jshint(OPTIONS))
            .pipe(jshint.reporter(stylish))
            .pipe(jshint.reporter('fail'));
    }

    //Public tasks
    //---------------------------------------------------------------------------
    gulp.task('CheckCodeStyle', function () {
        return checkCodeStyle(ALL_TARGETS);
    });

    gulp.task('CheckCodeStyle-HammerheadHealthMonitor', function () {
        return checkCodeStyle(HAMERHEAD_HEALTH_MONITOR_TARGETS);
    });
})();

//============================================================================================================

(function BUILD_AND_RUN() {
    gulp.task('Build', [
        'Hammerhead-Build',
        'ControlPanel-Build',
        'TestCafeClient-Build'
    ]);

    var runTaskDeps = env['check-style'] ? ['Build', 'CheckCodeStyle'] : ['Build'];

    //Public tasks
    //---------------------------------------------------------------------------
    gulp.task('Run', runTaskDeps, function () {
        var TestCafe = require('./lib/server/test_cafe');

        new TestCafe(null, true);

        if (!env.release) {
            gulp.watch('control_panel/**', ['ControlPanel-Build']);
            gulp.watch('hammerhead/client/**', ['Hammerhead-Build']);
            gulp.watch('client/**', ['TestCafeClient-Build']);
        }

        return HangPromise;
    });
})();

//============================================================================================================

(function TESTING() {
    //Public tasks
    //---------------------------------------------------------------------------
    gulp.task('Nodeunit', ['Build'], function () {
        return gulp
            .src('test/nodeunit/**/*_test.js')
            .pipe(nodeunit());
    });

    gulp.task('QUnit', ['Build'], function () {
        gulp.watch('hammerhead/client/**', ['Hammerhead-Build']);
        gulp.watch('client/**', ['TestCafeClient-Build']);

        require('./test/qunit/server.js').start();

        return HangPromise;
    });

    gulp.task('FunctionalTests:ControlPanel', ['Build'], function () {
        require('./test/functional/control_panel/server.js');

        return HangPromise;
    });

    gulp.task('FunctionalTests:TestCafe', ['Build'], function () {
        require('./test/functional/testcafe/run_environment.js').run(false);

        return HangPromise;
    });

    gulp.task('FunctionalTests:TestRunner', ['Build'], function () {
        require('./test/functional/testcafe/run_environment.js').runTestRunnerTests(function (message) {
            console.log(message);
            process.exit();
        });

        return HangPromise;
    });

    gulp.task('HammerheadPlayground', ['Hammerhead-Build'], function () {
        require('./test/hammerhead_playground/server.js').start();

        return HangPromise;
    });

    gulp.task('RecorderPlayground', ['Build'], function () {
        require('./test/recorder_playground/server.js').start();

        return HangPromise;
    });

    gulp.task('HammerhedHealthMonitor', ['Hammerhead-Build', 'CheckCodeStyle-HammerheadHealthMonitor'], function () {
        require('./test/hammerhead_health_monitor/index.js');

        return HangPromise;
    });
})();

//============================================================================================================

(function FARM_TESTING() {
    function execRunner(runnerPath) {
        require(runnerPath);

        return HangPromise;
    }

    //Public tasks
    //---------------------------------------------------------------------------
    gulp.task('Farm-Nodeunit', ['Build'], function () {
        return gulp
            .src('test/nodeunit/**/*_test.js')
            .pipe(nodeunit({
                reporter: 'junit',
                reporterOptions: {
                    output: './'
                }
            }));
    });


    //Qunit
    gulp.task('Farm-QUnit:IE', ['Build'], function () {
        return execRunner('./test/cc_run_client.js');
    });

    gulp.task('Farm-QUnit:Firefox', ['Build'], function () {
        return execRunner('./test/cc_run_client(firefox).js');
    });

    gulp.task('Farm-QUnit:Chrome', ['Build'], function () {
        return execRunner('./test/cc_run_client(chrome).js');
    });

    gulp.task('Farm-QUnit:Edge', ['Build'], function () {
        return execRunner('./test/cc_run_client(edge).js');
    });


    //FunctionalTests:ASPxPerience
    gulp.task('Farm-FunctionalTests:ASPxPerience:Chrome', ['Build'], function () {
        return execRunner('./bin/cc_run(chrome).js');
    });

    gulp.task('Farm-FunctionalTests:ASPxPerience:IE', ['Build'], function () {
        return execRunner('./bin/cc_run(ie).js');
    });

    gulp.task('Farm-FunctionalTests:ASPxPerience:Firefox', ['Build'], function () {
        return execRunner('./bin/cc_run(ff).js');
    });

    gulp.task('Farm-FunctionalTests:ASPxPerience:IE10', ['Build'], function () {
        return execRunner('./bin/cc_run(ie10).js');
    });

    gulp.task('Farm-FunctionalTests:ASPxPerience:Edge', ['Build'], function () {
        return execRunner('./bin/cc_run(edge).js');
    });

    //FunctionalTests:ASPDemos
    gulp.task('Farm-FunctionalTests:ASPDemos', ['Build'], function () {
        return execRunner('./bin/cc_run.js');
    });


    //FunctionalTests:ControlPanel
    gulp.task('Farm-FunctionalTests:ControlPanel:Chrome', ['Build'], function () {
        return execRunner('./test/cc_run(chrome).js');
    });

    gulp.task('Farm-FunctionalTests:ControlPanel:Firefox', ['Build'], function () {
        return execRunner('./test/cc_run(ff).js');
    });

    gulp.task('Farm-FunctionalTests:ControlPanel:IE', ['Build'], function () {
        return execRunner('./test/cc_run(ie10_11).js');
    });


    //FunctionalTests:TestCafe
    gulp.task('Farm-FunctionalTests:TestCafe:Chrome', ['Build'], function () {
        return execRunner('./test/cc_run_functional(chrome).js');
    });

    gulp.task('Farm-FunctionalTests:TestCafe:Firefox', ['Build'], function () {
        return execRunner('./test/cc_run_functional(ff).js');
    });

    gulp.task('Farm-FunctionalTests:TestCafe:IE11', ['Build'], function () {
        return execRunner('./test/cc_run_functional(ie11).js');
    });

    //FunctionalTests:TestRunner
    gulp.task('Farm-FunctionalTests:TestRunner', ['Build'], function () {
        return execRunner('./test/functional/testcafe/cc_run_environment.js');
    });

    //HealthMonitor
    gulp.task('Farm-HammerheadHealthMonitor', ['Hammerhead-Build'], function () {
        return execRunner('./test/hammerhead_health_monitor/farm_health_monitor.js');
    });


    //Install tests
    //NOTE: use --public-fix flag for testing the public fix build.
    gulp.task('Farm-InstallTests', ['Build'], function () {
        return execRunner('./test/install/run.js');
    });
})();

//============================================================================================================

(function DISTRIBUTIVE() {
    var OUTPUT_DIR = '_dist_',
        WIN_OUTPUT = path.join(OUTPUT_DIR, '/win'),
        LINUX32_OUTPUT = path.join(OUTPUT_DIR, '/linux32'),
        LINUX64_OUTPUT = path.join(OUTPUT_DIR, '/linux64'),
        MAC_TMP_OUTPUT = path.join(OUTPUT_DIR, '/mac/dist'),
        MAC_APP_OUTPUT = path.join(MAC_TMP_OUTPUT, '/TestCafe.app/Contents/Resources/testcafe'),
        MAC_SOURCE_TGZ = path.join(OUTPUT_DIR, '/mac/mac_dist.tgz'),
        MAC_DMG_TMPL_SRC = path.join(OUTPUT_DIR, './mac/dmg_tmpl'),
        MAC_DMG_SOURCE_TGZ = path.join(OUTPUT_DIR, '/mac_dmg_source.tgz'),
        NPM_OUTPUT = path.join(OUTPUT_DIR, '/npm'),
        BUILD_FILES = [
            '_compiled_/**',
            'hammerhead/lib/**',
            'hammerhead/shared/**',
            'hammerhead/index.js',
            'control_panel/views/**',
            'control_panel/lib/**',
            'control_panel/index.js',
            'lib/**',
            'examples/**',
            'index.js',
            'package.json'
        ];

    var MSBUILD_PATH = 'C:\\Windows\\Microsoft.NET\\Framework\\v4.0.30319\\msbuild';

    //NOTE: this vars will be setup by 'Distributive-SetupEnv' task
    var PACKAGE_JSON_VERSION = null,
        PUBLIC_VERSION = null,
        BUILD_FILES_WITH_MODULES = null,
        NIX_BUILD_FILES_WITH_MODULES = null;

    //Internal tasks
    //---------------------------------------------------------------------------
    gulp.task('Distributive-SetupEnv', function () {
        return makePromise(function (done, fail) {
            del([OUTPUT_DIR], function () {
                fs.readFile(path.join(__dirname, 'package.json'), function (err, cfg) {
                    if (err) {
                        fail(err);
                        return;
                    }

                    cfg = JSON.parse(cfg.toString());

                    //NOTE: time stamp is days since 01/01/1970 GMT
                    var timestamp = Math.round(new Date().getTime() / (24 * 3600 * 1000)) + '';

                    //NOTE: package version string should concatenate timestamp with last version digit
                    //to support NPM version convention (e.g. 13.2.338705)
                    PACKAGE_JSON_VERSION = env['public-fix'] ? (cfg.version + timestamp) : cfg.version;

                    //NOTE: public version string should separate timestamp with dot to follow our support center
                    //public fix versions convention (e.g. 13.2.3.38705)
                    PUBLIC_VERSION = env['public-fix'] ? (cfg.version + '.' + timestamp) : cfg.version;

                    BUILD_FILES_WITH_MODULES = Object.keys(cfg.dependencies)
                        .map(function (dep) {
                            return 'node_modules/' + dep + '/**';
                        })
                        .concat(BUILD_FILES);

                    NIX_BUILD_FILES_WITH_MODULES = BUILD_FILES_WITH_MODULES.concat(['!lib/test_runner/shot_window.exe']);

                    //NOTE: the following file should be included to the nix build only
                    BUILD_FILES_WITH_MODULES = BUILD_FILES_WITH_MODULES.concat('!lib/test_runner/shot_window');

                    done();
                });
            });
        });
    });

    function createCopyBuildTask(src) {
        var removeCommentsFilter = filter([
            'lib/**/*.js',
            'hammerhead/**/*.js',
            'control_panel/**/*.js'
        ]);

        var packageJsonFilter = filter('package.json');

        return gulp
            .src(src, {base: './'})

            //NOTE: remove comments from filtered files
            .pipe(removeCommentsFilter)
            .pipe(uglify({
                preserveComments: false,
                compress: false,
                mangle: false,
                output: {
                    beautify: true
                }
            }))
            .pipe(removeCommentsFilter.restore({end: true}))

            //NOTE: edit package.json
            .pipe(packageJsonFilter)
            .pipe(editJson(function (pkg) {
                delete pkg.devDependencies;
                pkg.version = PACKAGE_JSON_VERSION;

                return pkg;
            }))
            .pipe(packageJsonFilter.restore({end: true}));

    }

    gulp.task('Distributive-CopyBuild', ['Distributive-SetupEnv', 'Build', 'Distributive-CopyMacApp'], function () {
        var nix = createCopyBuildTask(NIX_BUILD_FILES_WITH_MODULES)
            .pipe(gulp.dest(LINUX32_OUTPUT))
            .pipe(gulp.dest(LINUX64_OUTPUT))
            .pipe(gulp.dest(MAC_APP_OUTPUT));

        var win = createCopyBuildTask(BUILD_FILES_WITH_MODULES).pipe(gulp.dest(WIN_OUTPUT));
        var npm = createCopyBuildTask(BUILD_FILES).pipe(gulp.dest(NPM_OUTPUT));

        return mergeTasks(nix, win, npm);
    });

    gulp.task('Distributive-CopyMacApp', ['Distributive-SetupEnv'], function () {
        return gulp
            .src([
                'runners/mac/app/**'
            ])
            .pipe(gulp.dest(MAC_TMP_OUTPUT));
    });

    gulp.task('Distributive-CopyBinFiles', ['Distributive-SetupEnv', 'Distributive-CopyMacApp'], function () {
        var win = gulp.src('bin/win/**').pipe(gulp.dest(WIN_OUTPUT)),
            linux32 = gulp.src('bin/linux32/**').pipe(gulp.dest(LINUX32_OUTPUT)),
            linux64 = gulp.src('bin/linux64/**').pipe(gulp.dest(LINUX64_OUTPUT)),
            macInstaller = gulp
                .src([
                    'runners/mac/dmg/**'
                ])
                .pipe(gulp.dest(MAC_TMP_OUTPUT)),
            macApp = gulp.src([
                'bin/mac/**',
                '!bin/mac/cli.sh'              //TODO: disable in next minor version
            ])
                .pipe(gulp.dest(MAC_APP_OUTPUT));

        mergeTasks(win, linux32, linux64, macInstaller, macApp);
    });

    gulp.task('Distributive-CopyDocs', ['Distributive-SetupEnv', 'Distributive-CopyMacApp'], function () {
        var eula = gulp
            .src([
                'docs/NODEJS_EULA.txt',
                'docs/TESTCAFE_EULA.txt'
            ])
            .pipe(gulp.dest(WIN_OUTPUT))
            .pipe(gulp.dest(LINUX32_OUTPUT))
            .pipe(gulp.dest(LINUX64_OUTPUT))
            .pipe(gulp.dest(MAC_APP_OUTPUT));

        var eulaInstaller = gulp
            .src([
                'docs/TESTCAFE_EULA.txt',
                'docs/NODEJS_EULA.txt'
            ])
            .pipe(concat('TESTCAFE_EULA_INSTALLER.txt', {newLine: '\n\n'}))
            .pipe(gulp.dest(OUTPUT_DIR));

        var readmePdf = gulp
            .src('docs/Readme.pdf')
            .pipe(gulp.dest(LINUX32_OUTPUT))
            .pipe(gulp.dest(LINUX64_OUTPUT));

        var npmReadme = gulp
            .src('docs/README_NPM.txt')
            .pipe(rename('README.txt'))
            .pipe(gulp.dest(NPM_OUTPUT));

        var npmEula = gulp
            .src('docs/TESTCAFE_EULA.txt')
            .pipe(gulp.dest(NPM_OUTPUT));

        return mergeTasks(eula, eulaInstaller, readmePdf, npmReadme, npmEula);
    });

    function buildDmgTmpl(sourceTgz, outputDir, buildName) {
        return makePromise(function (done, fail) {
            require('./runners/mac/dmg/build_tmpl.js').buildMacBySSH(sourceTgz, outputDir, buildName, done, fail);
        });
    }

    function buildMacDist(sourceTgz, outputDir, buildName) {
        return makePromise(function (done, fail) {
            require('./runners/mac/dmg/build_installer.js').buildMacBySSH(sourceTgz, outputDir, buildName, done, fail);
        });
    }

    function msbuild(projectFile, projectPath, outputDir) {
        return makePromise(function (done, fail) {
            var Process = require('child_process'),
                path = require('path'),
                fs = require('fs'),

                projectFilePath = path.join(__dirname, projectPath, projectFile),
                assemblyInfoPath = path.join(projectPath, 'Properties/AssemblyInfo.cs'),

                buildOptions = [
                    '/property:Configuration=Release;Platform="x86";DebugSymbols=false;DebugType=None,OutputPath=',
                    path.join(__dirname, outputDir)
                ].join(''),


                cmd = [
                    MSBUILD_PATH,
                    projectFilePath,
                    buildOptions
                ].join(' ');


            fs.readFile(assemblyInfoPath, 'utf8', function (readErr, assemblyInfoContent) {
                assemblyInfoContent = assemblyInfoContent.toString();

                var savedAssemblyInfoContent = assemblyInfoContent,
                    publicVersion = PUBLIC_VERSION;

                //NOTE: if product version has 3 numbers add ".0", e.g. 13.1.3 -> 13.1.3.0
                if (publicVersion.split('.').length === 3)
                    publicVersion += '.0';

                assemblyInfoContent = assemblyInfoContent
                    .replace(/\[assembly: AssemblyVersion\("[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+"\)\]/,
                    '[assembly: AssemblyVersion("' + publicVersion + '")]')
                    .replace(/\[assembly: AssemblyFileVersion\("[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+"\)\]/,
                    '[assembly: AssemblyFileVersion("' + publicVersion + '")]');

                fs.writeFile(assemblyInfoPath, assemblyInfoContent, function (writeErr) {
                    if (writeErr) {
                        fail('Error writing AssemblyInfo.cs: ' + writeErr);
                        return;
                    }

                    Process.exec(cmd, function (execErr, stdout) {
                        if (execErr) {
                            fail(stdout.toString());
                            return;
                        }

                        fs.writeFile(assemblyInfoPath, savedAssemblyInfoContent, function (restoreErr) {
                            if (restoreErr)
                                fail('Error restoring AssemblyInfo.cs: ' + restoreErr);
                            else
                                done();

                        });
                    });
                });
            });
        });
    }

    gulp.task('Distributive-BuildWinRunner', ['Distributive-SetupEnv'], function () {
        return msbuild('TestCafe.csproj', 'runners/win', WIN_OUTPUT);
    });

    gulp.task('Distributive-CopyDmgTmplSrc', function () {
        var resources = gulp.src([
            'runners/mac/dmg/tmpl/**',
            '!runners/mac/dmg/tmpl/*.zip'
        ]).pipe(gulp.dest(MAC_DMG_TMPL_SRC));
        var buildScript = gulp.src('runners/mac/dmg/*.sh')
            .pipe(gulp.dest(MAC_DMG_TMPL_SRC));

        return mergeTasks(resources, buildScript);
    });

    gulp.task('Distributive-CreateDmgTmplArchive', ['Distributive-CopyDmgTmplSrc'], function () {
        return createArchive(MAC_DMG_TMPL_SRC, './mac_dmg_source.tgz');
    });

    gulp.task('Distributive-BuildDmgTmpl', [
        'Distributive-CopyDmgTmplSrc',
        'Distributive-CreateDmgTmplArchive'
    ], function () {
        return buildDmgTmpl(MAC_DMG_SOURCE_TGZ, OUTPUT_DIR, 'testcafe_tmpl.zip');
    });

    gulp.task('Distributive-BuildMacRunner', [
        'Distributive-CopyBuild',
        'Distributive-CopyBinFiles',
        'Distributive-CopyDocs',
        'Distributive-CreateAppArchiveMac'
    ], function () {
        return buildMacDist(MAC_SOURCE_TGZ, OUTPUT_DIR, 'DevExpressTestCafe-Mac-' + PUBLIC_VERSION + '.dmg');
    });

    function createArchive(target, outFileName, isFile) {
        var files = isFile ? target : target + '/**/*';

        return gulp
            .src(files)
            .pipe(tar(outFileName))
            .pipe(gzip({
                append: false   //NOTE: don't append .gz to the file name
            }))
            .pipe(gulp.dest(OUTPUT_DIR));
    }

    gulp.task('Distributive-CreateLinuxArchives', [
            'Distributive-CopyBuild',
            'Distributive-CopyBinFiles',
            'Distributive-CopyDocs'
        ],
        function () {
            var linux32 = createArchive(LINUX32_OUTPUT, 'DevExpressTestCafe-Linux32-' + PUBLIC_VERSION + '.tgz');

            var linux64 = createArchive(LINUX64_OUTPUT, 'DevExpressTestCafe-Linux64-' + PUBLIC_VERSION + '.tgz');

            return mergeTasks(linux32, linux64);
        });

    gulp.task('Distributive-CreateAppArchiveMac', [
        'Distributive-CopyBuild',
        'Distributive-CopyBinFiles',
        'Distributive-CopyDocs'
    ], function () {
        return createArchive(MAC_TMP_OUTPUT, './mac/mac_dist.tgz');
    });

    //Public tasks
    //---------------------------------------------------------------------------
    gulp.task('Distributive', [
        'CheckCodeStyle',
        'Distributive-CopyMacApp',
        'Distributive-CopyBuild',
        'Distributive-CopyBinFiles',
        'Distributive-CopyDocs',
        'Distributive-BuildWinRunner',
        'Distributive-CreateAppArchiveMac',
        'Distributive-BuildMacRunner',
        'Distributive-CreateLinuxArchives'
    ]);
})();
