TestCafeClient.define('Automation', function (require, exports) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        Settings = require('Settings'),
        $ = Hammerhead.$,
        async = Hammerhead.async,
        NativeMethods = Hammerhead.NativeMethods,
        EventSandbox = Hammerhead.EventSandbox,
        EventSimulator = Hammerhead.EventSimulator,
        Util = Hammerhead.Util,
        JavascriptExecutor = require('Base.JavascriptExecutor'),

        ClickPlayback = require('Automation.Click.Playback'),
        RClickPlayback = require('Automation.RClick.Playback'),
        DblClickPlayback = require('Automation.DblClick.Playback'),
        DragPlayback = require('Automation.Drag.Playback'),
        SelectPlayback = require('Automation.Select.Playback'),
        PressPlayback = require('Automation.Press.Playback'),
        TypePlayback = require('Automation.Type.Playback'),
        HoverPlayback = require('Automation.Hover.Playback');

    //Const
    var REAL_ACTION_EVENTS_REGEXP = /blur|focus|(dbl)?click|contextmenu|key|mouse|pointer/i;

    //Default action descriptors
    var defaultActionDescriptor = {
            type: '',
            serviceInfo: {
                prevPageState: null,
                isDeferred: false
            }
        },

        defaultElementActionDescriptor = $.extend(true, {}, defaultActionDescriptor, {
            element: null,
            selector: null,
            serviceInfo: {
                selectors: []
            }
        });

    exports.AUTOMATION_RUNNERS = 'tc-ar-73630b99';

    exports.SUPPORTED_SHORTCUTS = [
        'ctrl+a',
        'backspace',
        'delete',
        'left',
        'right',
        'up',
        'down',
        'shift+left',
        'shift+right',
        'shift+up',
        'shift+down',
        'shift+home',
        'shift+end',
        'home',
        'end',
        'enter',
        'tab',
        'shift+tab'
    ];

    exports.ADD_ACTION_SHORTCUTS = {
        wait: 'Ctrl+Q',
        hover: 'Ctrl+Space',
        screenshot: 'Ctrl+M'
    };

    exports.defaultMouseActionDescriptor = $.extend(true, {}, defaultElementActionDescriptor, {
        apiArguments: {
            options: {
                ctrl: false,
                alt: false,
                shift: false,
                meta: false,
                offsetX: '',
                offsetY: ''
            }
        },
        serviceInfo: {
            useOffsets: false
        }
    });

    exports.defaultDragActionDescriptor = $.extend(true, {}, exports.defaultMouseActionDescriptor, {
        type: 'drag',
        apiArguments: {
            dragOffsetX: 0,
            dragOffsetY: 0
        },
        startPosition: null,
        endPosition: null
    });

    exports.defaultHoverActionDescriptor = $.extend(true, {}, defaultElementActionDescriptor, {
        type: 'hover'
    });

    exports.defaultTypeActionDescriptor = $.extend(true, {}, exports.defaultMouseActionDescriptor, {
        type: 'type',
        apiArguments: {
            text: '',
            options: {
                replace: false,
                caretPos: ''
            }
        }
    });

    exports.defaultPressActionDescriptor = $.extend(true, {}, defaultActionDescriptor, {
        type: 'press',
        apiArguments: {
            keysCommand: ''
        }
    });

    exports.defaultWaitActionDescriptor = {
        type: 'wait',
        apiArguments: {
            ms: ''
        }
    };

    exports.defaultScreenshotActionDescriptor = {
        type: 'screenshot'
    };

    exports.defaultSelectActionDescriptor = $.extend(true, {}, defaultElementActionDescriptor, {
        type: 'select',
        apiArguments: {
            startPos: null,
            endPos: null
        }
    });

    //Init
    var inSimulation = false;

    //NOTE: when test is run we should block real events (from mouse, keyboard), because it may lead to
    // unexpected test result.
    var preventRealEvtHandler = function (e, dispatched, preventDefault) {
        var target = e.target || e.srcElement;

        if (REAL_ACTION_EVENTS_REGEXP.test(e.type) && !dispatched && (!Settings.RECORDING || Settings.PLAYBACK || inSimulation) && !(Settings.PLAYBACK && Util.isShadowUIElement(target))) {
            //If an element loses focus because of becoming invisible, a blur event is raised. We must not prevent this blur event.
            //In the IE an element loses focus only if css display property is set to 'none', other ways of making
            // element invisible don't lead to blurring
            if (e.type === 'blur') {
                var $target = $(target);
                if (Util.isIE && Util.browserVersion < 12) {
                    if ((!Util.isWindowInstance(target) && $target.css('display') === 'none') || $target.parents().filter(function () {
                        return this.style.display === 'none';
                    }).length)
                    //B254768 - reason of setTimeout method using
                        NativeMethods.setTimeout.call(window, function () {
                            EventSimulator.blur($target[0]);
                        }, 0);
                }
                //NOTE: fix for jQuery bug. An exception raises when call .is(':visible') for window or document on page loading (e.ownerDocument is null) 
                else if (target !== window && target !== window.document && !$target.is(':visible'))
                    return;
            }

            preventDefault();
        }
    };

    var initialized = false;

    exports.init = function () {
        if (initialized)
            return;

        EventSandbox.addFirstInternalHandler(window, Util.RECORDING_LISTENED_EVENTS, preventRealEvtHandler);

        initialized = true;
    };

    //Running
    var runners = window[exports.AUTOMATION_RUNNERS] = {
        click: {
            playback: ClickPlayback
        },
        rclick: {
            playback: RClickPlayback
        },
        dblclick: {
            playback: DblClickPlayback
        },
        drag: {
            playback: DragPlayback
        },
        select: {
            playback: SelectPlayback
        },
        press: {
            playback: PressPlayback
        },
        type: {
            playback: TypePlayback
        },
        hover: {
            playback: HoverPlayback
        }
    };

    function runAutomation(descriptor, runner, callback) {
        if (/click|drag|select|type|hover/.test(descriptor.type) && !descriptor.element) {
            callback();
            return;
        }

        switch (descriptor.type) {
            case 'click':
                //NOTE: We should send previous selected index to be able to determine whether to send the change event
                runner.run(descriptor.element, descriptor.apiArguments.options, callback, {
                    prevSelectedIndex: descriptor.serviceInfo.prevPageState && descriptor.serviceInfo.prevPageState.affectedElementSelectedIndex ?
                        descriptor.serviceInfo.prevPageState.affectedElementSelectedIndex : null});
                break;

            case 'rclick':
                runner.run(descriptor.element, descriptor.apiArguments.options, callback);
                break;

            case 'dblclick':
                runner.run(descriptor.element, descriptor.apiArguments.options, callback);
                break;

            case 'drag':
                var to = {
                    dragOffsetX: descriptor.apiArguments.dragOffsetX,
                    dragOffsetY: descriptor.apiArguments.dragOffsetY
                };

                runner.run(descriptor.element, to, descriptor.apiArguments.options, callback, {
                    screenPointTo: descriptor.endPosition});
                break;

            case 'select':
                var positions = {
                    startPos: descriptor.apiArguments.startPos,
                    endPos: descriptor.apiArguments.endPos
                };

                runner.run(descriptor.element, positions, callback, {
                    screenPointTo: descriptor.endPosition
                });
                break;

            case 'type':
                runner.run(descriptor.element, descriptor.apiArguments.text,
                    descriptor.apiArguments.options, callback);
                break;

            case 'press':
                runner.run(descriptor.apiArguments.keysCommand, callback);
                break;

            case 'hover':
                runner.run(descriptor.element, descriptor.apiArguments.options, callback);
                break;

            default:
                callback();
                break;
        }
    }

    function ensureElements(descriptor) {
        if (descriptor.selector) {
            var $elements = JavascriptExecutor.parseSelectorSync(descriptor.selector).$elements;

            if ($elements && $elements.length) {
                if ($elements.length === 1)
                    descriptor.element = $elements[0];
                else
                    descriptor.elements = $elements.toArray();
            }
        }
    }

    exports.playback = function (descriptor, callback) {
        var curDescriptor = null,
            curIndex = 0;

        if (!runners[descriptor.type]) {
            callback();
            return;
        }

        inSimulation = true;

        ensureElements(descriptor);

        //NOTE: Descriptor contains array of elements for action click on select+option
        //we should call action playback for each array element
        async.forEachSeries(
            descriptor.elements || [],
            function (el, seriesCallback) {
                curDescriptor = $.extend(curDescriptor, descriptor);
                curDescriptor.element = el;
                delete curDescriptor.elements;

                if (curIndex === descriptor.elements.length - 1) {
                    seriesCallback();
                    return;
                }

                runAutomation(curDescriptor, runners[curDescriptor.type].playback, function () {
                    curDescriptor = {};
                    curIndex++;
                    seriesCallback();
                });

            },
            function () {
                curDescriptor = curDescriptor || descriptor;

                runAutomation(curDescriptor, runners[curDescriptor.type].playback, function () {
                    inSimulation = false;

                    callback();
                });
            }
        );
    };
});
