TestCafeClient.define('Automation.Press.KeyPressSimulator', function (require, exports) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        async = Hammerhead.async,
        Util = Hammerhead.Util,
        NativeMethods = Hammerhead.NativeMethods,
        EventSandbox = Hammerhead.EventSandbox,
        EventSimulator = Hammerhead.EventSimulator,
        TextSelection = Hammerhead.TextSelection,
        TypeCharPlaybackAutomation = require('Automation.TypeChar.Playback'),
        AutomationUtil = require('Automation.Util');

    //consts
    var KEY_PRESS_DELAY = 80;

    //utils
    var keyHelper = function (key, originKey) {
        var isChar = key.length === 1 || key === 'space',
            sanitizedKey = isChar ? key : key.toLowerCase(),
            beforeKeydownActiveElement = null,
            beforeKeypressActiveElement = null;

        if (Util.KEYS_MAPS.MODIFIERS_MAP[sanitizedKey])
            sanitizedKey = Util.KEYS_MAPS.MODIFIERS_MAP[sanitizedKey];

        var keyCode = null,
            keyProperty = null,
            modifierKeyCode = Util.KEYS_MAPS.MODIFIERS[sanitizedKey],
            specialKeyCode = Util.KEYS_MAPS.SPECIAL_KEYS[sanitizedKey];

        if (isChar && key !== 'space')
            keyCode = Util.getKeyCodeByChar(sanitizedKey);
        else if (modifierKeyCode)
            keyCode = modifierKeyCode;
        else if (specialKeyCode)
            keyCode = specialKeyCode;

        function needRaiseKeyPressEvent () {
            if (isChar)
                return true;

            //for enter
            if (Util.isWebKit || Util.isMSEdge)
                return specialKeyCode === 13;

            //for enter and esc
            if (Util.isIE && Util.browserVersion <= 11)
                return specialKeyCode === 13 || specialKeyCode === 27;

            if (Util.isMozilla)
                return specialKeyCode && sanitizedKey !== 'capslock';

            return false;
        }

        return {
            down: function (modifiersState) {
                var activeElement = Util.getActiveElement();
                beforeKeydownActiveElement = activeElement;

                if (modifierKeyCode)
                    modifiersState[sanitizedKey] = true;

                keyProperty = Util.KEYS_MAPS.KEY_PROPERTY[originKey] || originKey;

                if(modifiersState.shift && Util.isLetter(key))
                    keyProperty = Util.changeLetterCase(keyProperty);

                return EventSimulator.keydown(activeElement, $.extend({
                    keyCode: keyCode,
                    key: keyProperty
                }, modifiersState));
            },

            press: function (modifiersState) {
                var activeElement = Util.getActiveElement();

                //T559796 - The Backspace key is not processed on running a test in FireFox
                if (!needRaiseKeyPressEvent())
                    return true;

                function getChar(key) {
                    if (key === 'space')
                        return ' ';

                    if (modifiersState.shift) {
                        if (Util.isLetter(key))
                            return Util.changeLetterCase(key);

                        if (Util.KEYS_MAPS.REVERSED_SHIFT_MAP[key])
                            return Util.KEYS_MAPS.REVERSED_SHIFT_MAP[key];
                    }
                    return key;
                }

                var character = isChar ? getChar(sanitizedKey) : null,
                    charCode = specialKeyCode || character.charCodeAt(0);

                if (Util.isWebKit && activeElement !== beforeKeydownActiveElement && Util.isElementInIframe(activeElement) !== Util.isElementInIframe(beforeKeydownActiveElement))
                    return true;

                beforeKeypressActiveElement = activeElement;

                var raiseDefault = EventSimulator.keypress(
                    activeElement,
                    $.extend({
                        keyCode: keyCode,
                        charCode: charCode,
                        key: keyProperty
                    }, modifiersState)
                );

                activeElement = Util.getActiveElement();

                if (raiseDefault) {
                    if (character && !(modifiersState.ctrl || modifiersState.alt)) {
                        //T210448 - Unnecessary typing occurs if element was changed after keydown/keypress event
                        try {
                            if (beforeKeypressActiveElement === activeElement ||
                                (!(Util.isMozilla && !Util.isEditableElement(beforeKeypressActiveElement)) && !(Util.isWebKit && Util.isElementInIframe(activeElement) !== Util.isElementInIframe(beforeKeypressActiveElement) && !Util.isEditableElement(beforeKeypressActiveElement)))) {
                                var elementForTyping = activeElement;

                                if (!Util.isIE && activeElement !== beforeKeypressActiveElement && Util.isEditableElement(beforeKeypressActiveElement) && Util.isEditableElement(activeElement))
                                    elementForTyping = beforeKeypressActiveElement;

                                TypeCharPlaybackAutomation.run(elementForTyping, character);
                            }
                        }
                        catch (err) {
                        }
                    }

                    if (sanitizedKey === 'enter' && activeElement.tagName &&
                        activeElement.tagName.toLowerCase() === 'input' && /button|submit|reset/.test(activeElement.type)) {
                        activeElement.click();
                    }
                }

                return raiseDefault;
            },

            up: function (modifiersState) {
                if (modifierKeyCode)
                    modifiersState[sanitizedKey] = false;

                var raiseDefault = EventSimulator.keyup(Util.getActiveElement(), $.extend({
                        keyCode: keyCode,
                        key: keyProperty
                    }, modifiersState)),

                    activeElement = Util.getActiveElement();

                if (raiseDefault && sanitizedKey === 'space' && activeElement.tagName &&
                    activeElement.tagName.toLowerCase() === 'input' && /button|submit|reset|radio|checkbox/.test(activeElement.type)) {
                    activeElement.click();
                }

                return raiseDefault;
            },

            getKey: function () {
                return sanitizedKey;
            }
        };
    };

    function getLineIndentInTextarea(textarea) {
        var textareaValue = textarea.value,
            inverseSelection = TextSelection.hasInverseSelection(textarea),
            cursorPosition = inverseSelection ?
                TextSelection.getSelectionStart(textarea) :
                TextSelection.getSelectionEnd(textarea);

        if (!textareaValue || !cursorPosition)
            return 0;

        return Util.getTextareaIndentInLine(textarea, cursorPosition);
    }

    //api
    var supportedShortcutHandlers = (function () {
        //utils
        var curTextareaElement = null,
            curTextareaCursorIndent = null;

        function onTextAreaBlur() {
            curTextareaElement = null;
            curTextareaCursorIndent = null;
            NativeMethods.removeEventListener.call(this, 'blur', onTextAreaBlur, true);
        }

        function updateTextAreaIndent(element) {
            if (element.tagName.toLowerCase() === 'textarea') {
                if (curTextareaElement !== element) {
                    NativeMethods.addEventListener.call(element, 'blur', onTextAreaBlur, true);
                    curTextareaElement = element;
                }

                curTextareaCursorIndent = getLineIndentInTextarea(element);
            }
        }

        function moveTextAreaCursorUp(element, withSelection) {
            var textareaValue = element.value;

            if (textareaValue) {
                var hasInverseSelection = TextSelection.hasInverseSelection(element),
                    start = TextSelection.getSelectionStart(element),
                    end = TextSelection.getSelectionEnd(element),
                    partBeforeSelection = textareaValue.substring(0, hasInverseSelection ? start : end),
                    topIndex = partBeforeSelection.lastIndexOf('\n'),
                    top = partBeforeSelection.substring(0, topIndex);

                if (curTextareaCursorIndent === null || curTextareaElement !== element)
                    updateTextAreaIndent(element);

                var newPosition = Math.min(top.lastIndexOf('\n') + 1 + curTextareaCursorIndent, top.length);

                moveTextAreaCursor(element, start, end, hasInverseSelection, newPosition, withSelection);
            }
        }

        function moveTextAreaCursorDown(element, withSelection) {
            var textareaValue = element.value;

            if (textareaValue) {
                var hasInverseSelection = TextSelection.hasInverseSelection(element),
                    start = TextSelection.getSelectionStart(element),
                    end = TextSelection.getSelectionEnd(element),
                    last = textareaValue.substring(hasInverseSelection ? start : end),
                    nextIndex = last.indexOf('\n') === -1 ? last.length : last.indexOf('\n') + 1,
                    bottom = last.substring(nextIndex),
                    newPosition = (hasInverseSelection ? start : end) + nextIndex,
                    maxIndent = bottom.indexOf('\n') === -1 ? bottom.length : bottom.indexOf('\n');

                if (curTextareaCursorIndent === null || curTextareaElement !== element)
                    updateTextAreaIndent(element);

                if (curTextareaCursorIndent >= maxIndent)
                    newPosition += maxIndent;
                else
                    newPosition += curTextareaCursorIndent;

                moveTextAreaCursor(element, start, end, hasInverseSelection, newPosition, withSelection);
            }
        }

        function moveTextAreaCursor(element, start, end, hasInverseSelection, newPosition, withSelection) {
            var newStart = null,
                newEnd = null,
                inverse = null;

            if (withSelection) {
                if (start === end) {
                    if (newPosition < start) {
                        newStart = newPosition;
                        newEnd = start;
                        inverse = true;
                    }
                    else {
                        newStart = start;
                        newEnd = newPosition;
                    }
                }
                else {
                    if (!hasInverseSelection) {
                        if (newPosition < start) {
                            newStart = newPosition;
                            newEnd = start;
                            inverse = true;
                        }
                        else {
                            newStart = start;
                            newEnd = newPosition;
                        }
                    }
                    else {
                        if (newPosition > end) {
                            newStart = end;
                            newEnd = newPosition;
                        }
                        else {
                            newStart = newPosition;
                            newEnd = end;
                            inverse = true;
                        }
                    }
                }
            }
            else
                newEnd = newStart = newPosition;

            TextSelection.select(element, newStart, newEnd, inverse);
        }

        function setElementValue(element, value, selectionStart, selectionEnd) {
            if(value.charAt(0) === '-' && value.charAt(1) === '.')
                value = value.substring(1);

            if(value.charAt(value.length - 1) === '.')
                value = value.substring(0, value.length - 1);

            AutomationUtil.setValue(element, value);
            TextSelection.select(element, selectionStart, selectionEnd);
            EventSimulator.input(element);
        }

        //shortcuts
        function selectAll(element, callback) {
            if (Util.isEditableElement(element))
                TextSelection.select(element);

            callback();
        }

        function backspace(element, callback) {
            if (Util.isTextEditableElementAndEditingAllowed(element)) {
                var startSelection = TextSelection.getSelectionStart(element),
                    endSelection   = TextSelection.getSelectionEnd(element),

                    value    = element.value.replace(/\r\n/g, '\n'),
                    newValue = '';


                if (endSelection === startSelection) {
                    if (startSelection > 0) {
                        newValue = value.substring(0, startSelection - 1) + value.substring(endSelection, value.length);
                        setElementValue(element, newValue, startSelection - 1, startSelection - 1);
                    }
                }
                else {
                    newValue = value.substring(0, startSelection) + value.substring(endSelection, value.length);
                    setElementValue(element, newValue, startSelection, startSelection);
                }
            }
            else if (Util.isContentEditableElement(element))
                TextSelection.deleteSelectionContents(element);

            callback();
        }

        function del(element, callback) {
            if (Util.isTextEditableElementAndEditingAllowed(element)) {
                var startSelection = TextSelection.getSelectionStart(element),
                    endSelection   = TextSelection.getSelectionEnd(element),

                    value    = element.value.replace(/\r\n/g, '\n'),
                    newValue = '';

                if (endSelection === startSelection) {
                    if (startSelection < value.length) {
                        newValue = value.substring(0, startSelection) + value.substring(endSelection + 1, value.length);
                        setElementValue(element, newValue, startSelection, startSelection);
                    }
                }
                else {
                    newValue = value.substring(0, startSelection) + value.substring(endSelection, value.length);
                    setElementValue(element, newValue, startSelection, startSelection);
                }
            }
            else if (Util.isContentEditableElement(element))
                TextSelection.deleteSelectionContents(element);

            callback();
        }

        function left(element, callback) {
            var startSelection = null,
                endSelection = null;

            if (Util.isTextEditableElement(element)) {
                startSelection = TextSelection.getSelectionStart(element);
                endSelection = TextSelection.getSelectionEnd(element);

                var newPosition = startSelection ?
                    startSelection === endSelection ? startSelection - 1 : startSelection :
                    0;

                TextSelection.select(element, newPosition, newPosition);
                updateTextAreaIndent(element);
            }
            else if (Util.isContentEditableElement(element)) {
                startSelection = TextSelection.getSelectionStart(element);
                endSelection = TextSelection.getSelectionEnd(element);

                //NOTE: we only remove selection
                if (startSelection !== endSelection) {
                    var selection = TextSelection.getSelectionByElement(element),
                        inverseSelection = TextSelection.hasInverseSelectionContentEditable(element),
                        startNode = inverseSelection ? selection.focusNode : selection.anchorNode,
                        startOffset = inverseSelection ? selection.focusOffset : selection.anchorOffset;

                    TextSelection.selectByNodesAndOffsets(startNode, startOffset, startNode, startOffset, true, false);
                }
            }

            callback();
        }

        function right(element, callback) {
            var startSelection = null,
                endSelection = null;

            if (Util.isTextEditableElement(element)) {
                startSelection = TextSelection.getSelectionStart(element);
                endSelection = TextSelection.getSelectionEnd(element);

                var newPosition = startSelection === element.value.length ?
                    startSelection :
                    startSelection === endSelection ? startSelection + 1 : endSelection;

                TextSelection.select(element, newPosition, newPosition);
                updateTextAreaIndent(element);
            }
            else if (Util.isContentEditableElement(element)) {
                startSelection = TextSelection.getSelectionStart(element);
                endSelection = TextSelection.getSelectionEnd(element);

                //NOTE: we only remove selection
                if (startSelection !== endSelection) {
                    var selection = TextSelection.getSelectionByElement(element),
                        inverseSelection = TextSelection.hasInverseSelectionContentEditable(element),
                        endNode = inverseSelection ? selection.anchorNode : selection.focusNode,
                        endOffset = inverseSelection ? selection.anchorOffset : selection.focusOffset;

                    TextSelection.selectByNodesAndOffsets(endNode, endOffset, endNode, endOffset, true, false);
                }
            }

            callback();
        }

        function up(element, callback) {
            if (Util.isWebKit && element.tagName && element.tagName.toLowerCase() === 'input') {
                home(element, callback);
                return;
            }

            if (element.tagName && element.tagName.toLowerCase() === 'textarea')
                moveTextAreaCursorUp(element, false);

            callback();
        }

        function down(element, callback) {
            if (Util.isWebKit && element.tagName && element.tagName.toLowerCase() === 'input') {
                end(element, callback);
                return;
            }

            if (element.tagName && element.tagName.toLowerCase() === 'textarea') {
                moveTextAreaCursorDown(element, false);
            }

            callback();
        }

        function home(element, callback, withSelection) {
            if (Util.isTextEditableElement(element)) {
                var elementValue = element.value,
                    selectionStartPosition = TextSelection.getSelectionStart(element),
                    selectionEndPosition = TextSelection.getSelectionEnd(element),
                    inverseSelection = TextSelection.hasInverseSelection(element),

                    isSingleLineSelection = element.tagName.toLocaleLowerCase() !== 'textarea' ? true :
                        Util.getTextareaLineNumberByPosition(element, selectionStartPosition) === Util.getTextareaLineNumberByPosition(element, selectionEndPosition),

                    referencePosition = null;

                if (isSingleLineSelection)
                    referencePosition = inverseSelection ? selectionEndPosition : selectionStartPosition;
                else
                    referencePosition = inverseSelection ? selectionStartPosition : selectionEndPosition;

                var partBefore = elementValue.substring(0, referencePosition),
                    newPosition = partBefore.lastIndexOf('\n') === -1 ? 0 : partBefore.lastIndexOf('\n') + 1;

                if (isSingleLineSelection)
                    TextSelection.select(element, newPosition, withSelection ? referencePosition : newPosition, withSelection);
                else
                    TextSelection.select(element, inverseSelection ? newPosition : selectionStartPosition, inverseSelection ? selectionEndPosition : newPosition, inverseSelection);
            }

            callback();
        }

        function end(element, callback, withSelection) {
            if (Util.isTextEditableElement(element)) {
                var elementValue = element.value,
                    selectionStartPosition = TextSelection.getSelectionStart(element),
                    selectionEndPosition = TextSelection.getSelectionEnd(element),
                    inverseSelection = TextSelection.hasInverseSelection(element),

                    isSingleLineSelection = element.tagName.toLocaleLowerCase() !== 'textarea' ? true :
                        Util.getTextareaLineNumberByPosition(element, selectionStartPosition) === Util.getTextareaLineNumberByPosition(element, selectionEndPosition),

                    referencePosition = null;

                if (isSingleLineSelection)
                    referencePosition = inverseSelection ? selectionEndPosition : selectionStartPosition;
                else
                    referencePosition = inverseSelection ? selectionStartPosition : selectionEndPosition;

                var partAfter = elementValue.substring(referencePosition),
                    newPosition = referencePosition;

                newPosition += partAfter.indexOf('\n') === -1 ? partAfter.length : partAfter.indexOf('\n');

                if (isSingleLineSelection)
                    TextSelection.select(element, withSelection ? referencePosition : newPosition, newPosition);
                else
                    TextSelection.select(element, inverseSelection ? newPosition : selectionStartPosition, inverseSelection ? selectionEndPosition : newPosition, inverseSelection);
            }
            callback();
        }

        function shiftUp(element, callback) {
            if (Util.isWebKit && element.tagName && element.tagName.toLowerCase() === 'input') {
                shiftHome(element, callback);
                return;
            }

            if (element.tagName && element.tagName.toLowerCase() === 'textarea')
                moveTextAreaCursorUp(element, true);

            callback();
        }

        function shiftDown(element, callback) {
            if (Util.isWebKit && element.tagName && element.tagName.toLowerCase() === 'input') {
                shiftEnd(element, callback);
                return;
            }

            if (element.tagName && element.tagName.toLowerCase() === 'textarea')
                moveTextAreaCursorDown(element, true);

            callback();
        }

        function shiftLeft(element, callback) {
            if (Util.isTextEditableElement(element)) {
                var start = TextSelection.getSelectionStart(element),
                    end = TextSelection.getSelectionEnd(element);

                if (start === end || TextSelection.hasInverseSelection(element))
                    TextSelection.select(element, Math.max(start - 1, 0), end, true);
                else
                    TextSelection.select(element, start, Math.max(end - 1, 0), end - 1 < start);

                updateTextAreaIndent(element);
            }

            callback();
        }

        function shiftRight(element, callback) {
            if (Util.isTextEditableElement(element)) {
                var start = TextSelection.getSelectionStart(element),
                    end = TextSelection.getSelectionEnd(element);

                if (start === end || !TextSelection.hasInverseSelection(element))
                    TextSelection.select(element, start, Math.min(end + 1, element.value.length));
                else
                    TextSelection.select(element, Math.min(start + 1, element.value.length), end, start + 1 < end);

                updateTextAreaIndent(element);
            }

            callback();
        }

        function shiftHome(element, callback) {
            home(element, callback, true);
        }

        function shiftEnd(element, callback) {
            end(element, callback, true);
        }

        function enter(element, callback) {
            //submit form on enter pressed
            if (/input/i.test(element.tagName)) {
                if (!Util.isIE)
                    EventSandbox.processElementChanging(element);

                var $form = $(element).parents('form:first');
                if ($form.length) {

                    //if user presses enter when form input is focused, and the form has a submit button,
                    //  browser sends click event to the submit button
                    var $submitButton = $form.find('input[type=submit]:enabled,button[type=submit]:enabled');
                    if (!$submitButton.length)
                    //if button type is not declared, it becomes submit button by default
                        $submitButton = $form.find('button:enabled').filter(function () {
                            return this.type === 'submit';
                        });
                    if ($submitButton.length)
                        EventSimulator.click($submitButton[0]);

                    else {
                        //the form is also submitted on enter press if there is only one input of the following types on it
                        //  and this input is focused (http://www.w3.org/TR/html5/forms.html#implicit-submission)
                        var textInputTagNameRegExp = Util.isWebKit ? /^(text|search|url|number|tel|password|number|email)$/i :
                            /^(text|search|url|number|tel|password|number|email|date|time)$/i;

                        if (textInputTagNameRegExp.test(element.type)) {
                            var $textInputs = $form.find('input').filter(function () {
                                return textInputTagNameRegExp.test(this.type);
                            });

                            if ($textInputs.length === 1 && $textInputs[0] === element && (!element.validity || element.validity.valid)) {
                                if (EventSimulator.submit($form[0]))
                                    $form[0].submit();
                            }
                        }
                    }
                }
            }
            else if (element.tagName && element.tagName.toLowerCase() === 'textarea') {
                var startSelection = TextSelection.getSelectionStart(element),
                    elementValue   = element.value,
                    newPosition    = TextSelection.getSelectionStart(element) + 1,
                    newValue       = elementValue.substring(0, startSelection) + String.fromCharCode(10) + elementValue.substring(startSelection);

                setElementValue(element, newValue, newPosition, newPosition);
            }
            //S173120
            else if (element.tagName && element.tagName.toLowerCase() === 'a')
                EventSimulator.click(element);

            callback();
        }

        function focusNextElement(element, callback) {
            var nextElement = Util.getNextFocusableElement(element);

            if(!nextElement)
                return;

            EventSandbox.focus(nextElement, function () {
                if (Util.isTextEditableInput(nextElement))
                    TextSelection.select(nextElement);

                callback();
            });
        }

        function focusPrevElement(element, callback) {
            var prevElement = Util.getNextFocusableElement(element, true);

            if(!prevElement)
                return;

            EventSandbox.focus(prevElement, function () {
                if (Util.isTextEditableInput(prevElement))
                    TextSelection.select(prevElement);

                callback();
            });
        }

        return {
            'ctrl+a': selectAll,
            'backspace': backspace,
            'delete': del,
            'left': left,
            'right': right,
            'up': up,
            'down': down,
            'shift+left': shiftLeft,
            'shift+right': shiftRight,
            'shift+up': shiftUp,
            'shift+down': shiftDown,
            'shift+home': shiftHome,
            'shift+end': shiftEnd,
            'home': home,
            'end': end,
            'enter': enter,
            'tab': focusNextElement,
            'shift+tab': focusPrevElement
        };
    })();

    //api
    exports.press = function (keysString, actionCallback) {
        var shortcuts = Util.getShortcutsByKeyCombination(supportedShortcutHandlers, keysString.toLowerCase()),
            keysCombinationHandlers = {},
            processedString = '',
            notProcessedString = keysString,

            i = 0;

        for (i = 0; i < shortcuts.length; i++) {
            var shortcut = shortcuts[i].toLowerCase(),
                position = notProcessedString.indexOf(shortcut),
                length = shortcut.length;

            processedString += notProcessedString.substring(0, position + length);

            keysCombinationHandlers[processedString] = Util.getShortcutHandlerByKeyCombination(supportedShortcutHandlers, shortcut);

            notProcessedString = notProcessedString.substring(position + length);
        }

        var modifiersState = { ctrl: false, alt: false, shift: false, meta: false },
            keys = Util.getArrayByKeyCombination(keysString),
            keysForKeyEvent = keys.slice();

        //NOTE: check 'shift' modifier in keys
        for (i = 0; i < keys.length; i++) {
            var key = keys[i];

            if (key.toLowerCase() === 'shift') {
                var nextKey = keys[i + 1];

                if (!nextKey)
                    continue;

                if (Util.KEYS_MAPS.SHIFT_MAP[nextKey])
                    keys[i + 1] = Util.KEYS_MAPS.SHIFT_MAP[nextKey];
                else if (Util.KEYS_MAPS.REVERSED_SHIFT_MAP[nextKey])
                    keysForKeyEvent[i + 1] = Util.KEYS_MAPS.REVERSED_SHIFT_MAP[nextKey];
            }

            if (Util.KEYS_MAPS.SHIFT_MAP[key] && (!keys[i - 1] || keys[i - 1].toLowerCase() !== 'shift')) {
                keys[i] = Util.KEYS_MAPS.SHIFT_MAP[key];

                keys.splice(i, 0, 'shift');
                keysForKeyEvent.splice(i, 0, 'shift');
                i++;
            }
        }

        var keysHelpers = $.map(keys, function (key, index) {
            return keyHelper(key, keysForKeyEvent[index]);
        });

        //press keys
        var pressedKeyHelpers = [],
            currentCombination = '';

        async.series({
            keydown: function (seriesCallback) {
                async.forEachSeries(
                    keysHelpers,
                    function (helper, helperCallback) {
                        var preventDefault = !helper.down(modifiersState),
                            key = helper.getKey();

                        pressedKeyHelpers.push(helper);
                        currentCombination += (currentCombination ? '+' : '') + key;

                        var callback = function () {
                            NativeMethods.setTimeout.call(window, helperCallback, KEY_PRESS_DELAY);
                        };

                        if (preventDefault)
                            callback();
                        else {
                            var currentShortcutHandler = keysCombinationHandlers[currentCombination];

                            if (!currentShortcutHandler || Util.isMozilla || /^(enter|esc)$/.test(key))  //B254435
                                preventDefault = !helper.press(modifiersState);

                            if (!preventDefault && currentShortcutHandler)
                                currentShortcutHandler(Util.getActiveElement(), callback);
                            else
                                callback();
                        }
                    },
                    function () {
                        seriesCallback();
                    }
                );
            },

            keyup: function () {
                async.whilst(
                    function () {
                        return pressedKeyHelpers.length;
                    },
                    function (callback) {
                        pressedKeyHelpers.pop().up(modifiersState);
                        NativeMethods.setTimeout.call(window, callback, KEY_PRESS_DELAY);
                    },
                    function () {
                        actionCallback();
                    }
                );
            }
        });
    };
});