TestCafeClient.define('Automation.Scroll.Playback', function (require, exports) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        async = Hammerhead.async,
        Util = Hammerhead.Util,
        MessageSandbox = Hammerhead.MessageSandbox,
        NativeMethods = Hammerhead.NativeMethods,
        CrossDomainMessages = require('Base.CrossDomainMessages');

    var DEFAULT_MAX_SCROLL_MARGIN = 50,
        SCROLL_MARGIN_INCREASE_STEP = 20,
        MIN_SCROLL_STEP = 10,
        SCROLL_SPEED = 1,
        SCROLL_DELAY = 10;

    function isElementHiddenByFixed (target, offsetX, offsetY, doc) {
        doc = doc || document;
        
        var clientDimensions = Util.getClientDimensions(target);
        var elementInPoint   = Util.getElementFromPoint(clientDimensions.left +
                                                                 offsetX, clientDimensions.top + offsetY);

        return elementInPoint && Util.getComputedStyle(elementInPoint, doc).position === 'fixed';
    }

    function isScrollMarginTooBig (maxScrollMargin) {
        var minWindowDimension = Math.min(window.innerWidth, window.innerHeight);

        return maxScrollMargin >= minWindowDimension / 2;
    }

    function getScrollOption($currentScrollable, target, offsetX, offsetY, maxScrollMargin) {
        var targetDimensions = Util.getClientDimensions(target),
            scrollableDimensions = Util.getClientDimensions($currentScrollable[0]),

            scrollTop = null,
            scrollLeft = null,
            maxLeftScrollMargin = Math.min(maxScrollMargin, Math.floor(scrollableDimensions.width / 2)),
            maxTopScrollMargin = Math.min(maxScrollMargin, Math.floor(scrollableDimensions.height / 2)),

            isScrollDown = -1,
            isScrollRight = -1;

        var relation = {
            top: targetDimensions.top - (scrollableDimensions.top + scrollableDimensions.border.top + maxTopScrollMargin),
            bottom: scrollableDimensions.bottom - scrollableDimensions.border.bottom - scrollableDimensions.scrollbar.bottom - maxTopScrollMargin - targetDimensions.bottom,
            left: targetDimensions.left - (scrollableDimensions.left + scrollableDimensions.border.left + maxLeftScrollMargin),
            right: scrollableDimensions.right - scrollableDimensions.border.right - scrollableDimensions.scrollbar.right - maxLeftScrollMargin - targetDimensions.right
        };

        // vertical scroll
        if (relation.top < 0)
            scrollTop = Math.round(scrollableDimensions.scroll.top + relation.top - maxTopScrollMargin);
        else if (relation.top > 0 && relation.bottom < 0) {
            scrollTop = Math.round(scrollableDimensions.scroll.top + Math.min(relation.top, -relation.bottom) + maxTopScrollMargin);
            isScrollDown = 1;
        }

        // horizontal scroll
        if (relation.left < 0)
            scrollLeft = Math.round(scrollableDimensions.scroll.left + relation.left - maxLeftScrollMargin);
        else if (relation.left > 0 && relation.right < 0) {
            scrollLeft = Math.round(scrollableDimensions.scroll.left + Math.min(relation.left, -relation.right) + maxLeftScrollMargin);
            isScrollRight = 1;
        }

        //NOTE: we should check: can we show the full element
        var targetElementWidth = Math.abs(targetDimensions.right - targetDimensions.left),
            targetElementHeight = Math.abs(targetDimensions.bottom - targetDimensions.top);

        if (scrollableDimensions.width <= targetElementWidth && offsetX >= scrollableDimensions.width - (scrollableDimensions.border.left + scrollableDimensions.border.right))
            scrollLeft = (scrollLeft === null ? offsetX : scrollLeft + offsetX - isScrollRight * maxLeftScrollMargin) - maxLeftScrollMargin;

        if (scrollableDimensions.height <= targetElementHeight && offsetY >= scrollableDimensions.height - (scrollableDimensions.border.top + scrollableDimensions.border.bottom))
            scrollTop = (scrollTop === null ? offsetY : scrollTop + offsetY - isScrollDown * maxTopScrollMargin) - maxTopScrollMargin;

        if (scrollLeft !== null || scrollTop !== null) {
            return {
                top: scrollTop,
                left: scrollLeft
            };
        }

        return null;
    }

    function scrollElement($el, left, top, scrollCallback) {
        var leftScrollStep = null,
            topScrollStep = null,

            elementScroll = Util.getElementScroll($el),
            startLeftScroll = elementScroll.left,
            startTopScroll = elementScroll.top,

            scrollLeftDirection = 0,
            scrollTopDirection = 0,
            currentScrollLeft = null,
            currentScrollTop = null,

            lastScrollLeft = null,
            lastScrollTop = null,

            requestAnimationFrame = window.requestAnimationFrame || window.webkitRequestAnimationFrame,
            scrollingStep = null,
            startScrollingStamp = null;

        function getScrollDirection(newScrollValue, isLeft) {
            if (newScrollValue === null)
                return 0;
            if (newScrollValue > (isLeft ? startLeftScroll : startTopScroll))
                return 1;
            return -1;
        }

        function assignScrollValues() {
            $el.scrollLeft(currentScrollLeft);
            $el.scrollTop(currentScrollTop);

            lastScrollLeft = currentScrollLeft;
            lastScrollTop = currentScrollTop;
        }

        scrollLeftDirection = getScrollDirection(left, true);
        scrollTopDirection = getScrollDirection(top, false);

        //we assign this minimum possible scroll step to avoid too slow action
        leftScrollStep = Math.max(Math.abs(left - startLeftScroll) / SCROLL_SPEED, MIN_SCROLL_STEP);
        topScrollStep = Math.max(Math.abs(top - startTopScroll) / SCROLL_SPEED, MIN_SCROLL_STEP);

        async.whilst(
            function () {
                if (!scrollLeftDirection && !scrollTopDirection)
                    return false;

                currentScrollLeft = Math.round(Util.getElementScroll($el).left + scrollLeftDirection * leftScrollStep);
                currentScrollLeft = scrollLeftDirection === 1 ? Math.min(currentScrollLeft, left) : Math.max(currentScrollLeft, left);

                currentScrollTop = Math.round(Util.getElementScroll($el).top + scrollTopDirection * topScrollStep);
                currentScrollTop = scrollTopDirection === 1 ? Math.min(currentScrollTop, top) : Math.max(currentScrollTop, top);

                return currentScrollLeft !== lastScrollLeft || currentScrollTop !== lastScrollTop;
            },

            function (callback) {
                //NOTE: for IOS > 6 (B254728)
                if (Util.isSafari && requestAnimationFrame) {
                    scrollingStep = function (timeStamp) {
                        if (!startScrollingStamp)
                            startScrollingStamp = timeStamp;

                        assignScrollValues();

                        if (timeStamp - startScrollingStamp >= SCROLL_SPEED)
                            callback();
                        else
                            requestAnimationFrame(scrollingStep);
                    };

                    requestAnimationFrame(scrollingStep);
                }
                else {
                    assignScrollValues();
                    NativeMethods.setTimeout.call(window, function () {
                        callback();
                    }, SCROLL_DELAY);
                }
            },

            function () {
                scrollCallback();
            }
        );
    }

    exports.run = function (to, actionOptions, currentDocument, actionCallback) {
        var afterScrollDelay = Util.isTouchDevice && Util.isMozilla ? 200 : 0,
            isDomElement = Util.isDomElement(to),
            isHtmlElement = isDomElement && /html/i.test(to.tagName),
            scrollableParents = [],
            pointTo = null,

            $target = null,
            targetRect = null,
            targetWidth = null,
            targetHeight = null,
            offsetX = null,
            offsetY = null,

            maxScrollMargin     = actionOptions && actionOptions.maxScrollMargin || DEFAULT_MAX_SCROLL_MARGIN,
            maxLeftScrollMargin = null,
            maxTopScrollMargin  = null,


            ownScroll = null,
            newOwnScrollLeft = null,
            newOwnScrollTop = null,

            currentOffsetX = 0,
            currentOffsetY = 0,
            windowTopResponse = null;

        function considerWindowTopScroll(msg, callback) {
            windowTopResponse = function (e) {
                if (e.message.cmd === CrossDomainMessages.SCROLL_TOP_WINDOW_RESPONSE_CMD) {
                    MessageSandbox.off(MessageSandbox.SERVICE_MSG_RECEIVED, windowTopResponse);
                    windowTopResponse = null;

                    callback();
                }
            };

            MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, windowTopResponse);

            MessageSandbox.sendServiceMsg(msg, window.top);
        }

        function scrollParents (scrollParentsCallback) {
            scrollableParents = isDomElement ? Util.getScrollableParents(to, currentDocument) : [document.documentElement];

            if (!scrollableParents.length) {
                scrollParentsCallback();
                return;
            }

            var currentTarget = pointTo ? pointTo : to;

            currentOffsetX = $target ? offsetX - $target.scrollLeft() : offsetX;
            currentOffsetY = $target ? offsetY - $target.scrollTop() : offsetY;


            async.forEachSeries(scrollableParents,
                function (el, callback) {
                    var $el = $(el),
                        scrollOption = getScrollOption($el, currentTarget, Math.max(0, currentOffsetX), Math.max(0, currentOffsetY), maxScrollMargin);

                    function parentScrollCallback() {
                        var newTargetDimensions = Util.getClientDimensions(currentTarget),
                            newScrollableDimensions = Util.getClientDimensions($el[0]);

                        currentOffsetX = newTargetDimensions.left - newScrollableDimensions.left + newScrollableDimensions.border.left + offsetX;
                        currentOffsetY = newTargetDimensions.top - newScrollableDimensions.top + newScrollableDimensions.border.top + offsetY;

                        currentTarget = el;
                        callback();
                    }

                    if (!scrollOption) {
                        parentScrollCallback();
                        return;
                    }

                    async.series({
                        scrollBody: function (callback) {
                            var body = /html/i.test(el.tagName) && $el.find('body');

                            if (body.length)
                                scrollElement(body, scrollOption.left, scrollOption.top, callback);
                            else
                                callback();
                        },
                        ownScroll: function () {
                            scrollElement($el, scrollOption.left, scrollOption.top, parentScrollCallback);
                        }
                    });
                },

                function () {
                    scrollParentsCallback();
                });
        }

        async.series({
            scrollElement: function (scrollElementCallback) {
                if (!isDomElement) {
                    scrollElementCallback();
                    return;
                }

                $target = $(to);
                targetRect = $target[0].getBoundingClientRect();
                targetWidth = isHtmlElement ? $target[0].clientWidth : targetRect.width;
                targetHeight = isHtmlElement ? $target[0].clientHeight : targetRect.height;
                offsetX = actionOptions && typeof actionOptions.offsetX !== 'undefined' ? actionOptions.offsetX : Math.round(targetWidth / 2);
                offsetY = actionOptions && typeof actionOptions.offsetY !== 'undefined' ? actionOptions.offsetY : Math.round(targetHeight / 2);
                maxLeftScrollMargin = Math.min(maxScrollMargin, Math.floor(targetWidth / 2));
                maxTopScrollMargin = Math.min(maxScrollMargin, Math.floor(targetHeight / 2));
                ownScroll = Util.getElementScroll($target);

                if (!Util.hasScroll(to, currentDocument)) {
                    scrollElementCallback();
                    return;
                }

                if (offsetX >= ownScroll.left + targetWidth)
                    newOwnScrollLeft = offsetX - maxLeftScrollMargin;
                else if (offsetX <= ownScroll.left)
                    newOwnScrollLeft = offsetX - maxLeftScrollMargin;

                if (offsetY >= ownScroll.top + targetHeight)
                    newOwnScrollTop = offsetY - maxTopScrollMargin;
                else if (offsetY <= ownScroll.top)
                    newOwnScrollTop = offsetY - maxTopScrollMargin;

                async.series({
                    scrollBody: function (callback) {
                        if (isHtmlElement)
                            scrollElement($target.find('body'), newOwnScrollLeft, newOwnScrollTop, callback);
                        else
                            callback();
                    },
                    ownScroll: function (callback) {
                        scrollElement($target, newOwnScrollLeft, newOwnScrollTop, callback);
                    },
                    callback: function () {
                        scrollElementCallback();
                    }
                });
            },

            scrollElementParents: scrollParents,

            checkScrolling: function (checkScrollingCallback) {
                async.whilst(
                    function () {
                        return !isScrollMarginTooBig(maxScrollMargin) && isElementHiddenByFixed(to, offsetX, offsetY, currentDocument);
                    },

                    function  (iterationCallback) {
                        maxScrollMargin += SCROLL_MARGIN_INCREASE_STEP;

                        scrollParents(iterationCallback);
                    },
                    
                    checkScrollingCallback
                );
            },

            callback: function () {
                NativeMethods.setTimeout.call(window, function () {
                    if (window.top === window.self) {
                        actionCallback();
                        return;
                    }

                    if (isDomElement) {
                        var msg = {
                            cmd: CrossDomainMessages.SCROLL_TOP_WINDOW_REQUEST_CMD,
                            options: {
                                offsetX:         Math.max(0, currentOffsetX),
                                offsetY:         Math.max(0, currentOffsetY),
                                maxScrollMargin: maxScrollMargin
                            }
                        };

                        considerWindowTopScroll(msg, actionCallback);
                    }
                    else {
                        windowTopResponse = function (e) {
                            if (e.message.cmd === CrossDomainMessages.GET_IFRAME_POSITION_DATA_RESPONSE_CMD) {
                                MessageSandbox.off(MessageSandbox.SERVICE_MSG_RECEIVED, windowTopResponse);
                                windowTopResponse = null;

                                var documentScroll = Util.getElementScroll($(document)),

                                    msg = {
                                        cmd: CrossDomainMessages.SCROLL_TOP_WINDOW_REQUEST_CMD,
                                        point: {
                                            x: to.x - documentScroll.left + e.message.scroll.left,
                                            y: to.y - documentScroll.top + e.message.scroll.top,

                                            maxScrollMargin: maxScrollMargin
                                        }
                                    };

                                considerWindowTopScroll(msg, actionCallback);
                            }
                        };

                        MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, windowTopResponse);

                        MessageSandbox.sendServiceMsg({cmd: CrossDomainMessages.GET_IFRAME_POSITION_DATA_REQUEST_CMD}, window.top);
                    }
                }, afterScrollDelay);
            }
        });
    };
});