TestCafeClient.define('Automation.TypeChar.Playback', function (require, exports) {
    var Hammerhead            = HammerheadClient.get('Hammerhead'),
        $                     = Hammerhead.$,
        Util                  = Hammerhead.Util,
        NativeMethods         = Hammerhead.NativeMethods,
        EventSandbox          = Hammerhead.EventSandbox,
        EventSimulator        = Hammerhead.EventSimulator,
        TextSelection         = Hammerhead.TextSelection,
        ContentEditableHelper = Hammerhead.ContentEditableHelper,
        AutomationUtil        = require('Automation.Util');

    function getCorrectElementSelection (element) {
        var currentSelection = TextSelection.getSelectionByElement(element),
            inverseSelection = TextSelection.hasInverseSelectionContentEditable(element);

        if (TextSelection.hasElementContainsSelection(element))
            currentSelection = ContentEditableHelper.getSelection(element, currentSelection, inverseSelection);
        else {
            //NOTE: if we type text in element which don't contains selection
            // we think selectionStart and selectionEnd positions are null in this element.
            //So we calculate the necessary start and end nodes and offsets
            var startSelectionPosition = ContentEditableHelper.calculateNodeAndOffsetByPosition(element, 0),
                endSelectionPosition   = ContentEditableHelper.calculateNodeAndOffsetByPosition(element, 0);

            currentSelection             = {};
            currentSelection.startNode   = startSelectionPosition.node;
            currentSelection.startOffset = startSelectionPosition.offset;
            currentSelection.endNode     = endSelectionPosition.node;
            currentSelection.endOffset   = endSelectionPosition.offset;
        }
        return currentSelection;
    }

    function typeCharToContentEditable (element, characters) {
        var inputEventRaised = false;

        function onInput () {
            inputEventRaised = true;
        }

        EventSandbox.addInternalEventListener(window, ['input'], onInput);

        var chars            = characters === ' ' ? String.fromCharCode(160) : characters,
            currentSelection = getCorrectElementSelection(element),
            startNode        = currentSelection.startNode,
            startOffset      = currentSelection.startOffset,
            endNode          = currentSelection.endNode,
            endOffset        = currentSelection.endOffset;

        if (!startNode || !Util.isContentEditableElement(startNode) || !endNode ||
            !Util.isContentEditableElement(endNode))
            return;

        var firstNonWhitespaceSymbolIndex = ContentEditableHelper.getFirstNonWhitespaceSymbolIndex(startNode.nodeValue),
            lastNonWhitespaceSymbolIndex  = ContentEditableHelper.getLastNonWhitespaceSymbolIndex(startNode.nodeValue),
            startNodeParent               = startNode.parentNode,
            needDeleteSelectedContent     = !Util.isTheSameNode(startNode, endNode),
            inverseSelection              = null,

            correctedStartNode            = null,
            correctedStartOffset          = null,
            correctedEndOffset            = null,
            newCaretPos                   = null,

            oldStartNodeValue             = null;

        if (needDeleteSelectedContent) {
            TextSelection.deleteSelectionContents(element);

            //NOTE: After deleting selection contents we should refresh stored startNode
            //because contentEditable element's content could change and we can not find parentElements of nodes
            //In MSEdge 'parentElement' for deleted element isn't undefined
            if (Util.isWebKit || !(startNode.parentNode && startNode.parentElement) ||
                !Util.isElementContainsNode(element, startNode)) {
                currentSelection = getCorrectElementSelection(element);
                inverseSelection = TextSelection.hasInverseSelectionContentEditable(element);
                startNode        = inverseSelection ? currentSelection.endNode : currentSelection.startNode;
                startOffset      = inverseSelection ? currentSelection.endOffset : currentSelection.startOffset;

                if (!startNode || !Util.isContentEditableElement(startNode))
                    return;
            }
            endOffset = startOffset;
        }

        if (!Util.isRenderedNode(startNode)) {
            return;
        }

        //NOTE: We can type only in text nodes(except if selected node is 'br' node, then we use special behavior)
        if (startNode.nodeType === 1) {
            correctedStartNode = document.createTextNode(chars);
            if (startNode.tagName.toLowerCase() === 'br')
                startNodeParent.insertBefore(correctedStartNode, startNode);
            else
                startNode.appendChild(correctedStartNode);
            newCaretPos = chars.length;
            TextSelection.selectByNodesAndOffsets(correctedStartNode, newCaretPos, correctedStartNode, newCaretPos);

            NativeMethods.setTimeout.call(window, function () {
                if (!inputEventRaised)
                    EventSimulator.input(element);

                EventSandbox.removeInternalEventListener(window, ['input'], onInput);
            }, 0);
            return;
        }

        if (startOffset < firstNonWhitespaceSymbolIndex && startOffset !== 0) {
            correctedStartOffset = firstNonWhitespaceSymbolIndex;
            correctedEndOffset   = endOffset + (firstNonWhitespaceSymbolIndex - startOffset);
        }
        else if (endOffset > lastNonWhitespaceSymbolIndex && endOffset !== startNode.nodeValue.length) {
            correctedStartOffset = startNode.nodeValue.length;
            correctedEndOffset   = endOffset + (startNode.nodeValue.length - startOffset);
        }
        else {
            correctedStartOffset = startOffset;
            correctedEndOffset   = endOffset;
        }

        correctedStartNode           = correctedStartNode || startNode;
        oldStartNodeValue            = correctedStartNode.nodeValue;
        correctedStartNode.nodeValue = oldStartNodeValue.substring(0, correctedStartOffset) + chars +
                                       oldStartNodeValue.substring(correctedEndOffset, oldStartNodeValue.length);
        newCaretPos                  = correctedStartOffset + chars.length;
        TextSelection.selectByNodesAndOffsets(correctedStartNode, newCaretPos, correctedStartNode, newCaretPos);

        NativeMethods.setTimeout.call(window, function () {
            if (!inputEventRaised)
                EventSimulator.input(element);

            EventSandbox.removeInternalEventListener(window, ['input'], onInput);
        }, 0);
    }

    exports.run = function (element, characters) {
        if (Util.isContentEditableElement(element)) {
            typeCharToContentEditable(element, characters);
            return;
        }

        if (!Util.isTextEditableElementAndEditingAllowed(element))
            return;

        var startSelection   = TextSelection.getSelectionStart(element),
            endSelection     = TextSelection.getSelectionEnd(element),

            //NOTE: attribute 'maxlength' doesn't work in all browsers. In IE still don't support input with type 'number'
            isNumberInput    = Util.isInputTypeNumber(element),
            elementMaxLength = !Util.isIE && isNumberInput ?
                               null : parseInt($(element).attr('maxLength')),

            value            = element.value;

        if (elementMaxLength === null || isNaN(elementMaxLength) || elementMaxLength > value.length) {
            if (isNumberInput) {
                if (Util.isIOS && value[value.length - 1] === '.') {    //B254013
                    startSelection += 1;
                    endSelection += 1;
                }
            }

            AutomationUtil.setValue(element, value.substring(0, startSelection) + characters +
                            value.substring(endSelection, value.length));

            TextSelection.select(element, startSelection + characters.length, startSelection + characters.length);
        }

        //NOTE: B253410, T138385 (we should simulate 'input' event after type char in 'type' and 'press' action)
        EventSimulator.input(element);
    };
});