TestCafeClient.define('Automation.Util', function (require, exports) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        ContentEditableHelper = Hammerhead.ContentEditableHelper,
        TextSelection = Hammerhead.TextSelection,
        CursorWidget = require('UI.Cursor'),
        Util = Hammerhead.Util,
        EventSandbox = Hammerhead.EventSandbox,
        NativeMethods = Hammerhead.NativeMethods;

    var $document = $(document);

    exports.getMouseActionPoint = function (el, actionOptions, convertToScreen) {
        var elementOffset = Util.getOffsetPosition(el),
            left = el === $document[0].documentElement ? 0 : elementOffset.left,
            top = el === $document[0].documentElement ? 0 : elementOffset.top,
            elementScroll = Util.getElementScroll($(el)),
            point = Util.findCenter(el);

        if (actionOptions && typeof actionOptions.offsetX !== 'undefined' && !isNaN(parseInt(actionOptions.offsetX)))
            point.x = left + (actionOptions.offsetX || 0);

        if (actionOptions && typeof actionOptions.offsetY !== 'undefined' && !isNaN(parseInt(actionOptions.offsetY)))
            point.y = top + (actionOptions.offsetY || 0);

        if (convertToScreen) {
            if (!/html/i.test(el.tagName) && Util.hasScroll(el, Util.findDocument(el))) {
                point.x -= elementScroll.left;
                point.y -= elementScroll.top;
            }
            return Util.offsetToClientCoords(point);
        }

        return point;
    };

    exports.getEventOptionCoordinates = function (element, screenPoint) {
        var clientPoint = {
            x: screenPoint.x,
            y: screenPoint.y
        };

        if (Util.isElementInIframe(element)) {
            var currentIFrame = Util.getIFrameByElement(element);
            if (currentIFrame) {
                var iFramePosition = Util.getOffsetPosition(currentIFrame),
                    iFrameBorders = Util.getBordersWidth($(currentIFrame)),
                    iframeClientPosition = Util.offsetToClientCoords({x: iFramePosition.left, y: iFramePosition.top});

                clientPoint.x -= (iframeClientPosition.x + iFrameBorders.left);
                clientPoint.y -= (iframeClientPosition.y + iFrameBorders.top);
            }
        }

        return clientPoint;
    };

    exports.focusAndSetSelection = function (element, options, needFocus, callback) {
        var activeElement = Util.getActiveElement(),
            isTextEditable = Util.isTextEditableElement(element),
            isContentEditable = Util.isContentEditableElement(element),
            focusableElement = isContentEditable ? ContentEditableHelper.findContentEditableParent(element) : element,
            contentEditableParent = null,
            needSelection = isTextEditable || isContentEditable,
            $labelWithForAttr = $(element).closest('label[for]');

        //NOTE: in WebKit if selection was never set in an input element, focus method selects all text of this element
        if (needFocus && Util.isWebKit && isTextEditable)
            TextSelection.select(element, 0, 0);
        //NOTE: we should call focus for input element after click on label with attribute 'for' (on recording)
        //T253883 - Playback - It is impossible to type a password
        if ($labelWithForAttr.length && !Util.isElementFocusable($(element))) {
            if (needFocus)
                exports.focusElementByLabel($labelWithForAttr, callback);
            else
                callback();
        }
        else
            EventSandbox.focus(focusableElement, function () {
                //NOTE: if some other element was focused in the focus event handler we should not set selection
                if (!isContentEditable && needFocus && element !== Util.getActiveElement()) {
                    callback();
                    return;
                }

                if (needSelection) {
                    if (isContentEditable && isNaN(parseInt(options.caretPos)))
                        TextSelection.setCursorToLastVisiblePosition(element);
                    else {
                        var position = isNaN(parseInt(options.caretPos)) ? element.value.length : options.caretPos;
                        TextSelection.select(element, position, position);
                    }
                }
                else {
                    //NOTE: if focus is called for not contentEditable element (like 'img' or 'button') inside contentEditable parent
                    // we should try to set right window selection. Generally we can't set right window selection object because
                    // after selection setup window.getSelection method returns  a different object depending on the browser.
                    contentEditableParent = ContentEditableHelper.findContentEditableParent(focusableElement);
                    if (contentEditableParent)
                        TextSelection.setCursorToLastVisiblePosition(focusableElement);
                }
                //we can't avoid element focusing because set selection methods lead to focusing.
                // So we just focus previous active element without handlers if we don't need focus here
                if (!needFocus && activeElement !== Util.getActiveElement()) {
                    EventSandbox.focus(activeElement, callback, true, true);
                }
                else
                    callback();
            }, !needFocus, true);
    };

    //NOTE: in all browsers after simulation of  click on label with attribute 'for' occurs focus event
    // for related element, except Mozilla (so we should call focus)
    exports.focusLabelChildElement = function (element, callback) {
        var $labelWithForAttr = $(element).closest('label[for]');

        if ($labelWithForAttr.length && !Util.isElementFocusable($(element))) {
            var $focusableElements = $('#' + $labelWithForAttr.attr('for'), Util.findDocument($labelWithForAttr[0]));

            if ($focusableElements.length) {
                var focusableElement = $focusableElements[0];

                if (Util.getActiveElement() !== focusableElement) {
                    EventSandbox.focus(focusableElement, callback, false, true);

                    return;
                }
            }
        }
        if (callback)
            callback();
    };

    exports.focusElementByLabel = function ($label, callback) {
        var $focusableElements = $('#' + $label.attr('for'), Util.findDocument($label[0]));

        if ($focusableElements.length) {
            var focusableElement = $focusableElements[0];

            if (Util.getActiveElement() !== focusableElement) {
                EventSandbox.focus(focusableElement, callback, false, true);
                return;
            }
        }

        if (callback)
            callback();
    };

    exports.getDragEndPoint = function (startPosition, to, currentDocument) {
        var dragInIFrame = currentDocument !== document,
            pointTo = {
                x: startPosition.x + Math.floor(to.dragOffsetX),
                y: startPosition.y + Math.floor(to.dragOffsetY)
            },
            maxX = 0,
            maxY = 0;

        if (dragInIFrame) {
            var currentIFrame = Util.getIFrameByElement(currentDocument);
            if (currentIFrame) {
                var iFrameOffset = Util.getOffsetPosition(currentIFrame),
                    iFrameBorders = Util.getBordersWidth($(currentIFrame));

                maxX = iFrameOffset.left + iFrameBorders.left;
                maxY = iFrameOffset.top + iFrameBorders.top;
            }
        }

        maxX += $(currentDocument).width();
        maxY += $(currentDocument).height();
        pointTo.x = pointTo.x < 0 ? 0 : pointTo.x;
        pointTo.x = pointTo.x > maxX ? maxX : pointTo.x;
        pointTo.y = pointTo.y < 0 ? 0 : pointTo.y;
        pointTo.y = pointTo.y > maxY ? maxY : pointTo.y;

        return pointTo;
    };

    exports.getElementUnderCursor = function (x, y, currentDocument, target) {
        var topElement = CursorWidget.getElementUnderCursor(x, y, currentDocument);

        if (!target || !Util.isDomElement(target) || topElement === target || !topElement)
            return topElement;

        //NOTE: T299665 - Incorrect click on image with associated map element in Mozilla
        if (Util.isMozilla && target.tagName.toLowerCase() === 'area' && topElement.tagName.toLowerCase() === 'img') {
            var mapElement = $(target).closest('map')[0];

            return mapElement && mapElement.name === topElement.useMap.substring(1) ? target : topElement;
        }

        if ((target.tagName.toLowerCase() !== 'a' && !$(target).parents('a').length) || !$(target).text().length)
            return topElement;

        //NOTE: we caught link's child with text so it fit for redirect
        if (target.tagName.toLowerCase() === 'a' && $(target).has(topElement).length && $(topElement).text().length)
            return topElement;

        //NOTE: try to find myltiline link by its rectangle (T163678)
        var linkRect = target.getBoundingClientRect(),
            curTopElement = CursorWidget.getElementUnderCursor(linkRect.right - 1, linkRect.top + 1, currentDocument);

        if ((curTopElement && curTopElement === target) ||
            ($(target).has(curTopElement).length && $(curTopElement).text().length))
            return curTopElement;
        else {
            curTopElement = CursorWidget.getElementUnderCursor(linkRect.left + 1, linkRect.bottom - 1);

            if ((curTopElement && curTopElement === target) ||
                ($(target).has(curTopElement).length && $(curTopElement).text().length))
                return curTopElement;
        }

        return topElement;
    };

    exports.setValue = function (element, value) {
        if (NativeMethods.inputValueSetter && Util.isInputElement(element))
            NativeMethods.inputValueSetter.call(element, value);
        else if (NativeMethods.textAreaValueSetter && Util.isTextAreaElement(element))
            NativeMethods.textAreaValueSetter.call(element, value);
        else
            element.value = value;
    };
});
