TestCafeClient.define('TestRunner.API.Actions', function (require, exports) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        async = Hammerhead.async,
        Util = Hammerhead.Util,
        JSProcessor = Hammerhead.JSProcessor,
        NativeMethods = Hammerhead.NativeMethods,
        SharedErrors = require('Shared.Errors'),
        SharedConst = require('Shared.Const'),
        Settings = require('Settings'),
        SourceIndexTracker = require('TestRunner.SourceIndexTracker'),
        ContentEditableHelper = Hammerhead.ContentEditableHelper,
        Automation = require('Automation'),
        ClickPlaybackAutomation = require('Automation.Click.Playback'),
        RClickPlaybackAutomation = require('Automation.RClick.Playback'),
        DblClickPlaybackAutomation = require('Automation.DblClick.Playback'),
        DragPlaybackAutomation = require('Automation.Drag.Playback'),
        SelectPlaybackAutomation = require('Automation.Select.Playback'),
        TypePlaybackAutomation = require('Automation.Type.Playback'),
        PressPlaybackAutomation = require('Automation.Press.Playback'),
        HoverPlaybackAutomation = require('Automation.Hover.Playback');

    //NOTE: published for tests purposes only
    exports.ELEMENT_AVAILABILITY_WAITING_TIMEOUT = 10000;

    var ELEMENT_AVAILABILITY_WAITING_DELAY = 200,
        WAIT_FOR_DEFAULT_TIMEOUT = 10000,
        CHECK_CONDITION_INTERVAL = 50;

    //Global
    var testIterator = null;

    function failWithError(code, additionalParams) {
        var err = {
            code: code,
            stepName: Settings.CURRENT_TEST_STEP_NAME,
            __sourceIndex: SourceIndexTracker.currentIndex
        };

        if (additionalParams) {
            Util.forEachKey(additionalParams, function (param) {
                err[param] = additionalParams[param];
            });
        }

        testIterator.onError(err);
    }

    function ensureElementsExist(item, actionName, callback) {
        var success = false;

        var ensureExists = function () {
            var array = null;

            if (typeof item === 'function') {
                var res = item();

                array = Util.ensureArray(res);

                if (res && !(Util.isJQueryObj(res) && !res.length) && array.length) {
                    callback(array);
                    return true;
                }
            }
            else if (typeof item === 'string') {
                array = Util.parseActionArgument(item, actionName);

                if (array && array.length) {
                    callback(array);
                    return true;
                }
            }

            return false;
        };

        if (ensureExists())
            return;

        var interval = NativeMethods.setInterval.call(window, function () {
            if (ensureExists()) {
                success = true;
                window.clearInterval(interval);
            }
        }, ELEMENT_AVAILABILITY_WAITING_DELAY);

        NativeMethods.setTimeout.call(window, function () {
            if (!success) {
                window.clearInterval(interval);
                failWithError(SharedErrors.API_EMPTY_FIRST_ARGUMENT, {action: actionName});
            }
        }, exports.ELEMENT_AVAILABILITY_WAITING_TIMEOUT);
    }

    function ensureElementVisibility(element, actionName, callback) {
        var success = false;

        if (Util.isElementVisible(element) || element.tagName.toLowerCase() === 'option' ||
            element.tagName.toLowerCase() === 'optgroup') {
            callback();
            return;
        }

        var interval = NativeMethods.setInterval.call(window, function () {
            if (Util.isElementVisible(element)) {
                success = true;
                window.clearInterval(interval);
                callback();
            }
        }, ELEMENT_AVAILABILITY_WAITING_DELAY);

        NativeMethods.setTimeout.call(window, function () {
            if (!success) {
                window.clearInterval(interval);

                failWithError(SharedErrors.API_INVISIBLE_ACTION_ELEMENT, {
                    element: Util.getElementDescription(element),
                    action: actionName
                });
            }
        }, exports.ELEMENT_AVAILABILITY_WAITING_TIMEOUT);
    }

    function actionArgumentsIterator(actionName) {
        var runAction = null;

        var iterate = function (item, iterationCallback) {
            if ($.isArray(item))
                extractArgs(item, iterationCallback);
            else if (typeof item === 'function') {
                ensureElementsExist(item, actionName, function (elementsArray) {
                    extractArgs(elementsArray, iterationCallback);
                });
            }
            else if (typeof item === 'string') {
                ensureElementsExist(item, actionName, function (elementsArray) {
                    runAction(elementsArray, iterationCallback);
                });
            }
            else {
                var elementsArray = Util.parseActionArgument(item, actionName);
                if (!elementsArray || elementsArray.length < 1)
                    failWithError(SharedErrors.API_EMPTY_FIRST_ARGUMENT, {action: actionName});
                else
                    runAction(elementsArray, iterationCallback);
            }
        };

        var extractArgs = function (items, callback) {
            if (!items.length) {
                failWithError(SharedErrors.API_EMPTY_FIRST_ARGUMENT, {action: actionName});
            }
            else {
                async.forEachSeries(
                    items,
                    function (item, seriaCallback) {
                        iterate(item, seriaCallback);
                    },
                    function () {
                        callback();
                    }
                );
            }
        };

        return {
            run: function (items, actionRunner, callback) {
                onTargetWaitingStarted();
                runAction = actionRunner;
                extractArgs(items, callback);
            }
        };
    }

    function pressActionArgumentsIterator() {
        return {
            run: function (items, actionRunner, callback) {
                actionRunner(items, callback);
            }
        };
    }

    function onTargetWaitingStarted(isWaitAction) {
        testIterator.onActionTargetWaitingStarted({
            maxTimeout: exports.ELEMENT_AVAILABILITY_WAITING_TIMEOUT,
            isWaitAction: isWaitAction
        });
    }

    function onTargetWaitingFinished() {
        testIterator.onActionRun();
    }

    exports.init = function (iterator) {
        testIterator = iterator;
    };

    exports.click = function (what, options) {
        var actionStarted = false,
            elements = Util.ensureArray(what);

        testIterator.asyncActionSeries(
            elements,
            actionArgumentsIterator('click').run,
            function (element, callback, iframe) {
                ensureElementVisibility(element, 'click', function () {
                    function onerror(err) {
                        failWithError(err.code, {
                            element: err.element,
                            action: 'click'
                        });
                    }

                    if (!actionStarted) {
                        actionStarted = true;
                        onTargetWaitingFinished();
                    }

                    if (iframe)
                        iframe.contentWindow[Automation.AUTOMATION_RUNNERS].click.playback.run(element, options ||
                        {}, callback, onerror);
                    else
                        ClickPlaybackAutomation.run(element, options || {}, callback, onerror);
                });
            });
    };

    exports.rclick = function (what, options) {
        var actionStarted = false,
            elements = Util.ensureArray(what);

        testIterator.asyncActionSeries(
            elements,
            actionArgumentsIterator('rclick').run,
            function (element, callback, iframe) {
                ensureElementVisibility(element, 'rclick', function () {
                    if (!actionStarted) {
                        actionStarted = true;
                        onTargetWaitingFinished();
                    }

                    if (iframe)
                        iframe.contentWindow[Automation.AUTOMATION_RUNNERS].rclick.playback.run(element, options ||
                        {}, callback);
                    else
                        RClickPlaybackAutomation.run(element, options || {}, callback);
                });
            });
    };

    exports.dblclick = function (what, options) {
        var actionStarted = false,
            elements = Util.ensureArray(what);

        testIterator.asyncActionSeries(
            elements,
            actionArgumentsIterator('dblclick').run,
            function (element, callback, iframe) {
                ensureElementVisibility(element, 'dblclick', function () {
                    if (!actionStarted) {
                        actionStarted = true;
                        onTargetWaitingFinished();
                    }

                    if (iframe)
                        iframe.contentWindow[Automation.AUTOMATION_RUNNERS].dblclick.playback.run(element, options ||
                        {}, callback);
                    else
                        DblClickPlaybackAutomation.run(element, options || {}, callback);
                });
            });
    };

    exports.drag = function (what) {
        var actionStarted = false,
            args = arguments,
            elements = Util.ensureArray(what);

        var secondArgIsCoord = !(isNaN(parseInt(args[1])));

        var to = args.length > 2 && secondArgIsCoord ? {dragOffsetX: args[1], dragOffsetY: args[2]} : args[1];

        if (!to) {
            failWithError(SharedErrors.API_INCORRECT_DRAGGING_SECOND_ARGUMENT);
            return;
        }
        if (Util.isJQueryObj(to)) {
            if (to.length < 1) {
                failWithError(SharedErrors.API_INCORRECT_DRAGGING_SECOND_ARGUMENT);
                return;
            }
            else
                to = to[0];
        }
        else if (!Util.isDomElement(to) && (isNaN(parseInt(to.dragOffsetX)) || isNaN(parseInt(to.dragOffsetY)))) {
            failWithError(SharedErrors.API_INCORRECT_DRAGGING_SECOND_ARGUMENT);
            return;
        }

        var options = secondArgIsCoord ? args[3] : args[2];

        testIterator.asyncActionSeries(
            elements,
            actionArgumentsIterator('drag').run,
            function (element, callback, iframe) {
                ensureElementVisibility(element, 'drag', function () {
                    if (!actionStarted) {
                        actionStarted = true;
                        onTargetWaitingFinished();
                    }

                    if (iframe)
                        iframe.contentWindow[Automation.AUTOMATION_RUNNERS].drag.playback.run(element, to, options ||
                        {}, callback);
                    else
                        DragPlaybackAutomation.run(element, to, options || {}, callback);
                });
            });
    };

    exports.select = function () {
        var actionStarted = false,
            elements = Util.ensureArray(arguments[0]),
            args = $.makeArray(arguments).slice(1),
            secondArg = null,
            options = {},
            error = false,
            commonParent = null;

        if (!arguments[0])
            failWithError(SharedErrors.API_INCORRECT_SELECT_ACTION_ARGUMENTS);

        if (args.length === 1) {
            //NOTE: second action argument is jquery object
            if (Util.isJQueryObj(args[0])) {
                if (args[0].length < 1) {
                    failWithError(SharedErrors.API_INCORRECT_SELECT_ACTION_ARGUMENTS);
                    return;
                }
                else
                    secondArg = args[0][0];
            }
            else
                secondArg = args[0];
        }

        //NOTE: second action argument is dom element or node
        if (args.length === 1 && (Util.isDomElement(secondArg) || Util.isTextNode(secondArg))) {
            if (Util.isNotVisibleNode(secondArg))
                error = true;
            else {
                options.startNode = Util.isJQueryObj(elements[0]) ? elements[0][0] : elements[0];
                options.endNode = secondArg;

                if (!Util.isContentEditableElement(options.startNode) || !Util.isContentEditableElement(options.endNode))
                    error = true;
                else {
                    //NOTE: We should find element for perform select action
                    commonParent = ContentEditableHelper.getNearestCommonAncestor(options.startNode, options.endNode);
                    if (!commonParent)
                        error = true;
                    else if (Util.isTextNode(commonParent)) {
                        if (!commonParent.parentElement)
                            error = true;
                        else
                            elements = [commonParent.parentElement];
                    }
                    else
                        elements = [commonParent];
                }
            }
        }
        else {
            $.each(args, function (index, value) {
                if (isNaN(parseInt(value)) || (args.length > 1 && value < 0)) {
                    error = true;
                    return false;
                }
            });
        }

        if (error) {
            failWithError(SharedErrors.API_INCORRECT_SELECT_ACTION_ARGUMENTS);
            return;
        }

        testIterator.asyncActionSeries(
            elements,
            actionArgumentsIterator('select').run,
            function (element, callback, iframe) {
                if (args.length === 1 && !options.startNode)
                    options = {offset: args[0]};
                else if (args.length === 2 || (args.length > 2 && element.tagName.toLowerCase() !== 'textarea'))
                    options = {
                        startPos: args[0],
                        endPos: args[1]
                    };
                else if (args.length > 2)
                    options = {
                        startLine: args[0],
                        startPos: args[1],
                        endLine: args[2],
                        endPos: args[3]
                    };

                ensureElementVisibility(element, 'select', function () {
                    if (!actionStarted) {
                        actionStarted = true;
                        onTargetWaitingFinished();
                    }

                    if (iframe)
                        iframe.contentWindow[Automation.AUTOMATION_RUNNERS].select.playback.run(element, options, callback);
                    else
                        SelectPlaybackAutomation.run(element, options, callback);
                });
            });
    };

    exports.type = function (what, text, options) {
        if (!text) {
            failWithError(SharedErrors.API_EMPTY_TYPE_ACTION_ARGUMENT);
            return;
        }

        var actionStarted = false,
            elements = Util.ensureArray(what);

        testIterator.asyncActionSeries(
            elements,
            actionArgumentsIterator('type').run,
            function (element, callback, iframe) {
                ensureElementVisibility(element, 'type', function () {
                    if (!actionStarted) {
                        actionStarted = true;
                        onTargetWaitingFinished();
                    }

                    if (iframe)
                        iframe.contentWindow[Automation.AUTOMATION_RUNNERS].type.playback.run(element, text, options ||
                        {}, callback);
                    else
                        TypePlaybackAutomation.run(element, text, options || {}, callback);
                });
            });
    };

    exports.hover = function (what, options) {
        var actionStarted = false,
            elements = Util.ensureArray(what);

        testIterator.asyncActionSeries(
            elements,
            actionArgumentsIterator('hover').run,
            function (element, callback, iframe) {
                ensureElementVisibility(element, 'hover', function () {
                    if (!actionStarted) {
                        actionStarted = true;
                        onTargetWaitingFinished();
                    }

                    if (iframe)
                        iframe.contentWindow[Automation.AUTOMATION_RUNNERS].hover.playback.run(element, callback);
                    else
                        HoverPlaybackAutomation.run(element, options || {}, callback);
                });
            });
    };

    exports.press = function () {
        testIterator.asyncActionSeries(
            arguments,
            pressActionArgumentsIterator().run,
            function (keys, callback) {
                if (Util.parseKeysString(keys).error)
                    failWithError(SharedErrors.API_INCORRECT_PRESS_ACTION_ARGUMENT);

                else
                    PressPlaybackAutomation.run(keys, callback);
            });
    };

    //wait
    var conditionIntervalId = null;

    function startConditionCheck(condition, onConditionReached) {
        conditionIntervalId = NativeMethods.setInterval.call(window, function () {
            if (testIterator.callWithSharedDataContext(condition))
                onConditionReached();
        }, CHECK_CONDITION_INTERVAL);
    }

    function stopConditionCheck() {
        if (conditionIntervalId !== null) {
            window.clearInterval(conditionIntervalId);
            conditionIntervalId = null;
        }
    }

    exports.wait = function (ms, condition) {
        condition = typeof(condition) === 'function' ? condition : null;

        if (typeof ms !== 'number' || ms < 0) {
            failWithError(SharedErrors.API_INCORRECT_WAIT_ACTION_MILLISECONDS_ARGUMENT);
            return;
        }

        testIterator.asyncAction(function (iteratorCallback) {
            testIterator.expectInactivity(ms, function () {
                function onConditionReached() {
                    window.clearTimeout(timeout);
                    stopConditionCheck();
                    iteratorCallback();
                }

                var timeout = NativeMethods.setTimeout.call(window, onConditionReached, ms || 0);

                if (condition)
                    startConditionCheck(condition, onConditionReached);
            });
        });
    };

    exports.waitFor = function (event, timeout) {
        var waitForElements = Util.isStringOrStringArray(event, true),
            timeoutExceeded = false;

        if (typeof event !== 'function' && !waitForElements) {
            failWithError(SharedErrors.API_INCORRECT_WAIT_FOR_ACTION_EVENT_ARGUMENT);
            return;
        }

        if (typeof timeout === 'undefined')
            timeout = WAIT_FOR_DEFAULT_TIMEOUT;

        if (typeof timeout !== 'number' || timeout < 0) {
            failWithError(SharedErrors.API_INCORRECT_WAIT_FOR_ACTION_TIMEOUT_ARGUMENT);
            return;
        }

        onTargetWaitingStarted(true);

        testIterator.asyncAction(function (iteratorCallback) {
            var timeoutID = NativeMethods.setTimeout.call(window, function () {
                if (waitForElements)
                    stopConditionCheck();

                timeoutExceeded = true;
                failWithError(SharedErrors.API_WAIT_FOR_ACTION_TIMEOUT_EXCEEDED);
            }, timeout);

            function onConditionReached() {
                if (timeoutExceeded)
                    return;

                if (waitForElements)
                    stopConditionCheck();

                window.clearTimeout(timeoutID);
                onTargetWaitingFinished();
                iteratorCallback();
            }


            testIterator.expectInactivity(timeout, function () {
                var condition = null;

                if (waitForElements) {
                    if (typeof event === 'string') {
                        condition = function () {
                            return !!$(event).length;
                        };
                    }
                    else {
                        condition = function () {
                            var elementsExist = true;

                            for (var i = 0; i < event.length; i++) {
                                if (!$(event[i]).length) {
                                    elementsExist = false;
                                    break;
                                }
                            }

                            return elementsExist;
                        };
                    }

                    startConditionCheck(condition, onConditionReached);
                }
                else {
                    testIterator.callWithSharedDataContext(function () {
                        event.call(this, function () {
                            onConditionReached();
                        });
                    });
                }
            });
        });
    };

    exports.navigateTo = function (url) {
        var NAVIGATION_DELAY = 1000;

        testIterator.asyncAction(function (iteratorCallback) {
            window[JSProcessor.SET_PROPERTY_METH_NAME](window, 'location', url);

            //NOTE: give browser some time to navigate
            NativeMethods.setTimeout.call(window, iteratorCallback, NAVIGATION_DELAY);
        });
    };

    exports.upload = function (what, path) {
        var actionStarted = false,
            elements = Util.ensureArray(what);

        if (!Util.isStringOrStringArray(path) && path)
            failWithError(SharedErrors.API_UPLOAD_INVALID_FILE_PATH_ARGUMENT);

        testIterator.asyncActionSeries(
            elements,
            actionArgumentsIterator('upload').run,
            function (element, callback) {
                if (!Util.isFileInput(element))
                    failWithError(SharedErrors.API_UPLOAD_ELEMENT_IS_NOT_FILE_INPUT);

                else {
                    if (!actionStarted) {
                        actionStarted = true;
                        onTargetWaitingFinished();
                    }

                    Hammerhead.upload(element, path, function (errs) {
                        if (errs.length) {
                            var errPaths = errs.map(function (err) {
                                return err.filePath;
                            });

                            failWithError(SharedErrors.API_UPLOAD_CAN_NOT_FIND_FILE_TO_UPLOAD, {filePaths: errPaths});
                        }

                        else
                            callback();
                    });
                }
            }
        );
    };

    exports.screenshot = function () {
        testIterator.asyncAction(function (iteratorCallback) {
            testIterator.takeScreenshot(function () {
                iteratorCallback();
            }, false);
        });
    };

    //NOTE: add sourceIndex wrapper
    SourceIndexTracker.wrapTrackableMethods(exports, SharedConst.ACTION_FUNC_NAMES);
});
