TestCafeClient.define('TestRunner.API.Dialogs', function (require, exports) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        EventEmitter = Hammerhead.Util.EventEmitter;

    exports.UNEXPECTED_DIALOG_ERROR_EVENT = 'unexpectedDialogError';
    exports.WAS_NOT_EXPECTED_DIALOG_ERROR_EVENT = 'wasNotExpectedDialogError';
    exports.DIALOGS_INFO_CHANGED_EVENT = 'dialogsInfoChangedEvent';

    var dialogsInfo = null; //NOTE: call the onDialogsInfoChanged function when you change this
    var overriddenDialogs = {
        alert:   function () {
        },
        confirm: function () {
        },
        prompt:  function () {
        }
    };

    var beforeUnloadEventWasRaised = false;

    var eventEmitter = new EventEmitter();

    exports.on = function (event, handler) {
        eventEmitter.on(event, handler);
    };

    function initDialogsInfo (info) {
        if (!info) {
            dialogsInfo = {
                expectAlertCount:           0,
                expectConfirmCount:         0,
                expectPromptCount:          0,
                expectedConfirmRetValues:   [],
                expectedPromptRetValues:    [],
                expectBeforeUnload:         false,
                alerts:                     [],
                confirms:                   [],
                prompts:                    [],
                beforeUnloadDialogAppeared: false
            };
        }
        else
            dialogsInfo = info;
    }

    function onDialogsInfoChanged () {
        eventEmitter.emit(exports.DIALOGS_INFO_CHANGED_EVENT, {
            info: dialogsInfo
        });
    }

    function beforeUnloadHandler (e) {
        dialogsInfo.beforeUnloadDialogAppeared = !!e.prevented;
        beforeUnloadEventWasRaised = true;

        if (dialogsInfo.beforeUnloadDialogAppeared)
            onDialogsInfoChanged();

        if (dialogsInfo.beforeUnloadDialogAppeared && !dialogsInfo.expectBeforeUnload)
            sendUnexpectedDialogError('beforeUnload', e.returnValue === true ? '' : e.returnValue.toString());
    }

    function overrideDialogs () {
        overriddenDialogs = {
            alert: function (message) {
                dialogsInfo.alerts.push(message);
                onDialogsInfoChanged();

                if (dialogsInfo.alerts.length > dialogsInfo.expectAlertCount)
                    sendUnexpectedDialogError('alert', message);
            },

            confirm: function (message) {
                var returnValue = dialogsInfo.expectedConfirmRetValues[dialogsInfo.confirms.length];

                dialogsInfo.confirms.push(message);
                onDialogsInfoChanged();

                if (dialogsInfo.confirms.length > dialogsInfo.expectConfirmCount)
                    sendUnexpectedDialogError('confirm', message);

                return returnValue;
            },

            prompt: function (message) {
                var returnValue = dialogsInfo.expectedPromptRetValues[dialogsInfo.prompts.length];

                dialogsInfo.prompts.push(message);
                onDialogsInfoChanged();

                if (dialogsInfo.prompts.length > dialogsInfo.expectPromptCount)
                    sendUnexpectedDialogError('prompt', message);

                return returnValue;
            }
        };
    }

    function sendUnexpectedDialogError (dialog, message) {
        //NOTE: the following dialogs are not raised in browsers after before unload event
        if (/alert|confirm|prompt/.test(dialog) && beforeUnloadEventWasRaised)
            return;

        eventEmitter.emit(exports.UNEXPECTED_DIALOG_ERROR_EVENT, {
            dialog:  dialog,
            message: message
        });
    }

    function sendWasNotExpectedDialog (dialog) {
        eventEmitter.emit(exports.WAS_NOT_EXPECTED_DIALOG_ERROR_EVENT, {
            dialog: dialog
        });
    }

    exports.init = function (info) {
        Hammerhead.on(Hammerhead.BEFORE_UNLOAD_EVENT, beforeUnloadHandler);

        beforeUnloadEventWasRaised = false;

        initDialogsInfo(info);

        window.alert = function (message) {
            overriddenDialogs.alert(message);
        };

        window.confirm = function (message) {
            return overriddenDialogs.confirm(message);
        };

        window.prompt = function (message) {
            return overriddenDialogs.prompt(message);
        };

        overrideDialogs();

        exports.destroy = function () {
            Hammerhead.off(Hammerhead.BEFORE_UNLOAD_EVENT, beforeUnloadHandler);
        };

        exports.handleAlert = function () {
            dialogsInfo.expectAlertCount++;
            onDialogsInfoChanged();
        };

        exports.handleConfirm = function (value) {
            dialogsInfo.expectConfirmCount++;
            dialogsInfo.expectedConfirmRetValues.push(!(!value || value === 'Cancel'));
            onDialogsInfoChanged();
        };

        exports.handlePrompt = function (value) {
            dialogsInfo.expectPromptCount++;
            dialogsInfo.expectedPromptRetValues.push((value || value === '') ? value : null);
            onDialogsInfoChanged();
        };

        exports.handleBeforeUnload = function () {
            dialogsInfo.expectBeforeUnload = true;
            onDialogsInfoChanged();
        };

        exports.resetHandlers = function () {
            initDialogsInfo();
            overrideDialogs();
            onDialogsInfoChanged();
        };

        exports.hasUnexpectedBeforeUnloadDialog = function () {
            return dialogsInfo && dialogsInfo.beforeUnloadDialogAppeared && !dialogsInfo.expectBeforeUnload;
        };

        exports.checkExpectedDialogs = function () {
            if (dialogsInfo.expectAlertCount && dialogsInfo.alerts.length < dialogsInfo.expectAlertCount)
                sendWasNotExpectedDialog('alert');

            if (dialogsInfo.expectConfirmCount && dialogsInfo.confirms.length < dialogsInfo.expectConfirmCount)
                sendWasNotExpectedDialog('confirm');

            if (dialogsInfo.expectPromptCount && dialogsInfo.prompts.length < dialogsInfo.expectPromptCount)
                sendWasNotExpectedDialog('prompt');

            if (dialogsInfo.alerts.length > dialogsInfo.expectAlertCount)
                sendUnexpectedDialogError('alert', dialogsInfo.alerts[dialogsInfo.expectAlertCount]);

            if (dialogsInfo.confirms.length > dialogsInfo.expectConfirmCount)
                sendUnexpectedDialogError('confirm', dialogsInfo.confirms[dialogsInfo.expectConfirmCount]);

            if (dialogsInfo.prompts.length > dialogsInfo.expectPromptCount)
                sendUnexpectedDialogError('prompt', dialogsInfo.prompts[dialogsInfo.expectPromptCount]);

            if (dialogsInfo.expectBeforeUnload && !dialogsInfo.beforeUnloadDialogAppeared)
                sendWasNotExpectedDialog('beforeUnload');
        };
    };
});