TestCafeClient.define('TestRunner.APIv2.Select', function (require, exports) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        Util = Hammerhead.Util;

    //Const
    //------------------------------------------------------------------------------------------------
    exports.REPRODUCE_SELECTION_CHAIN_FUNC_KEY = '__#TC_REPRODUCE_SELECTION_CHAIN_FUNC_KEY';


    //Selector token types
    //------------------------------------------------------------------------------------------------
    var START_TOKEN = 0,
        FILTER_TOKEN = 1,
        DESCENDANTS_SEARCH_TOKEN = 2,
        JQUERY_OBJECT_TOKEN = 3,
        INDEX_TOKEN = 4;


    //Convert selecting tokens sequence into reusable selecting function
    //------------------------------------------------------------------------------------------------
    function createSelectingFunc($, tokens) {
        return function () {
            var $result = null;

            for (var i = 0; i < tokens.length; i++) {
                var type = tokens[i].type,
                    selector = tokens[i].selector,
                    fnSelector = typeof selector === 'function';

                if (type === START_TOKEN)
                    $result = fnSelector ? $('*').filter(selector) : $(selector);

                else if (type === FILTER_TOKEN)
                    $result = $result.filter(selector);

                else if (type === DESCENDANTS_SEARCH_TOKEN)
                    $result = fnSelector ? $result.find('*').filter(selector) : $result.find(selector);

                else if (type === JQUERY_OBJECT_TOKEN)
                    $result = selector;

                else if (type === INDEX_TOKEN)
                    $result = $result.eq(selector);
            }

            return $result;
        };
    }

    //Selectors
    //------------------------------------------------------------------------------------------------
    var Selectors = function (hostOperator) {
        this.hostOperator = hostOperator;
    };

    Selectors.prototype.id = function (id) {
        if (typeof id === 'string')
            return this.hostOperator.$('#' + id);

        if (id instanceof RegExp)
            return this.attr('id', id);

        return this.hostOperator.$('');
    };

    Selectors.prototype.tag = function (tagName) {
        return this.hostOperator.$(tagName);
    };

    Selectors.prototype.class = function (className) {
        return this.hostOperator.$('.' + className);
    };

    Selectors.prototype.text = function (text) {
        var isRegExp = text instanceof RegExp;

        if (typeof text !== 'string' && !isRegExp)
            return this.hostOperator.$('');

        //NOTE: this is different from ':containsExcludeChildren'. Now we can don't care about escaping the string
        //and we can pass regexp.
        return this.hostOperator.$(function () {
            var element = this,
                elementText = '',
                tn = element.tagName.toLowerCase();

            if (tn === 'script' || tn === 'style')
                return false;

            for (var i = 0; i < element.childNodes.length; i++) {
                if (element.childNodes[i].nodeType === 3 && element.childNodes[i].data)
                    elementText += element.childNodes[i].data;
            }

            if (isRegExp)
                return text.test(elementText);
            else
                return elementText.indexOf(text) > -1;
        });
    };

    Selectors.prototype.attr = function (name, value) {
        if (typeof name !== 'string')
            return this.hostOperator.$('');

        if (typeof value === 'string')
            return this.hostOperator.$('[' + name + '=\'' + value + '\']');

        if (value instanceof RegExp) {
            return this.hostOperator.$(function () {
                return this.hasAttribute(name) && value.test(this.getAttribute(name));
            });
        }

        return this.hostOperator.$('[' + name + ']');
    };

    Selectors.prototype.name = function (value) {
        if (typeof value !== 'string' && !(value instanceof RegExp))
            return this.hostOperator.$('');

        return this.attr('name', value);
    };


    //Base class for all operators
    //------------------------------------------------------------------------------------------------
    var OperatorBase = function (jQuery, chainTokens, $cachedSelection) {
        this.chainTokens = chainTokens || [];
        this.jQuery = jQuery;
        this.allowCacheForAndOperator = true;
        this.$cachedSelection = $cachedSelection;
    };

    //Pure virtual
    OperatorBase.prototype._addSelectorToTokenSequence = null;

    OperatorBase.prototype.$ = function (selector) {
        var $result = null;

        //NOTE: slice token sequence, so mutations in this operator will not affect parent operators
        //and current operator (previous chain parts can be reused).
        var tokens = this.chainTokens.slice(0);
        this._addSelectorToTokenSequence(selector, tokens);

        //NOTE: now let's create function which will be able to reproduce whole selection chain.
        //It will be used later by actions API to wait for element to appear.
        var reproduceSelectionChainFunc = createSelectingFunc(this.jQuery, tokens);

        //NOTE: now let's get result of the current operator. If we have cached previous result we will
        //create selecting function using it, so we will not reproduce whole chain and just apply latest
        //operator to the cached result.
        if (this.$cachedSelection) {
            tokens = [
                {type: JQUERY_OBJECT_TOKEN, selector: this.$cachedSelection}
            ];

            this._addSelectorToTokenSequence(selector, tokens);
            $result = createSelectingFunc(this.jQuery, tokens)();
        }

        else
            $result = reproduceSelectionChainFunc();

        //NOTE: since we have already selected something we should append 'in' and 'and' operators to result
        $result.in = new InOperator(this.jQuery, tokens, $result);
        $result.and = new AndOperator(this.jQuery, tokens, this.allowCacheForAndOperator && $result);
        $result[exports.REPRODUCE_SELECTION_CHAIN_FUNC_KEY] = reproduceSelectionChainFunc;

        return $result;
    };


    //Base class for indexable operators
    //------------------------------------------------------------------------------------------------
    var IndexableOperatorBase = function (jQuery, chainTokens, $cachedSelection) {
        OperatorBase.call(this, jQuery, chainTokens, $cachedSelection);
    };

    Util.inherit(IndexableOperatorBase, OperatorBase);

    //TODO define it on 'in' and 'select' operators prototype
  /*  Object.defineProperties(IndexableOperatorBase.prototype, {
        //NOTE: just a shortcut for .nth(0)
        first: {
            enumerable: true,

            get: function () {
                return this.nth(0);
            }
        }
    });*/

    IndexableOperatorBase.prototype.nth = function (index) {
        var tokens = this.chainTokens.slice(0);

        //NOTE: indexing token should be first, since we haven't applied general selector, whitch will add START_TOKEN:
        //select.first -> [INDEX_TOKEN]
        //select.first.by.id() -> [START_TOKEN, INDEX_TOKEN] - index token pushed forward by start
        //Same goes for 'in' operator.
        tokens.splice(0, 0, {
            type: INDEX_TOKEN,
            selector: index
        });

        //NOTE: recreate yourself with applied indexer, so chain before indexer can be reused.
        var indexedOp = new this.constructor(this.jQuery, tokens, null);

        //NOTE: Also, disable cache for 'and' operator:

        //var R1 = select.nth(3).by.class('yo') -> $('.yo').eq(3);
        //var R2 = R1.and.class('beep') -> $('.yo').eq(3).filter('.beep') - WRONG!

        //Correct expression should be:

        //var R2 = $('.yo').filter('.beep').eq(3)

        //So, we should rebuild chain to achieve this.
        indexedOp.allowCacheForAndOperator = false;

        return indexedOp;
    };


    //Select
    //------------------------------------------------------------------------------------------------
    var SelectOperator = exports.SelectOperator = function (jQuery) {
        IndexableOperatorBase.call(this, jQuery);

        //NOTE: normally you should use .by, but we will make things more relaxing, to make user care less about words
        this.by = this.with = new Selectors(this);
    };

    Util.inherit(SelectOperator, IndexableOperatorBase);

    SelectOperator.prototype._addSelectorToTokenSequence = function (selector, tokens) {
        tokens.splice(0, 0, {
            type: START_TOKEN,
            selector: selector
        });
    };


    //In
    //------------------------------------------------------------------------------------------------
    var InOperator = function (jQuery, chainTokens, $cachedSelection) {
        IndexableOperatorBase.call(this, jQuery, chainTokens, $cachedSelection);

        //NOTE: 'and' operator can't use cache after 'in' operator:

        //var R1 = select.by.class('yo') -> $('.yo')
        //var R2 = R1.in.with.class('beep') -> $('.beep').find($('.yo'))
        //var R3 = R2.and.with.tag('div') -> $('div').filter($('.beep').find($('.yo'))) - WRONG!

        //The expression above expects parent selection [$('div')] to contain result of previous
        //parent selection [$('.beep').find($('.yo'))] at the same level of hierarchy, which obviously
        //will not work. The correct expression should be:

        //var R3 = $('div').filter($('.beep')).find($('.yo'))

        //But we can't achieve it without rebuilding whole chain, so caching should be disabled.
        this.allowCacheForAndOperator = false;

        //NOTE: normally you should use .by, but we will make things more relaxing, to make user care less about words
        this.with = this.by = new Selectors(this);
    };

    Util.inherit(InOperator, IndexableOperatorBase);

    InOperator.prototype._addSelectorToTokenSequence = function (selector, tokens) {
        //NOTE: 'in' operator means that we filtering selection parent. So we need to reorder tokens
        //since we need to select parent first in jQuery. So first non-indexing token becomes DESCENDANTS_SEARCH_TOKEN
        //while 'in'-operator selector becomes new 'start'.
        for (var i = 0; i < tokens.length; i++) {
            if (tokens[i].type !== INDEX_TOKEN) {
                tokens[i].type = DESCENDANTS_SEARCH_TOKEN;
                break;
            }
        }

        tokens.splice(0, 0, {
            type: START_TOKEN,
            selector: selector
        });
    };


    //Add
    //------------------------------------------------------------------------------------------------
    var AndOperator = function (jQuery, chainTokens, $cachedSelection) {
        //NOTE: 'and' operator doesn't have intermediate pronoun (like 'by' or 'with'), so it's a director
        //descendant of Selectors. However we will add .by and .with, to make user care less about words
        Selectors.call(this, this);
        OperatorBase.call(this, jQuery, chainTokens, $cachedSelection);

        this.with = this.by = this;
    };

    Util.inherit(AndOperator, OperatorBase);
    Util.inherit(AndOperator, Selectors);

    AndOperator.prototype._addSelectorToTokenSequence = function (selector, tokens) {
        //NOTE: 'and' operator should be applied to the latest 'select'-operator or 'in'-operator
        //result. Since both results are always in the current 'start' we should insert token
        //as the second in the sequence.
        tokens.splice(1, 0, {
            type: FILTER_TOKEN,
            selector: selector
        });
    };

});