TestCafeClient.define('TestRunner.TestRunner', function (require) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        Util = Hammerhead.Util,
        NativeMethods = Hammerhead.NativeMethods,
        Transport = require('Base.Transport'),
        ServiceCommands = require('Shared.ServiceCommands'),
        SharedErrors = require('Shared.Errors'),
        Settings = require('Settings'),

        TestRunnerBase = require('TestRunner.TestRunnerBase');

    var WAITING_FOR_SERVICE_MESSAGES_COMPLETED_DELAY = 1000;

    var TestRunner = this.exports = function (startedCallback) {
        var testRunner = this;

        TestRunnerBase.apply(this, [startedCallback]);

        Transport.startInactivityMonitor(function () {
            testRunner._onError({
                code: SharedErrors.TEST_INACTIVITY
            });
        });
    };

    Util.inherit(TestRunner, TestRunnerBase);

    TestRunner.prototype._onTestComplete = function (e) {
        Transport.waitForServiceMessagesCompleted(function () {
            var testCompleteMsg = {
                cmd: ServiceCommands.TEST_COMPLETE
            };

            Transport.asyncServiceMsg(testCompleteMsg, function () {
                e.callback();
                Transport.switchToWorkerIdle();
            });
        }, WAITING_FOR_SERVICE_MESSAGES_COMPLETED_DELAY);
    };

    TestRunner.prototype._onNextStepStarted = function (e) {
        var nextStepMsg = {
            cmd: ServiceCommands.SET_NEXT_STEP,
            nextStep: e.nextStep
        };

        Transport.asyncServiceMsg(nextStepMsg, e.callback);
    };

    //NOTE: decrease step counter while an action is waiting for element available and decrease it when action running started (T230851)
    TestRunner.prototype._onActionTargetWaitingStarted = function (e) {
        TestRunnerBase.prototype._onActionTargetWaitingStarted.apply(this, [e]);
    };

    TestRunner.prototype._onActionRun = function () {
        TestRunnerBase.prototype._onActionRun.apply(this, []);
    };

    TestRunner.prototype._onError = function (err) {
        var testRunner = this;

        if (this.stopped)
            return;

        //NOTE: we should stop testIterator to prevent playback after an error is occurred
        this.testIterator.stop();

        TestRunnerBase.prototype._onError.call(this, err);

        if (!Settings.TAKE_SCREENSHOT_ON_FAILS) {
            this.stopped = true;
            Transport.fail(err);
            return;
        }

        var setErrorMsg = {
            cmd: ServiceCommands.SET_TEST_ERROR,
            err: err
        };

        Transport.asyncServiceMsg(setErrorMsg);

        this._onTakeScreenshot({
            isFailedStep: true,
            withoutStepName: !(SharedErrors.hasErrorStepName(err) && Hammerhead.Errs.hasErrorStepName(err)),
            callback: function () {
                testRunner.stopped = true;
                Transport.fail(err);
            }
        });
    };

    TestRunner.prototype._onAssertionFailed = function (e, inIFrame) {
        this.testIterator.state.needScreeshot = !inIFrame;
        Transport.assertionFailed(e.err);
    };

    TestRunner.prototype._onSetStepsSharedData = function (e) {
        var msg = {
            cmd: ServiceCommands.SET_STEPS_SHARED_DATA,
            stepsSharedData: e.stepsSharedData
        };

        Transport.asyncServiceMsg(msg, function () {
            e.callback();
        });
    };

    TestRunner.prototype._onGetStepsSharedData = function (e) {
        var msg = {cmd: ServiceCommands.GET_STEPS_SHARED_DATA};

        Transport.asyncServiceMsg(msg, e.callback);
    };

    TestRunner.prototype._onExpectInactivity = function (e) {
        Transport.expectInactivity(e.duration, e.callback);
    };

    TestRunner.prototype._onTakeScreenshot = function (e) {
        var savedTitle = document.title,
            windowMark = '[tc-' + Date.now() + ']',
            browserName = null,
            callback = e && e.callback ? e.callback : function () {
            },
            testRunner = this;

        testRunner.eventEmitter.emit(TestRunnerBase.SCREENSHOT_CREATING_STARTED_EVENT, {});

        if (Util.isMSEdge)
            browserName = 'MSEDGE';
        else if (Util.isSafari)
            browserName = 'SAFARI';
        else if (Util.isOpera || Util.isOperaWithWebKit)
            browserName = 'OPERA';
        else if (Util.isWebKit)
            browserName = 'CHROME';
        else if (Util.isMozilla)
            browserName = 'FIREFOX';
        else if (Util.isIE)
            browserName = 'IE';

        var msg = {
            cmd: ServiceCommands.TAKE_SCREENSHOT,
            windowMark: windowMark,
            browserName: browserName,
            isFailedStep: e.isFailedStep,
            withoutStepName: e.withoutStepName,
            url: window.location.toString()
        };

        var assignedTitle = savedTitle + windowMark,
            checkTitleIntervalId = NativeMethods.setInterval.call(window, function () {
                if (document.title !== assignedTitle) {
                    savedTitle = document.title;
                    document.title = assignedTitle;
                }
            }, 50);

        document.title = assignedTitle;

        //NOTE: we should set timeouts to changing of document title
        //in any case we are waiting response from server
        NativeMethods.setTimeout.call(window, function () {
            Transport.asyncServiceMsg(msg, function () {
                window.clearInterval(checkTitleIntervalId);
                checkTitleIntervalId = null;
                document.title = savedTitle;
                testRunner.eventEmitter.emit(TestRunnerBase.SCREENSHOT_CREATING_FINISHED_EVENT, {});

                NativeMethods.setTimeout.call(window, function () {
                    callback();
                }, 100);
            });
        }, 500);
    };

    TestRunner.prototype._onDialogsInfoChanged = function (info) {
        Transport.asyncServiceMsg({
            cmd: ServiceCommands.NATIVE_DIALOGS_INFO_SET,
            info: info,
            timeStamp: Date.now()
        });
    };
});
