TestCafeClient.define('TestRunner.TestRunnerBase', function (require) {
    var Hammerhead               = HammerheadClient.get('Hammerhead'),
        $                        = Hammerhead.$,
        Util                     = Hammerhead.Util,
        MessageSandbox           = Hammerhead.MessageSandbox,
        NativeMethods            = Hammerhead.NativeMethods,
        jQuerySelectorExtensions = require('Base.jQueryExtensions.Selectors'),
        Settings                 = require('Settings'),
        Transport                = require('Base.Transport'),
        CrossDomainMessages      = require('Base.CrossDomainMessages'),
        SharedErrors             = require('Shared.Errors'),
        ServiceCommands          = require('Shared.ServiceCommands'),

        TestIterator             = require('TestRunner.TestIterator'),
        ActionsAPI               = require('TestRunner.API.Actions'),
        DialogsAPI               = require('TestRunner.API.Dialogs'),
        AssertionsAPI            = require('TestRunner.API.Assertions'),

        Automation               = require('Automation'),
        ActionBarrier            = require('ActionBarrier'),
        ModalBackground          = require('UI.ModalBackground'),
        CursorWidget             = require('UI.Cursor'),
        AutomationIFrameBehavior = require('Automation.IFrameBehavior');

    //Const
    var PAGE_LOAD_TIMEOUT                  = 3000,
        ANIMATIONS_WAIT_DELAY              = 200,
        CHECK_FILE_DOWNLOADING_DELAY       = 500,
        IFRAME_EXISTENCE_WATCHING_INTERVAL = 1000;

    var STORAGE_ELEMENT_WAITING_KEY = 'runner|element-waiting|' + Settings.JOB_UID;

    //Util
    function waitPageLoad (callback) {
        var internal$       = Hammerhead.$,
            loaded          = false,
            callbackWrapper = function () {
                if (!loaded) {
                    loaded = true;
                    callback();
                }
            };

        internal$(window).load(callbackWrapper);
        internal$(document).ready(function () {
            //NOTE: an iFrame may be removed in this moment
            if (Util.isIFrameWindowInDOM(window))
                NativeMethods.setTimeout.call(window, callbackWrapper, PAGE_LOAD_TIMEOUT);
        });
    }

    //Init
    var TestRunnerBase = this.exports = function () {
        var testRunner = this;

        this.eventEmitter = new Util.EventEmitter();
        this.testIterator = new TestIterator(pingIFrame);

        this.executingStepInIFrameWindow = null;
        this.stopped                     = false;
        this.listenNativeDialogs         = false;
        this.isFileDownloadingIntervalID = null;

        this.actionTargetWaitingStarted = window.top === window.self &&
                                          !!window.sessionStorage.getItem(STORAGE_ELEMENT_WAITING_KEY, true);
        window.sessionStorage.removeItem(STORAGE_ELEMENT_WAITING_KEY);

        this.assertionsAPI = new AssertionsAPI(function (err) {
            testRunner.testIterator.onAssertionFailed(err);
        });

        ActionsAPI.init(this.testIterator);

        this._initNativeDialogs();

        Automation.init();
        this._initBarrier();

        this._initApi();
        this._initIFrameBehavior();

        Hammerhead.on(Hammerhead.UNCAUGHT_JS_ERROR, function (err) {
            //NOTE: in this case we should to stop test iterator in iFrame
            if (err.inIFrame && !Settings.PLAYBACK)
                testRunner.testIterator.stop();
            else if (Settings.FAIL_ON_JS_ERRORS || Settings.RECORDING) {
                testRunner._onError({
                    code:      SharedErrors.UNCAUGHT_JS_ERROR,
                    scriptErr: err.msg,
                    pageError: true,
                    pageUrl:   err.pageUrl
                });
            }
        });

        testRunner.testIterator.on(TestIterator.ERROR_EVENT, function (e) {
            testRunner._onError(e);
        });

        testRunner.act._onJSError = function (err) {
            testRunner._onError({
                code:      SharedErrors.UNCAUGHT_JS_ERROR,
                scriptErr: (err && err.message) || err
            });
        };

        testRunner.act._start = function (stepNames, testSteps, nextStep, skipPageWaiting) {
            if (nextStep && testRunner.actionTargetWaitingStarted) {
                nextStep                              = nextStep - 1;
                Settings.CURRENT_TEST_STEP_NAME       = stepNames[nextStep];
                testRunner.actionTargetWaitingStarted = false;
            }

            //NOTE: start test execution only when all content is loaded or if loading
            //timeout is reached (whichever comes first).
            testRunner._prepareStepsExecuting(function () {
                delete testRunner.act._onJSError;
                delete testRunner.act._start;

                testRunner.eventEmitter.emit(testRunner.TEST_STARTED_EVENT, {
                    nextStep: nextStep
                });

                ModalBackground.hide();

                testRunner.testIterator.on(TestIterator.TEST_COMPLETE_EVENT, function (e) {
                    testRunner._onTestComplete(e);
                });

                testRunner.testIterator.on(TestIterator.NEXT_STEP_STARTED_EVENT, function (e) {
                    testRunner._onNextStepStarted(e);
                    testRunner._clearFileDownloadingInterval();
                });

                testRunner.testIterator.on(TestIterator.ACTION_TARGET_WAITING_STARTED_EVENT, function (e) {
                    testRunner._onActionTargetWaitingStarted(e);
                });

                testRunner.testIterator.on(TestIterator.ACTION_RUN_EVENT, function () {
                    testRunner._onActionRun();
                });

                testRunner.testIterator.on(TestIterator.ASSERTION_FAILED_EVENT, function (e) {
                    testRunner._onAssertionFailed(e);
                });

                testRunner.testIterator.on(TestIterator.SET_STEPS_SHARED_DATA_EVENT, function (e) {
                    testRunner._onSetStepsSharedData(e);
                });

                testRunner.testIterator.on(TestIterator.GET_STEPS_SHARED_DATA_EVENT, function (e) {
                    testRunner._onGetStepsSharedData(e);
                });
                testRunner.testIterator.on(TestIterator.EXPECT_INACTIVITY_EVENT, function (e) {
                    testRunner._onExpectInactivity(e);
                });

                testRunner.testIterator.on(TestIterator.TAKE_SCREENSHOT_EVENT, function (e) {
                    testRunner._onTakeScreenshot(e);
                });

                testRunner.testIterator.on(TestIterator.BEFORE_UNLOAD_EVENT_RAISED, function () {
                    testRunner._onBeforeUnload();
                });

                testRunner.testIterator.on(TestIterator.UNLOAD_EVENT_RAISED, function () {
                    testRunner._clearFileDownloadingInterval();
                });

                testRunner.listenNativeDialogs = true;

                testRunner.testIterator.start(stepNames, testSteps, DialogsAPI.resetHandlers,
                    DialogsAPI.checkExpectedDialogs, nextStep);
            }, skipPageWaiting);
        };
    };

    TestRunnerBase.prototype.run = function (stepNames, testSteps, nextStep) {
        this.testIterator.runSteps(stepNames, testSteps, DialogsAPI.resetHandlers,
            DialogsAPI.checkExpectedDialogs, nextStep);
    };

    TestRunnerBase.prototype._destroy = function () {
        DialogsAPI.destroy();

        this._destroyIFrameBehavior();

        Transport.stopInactivityMonitor();
    };

    TestRunnerBase.prototype._initBarrier = function () {
        ActionBarrier.init();
    };

    TestRunnerBase.prototype._initIFrameBehavior = function () {
        var testRunner = this;

        AutomationIFrameBehavior.init();

        function onMessage (e) {
            var message = e.message,
                msg     = null;

            switch (message.cmd) {
                case TestRunnerBase.IFRAME_STEP_COMPLETED_CMD:
                    testRunner._onIFrameStepExecuted();
                    testRunner._clearIFrameExistenceWatcherInterval();
                    break;

                case TestRunnerBase.IFRAME_ERROR_CMD:
                    if (message.err.stepNum === -1) {
                        message.err.stepNum  = testRunner.testIterator.getCurrentStepNum();
                        message.err.stepName = testRunner.testIterator.getCurrentStep();
                    }

                    testRunner._onError(message.err);
                    testRunner._clearIFrameExistenceWatcherInterval();
                    break;

                case TestRunnerBase.IFRAME_FAILED_ASSERTION_CMD:
                    testRunner.executingStepInIFrameWindow = null;

                    testRunner._onAssertionFailed(message.err, true);
                    break;

                case TestRunnerBase.IFRAME_GET_SHARED_DATA_REQUEST_CMD:
                    msg = {
                        cmd:        TestRunnerBase.IFRAME_GET_SHARED_DATA_RESPONSE_CMD,
                        sharedData: testRunner.testIterator.getSharedData()
                    };

                    MessageSandbox.sendServiceMsg(msg, e.source);
                    break;

                case TestRunnerBase.IFRAME_SET_SHARED_DATA_CMD:
                    testRunner.testIterator.setSharedData(message.sharedData);
                    break;

                case TestRunnerBase.IFRAME_EXPECT_INACTIVITY_CMD:
                    testRunner._onExpectInactivity({
                        duration: message.duration,
                        callback: function () {
                        }
                    });
                    break;

                case TestRunnerBase.IFRAME_NEXT_STEP_STARTED_CMD:
                    testRunner.executingStepInIFrameWindow = e.source;
                    testRunner._clearFileDownloadingInterval();

                    break;

                case TestRunnerBase.IFRAME_ACTION_TARGET_WAITING_STARTED_CMD:
                    testRunner.actionTargetWaitingStarted = true;
                    testRunner._onActionTargetWaitingStarted(e.message.params);
                    break;

                case TestRunnerBase.IFRAME_ACTION_RUN_CMD:
                    testRunner.actionTargetWaitingStarted = false;
                    testRunner._onActionRun();
                    break;

                case CrossDomainMessages.IFRAME_TEST_RUNNER_WAITING_STEP_COMPLETION_REQUEST_CMD:
                    if (testRunner.testIterator.waitedIFrame === Util.getIFrameByWindow(e.source) ||
                        testRunner.executingStepInIFrameWindow === e.source) {
                        msg = {
                            cmd: CrossDomainMessages.IFRAME_TEST_RUNNER_WAITING_STEP_COMPLETION_RESPONSE_CMD
                        };

                        MessageSandbox.sendServiceMsg(msg, e.source);
                    }
                    break;

                case CrossDomainMessages.IFRAME_TEST_RUNNER_STEP_COMPLETED_CMD:
                    if (testRunner.testIterator.waitedIFrame === Util.getIFrameByWindow(e.source))
                        testRunner.testIterator.iFrameActionCallback();
                    else if (testRunner.executingStepInIFrameWindow === e.source)
                        testRunner._onIFrameStepExecuted();
                    break;

                case TestRunnerBase.IFRAME_TAKE_SCREENSHOT_REQUEST_CMD:
                    testRunner._onTakeScreenshot({
                        isFailedStep: message.isFailedStep,
                        callback:     function () {
                            msg = {
                                cmd: TestRunnerBase.IFRAME_TAKE_SCREENSHOT_RESPONSE_CMD
                            };

                            MessageSandbox.sendServiceMsg(msg, e.source);
                        }
                    });
                    break;

                case TestRunnerBase.IFRAME_NATIVE_DIALOGS_INFO_CHANGED_CMD:
                    testRunner._onDialogsInfoChanged(message.info);
                    break;

                case TestRunnerBase.IFRAME_BEFORE_UNLOAD_REQUEST_CMD:
                    testRunner._onActionRun();
                    testRunner._onBeforeUnload(true, function (res) {
                        msg = {
                            cmd: TestRunnerBase.IFRAME_BEFORE_UNLOAD_RESPONSE_CMD,
                            res: res
                        };
                        MessageSandbox.sendServiceMsg(msg, e.source);
                    });
                    break;
            }
        }

        MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, onMessage);

        //NOTE: for test purposes
        testRunner._destroyIFrameBehavior = function () {
            AutomationIFrameBehavior.destroy();
            MessageSandbox.off(MessageSandbox.SERVICE_MSG_RECEIVED, onMessage);
        };
    };

    TestRunnerBase.prototype._prepareStepsExecuting = function (callback, skipPageWaiting) {
        function runCallback () {
            CursorWidget.init();
            callback();
        }

        if (skipPageWaiting)
            runCallback();
        else {
            waitPageLoad(function () {
                NativeMethods.setTimeout.call(window, function () {
                    Transport.batchUpdate(function () {
                        jQuerySelectorExtensions.init();

                        ActionBarrier.waitPageInitialization(function () {
                            CursorWidget.init();

                            callback();
                        });
                    });
                }, ANIMATIONS_WAIT_DELAY);
            });
        }
    };

    TestRunnerBase.WAITING_FOR_ACTION_TARGET_MESSAGE = 'Waiting for the target element of the next action to appear';

    TestRunnerBase.prototype.TEST_STARTED_EVENT                  = 'testStarted';
    TestRunnerBase.prototype.TEST_COMPLETED_EVENT                = 'testCompleted';
    TestRunnerBase.prototype.NEXT_STEP_STARTED_EVENT             = 'nextStepStarted';
    TestRunnerBase.prototype.ACTION_TARGET_WAITING_STARTED_EVENT = 'actionTargetWaitingStarted';
    TestRunnerBase.prototype.ACTION_RUN_EVENT                    = 'actionRun';
    TestRunnerBase.prototype.TEST_FAILED_EVENT                   = 'testFailed';

    TestRunnerBase.SCREENSHOT_CREATING_STARTED_EVENT  = 'screenshotCreatingStarted';
    TestRunnerBase.SCREENSHOT_CREATING_FINISHED_EVENT = 'screenshotCreatingFinished';

    TestRunnerBase.IFRAME_STEP_COMPLETED_CMD                = 'iframeStepCompleted';
    TestRunnerBase.IFRAME_ERROR_CMD                         = 'iframeError';
    TestRunnerBase.IFRAME_FAILED_ASSERTION_CMD              = 'iframeFailedAssertion';
    TestRunnerBase.IFRAME_GET_SHARED_DATA_REQUEST_CMD       = 'getSharedDataRequest';
    TestRunnerBase.IFRAME_GET_SHARED_DATA_RESPONSE_CMD      = 'getSharedDataResponse';
    TestRunnerBase.IFRAME_SET_SHARED_DATA_CMD               = 'setSharedData';
    TestRunnerBase.IFRAME_EXPECT_INACTIVITY_CMD             = 'expectInactivity';
    TestRunnerBase.IFRAME_NEXT_STEP_STARTED_CMD             = 'nextStepStarted';
    TestRunnerBase.IFRAME_ACTION_TARGET_WAITING_STARTED_CMD = 'actionTargetWaitingStarted';
    TestRunnerBase.IFRAME_ACTION_RUN_CMD                    = 'actionRun';
    TestRunnerBase.IFRAME_TAKE_SCREENSHOT_REQUEST_CMD       = 'takeScreenshotRequest';
    TestRunnerBase.IFRAME_TAKE_SCREENSHOT_RESPONSE_CMD      = 'takeScreenshotResponse';
    TestRunnerBase.IFRAME_NATIVE_DIALOGS_INFO_CHANGED_CMD   = 'nativeDialogsInfoChanged';
    TestRunnerBase.IFRAME_BEFORE_UNLOAD_REQUEST_CMD         = 'iframeBeforeUnloadRequest';
    TestRunnerBase.IFRAME_BEFORE_UNLOAD_RESPONSE_CMD        = 'iframeBeforeUnloadResponse';

    TestRunnerBase.prototype.on = function (event, handler) {
        this.eventEmitter.on(event, handler);
    };

    function pingIFrame (iframe, callback) {
        MessageSandbox.pingIFrame(iframe, CrossDomainMessages.IFRAME_TEST_RUNNER_PING_DISPATCHER_CMD, callback);
    }

    TestRunnerBase.prototype._clearIFrameExistenceWatcherInterval = function () {
        if (this.iframeExistenceWatcherInterval !== -1) {
            window.clearInterval(this.iframeExistenceWatcherInterval);
            this.iframeExistenceWatcherInterval = -1;
        }
    };

    TestRunnerBase.prototype._runInIFrame = function (iframe, stepName, step, stepNum) {
        var runner = this;

        this.testIterator.state.inAsyncAction = true;

        var msg = {
            cmd:      CrossDomainMessages.IFRAME_TEST_RUNNER_RUN_CMD,
            stepName: stepName,
            step:     step.toString(),
            stepNum:  stepNum
        };

        this._clearIFrameExistenceWatcherInterval();

        function iframeExistenceWatcher () {
            if (!Util.isElementInDocument(iframe)) {
                runner._onIFrameStepExecuted();
                runner._clearIFrameExistenceWatcherInterval();
            }
        }

        pingIFrame(iframe, function (err) {
            if (err) {
                runner._onError({
                    code:     SharedErrors.IN_IFRAME_TARGET_LOADING_TIMEOUT,
                    stepName: Settings.CURRENT_TEST_STEP_NAME
                });
            }
            else {
                runner.iframeExistenceWatcherInterval = NativeMethods.setInterval.call(window, iframeExistenceWatcher, IFRAME_EXISTENCE_WATCHING_INTERVAL);
                MessageSandbox.sendServiceMsg(msg, iframe.contentWindow);
            }
        });
    };

    TestRunnerBase.prototype._ensureIFrame = function (arg) {
        if (!arg) {
            this._onError({
                code:     SharedErrors.API_EMPTY_IFRAME_ARGUMENT,
                stepName: Settings.CURRENT_TEST_STEP_NAME
            });
            return null;
        }

        if (Util.isDomElement(arg)) {
            if (arg.tagName && arg.tagName.toLowerCase() === 'iframe')
                return arg;
            else {
                this._onError({
                    code:     SharedErrors.API_IFRAME_ARGUMENT_IS_NOT_IFRAME,
                    stepName: Settings.CURRENT_TEST_STEP_NAME
                });
                return null;
            }
        }

        if (typeof arg === 'string')
            arg = $(arg);

        if (Util.isJQueryObj(arg)) {
            if (arg.length === 0) {
                this._onError({
                    code:     SharedErrors.API_EMPTY_IFRAME_ARGUMENT,
                    stepName: Settings.CURRENT_TEST_STEP_NAME
                });
                return null;
            }
            else if (arg.length > 1) {
                this._onError({
                    code:     SharedErrors.API_MULTIPLE_IFRAME_ARGUMENT,
                    stepName: Settings.CURRENT_TEST_STEP_NAME
                });
                return null;
            }
            else
                return this._ensureIFrame(arg[0]);
        }

        if (typeof arg === 'function')
            return this._ensureIFrame(arg());

        this._onError({
            code:     SharedErrors.API_INCORRECT_IFRAME_ARGUMENT,
            stepName: Settings.CURRENT_TEST_STEP_NAME
        });
        return null;
    };

    //API
    TestRunnerBase.prototype._initApi = function () {
        var testRunner = this;

        this.act = ActionsAPI;

        this.ok                 = function () {
            testRunner.assertionsAPI.ok.apply(testRunner.assertionsAPI, arguments);
        };
        this.notOk              = function () {
            testRunner.assertionsAPI.notOk.apply(testRunner.assertionsAPI, arguments);
        };
        this.eq                 = function () {
            testRunner.assertionsAPI.eq.apply(testRunner.assertionsAPI, arguments);
        };
        this.notEq              = function () {
            testRunner.assertionsAPI.notEq.apply(testRunner.assertionsAPI, arguments);
        };
        this.handleAlert        = DialogsAPI.handleAlert;
        this.handleConfirm      = DialogsAPI.handleConfirm;
        this.handlePrompt       = DialogsAPI.handlePrompt;
        this.handleBeforeUnload = DialogsAPI.handleBeforeUnload;
        this.inIFrame           = function (iFrameGetter, step) {
            return function () {
                var stepNum = testRunner.testIterator.state.step,
                    iFrame  = testRunner._ensureIFrame(iFrameGetter());

                if (iFrame)
                    testRunner._runInIFrame(iFrame, Settings.CURRENT_TEST_STEP_NAME, step, stepNum);
            };
        };
    };

    TestRunnerBase.prototype._initNativeDialogs = function () {
        //NOTE: this method should be synchronous because we should have this info before page scripts are executed
        var testRunner = this;

        if (Settings.NATIVE_DIALOGS_INFO)
            testRunner.listenNativeDialogs = true;

        DialogsAPI.init(Settings.NATIVE_DIALOGS_INFO);

        DialogsAPI.on(DialogsAPI.UNEXPECTED_DIALOG_ERROR_EVENT, function (e) {
            if (testRunner.listenNativeDialogs) {
                testRunner.testIterator.onError({
                    code:     SharedErrors.API_UNEXPECTED_DIALOG,
                    stepName: testRunner.testIterator.getCurrentStep(),
                    dialog:   e.dialog,
                    message:  e.message
                });
            }
        });

        DialogsAPI.on(DialogsAPI.WAS_NOT_EXPECTED_DIALOG_ERROR_EVENT, function (e) {
            if (testRunner.listenNativeDialogs) {
                testRunner.testIterator.onError({
                    code:     SharedErrors.API_EXPECTED_DIALOG_DOESNT_APPEAR,
                    stepName: testRunner.testIterator.getCurrentStep(),
                    dialog:   e.dialog
                });
            }
        });

        DialogsAPI.on(DialogsAPI.DIALOGS_INFO_CHANGED_EVENT, function (e) {
            testRunner._onDialogsInfoChanged(e.info);
        });
    };
    //Handlers
    TestRunnerBase.prototype._onTestComplete = function (e) {
        this.eventEmitter.emit(this.TEST_COMPLETED_EVENT, {});
        e.callback();
    };

    TestRunnerBase.prototype._onNextStepStarted = function (e) {
        e.callback();
    };

    TestRunnerBase.prototype._onActionTargetWaitingStarted = function (e) {
        window.sessionStorage.setItem(STORAGE_ELEMENT_WAITING_KEY, true);
        this.eventEmitter.emit(this.ACTION_TARGET_WAITING_STARTED_EVENT, e);
    };

    TestRunnerBase.prototype._onActionRun = function () {
        window.sessionStorage.removeItem(STORAGE_ELEMENT_WAITING_KEY);
        this.eventEmitter.emit(this.ACTION_RUN_EVENT, {});
    };

    TestRunnerBase.prototype._onError = function (err) {
        this.eventEmitter.emit(this.TEST_FAILED_EVENT, {
            stepNum: this.testIterator.state.step - 1,
            err:     err
        });
    };

    TestRunnerBase.prototype._onAssertionFailed = function () {
    };

    TestRunnerBase.prototype._onSetStepsSharedData = function (e) {
        e.callback();
    };

    TestRunnerBase.prototype._onGetStepsSharedData = function (e) {
        e.callback();
    };

    TestRunnerBase.prototype._onExpectInactivity = function (e) {
        e.callback();
    };

    TestRunnerBase.prototype._onTakeScreenshot = function (e) {
        if (e && e.callback)
            e.callback();
    };

    TestRunnerBase.prototype._onIFrameStepExecuted = function () {
        this.executingStepInIFrameWindow = null;

        if (this.actionTargetWaitingStarted) {
            this.actionTargetWaitingStarted = false;
            this.testIterator.runLast();
        }
        else
            this.testIterator.runNext();
    };

    TestRunnerBase.prototype._onDialogsInfoChanged = function () {
    };

    TestRunnerBase.prototype.setGlobalWaitFor = function (event, timeout) {
        this.testIterator.setGlobalWaitFor(event, timeout);
    };

    TestRunnerBase.prototype._onBeforeUnload = function (fromIFrame, callback) {
        var testRunner = this;

        //NOTE: we should expect file downloading request only after before unload event (T216625)
        Transport.asyncServiceMsg({ cmd: ServiceCommands.UNCHECK_FILE_DOWNLOADING_FLAG }, function () {

            //NOTE: we need check it to determinate file downloading
            testRunner.isFileDownloadingIntervalID = NativeMethods.setInterval.call(window, function () {
                Transport.asyncServiceMsg({ cmd: ServiceCommands.GET_AND_UNCHECK_FILE_DOWNLOADING_FLAG }, function (res) {
                    if (res) {
                        window.clearInterval(testRunner.isFileDownloadingIntervalID);
                        testRunner.isFileDownloadingIntervalID = null;

                        if (fromIFrame) {
                            callback(res);
                            return;
                        }

                        if (testRunner.testIterator.state.stepDelayTimeout) {
                            window.clearTimeout(testRunner.testIterator.state.stepDelayTimeout);
                            testRunner.testIterator.state.stepDelayTimeout = null;
                        }

                        testRunner.testIterator.state.pageUnloading = false;
                        testRunner.testIterator._runStep();
                    }
                });
            }, CHECK_FILE_DOWNLOADING_DELAY);
        });
    };

    TestRunnerBase.prototype._clearFileDownloadingInterval = function () {
        if (this.isFileDownloadingIntervalID) {
            window.clearInterval(this.isFileDownloadingIntervalID);
            this.isFileDownloadingIntervalID = null;
        }
    };
});