TestCafeClient.define('Recorder.EventParser', function (require, exports) {
    var Hammerhead            = HammerheadClient.get('Hammerhead'),
        $                     = Hammerhead.$,
        Util                  = Hammerhead.Util,
        EventSandbox          = Hammerhead.EventSandbox,
        TextSelection         = Hammerhead.TextSelection,
        ContentEditableHelper = Hammerhead.ContentEditableHelper,
        PageState             = Hammerhead.PageState,
        NativeMethods         = Hammerhead.NativeMethods,
        SharedConst           = require('Shared.Const'),
        SelectorGenerator     = require('Recorder.SelectorGenerator'),
        KeyEventParser        = require('Recorder.KeyEventParser'),
        Automation            = require('Automation');

    var DBLCLICK_WAITING_TIMEOUT = 500;

    //Helpers
    function createEventObject (ev, element) {
        var pageCoords         = Util.getEventPageCoordinates(ev),
            isIE10PointerEvent = /MSPointer(Down|Move|Up)/.test(ev.type),
            eventShortType     = isIE10PointerEvent ? ev.type.replace('MSPointer', '') : '';

        return {
            target:   element,
            type:     isIE10PointerEvent ? 'pointer' + eventShortType.toLowerCase() : ev.type,
            ctrlKey:  ev.ctrlKey,
            altKey:   ev.altKey,
            shiftKey: ev.shiftKey,
            metaKey:  ev.metaKey,
            pageX:    pageCoords.x,
            pageY:    pageCoords.y
        };
    }

    var deferredActionQueue = (function () {
        var actionQueue = [];

        function removeAction (index) {
            if (actionQueue[index].timeoutId) {
                window.clearTimeout(actionQueue[index].timeoutId);
                actionQueue[index].timeoutId = null;
            }

            actionQueue[index].isWaitingDblClick = false;
            actionQueue[index].actionCreator     = null;
            actionQueue[index].element           = null;
            actionQueue[index].actionDescriptor  = null;
        }

        function removeAllActions () {
            actionQueue = [];
        }

        function createDeferredAction (type, creator, element, actionDescriptor) {
            return {
                type:             type,
                actionCreator:    creator,
                element:          element,
                elementParent:    Util.isContentEditableElement(element) ? ContentEditableHelper.findContentEditableParent(element) : null,
                actionDescriptor: actionDescriptor,
                index:            actionQueue.length
            };
        }

        function createAction (index) {
            var action = actionQueue[index];

            if (action.actionCreator)
                action.actionCreator.call();

            if (action.type === 'type')
                keyEventParser.resetTyping();
            else
                mouseEventParser.resetLastElementSelectors();

            removeAction(index);
        }

        function createAllActions () {
            $.each(actionQueue, function (index) {
                createAction(index);
            });
            removeAllActions();
        }

        return {
            init:      function () {
                actionQueue = [];
            },
            push:      function (type, creator, element, actionDescriptor) {
                var newAction  = createDeferredAction(type, creator, element, actionDescriptor),
                    prevAction = actionQueue[newAction.index - 1];

                if (type === 'click') {
                    if (prevAction && prevAction.actionCreator) {
                        if (prevAction.isWaitingDblClick && prevAction.element === element) {
                            removeAction(newAction.index - 1);
                            return;
                        }
                        else
                            createAllActions();
                    }

                    newAction.isWaitingDblClick = true;
                    newAction.timeoutId         = NativeMethods.setTimeout.call(window, function () {
                        newAction.isWaitingDblClick = false;
                        newAction.timeoutId         = null;

                        if (!Util.isEditableElement(newAction.element))
                            createAllActions();
                    }, DBLCLICK_WAITING_TIMEOUT);
                }
                else if (prevAction && prevAction.actionCreator) {
                    //NOTE: During typing in child of contentEditable elements events 'keydown', 'keypress', 'keyup' occurs
                    //for contentEditable parent element.

                    //We have problem in situation when you made click on one element, then click on another and start typing.
                    //After close addActionDialog for first click contentEditable parent's content will change
                    //and we can not find prevAction element parent.
                    //Therefore, we should store contentEditable parent element in deferred action object.
                    if (prevAction.element === element ||
                        (Util.isContentEditableElement(element) && prevAction.elementParent === element))
                        removeAction(newAction.index - 1);
                    else
                        createAllActions();
                }

                actionQueue.push(newAction);
            },
            removeAll: removeAllActions,
            createAll: createAllActions,

            getLastActionType:       function () {
                if (!actionQueue.length)
                    return '';

                return actionQueue[actionQueue.length - 1].type;
            },
            getLastActionDescriptor: function () {
                if (!actionQueue.length)
                    return null;

                return actionQueue[actionQueue.length - 1].actionDescriptor;
            },
            isWaitingDblClick:       function () {
                return actionQueue.length ?
                       actionQueue[actionQueue.length - 1].isWaitingDblClick :
                       false;
            }
        };
    })();

    //options
    var actionParsedCallback            = null,
        executableShortcuts             = null,
        recordableShortcutList          = null,
        typingStateChangedCallback      = null,
        clickEditorStateChangedCallback = null;

    //state
    var preventCurrentEvent     = false,
        inParsing               = false,
        parsedActionType        = '',
        lastParsedActionElement = null,
        lastParsedActionType    = null,
        prevPageState           = null,
        savedPrevPageState      = null,
        typingActionNotified    = false,
        typingActionProcessed   = false,
        saveSelection           = false,
        checkDblclick           = false,
        clickedSelectElement    = null;

    //Parsers
    var mouseEventParser = (function () {
        var EVENT_CODES                 = {
                others:      0,
                mousedown:   1,
                pointerdown: 1,
                touchstart:  1,

                mouseup:   2,
                pointerup: 2,
                touchend:  2,

                mousemove:   'm',
                pointermove: 'p',
                click:       3,
                dblclick:    4,
                contextmenu: 5,
                touchmove:   6,

                shortmove:  7,
                mediummove: 8,
                longmove:   9
            },
            EVENT_QUEUE_TYPES           = {
                NONE:                  0,
                MOUSEMOVE:             1,
                POINTERMOVE:           2,
                MOUSEMOVE_POINTERMOVE: 3,
                POINTERMOVE_MOUSEMOVE: 4
            },
            CHECKED_MOUSE_ACTION_EVENTS = ['click', 'dblclick', 'contextmenu', 'mouseup', 'touchend', 'pointerup'];

        var state = {
            mousePressed:                     false,
            lastActionType:                   '',
            lastHoveredElement:               null,
            lastHoveredElementSelectors:      null,
            startDragElement:                 null,
            lastPressedElement:               null,
            lastMouseDownOffset:              null,
            lastMouseDownPosition:            null,
            lastMouseDownButton:              null,
            clickEditorStateChangedTimeoutId: null,
            preclickTimeout:                  false
        };

        //helpers
        function createEncodedEventRegExp (regExpString) {
            var event = '',
                res   = regExpString;

            for (event in EVENT_CODES) {
                if (EVENT_CODES.hasOwnProperty(event)) {
                    res = res.replace(new RegExp('\\b' + event + '\\b', 'g'), EVENT_CODES[event]);
                }
            }

            return new RegExp(res);
        }

        var beforeNativeDialogClickTemplate = createEncodedEventRegExp('mousedown(,mouseup)?$');

        var events = (function () {
            var eventsState = [],
                lastEvent   = '';

            return {
                add: function (e) {
                    var event = e.type;

                    //NOTE: replace combinations of events like 'pointerdown,mousedown' in IE to the 'mousedown'
                    if (event === 'mousedown' || event === 'pointerdown' || event === 'touchstart') {
                        event = 'mousedown';

                        if (lastEvent === 'mousedown')
                            return;
                    }
                    else if (event === 'mouseup' || event === 'pointerup' || event === 'touchend') {
                        event = 'mouseup';

                        if (lastEvent === 'mouseup')
                            return;
                    }

                    eventsState.push(EVENT_CODES[event] || EVENT_CODES.others);
                    lastEvent = event;
                },

                clear: function () {
                    eventsState = [];
                    lastEvent   = '';
                },

                getState: function () {
                    return eventsState;
                }
            };
        })();

        function createMouseActionDescriptor (ev, eventType, options, deferred) {
            //NOTE: create descriptor for click, rclick, dblclick, drag and select
            var element               = options.element || ev.target,
                isTextEditable        = Util.isTextEditableElement(element),
                offset                = options.offset,
                selectors             = options.selectors,
                startPosition         = options.startPosition,
                elementSelectors      = null,
                elementSelectionStart = null,
                elementValueLength    = null,
                actionType            = eventType ? eventType : ev.type,
                actionElement         = null,
                isInverse             = null,
                start                 = null,
                end                   = null,
                dragPosition          = null,

                isDragAction          = eventType === 'drag',
                actionDescriptor      = null;

            state.lastActionType = actionType;

            if (isDragAction) {
                //NOTE: When drag action recorded via finger (touch screen device)
                //lastHoveredElement isn't correct to creation of action descriptor
                //(startDragElement created from lastHoveredElement)
                actionElement = ev.type === 'touchend' || (Util.isIE && window.navigator.msMaxTouchPoints > 0) ?
                                state.lastPressedElement :
                                (state.startDragElement || element);

                //NOTE: we recognize drag in editable element as a select action
                //(select action is impossible on touch screen devices)
                if (Util.isEditableElement(actionElement) && ev.type !== 'touchend') {
                    actionDescriptor = $.extend(true, {}, Automation.defaultSelectActionDescriptor);

                    if (Util.isContentEditableElement(actionElement) &&
                        !TextSelection.hasElementContainsSelection(actionElement))
                        actionElement = ContentEditableHelper.getElementBySelection(TextSelection.getSelectionByElement(actionElement));

                    actionType       = 'select';
                    parsedActionType = 'select';
                    isInverse        = TextSelection.hasInverseSelection(actionElement);
                    start            = TextSelection.getSelectionStart(actionElement);
                    end              = TextSelection.getSelectionEnd(actionElement);

                    actionDescriptor.apiArguments.startPos = isInverse ? end : start;
                    actionDescriptor.apiArguments.endPos   = isInverse ? start : end;
                }
                else {
                    actionDescriptor = $.extend(true, {}, Automation.defaultDragActionDescriptor);

                    dragPosition = Util.getEventAbsoluteCoordinates(ev);

                    if (Util.isElementInIframe(element)) {
                        var currentIFrame = Util.getIFrameByElement(element);
                        if (currentIFrame) {
                            var iFrameOffset    = Util.getOffsetPosition(currentIFrame),
                                iFrameBorders   = Util.getBordersWidth($(currentIFrame)),
                                pageCoordinates = Util.getEventPageCoordinates(ev);

                            actionDescriptor.apiArguments.dragOffsetX = pageCoordinates.x -
                                                                        (startPosition.x - iFrameOffset.left -
                                                                        iFrameBorders.left);
                            actionDescriptor.apiArguments.dragOffsetY = pageCoordinates.y -
                                                                        (startPosition.y - iFrameOffset.top -
                                                                        iFrameBorders.top);
                        }
                    }
                    else {
                        actionDescriptor.apiArguments.dragOffsetX = dragPosition.x - startPosition.x;
                        actionDescriptor.apiArguments.dragOffsetY = dragPosition.y - startPosition.y;
                    }

                    dragPosition = Util.getEventAbsoluteCoordinates(ev);

                    //NOTE: we should save start drag position to be able to rollback drag action
                    //also we should save end drag position to be able to emulate 'mouseup' and 'click' actions
                    // with correct coordinates (after confirm drag action)
                    actionDescriptor.startPosition = options.startPosition;
                    actionDescriptor.endPosition   = dragPosition;
                }

                elementSelectors = SelectorGenerator.generate($(actionElement));

                state.startDragElement = null;
            }
            else {
                actionDescriptor = $.extend(true, {}, Automation.defaultMouseActionDescriptor, {
                    type: actionType
                });

                elementSelectors = selectors || SelectorGenerator.generate($(element));
            }

            if (offset && /click|drag/.test(actionType)) {
                actionDescriptor.apiArguments.options.offsetX = offset.x;
                actionDescriptor.apiArguments.options.offsetY = offset.y;
                actionDescriptor.serviceInfo.useOffsets       = isDragAction;
            }

            savedPrevPageState = prevPageState;

            if (/click|drag/.test(actionType)) {
                actionDescriptor.apiArguments.options.ctrl  = ev.ctrlKey;
                actionDescriptor.apiArguments.options.alt   = ev.altKey;
                actionDescriptor.apiArguments.options.shift = ev.shiftKey;
                actionDescriptor.apiArguments.options.meta  = ev.metaKey;
            }

            if ((isTextEditable || Util.isContentEditableElement(element)) && !/drag|select/.test(actionType)) {
                elementSelectionStart                          = TextSelection.getSelectionStart(element);
                elementValueLength                             = isTextEditable ? element.value.length : ContentEditableHelper.getContentEditableValue(element).length;
                actionDescriptor.apiArguments.options.caretPos = elementSelectionStart ===
                                                                 elementValueLength ? '' : elementSelectionStart;
            }

            return $.extend(true, {}, actionDescriptor, {
                element:     actionElement || element,
                selector:    elementSelectors[0].selector,
                serviceInfo: {
                    selectors:     elementSelectors,
                    prevPageState: prevPageState,
                    isDeferred:    deferred
                }
            });
        }

        //NOTE: encode events to increase regular expressions executing performance (T230234)
        var actionsTemplates = {
            preclick: createEncodedEventRegExp('mousedown,(shortmove,|mediummove,)?mouseup$'), //BB239376
            click:    createEncodedEventRegExp('mousedown,(shortmove,|mediummove,)?mouseup,click$'),
            dblclick: createEncodedEventRegExp('(mousedown,)?(shortmove,)?(mouseup,)?(click,)?dblclick$'),
            rclick:   createEncodedEventRegExp('mousedown,(shortmove,)?(mouseup),contextmenu$'),
            drag:     createEncodedEventRegExp('mousedown,longmove,mouseup(,click)?$')
        };

        function recognizeAction (ev, force) {
            //NOTE: if drag is ended on other element, click events is not raised
            //otherwise click event is raised
            var joinedState = events.getState().join(',');

            if (state.lastActionType === 'drag' &&
                (joinedState === EVENT_CODES.click + '' || joinedState === EVENT_CODES.mouseup + '')) {
                events.clear();
                return;
            }

            var eventsState          = events.getState(),
                processedState       = [],
                processedStateString = '',
                curEvent             = '',
                queueLength          = 0,
                curQueueType         = EVENT_QUEUE_TYPES.NONE;

            function stopQueue () {
                if (curQueueType === EVENT_QUEUE_TYPES.MOUSEMOVE_POINTERMOVE ||
                    curQueueType === EVENT_QUEUE_TYPES.POINTERMOVE_MOUSEMOVE)
                    queueLength = Math.floor(queueLength / 2);

                if (queueLength > 0 && queueLength <= 2)
                    processedState.push(EVENT_CODES.shortmove);
                else if (queueLength > 2 && queueLength <= 6)
                    processedState.push(EVENT_CODES.mediummove);
                else if (queueLength > 6)
                    processedState.push(EVENT_CODES.longmove);

                queueLength  = 0;
                curQueueType = EVENT_QUEUE_TYPES.NONE;
            }

            //NOTE: join long queue of mousemoves to one of the following events: 'shortmove' (0-2 mousemoves), 'mediummove' (3-6), 'longmove' (7+)
            for (var i = 0; i < eventsState.length; i++) {
                curEvent = eventsState[i];
                switch (curEvent) {
                    case EVENT_CODES.mousemove:
                    case EVENT_CODES.touchmove:
                        if (curQueueType === EVENT_QUEUE_TYPES.NONE) {
                            curQueueType = EVENT_QUEUE_TYPES.MOUSEMOVE;
                        }
                        else if (curQueueType === EVENT_QUEUE_TYPES.POINTERMOVE)
                            curQueueType = EVENT_QUEUE_TYPES.POINTERMOVE_MOUSEMOVE;
                        queueLength++;
                        break;
                    case EVENT_CODES.pointermove:
                        if (curQueueType === EVENT_QUEUE_TYPES.NONE) {
                            curQueueType = EVENT_QUEUE_TYPES.POINTERMOVE;
                        }
                        else if (curQueueType === EVENT_QUEUE_TYPES.MOUSEMOVE)
                            curQueueType = EVENT_QUEUE_TYPES.MOUSEMOVE_POINTERMOVE;
                        queueLength++;
                        break;
                    default:
                        if (curQueueType !== EVENT_QUEUE_TYPES.NONE)
                            stopQueue();

                        processedState.push(curEvent);
                        break;
                }
            }

            if (curQueueType !== EVENT_QUEUE_TYPES.NONE)
                stopQueue();

            processedStateString = processedState.join(',');

            $.each(actionsTemplates, function (key, value) {
                if (value.test(processedStateString)) {
                    actionCreators[key](ev, force);
                    return false;
                }
            });
        }

        function resetState () {
            state.mousePressed                     = false;
            state.lastActionType                   = '';
            state.lastHoveredElement               = null;
            state.lastHoveredElementSelectors      = null;
            state.startDragElement                 = null;
            state.lastPressedElement               = null;
            state.lastMouseDownOffset              = null;
            state.lastMouseDownPosition            = null;
            state.lastMouseDownButton              = null;
            state.clickEditorStateChangedTimeoutId = null;

            if (state.preclickTimeout) {
                window.clearTimeout(state.preclickTimeout);
                state.preclickTimeout = null;
            }
        }

        //actions creators
        function createMouseActionNotifier (ev, type, deferred) {
            var isClickOnEditableElement = ev.type === 'click' && Util.isEditableElement(ev.target),
                actionDescriptor         = createMouseActionDescriptor(ev, type,
                    {
                        element:       state.lastPressedElement,
                        offset:        state.lastMouseDownOffset,
                        selectors:     state.lastHoveredElementSelectors,
                        startPosition: state.lastMouseDownPosition
                    }, isClickOnEditableElement);

            if (deferred) {
                var deferredActionCreator = function () {
                    if (isClickOnEditableElement && typeof clickEditorStateChangedCallback === 'function') {
                        clickEditorStateChangedCallback.call(this, false);
                    }
                    actionParsedCallback(actionDescriptor);
                };

                if (isClickOnEditableElement && typeof clickEditorStateChangedCallback === 'function') {
                    state.clickEditorStateChangedTimeoutId = NativeMethods.setTimeout.call(window, function () {
                        clickEditorStateChangedCallback.call(this, true, function () {
                            deferredActionQueue.createAll();
                        });
                    }, 0);
                }

                return function () {
                    deferredActionQueue.push('click', deferredActionCreator, state.lastPressedElement, actionDescriptor);
                    events.clear();
                };
            }

            return function () {
                if (ev.type === 'dblclick' && Util.isEditableElement(ev.target) &&
                    typeof clickEditorStateChangedCallback === 'function') {
                    if (state.clickEditorStateChangedTimeoutId) {
                        window.clearTimeout(state.clickEditorStateChangedTimeoutId);
                        state.clickEditorStateChangedTimeoutId = null;
                    }

                    clickEditorStateChangedCallback.call(this, false);
                }

                deferredActionQueue.createAll();
                actionParsedCallback(actionDescriptor);

                events.clear();
            };
        }

        function createDeferredMouseAction (ev, type) {
            createMouseActionNotifier(ev, type, true)();
        }

        var actionCreators = {
            preclick: function (ev) {
                var target = ev.target;

                if (target !== state.lastPressedElement) {
                    ev.type = 'click';
                    actionCreators.click(ev, true);

                    return;
                }

                //T359154 - Test recorder doesn't record item selection in combobox
                var clickActionNotifier = createMouseActionNotifier(ev, 'click');

                if (state.preclickTimeout)
                    window.clearTimeout(state.preclickTimeout);

                state.preclickTimeout = NativeMethods.setTimeout.call(window, function () {
                    state.preclickTimeout = null;

                    if (!Util.isElementInDocument(target))
                        clickActionNotifier();
                }, 0);
            },

            click: function (ev, force) {
                if (state.preclickTimeout) {
                    window.clearTimeout(state.preclickTimeout);
                    state.preclickTimeout = null;
                }


                parsedActionType = 'click';
                createDeferredMouseAction(ev);

                if (force)
                    deferredActionQueue.createAll();

                events.clear();
            },

            dblclick: function (ev) {
                if (state.preclickTimeout) {
                    window.clearTimeout(state.preclickTimeout);
                    state.preclickTimeout = null;
                }

                parsedActionType = 'dblclick';
                createMouseActionNotifier(ev)();
            },

            rclick: function (ev) {
                if (state.preclickTimeout) {
                    window.clearTimeout(state.preclickTimeout);
                    state.preclickTimeout = null;
                }

                parsedActionType = 'rclick';
                createMouseActionNotifier(ev, 'rclick')();
            },

            drag: function (ev) {
                parsedActionType = 'drag';
                createMouseActionNotifier(ev, 'drag')();
            }
        };

        //NOTE: in touch devices we should not create drag action when events raised page scaling or scrolling
        var scalingScrollingChecker = (function () {
            var lastInnerHeight = null,
                lastInnerWidth  = null,
                lastScrollTop   = null,
                lastScrollLeft  = null,

                $window         = $(window);

            function save () {
                lastInnerHeight = window.innerHeight;
                lastInnerWidth  = window.innerWidth;
                lastScrollTop   = $window.scrollTop();
                lastScrollLeft  = $window.scrollLeft();
            }

            return {
                save:  save,
                check: function () {
                    var isZoomingOrScrolling = window.innerHeight !== lastInnerHeight ||
                                               window.innerWidth !== lastInnerWidth ||
                                               $window.scrollLeft() !== lastScrollLeft ||
                                               $window.scrollTop() !== lastScrollTop;
                    save();

                    return isZoomingOrScrolling;
                }
            };
        })();

        //NOTE: we should check events state to identify unfinished clicks after page unload
        var checkEventsState = function () {
            if (createEncodedEventRegExp('^mousedown(,mouseup)?$').test(events.getState().join(',')))
                actionParsedCallback(createMouseActionDescriptor({}, 'click', { element: state.lastPressedElement }));
        };

        //api
        return {
            init: function () {
            },

            destroy: function (skipDeferred) {
                if (!skipDeferred) {
                    deferredActionQueue.createAll();
                    checkEventsState();
                }
                else
                    deferredActionQueue.removeAll();

                events.clear();
                resetState();
            },

            parse: function (ev) {
                if (ev.button === Util.BUTTON.MIDDLE)
                    return;

                var eventElement     = ev.target || ev.srcElement,
                    $select          = null,
                    selectOffset     = null,
                    selectChildIndex = null;

                //NOTE: in mozilla only 'click' event is raised before 'contextmenu' event when mouse right button clicked
                if (Util.isMozilla && ev.type === 'click' && ev.button === Util.BUTTON.RIGHT)
                    return;

                if (/mousemove|touchmove|pointmove|MSPointerMove/.test(ev.type) && !state.mousePressed)
                    return;

                if (/mouseout|pointerout|MSPointerOut/.test(ev.type))
                    return;

                //T173548 - TD_14_2 Safari 8 - Wrong recording click on select and option
                if (Util.isSafari && ev.type === 'click' && Util.isSelectElement(eventElement) &&
                    clickedSelectElement && eventElement === clickedSelectElement) {
                    actionParsedCallback(mouseEventParser.createMouseActionDescriptor({}, 'click', {
                        element:   clickedSelectElement,
                        selectors: state.lastHoveredElementSelectors
                    }));
                    actionParsedCallback(mouseEventParser.createMouseActionDescriptor({}, 'click', { element: $(clickedSelectElement).find('option')[clickedSelectElement.selectedIndex] }));
                    clickedSelectElement = null;
                    events.clear();
                    return;
                }

                //NOTE: the 'mousedown' event is not raised when an option tag is clicked
                if ((/pointerdown|MSPointerDown/.test(ev.type) ||
                    ((!Util.isIE || Util.browserVersion === 9) && /mousedown|touchstart/.test(ev.type))) ||
                    (Util.isIE && ev.type === 'click' && ev.target.tagName.toLowerCase() === 'option')) {
                    state.lastPressedElement  = ev.target || ev.srcElement;
                    state.lastMouseDownButton = ev.button;

                    if (state.lastPressedElement === state.lastHoveredElement)
                        state.lastHoveredElementSelectors = SelectorGenerator.generate($(state.lastHoveredElement));
                    else
                        state.lastHoveredElementSelectors = SelectorGenerator.generate($(state.lastPressedElement));

                    var element       = ev.target || ev.srcElement,
                        elementOffset = Util.getOffsetPosition(element);

                    state.lastMouseDownPosition = Util.getEventAbsoluteCoordinates(ev);

                    if (Util.isWebKit && element.tagName.toLowerCase() === 'option') {
                        var select = Util.getSelectParent($(element));

                        if (!select) {
                            elementOffset.left = 0;
                            elementOffset.top  = 0;
                        }
                        else {
                            $select          = $(select);
                            selectOffset     = Util.getOffsetPosition($select[0]);
                            selectChildIndex = Util.getChildVisibleIndex($select, element);

                            elementOffset.left = selectOffset.left;
                            elementOffset.top  = selectOffset.top + selectChildIndex * Util.getOptionHeight($select);
                        }
                    }

                    //NOTE: We should save affected element to be able restore selected index
                    if (/pointerdown|MSPointerDown/.test(ev.type) ||
                        ((!Util.isIE || Util.browserVersion === 9) && /mousedown|touchstart/.test(ev.type))) {
                        if (Util.isSelectElement(ev.target))
                            PageState.addAffectedElement(prevPageState, ev.target);
                        else if (ev.target.tagName.toLowerCase() === 'option' && Util.getSelectParent($(ev.target)))
                            PageState.addAffectedElement(prevPageState, Util.getSelectParent($(ev.target)));
                    }

                    state.lastMouseDownOffset = {
                        x: state.lastMouseDownPosition.x > elementOffset.left ? state.lastMouseDownPosition.x -
                                                                                elementOffset.left : 0,
                        y: state.lastMouseDownPosition.y > elementOffset.top ? state.lastMouseDownPosition.y -
                                                                               elementOffset.top : 0
                    };

                    if (Util.isElementInIframe(element)) {
                        var currentIFrame = Util.getIFrameByElement(element);
                        if (currentIFrame) {
                            state.lastMouseDownOffset.x -= $(currentIFrame.contentWindow).scrollLeft();
                            state.lastMouseDownOffset.y -= $(currentIFrame.contentWindow).scrollTop();
                        }
                    }

                    //NOTE: for FF we should not record scroll element like drag action (B237789)
                    //than we doesn't set mousePressed state after mousedown on scrollbars
                    if (Util.isMozilla ||
                        (Util.isIE && Util.isSelectElement(element) && Util.getSelectSizeValue($(element)) > 1)) {
                        var elementDim = Util.getClientDimensions(element);
                        if (Util.hasScroll(element)) {
                            if (element.scrollWidth > element.clientWidth &&
                                state.lastMouseDownOffset.y >=
                                elementDim.height - elementDim.border.bottom - Util.getScrollbarSize())
                                return;

                            if (element.scrollHeight > element.clientHeight &&
                                state.lastMouseDownOffset.x >=
                                elementDim.width - elementDim.border.right - Util.getScrollbarSize())
                                return;
                        }
                        //NOTE: This hack specially for scroll option list of select elements
                        else if (Util.isMozilla && Util.isSelectElement(element) &&
                                 Util.getSelectVisibleChildren($(element)).length > Util.MAX_OPTION_LIST_LENGTH) {
                            if (state.lastMouseDownOffset.y >
                                elementDim.height + elementDim.border.top + elementDim.border.bottom &&
                                state.lastMouseDownOffset.x >=
                                elementDim.width - elementDim.border.right - Util.getScrollbarSize()) {
                                return;
                            }
                        }
                    }

                    if (ev.button === Util.BUTTON.LEFT || Util.hasTouchEvents)
                        state.mousePressed = true;
                }

                if (/touchstart/.test(ev.type)) {
                    scalingScrollingChecker.save();
                }

                if (/touchend/.test(ev.type) && scalingScrollingChecker.check()) {
                    events.clear();
                    return;
                }

                //NOTE: we don't listen 'mouseup' event when 'mousedown' event was raised before parsing started
                if (/mouseup|touchend|pointerup|MSPointerUp/.test(ev.type) &&
                    (ev.button === Util.BUTTON.LEFT || Util.hasTouchEvents)) {
                    if (!state.mousePressed)
                        return;
                    else
                        state.mousePressed = false;
                }

                if (Util.isIE && Util.isSelectElement(eventElement) &&
                    Util.getSelectSizeValue($(eventElement)) > 1 && ev.type === 'click') {
                    $select      = $(eventElement);
                    selectOffset = Util.getOffsetPosition($select[0]);

                    var $selectChildren  = Util.getSelectVisibleChildren($select),
                        eventCoordinates = Util.getEventAbsoluteCoordinates(ev),
                        selectBorders    = Util.getBordersWidth($select),
                        optionHeight     = Util.getOptionHeight($select);

                    eventCoordinates.x = eventCoordinates.x - selectOffset.left - selectBorders.left +
                                         $select.scrollLeft();
                    eventCoordinates.y = eventCoordinates.y - selectOffset.top - selectBorders.top +
                                         $select.scrollTop();

                    selectChildIndex = Math.floor(eventCoordinates.y / optionHeight);
                    eventElement     = $selectChildren[selectChildIndex];

                    var childOffsetTop = selectChildIndex * optionHeight;

                    state.lastPressedElement          = eventElement;
                    state.lastHoveredElementSelectors = SelectorGenerator.generate($(eventElement));
                    state.lastMouseDownOffset.x       = eventCoordinates.x > 0 ? eventCoordinates.x : 0;
                    state.lastMouseDownOffset.y       = eventCoordinates.y > childOffsetTop ? eventCoordinates.y -
                                                                                              childOffsetTop : 0;
                }

                var event = createEventObject(ev, eventElement);

                //NOTE: we clean accumulated 'click' events when 'dblclick' event is raised
                if (event.type === 'dblclick')
                    deferredActionQueue.removeAll();

                if (/pointerover|MSPointerOver/.test(event.type) ||
                    ((!Util.isIE || Util.browserVersion === 9) && event.type === 'mouseover')) {
                    state.lastHoveredElement = event.target;

                    //T198725 - Can't select element for the assertion
                    Util.storeElementAttributes(SharedConst.OLD_ATTR_VALUES, state.lastHoveredElement);

                    if (!state.mousePressed) {
                        if (Util.isIE && state.lastHoveredElement.tagName.toLowerCase() === 'iframe' &&
                            state.startDragElement &&
                            Util.getIFrameByElement(state.startDragElement) === state.lastHoveredElement)
                            return;
                        state.startDragElement = state.lastHoveredElement;
                    }

                    return;
                }

                //T182240 - Drag actions aren't recorded on http://dev.sencha.com/extjs/5.0.0/examples/organizer/organizer.html in IE
                //This check need for 'mouseover' in IE > 9
                if (event.type !== 'mouseover')
                    events.add(event);

                if ($.inArray(event.type, CHECKED_MOUSE_ACTION_EVENTS) > -1) {
                    var isClickOnLink = event.type === 'click' &&
                                        event.target.tagName &&
                                        event.target.tagName.toLocaleLowerCase() === 'a' &&
                                        event.target.href;

                    recognizeAction(event, isClickOnLink);
                }
            },

            resetLastElementSelectors: function () {
                state.lastHoveredElementSelectors = null;
            },

            createMouseActionDescriptor: createMouseActionDescriptor,

            checkActionsBeforeAdditionNativeDialogHandler: function () {
                if (beforeNativeDialogClickTemplate.test(events.getState().join(','))) {
                    actionParsedCallback(createMouseActionDescriptor({}, state.lastMouseDownButton ===
                                                                         Util.BUTTON.RIGHT ? 'rclick' : 'click', { element: state.lastPressedElement }));
                    events.clear();
                }
            }
        };
    })();

    var keyEventParser = (function () {
        var parser = null;

        var state = {
            typedSymbols:            [],
            inTyping:                false,
            isFirstTypedSymbol:      false,
            currentElement:          null,
            currentElementSelectors: null,
            initialElementValue:     null,
            lastHoveredElement:      null,
            lastEventType:           null
        };

        //helpers
        function hasSymbolsOrShiftOnly (keysCombination) {
            var keys = keysCombination.replace(/^\+/g, 'plus').replace(/\+\+/g, '+plus').split('+');
            $.map(keys, function (key, index) {
                    keys[index] = key.replace('plus', '+');
                }
            );

            for (var i = 0; i < keys.length; i++) {
                if (keys[i].length > 1 && keys[i] !== 'shift' && keys[i] !== 'space')
                    return false;
            }

            return true;
        }

        function resetState () {
            state.typedSymbols            = [];
            state.inTyping                = false;
            state.currentElement          = null;
            state.currentElementSelectors = null;
            state.initialElementValue     = null;
            state.lastEventType           = null;
        }

        //actions creators
        function createPressActionDescriptor (keys) {
            var descriptor = $.extend(true, {}, Automation.defaultPressActionDescriptor, {
                apiArguments: {
                    keysCommand: keys
                }
            });

            parsedActionType = 'press';
            deferredActionQueue.createAll();
            //NOTE: We should save page state only for using in FF
            // to be able restore selected index
            descriptor.serviceInfo.prevPageState = prevPageState;
            savedPrevPageState                   = prevPageState;

            actionParsedCallback(descriptor);
        }

        function createTypeActionDescriptor (el) {
            var elementSelectionStart      = null,
                elementValueLength         = null,
                descriptor                 = $.extend(true, {}, Automation.defaultTypeActionDescriptor, {
                    element: el
                }),

                secondTypingInSuccession   = typingActionNotified && !typingActionProcessed,
                saveActiveCancellingOption = (secondTypingInSuccession ||
                                              !((lastParsedActionType === 'click' || lastParsedActionType === 'type') &&
                                                el === lastParsedActionElement));

            if (saveActiveCancellingOption)
                savedPrevPageState = prevPageState;

            var pageState = saveActiveCancellingOption ? prevPageState : savedPrevPageState;

            if (state.inTyping && state.isFirstTypeSymbol)
                PageState.addAffectedElement(pageState, el);

            descriptor.apiArguments.text         = state.typedSymbols.join('');
            descriptor.serviceInfo.prevPageState = pageState;

            var lastActionDescriptor = deferredActionQueue.getLastActionDescriptor();

            if (lastActionDescriptor) {
                var lastOptions = lastActionDescriptor.apiArguments.options;

                if (lastOptions)
                    descriptor.apiArguments.options = lastOptions;

                descriptor.element = lastActionDescriptor.element;
            }
            else {
                elementSelectionStart                    = TextSelection.getSelectionStart(el);
                elementValueLength                       = Util.isTextEditableElement(el) ? el.value.length : ContentEditableHelper.getContentEditableValue(el).length;
                descriptor.apiArguments.options.caretPos = elementSelectionStart ===
                                                           elementValueLength ? '' : elementSelectionStart;
            }

            descriptor.serviceInfo.selectors = lastActionDescriptor ?
                                               lastActionDescriptor.serviceInfo.selectors :
                                               SelectorGenerator.generate($(state.lastHoveredElement));

            descriptor.selector = descriptor.serviceInfo.selectors[0].selector;

            return descriptor;
        }

        function createTypeAction (el, deferred) {
            parsedActionType = 'type';

            var typeActionDescriptor          = createTypeActionDescriptor(el),
                completeTypeActionIconCreated = false,
                typeActionCreated             = false;

            if (deferred) {
                var deferredActionCreator = function () {
                    typeActionCreated         = true;
                    state.initialElementValue = null;

                    if (typeof typingStateChangedCallback === 'function') {
                        if (completeTypeActionIconCreated)
                            typingStateChangedCallback.call(this, false);
                        else
                            typingStateChangedCallback.call(this, true);
                    }

                    actionParsedCallback(typeActionDescriptor);
                    typingActionNotified = true;
                };

                if (typeof typingStateChangedCallback === 'function') {
                    NativeMethods.setTimeout.call(window, function () { //B235930
                        if (!typeActionCreated) {
                            completeTypeActionIconCreated = true;

                            typingStateChangedCallback.call(this, true, function () {
                                deferredActionQueue.createAll();
                            });
                        }
                        else
                            typingStateChangedCallback.call(this, false);
                    }, 0);
                }

                deferredActionQueue.push('type', deferredActionCreator, el, typeActionDescriptor);
            }
            else
                deferredActionQueue.createAll();
        }

        //events handlers
        function symbolPressed (charCode, target) {
            if (state.createOnKeyPress)
                return;

            if (Util.isEditableElement(target)) {
                state.isFirstTypeSymbol = !state.inTyping;
                state.inTyping          = true;
                state.typedSymbols.push(String.fromCharCode(charCode));
                state.currentElement = target;
                createTypeAction(target, true);
            }
            else {
                state.inTyping          = false;
                state.isFirstTypeSymbol = false;
            }
        }

        function keysPressedCallback (keysCombination, force) {
            if (!force && state.inTyping && hasSymbolsOrShiftOnly(keysCombination))
                return;

            if (state.typedSymbols.length)
                createTypeAction(state.currentElement);

            //NOTE: we don't record separated key presses of 'Ctrl' and 'Shift'
            // because they are useless and we use this keys for control stepsPanel
            if (keysCombination !== 'ctrl' && keysCombination !== 'shift')
                createPressActionDescriptor(keysCombination.replace(/\n/g, 'enter')); //T230273
        }

        //api
        return {
            init: function () {
                var options = {
                    symbolPressed:    symbolPressed,
                    keysPressed:      keysPressedCallback,
                    shortcutHandlers: {}
                };

                var createExecutableShortcutHandler = function (shortcut) {
                    return {
                        start:    function () {
                            preventCurrentEvent = true;

                            if (typeof executableShortcuts[shortcut].start === 'function')
                                executableShortcuts[shortcut].start();
                            else if (typeof executableShortcuts[shortcut] === 'function')
                                executableShortcuts[shortcut]();
                        },
                        end:      function () {
                            preventCurrentEvent = true;

                            if (typeof executableShortcuts[shortcut].end === 'function')
                                executableShortcuts[shortcut].end();
                        },
                        multiple: false
                    };
                };

                for (var executableShortcut in executableShortcuts) {
                    if (executableShortcuts.hasOwnProperty(executableShortcut))
                        options.shortcutHandlers[executableShortcut.toLowerCase()] = createExecutableShortcutHandler(executableShortcut);
                }

                var accumulatedKeys = [];

                var createRecordableShortcutHandler = function () {
                    return {
                        start: function (keys) {
                            preventCurrentEvent = false;
                            accumulatedKeys.push(keys);
                        },
                        end:   function () {
                            //T182340 - Press 'Tab' and 'Shift+Tab' recording problem
                            if (accumulatedKeys && accumulatedKeys.length)
                                keysPressedCallback(accumulatedKeys.join(' '), true);
                            accumulatedKeys     = [];
                            preventCurrentEvent = false;
                        }
                    };
                };

                if (recordableShortcutList)
                    for (var i = 0; i < recordableShortcutList.length; i++)
                        options.shortcutHandlers[recordableShortcutList[i].toLowerCase()] = createRecordableShortcutHandler();

                parser = new KeyEventParser();
                parser.init(options);
            },

            destroy: function (skipDeferred) {
                if (!parser)
                    return;

                if (!skipDeferred) {
                    if (state.typedSymbols.length)
                        createTypeAction(state.currentElement);

                    //NOTE: this check for case when not symbol key was pressed and page was unloaded after that.
                    // Keydown and keypress (keypress in optional) are raised before unloding, but no keyup.
                    if (!parser.inTyping && !parser.suspend && parser.currentKeysCombination)
                        keysPressedCallback(parser.currentKeysCombination, true);
                }

                parser.clear();
                resetState();
            },

            clearState: function () {
                parser.clearState();
            },

            parse: function (ev) {
                var curElement = ev.target || ev.srcElement;

                if (!state.inTyping && ev.type === 'keydown') {
                    state.lastHoveredElement = curElement;

                    //T198725 - Can't select element for the assertion
                    Util.storeElementAttributes(SharedConst.OLD_ATTR_VALUES, state.lastHoveredElement);

                    //NOTE: We should save affected element only for FF
                    // to be able restore selected index
                    if (Util.isMozilla && Util.isSelectElement(curElement))
                        PageState.addAffectedElement(prevPageState, curElement);
                }
                //NOTE: There is different behavior when recording select an 'option' tag in 'select' element.
                // In webkit 'option' element is not visual and click on it raises 'change event on the 'select' element only.
                // In firefox, opera and ie it raises 'click' event on 'option' element.
                // Therefore in webkit and IE we should catch the 'change' event after click on a 'select' element.
                else if (Util.isSelectElement(curElement) && ev.type === 'change') {
                    if (Util.isWebKit && Util.getSelectSizeValue($(curElement)) === 1) {
                        deferredActionQueue.createAll();

                        //T173548 - TD_14_2 Safari 8 - Wrong recording click on select and option
                        if (Util.isSafari)
                            clickedSelectElement = curElement;
                        else {
                            //T122981: TD14_1_2 - Actions are not record and playback with select that has optgroup tags
                            actionParsedCallback(mouseEventParser.createMouseActionDescriptor({}, 'click', { element: $(curElement).find('option')[curElement.selectedIndex] }));
                        }
                    }

                }

                EventSandbox.watchElementEditing(curElement);
                state.lastEventType = ev.type;
                parser.pushEvent(ev);
            },

            resetTyping: function () {
                state.inTyping     = false;
                state.typedSymbols = [];
            },

            mouseActionParsed: function () {
                parser.mouseActionParsed();
            },

            createTypeAndDeferredActions: function () {
                if (state.typedSymbols.length)
                    createTypeAction(state.currentElement);

                deferredActionQueue.createAll();
            },

            checkActionsBeforeAdditionNativeDialogHandler: function () {
                if (state.lastEventType === 'keydown') {
                    if (parser.symbolPressed)
                        state.createOnKeyPress = true;

                    parser.endKeyPressing();
                }
            }
        };
    })();

    //API
    exports.init = function (callback, options) {
        actionParsedCallback = callback;

        if (options) {
            executableShortcuts             = options.executableShortcuts || {};
            recordableShortcutList          = options.recordableShortcutList || [];
            typingStateChangedCallback      = options.typingStateChangedCallback || null;
            clickEditorStateChangedCallback = options.clickEditorStateChangedCallback || null;
        }

        inParsing = true;

        $(document).ready(function () {
            prevPageState      = PageState.saveState();
            savedPrevPageState = prevPageState;
        });

        mouseEventParser.init();
        keyEventParser.init({});
        deferredActionQueue.init();
    };

    exports.destroy = function (skipDeferred) {
        inParsing = false;
        keyEventParser.destroy(skipDeferred);
        mouseEventParser.destroy(skipDeferred);
        actionParsedCallback    = null;
        parsedActionType        = '';
        prevPageState           = null;
        savedPrevPageState      = null;
        lastParsedActionElement = null;
        lastParsedActionType    = null;
        typingActionNotified    = false;
        typingActionProcessed   = false;
    };

    exports.resetParsing = function () {
        if (inParsing) {
            inParsing = false;
            keyEventParser.clearState();
        }
    };

    exports.parseEvent = function (ev) {
        inParsing = true;

        var isEditable = Util.isEditableElement(ev.target || ev.srcElement, true);

        //NOTE: we should save current selection before the new action will be raised
        if (saveSelection ||
            (checkDblclick && !deferredActionQueue.isWaitingDblClick() && !(ev.type === 'dblclick' && isEditable))) {
            prevPageState = PageState.saveState();

            saveSelection = false;
            checkDblclick = false;
        }

        if (/(^mouse\w+$)|(^touch\w+$)|(^pointer\w+$)|(^MSPointer\w+$)|^(dbl)?click$|^contextmenu$/.test(ev.type))
            mouseEventParser.parse(ev);

        if (/^key\w+$|^change$|^input$/.test(ev.type))
            keyEventParser.parse(ev);

        var curParsedAction = parsedActionType;

        if (parsedActionType) {
            lastParsedActionElement = ev.target || ev.srcElement;
            lastParsedActionType    = parsedActionType;

            if (typingActionNotified && !typingActionProcessed)
                typingActionProcessed = true;
            else if (typingActionNotified && typingActionProcessed) {
                typingActionProcessed = false;
                typingActionNotified  = false;
            }

            if (deferredActionQueue.getLastActionType() !== 'click')
                saveSelection = true;
            else
                checkDblclick = true;

            //B254740
            if (/click|drag|select/.test(parsedActionType))
                keyEventParser.mouseActionParsed();

            parsedActionType = '';
        }

        //NOTE: In webkit click on 'option' element raises 'change event on the 'select' element (without 'click' for option)
        // and we recognize it like click on option but we should prevent bubbling click for select element

        var prevent = preventCurrentEvent ||
                      (Util.isWebKit && lastParsedActionElement && Util.isSelectElement(lastParsedActionElement) &&
                       ev.type === 'click' && curParsedAction === '' &&
                       lastParsedActionType === 'click' && lastParsedActionElement === (ev.target || ev.srcElement));

        preventCurrentEvent = false;

        return prevent;
    };

    exports.createAllDeferred = function () {
        keyEventParser.createTypeAndDeferredActions();
    };

    exports.checkActionsBeforeAdditionNativeDialogHandler = function () {
        if (inParsing) {
            deferredActionQueue.createAll();

            mouseEventParser.checkActionsBeforeAdditionNativeDialogHandler();
            keyEventParser.checkActionsBeforeAdditionNativeDialogHandler();
        }
    };
});