TestCafeClient.define('Recorder.IFrameRecorder', function (require) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        MessageSandbox = Hammerhead.MessageSandbox,
        Util = Hammerhead.Util,
        JavascriptExecutor = require('Base.JavascriptExecutor'),

        Settings = require('Settings'),
        SelectorGenerator = require('Recorder.SelectorGenerator'),
        SharedConst = require('Shared.Const'),
        Automation = require('Automation'),
        EventListener = require('Recorder.EventListener'),
        RecorderBase = require('Recorder.RecorderBase'),
        RecorderUI = require('UI.Recorder'),
        RecorderUtil = require('Recorder.Util'),
        PageState = Hammerhead.PageState,
        ElementPicker = require('Recorder.ElementPicker');

    var IFrameRecorder = this.exports = function (storedNativeDialogs) {
        RecorderBase.apply(this, [storedNativeDialogs]);
    };

    Util.inherit(IFrameRecorder, RecorderBase);

    IFrameRecorder.prototype.completeInitialization = function () {
        var initializedMsg = {
            cmd: RecorderBase.IFRAME_RECORDER_INITIALIZED_CMD
        };

        MessageSandbox.sendServiceMsg(initializedMsg, window.top);

        var recorder = this;

        MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, function (e) {
            var message = e.message;

            switch (message.cmd) {
                case RecorderBase.START_IFRAME_RECORDER_CMD:
                    Settings.PLAYBACK = false;
                    recorder.start();

                    if (message.initNativeDialogs)
                        recorder.initNativeDialogsHandling();
                    break;

                case RecorderBase.IFRAME_CONFIRM_ACTION_RESPONSE_CMD:
                    if (typeof recorder.currentConfirmationCallback === 'function') {
                        recorder.currentConfirmationCallback(message.confirmed,
                            message.stepInfo ? RecorderUtil.parseStepInfo(message.stepInfo) : null);

                        recorder.currentConfirmationCallback = null;
                    }

                    break;

                case RecorderBase.IFRAME_TYPING_STATE_CHANGED_RESPONSE_CMD:
                    if (typeof recorder.currentCompleteTypingCallback === 'function') {
                        recorder.currentCompleteTypingCallback();
                        recorder.currentCompleteTypingCallback = null;
                    }
                    break;

                case RecorderBase.IFRAME_CLICK_EDITOR_STATE_CHANGED_RESPONSE_CMD:
                    if (typeof recorder.currentCompleteClickEditorCallback === 'function') {
                        recorder.currentCompleteClickEditorCallback();
                        recorder.currentCompleteClickEditorCallback = null;
                    }
                    break;

                case RecorderBase.IFRAME_START_PICKING_ELEMENT_CMD:
                    recorder._startPickingElement();
                    break;

                case RecorderBase.IFRAME_STOP_PICKING_ELEMENT_CMD:
                    recorder._stopPickingElement();
                    break;

                case RecorderBase.IFRAME_HOVER_ACTION_RECORDED_CMD:
                    recorder._hoverActionRecorded(message.elementSelectors, message.options);
                    break;

                case RecorderBase.IFRAME_CREATE_ALL_DEFERRED_ACTIONS_CMD:
                    //NOTE: save deferred actions in iframe
                    EventListener.setPreventingMode(true);

                    recorder._createAllDeferred();

                    EventListener.setPreventingMode(false);
                    break;
            }
        });

        this.currentConfirmationCallback = null;
        this.currentCompleteTypingCallback = null;
        this.currentCompleteClickEditorCallback = null;
    };

    IFrameRecorder.prototype.start = function (initNativeDialogs) {
        IFrameRecorder.base.start.call(this, initNativeDialogs);
    };

    IFrameRecorder.prototype._saveDialogHandler = function (dialog, retValue) {
        var sendSaveDialogHandlerRequest = function () {
            var msg = {
                cmd: RecorderBase.IFRAME_SAVE_DIALOG_HANDLER_CMD,
                dialog: dialog,
                retValue: retValue
            };

            MessageSandbox.sendServiceMsg(msg, window.top);
        };

        if (this.stepSaving && dialog !== 'beforeUnload') {
            var stepSavingResponse = function (e) {
                if (e.message.cmd === RecorderBase.IFRAME_SAVE_STEP_RESPONSE_CMD) {
                    MessageSandbox.off(MessageSandbox.SERVICE_MSG_RECEIVED, stepSavingResponse);

                    sendSaveDialogHandlerRequest();
                }
            };

            MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, stepSavingResponse);
        }
        else
            sendSaveDialogHandlerRequest();
    };

    IFrameRecorder.prototype._initUi = function () {
        var recorder = this;

        this.recordingManagerOptions.typingStateChangedCallback = function (isTypingStarted, completeCallback) {
            if (isTypingStarted)
                recorder.currentCompleteTypingCallback = completeCallback;

            var msg = {
                cmd: RecorderBase.IFRAME_TYPING_STATE_CHANGED_REQUEST_CMD,
                rect: Util.getElementClientRectangle(Util.getActiveElement()),
                isTypingStarted: isTypingStarted
            };

            MessageSandbox.sendServiceMsg(msg, window.top);
        };

        this.recordingManagerOptions.clickEditorStateChangedCallback = function (isClickOnEditorStarted, completeCallback) {
            if (isClickOnEditorStarted)
                recorder.currentCompleteClickEditorCallback = completeCallback;

            var msg = {
                cmd: RecorderBase.IFRAME_CLICK_EDITOR_STATE_CHANGED_REQUEST_CMD,
                rect: Util.getElementClientRectangle(Util.getActiveElement()),
                isClickOnEditorStarted: isClickOnEditorStarted
            };

            MessageSandbox.sendServiceMsg(msg, window.top);
        };

        function shortcutPressedStart(shortcut) {
            var msg = {
                cmd: RecorderBase.IFRAME_SHORTCUT_PRESSED_START_CMD,
                shortcut: shortcut
            };

            MessageSandbox.sendServiceMsg(msg, window.top);
        }

        function shortcutPressedEnd(shortcut) {
            var msg = {
                cmd: RecorderBase.IFRAME_SHORTCUT_PRESSED_END_CMD,
                shortcut: shortcut
            };

            MessageSandbox.sendServiceMsg(msg, window.top);
        }

        function getShortcutHandler(shortcut) {
            return {
                start: function () {
                    EventListener.setPreventingMode(true);

                    shortcutPressedStart(shortcut);
                },
                end: function () {
                    //HACK: if wait action shortcut was pressed in iframe we need create iframe's deferred actions immediately
                    //to save step (async action) before creation wait action
                    if (shortcut === Automation.ADD_ACTION_SHORTCUTS.wait) {
                        recorder._createAllDeferred();
                    }

                    shortcutPressedEnd(shortcut);
                    EventListener.setPreventingMode(false);
                }
            };
        }

        for (var shortcut in RecorderUI.Shortcuts) {
            if (RecorderUI.Shortcuts.hasOwnProperty(shortcut))
                recorder.recordingManagerOptions.executableShortcuts[shortcut] = getShortcutHandler(shortcut);
        }
    };

    IFrameRecorder.prototype._startPickingElement = function () {
        IFrameRecorder.base._startPickingElement.call(this);
        EventListener.pause();

        this.elementPickCallback = function (elementSelectors, iFrameSelectors, options) {
            MessageSandbox.sendServiceMsg({
                cmd: RecorderBase.IFRAME_ELEMENT_PICKED_CMD,
                elementSelector: elementSelectors,
                iFrameSelectors: iFrameSelectors,
                options: options
            }, window.top);
        };

        ElementPicker.start(this.elementPickCallback);
    };

    IFrameRecorder.prototype._stopPickingElement = function () {
        IFrameRecorder.base._stopPickingElement.call(this);
        EventListener.resume();
    };

    //Upload

    IFrameRecorder.prototype._initHammerheadEventHandling = function () {
        var recorder = this;

        Hammerhead.on(Hammerhead.FILE_UPLOADING_EVENT, function (fileNames, input, callback) {
            var selectors = SelectorGenerator.generate($(input)),
                uploadAllowedHandler = function (e) {
                    if (e.message.cmd === RecorderBase.IFRAME_UPLOAD_ALLOWED_CMD) {
                        MessageSandbox.off(MessageSandbox.SERVICE_MSG_RECEIVED, uploadAllowedHandler);
                        callback(function (errs) {
                            var filePaths = fileNames.map(function (name) {
                                return SharedConst.UPLOADED_FILES_PATH + name;
                            });

                            recorder._onActionParsed({
                                type: 'upload',
                                apiArguments: {
                                    files: filePaths
                                },
                                serviceInfo: {
                                    selectors: SelectorGenerator.generate($(input)),
                                    fileNames: fileNames
                                }
                            });

                            MessageSandbox.sendServiceMsg({
                                cmd: RecorderBase.IFRAME_UPLOAD_COMPLETED_CMD,
                                errs: errs
                            }, window.top);
                        });
                    }
                };

            MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, uploadAllowedHandler);
            MessageSandbox.sendServiceMsg({
                cmd: RecorderBase.IFRAME_UPLOAD_REQUEST_TO_UPLOAD_CMD,
                fileNames: fileNames,
                selectors: selectors
            }, window.top);
        });
    };

    //Recording management
    IFrameRecorder.prototype._confirmAction = function (parsedStepInfo, callback) {
        var recorder = this,
            pageState = PageState.saveState();

        var confirmActionRequestMsg = {
            cmd: RecorderBase.IFRAME_CONFIRM_ACTION_REQUEST_CMD,
            stepInfo: parsedStepInfo ? RecorderUtil.serializeStepInfo(parsedStepInfo) : null
        };

        this.currentConfirmationCallback = function () {
            var args = arguments;

            PageState.restoreState(pageState, true, function () {
                callback.apply(recorder, args);
            });
        };

        MessageSandbox.sendServiceMsg(confirmActionRequestMsg, window.top);
    };

    IFrameRecorder.prototype._saveStep = function (stepInfo) {
        this.stepSaving = true;

        var iFrameRecorder = this,
            stepSavingResponse = function (e) {
                if (e.message.cmd === RecorderBase.IFRAME_SAVE_STEP_RESPONSE_CMD) {
                    MessageSandbox.off(MessageSandbox.SERVICE_MSG_RECEIVED, stepSavingResponse);
                    iFrameRecorder.stepSaving = false;
                }
            };

        MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, stepSavingResponse);

        var saveStepMsg = {
            cmd: RecorderBase.IFRAME_SAVE_STEP_REQUEST_CMD,
            stepInfo: RecorderUtil.serializeStepInfo(stepInfo)
        };

        MessageSandbox.sendServiceMsg(saveStepMsg, window.top);
    };

    IFrameRecorder.prototype._hoverActionRecorded = function (elementSelectors, options) {
        var actionDescriptor = $.extend(Automation.defaultMouseActionDescriptor, {
            type: 'hover',
            element: JavascriptExecutor.parseSelectorSync(elementSelectors[0].selector).$elements[0],
            apiArguments: {
                options: options
            },
            selector: elementSelectors[0].selector,
            serviceInfo: {
                selectors: elementSelectors
            }
        });

        this._onActionParsed(actionDescriptor);
    };
});