TestCafeClient.define('Recorder.KeyEventParser', function () {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        Util = Hammerhead.Util;

    //consts
    var TAB_KEY_CODE = Util.KEYS_MAPS.SPECIAL_KEYS['tab'];

    //utils
    function inModifying(state) {
        if (!(state.ctrl || state.alt || state.meta) && state.shift)
            return false;

        return state.ctrl || state.alt || state.shift || state.meta;
    }

    function keyToString(keyCode) {
        var activeElement = Util.getActiveElement();

        if (keyCode === Util.KEYS_MAPS.SPECIAL_KEYS.enter && !Util.isShadowUIElement(activeElement) &&
            activeElement.tagName.toLowerCase() === 'textarea') {
            return String.fromCharCode(10); //newline character
        }

        if (keyCode === Util.KEYS_MAPS.SPECIAL_KEYS.space)
            return Util.KEYS_MAPS.REVERSED_SPECIAL_KEYS[keyCode];

        return Util.isCharByKeyCode(keyCode) ?
            String.fromCharCode(Util.KEYS_MAPS.SYMBOLS_KEYS_CHAR_CODES[keyCode] || keyCode).toLowerCase() :
            (Util.KEYS_MAPS.REVERSED_SPECIAL_KEYS[keyCode] || Util.KEYS_MAPS.REVERSED_MODIFIERS[keyCode]) || '';
    }

    function addKey(keyCode, keysCombination) {
        var res = keysCombination;

        if (res)
            res += '+';

        return res + keyToString(keyCode);
    }

    function escapeSpecialSymbols(str) {
        return str.replace(/\(|\)|\[|\]|\\|\.|\^|\$|\||\?|\+|\*/, function (symbol) {
            return '\\' + symbol;
        });
    }

    function removeKey(keyCode, keysCombination) {
        if (!keyCode)
            return keysCombination;

        var keyString = escapeSpecialSymbols(keyToString(keyCode));

        return keysCombination.replace(new RegExp('\\+?' + keyString + '($|\\+)'), '');
    }

    function containsKey(keyCode, keysCombination) {
        var keyString = escapeSpecialSymbols(keyToString(keyCode));

        return (new RegExp('(^|\\+)' + keyString + '($|\\+)')).test(keysCombination);
    }

    //KeyEventParser
    var KeyEventParser = this.exports = function () {
        //callbacks
        this.symbolPressedCallback = null;
        this.keysPressedCallback = null;
        this.shortcutHandlers = {};

        //state
        this.modifiersState = { ctrl: false, alt: false, shift: false, meta: false };
        this.currentKeysCombination = '';
        this.savedKeysCombinations = [];
        this.pressedKeys = [];
        this.symbolPressed = false;
        this.currentStartShortcutHandler = null;
        this.suspend = false;

        var parser = this;
        //NOTE: to prevent keydown without keyup collision (for example: if you press F12 after keydown event window
        // loses focus and key up event is not raised)
        $(window).bind('blur', function (e) {
            if ((e.target || e.srcElement) === window)
                parser.clearState();
        });
    };


    //event handlers
    var lastKeyCode = null,
        repeatedSymbolSaved = false;

    KeyEventParser.prototype._keyDownHandler = function (ev) {
        var keyCode = ev.keyCode;

        //multiple key press handling
        this.currentEndShortcutHandler = null;

        if (this.currentStartShortcutHandler) {
            if (containsKey(keyCode, this.currentKeysCombination)) {
                if (this.currentStartShortcutHandler.multiple) {
                    this.currentKeysCombination = removeKey(keyCode, this.currentKeysCombination);
                    lastKeyCode = null;
                }
                else
                    return;
            }

            else {
                var joinedCombination = addKey(keyCode, this.currentKeysCombination),
                    newCombinationWithModifier = addKey(keyCode, removeKey(lastKeyCode, this.currentKeysCombination)),

                    joinedShortcut = this.shortcutHandlers[joinedCombination],
                    newShortcut = this.shortcutHandlers[newCombinationWithModifier];

                if ((newShortcut && newShortcut.start) &&
                    (joinedCombination === newCombinationWithModifier || !(joinedShortcut && joinedShortcut))) {
                    this.currentKeysCombination = removeKey(lastKeyCode, this.currentKeysCombination);
                }
                else if (this.shortcutHandlers[addKey(keyCode, '')] && this.shortcutHandlers[addKey(keyCode, '')].start)
                    this.currentKeysCombination = '';
            }
        }

        this.currentStartShortcutHandler = null;

        if (Util.KEYS_MAPS.REVERSED_MODIFIERS[keyCode]) {
            if (this.modifiersState[Util.KEYS_MAPS.REVERSED_MODIFIERS[keyCode]])
                return;

            this.modifiersState[Util.KEYS_MAPS.REVERSED_MODIFIERS[keyCode]] = true;
        }

        this.symbolPressed = Util.isCharByKeyCode(keyCode);

        if (!this.symbolPressed && this.inTyping) {
            this.inTyping = false;
            this.currentKeysCombination = '';
        }

        if ((!this.symbolPressed || inModifying(this.modifiersState)) && lastKeyCode === keyCode) {
            if (!this.savedKeysCombinations.length)
                this.savedKeysCombinations.push(this.currentKeysCombination);

            this.savedKeysCombinations.push(this.currentKeysCombination);
            repeatedSymbolSaved = true;
        }
        else {
            this.currentKeysCombination = addKey(keyCode, this.currentKeysCombination);
            repeatedSymbolSaved = false;
        }

        lastKeyCode = keyCode;

        var shortcutHandler = Util.getShortcutHandlerByKeyCombination(this.shortcutHandlers, this.currentKeysCombination);

        if (shortcutHandler) {
            this.currentStartShortcutHandler = shortcutHandler;

            if (this.inTyping)
                this.currentKeysCombination = Util.getShortcutsByKeyCombination(this.shortcutHandlers, this.currentKeysCombination)[0];

            if (shortcutHandler.start)
                shortcutHandler.start(this.currentKeysCombination);

            this.inTyping = false;
        }
        //NOTE: when focus is in the last focusable element on the page and tab key was pressed, keyup event
        //is not raised on any dom element because of default browser behavior (except mozilla and opera).
        // It means we should create tab press action on keydown event
        else if (!(Util.isMozilla || $.browser.opera) &&
            keyCode === TAB_KEY_CODE &&
            this.currentKeysCombination === 'tab' &&
            Util.getNextFocusableElement(Util.getActiveElement()).tagName.toLowerCase() === 'body') {

            if (this.keysPressedCallback)
                this.keysPressedCallback(this.currentKeysCombination);

            this.currentKeysCombination = '';
            this.pressedKeys = [];

            return;
        }

        this.pressedKeys.push(keyCode);
    };

    KeyEventParser.prototype._keyPressHandler = function (ev) {
        this.lastCharCode = ev.charCode || ev.keyCode || this.lastCharCode;

        if (!this.symbolPressed || this.currentStartShortcutHandler)
            return;

        if (ev.type === 'input') {      //T260092
            if (this.keyPressStoppedFromModifiers) {
                this.keyPressStoppedFromModifiers = false;
                this.currentKeysCombination       = '';
            }
            else
                return;
        }
        else if (inModifying(this.modifiersState)) {
            this.keyPressStoppedFromModifiers = true;
            return;
        }

        this.inTyping = true;

        if (this.symbolPressedCallback)
            this.symbolPressedCallback(this.lastCharCode, ev.target || ev.srcElement);
    };

    KeyEventParser.prototype._keyUpHandler = function (ev) {
        var keyCode = ev.keyCode;

        if (Util.KEYS_MAPS.REVERSED_MODIFIERS[keyCode])
            this.modifiersState[Util.KEYS_MAPS.REVERSED_MODIFIERS[keyCode]] = false;

        var shortcutHandler = Util.getShortcutHandlerByKeyCombination(this.shortcutHandlers, this.currentKeysCombination);

        if (!this.suspend && shortcutHandler && shortcutHandler.end) {
            if (this.currentEndShortcutHandler !== shortcutHandler) {
                this.currentEndShortcutHandler = shortcutHandler;
                shortcutHandler.end();
            }
        }

        if (this.currentStartShortcutHandler && this.currentStartShortcutHandler.start)
            this.currentKeysCombination = removeKey(keyCode, this.currentKeysCombination);

        this.pressedKeys = $.map(this.pressedKeys, function (item) {
            return item === keyCode ? null : item;
        });

        if (!this.currentKeysCombination.length) {
            this.savedKeysCombinations = [];
            lastKeyCode = null;
            this.suspend = false;

            return;
        }

        if (!this.pressedKeys.length) {
            if (this.suspend)
                this.suspend = false;
            else if (this.keysPressedCallback) {
                if (!this.savedKeysCombinations.length)
                    this.keysPressedCallback(this.currentKeysCombination);
                else
                    this.keysPressedCallback(this.savedKeysCombinations.join(' '));
            }

            this.currentKeysCombination = '';
            this.savedKeysCombinations = [];
            lastKeyCode = null;
        }
        else {
            if (!Util.isCharByKeyCode(keyCode) || inModifying(this.modifiersState)) {
                if (lastKeyCode === keyCode)
                    lastKeyCode = null;

                if (this.lastKeysCombination !== this.currentKeysCombination && !repeatedSymbolSaved && !this.currentStartShortcutHandler)
                    this.savedKeysCombinations.push(this.currentKeysCombination);

                repeatedSymbolSaved = false;
                this.currentKeysCombination = removeKey(keyCode, this.currentKeysCombination);
                this.lastKeysCombination = this.currentKeysCombination;
            }
        }
    };

    //API
    KeyEventParser.prototype.init = function (options) {
        /*
         init: function(options)

         options: {
         symbolPressed(charCode, target), // raised when some writible symbol was pressed
         keysPressed(keysCombination), // raised when some key (or combination) was pressed (such as 'ctrl+a', 'left' etc.)
         shortcutHandlers {
         'shortcut': shortcutHandler // raised when shortcut is pressed
         or
         'shortcut': {
         start: startShortcutHandler, //raised when shortcut is pressed [optional]
         end: endShortcutHandler, //raised when shortcut keys is up [optional]
         multiple: true|false // raise handler again while keys are held (default: true) [optional]
         }
         }
         }
         */
        var parser = this;

        if (!options)
            return;

        if (options.symbolPressed)
            parser.symbolPressedCallback = options.symbolPressed;

        if (options.keysPressed)
            parser.keysPressedCallback = options.keysPressed;

        if (options.shortcutHandlers) {
            $.each(options.shortcutHandlers, function (keys, handlers) {
                if (handlers instanceof Function)
                    parser.shortcutHandlers[keys] = { start: handlers, multiple: true };
                else {
                    parser.shortcutHandlers[keys] = {
                        start: handlers.start,
                        end: handlers.end,
                        multiple: typeof handlers.multiple === 'undefined' ? true : handlers.multiple
                    };
                }
            });
        }
    };

    KeyEventParser.prototype.clearState = function () {
        this.pressedKeys = [];
        this.currentKeysCombination = '';
        this.modifiersState = { ctrl: false, alt: false, shift: false, meta: false };
        this.savedKeysCombinations = [];
        this.symbolPressed = false;
        this.currentStartShortcutHandler = null;
        lastKeyCode = null;
        repeatedSymbolSaved = false;
    };

    KeyEventParser.prototype.clear = function () {
        var parser = this;

        parser.clearState();
        parser.symbolPressedCallback = null;
        parser.keysPressedCallback = null;
        parser.shortcutHandlers = {};
    };

    KeyEventParser.prototype.pushEvent = function (ev) {
        if (ev.type === 'keydown' && !this.suspend)
            this._keyDownHandler(ev);

        if ((ev.type === 'keypress' || ev.type === 'input') && !this.suspend)
            this._keyPressHandler(ev);

        if (ev.type === 'keyup')
            this._keyUpHandler(ev);
    };

    KeyEventParser.prototype.mouseActionParsed = function () {
        //NOTE: (B254740) if some mouse action was parsed when some key is pressed we are waiting for the moment when
        // there is not any pressed key
        if (lastKeyCode)
            this.suspend = true;
    };

    KeyEventParser.prototype.endKeyPressing = function(){
        var keyEventParser = this;

        this.pressedKeys.forEach(function(key){
            keyEventParser._keyUpHandler({
                keyCode: key
            });
        });
    };
});