TestCafeClient.define('Recorder.PropertyListGenerator', function (require, exports) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        Util = Hammerhead.Util,
        MessageSandbox = Hammerhead.MessageSandbox,
        IFrameMessages = require('Base.CrossDomainMessages'),
        JavascriptExecutor = require('Base.JavascriptExecutor');

    function getGeneralPropertyConstructorsForJQueryObject() {
        return [
            {
                name: 'exists',
                isSuitable: function () {
                    return true;
                },
                getValue: function (obj) {
                    return obj.length > 0;
                },
                generateGetter: function (selector) {
                    return selector + '.length > 0';
                },
                isDefault: true
            },
            {
                name: 'text',
                isSuitable: function (obj) {
                    return !!obj.length;
                },
                getValue: function (obj) {
                    return obj.text();
                },
                generateGetter: function (selector) {
                    return selector + '.text()';
                }
            },
            {
                name: 'value',
                isSuitable: function (obj) {
                    return obj.length === 1 && hasValueProperty(obj[0]);
                },
                getValue: function (obj) {
                    return obj.val();
                },
                generateGetter: function (selector) {
                    return selector + '.val()';
                }
            },
            {
                name: 'selectedIndex',
                isSuitable: function (obj) {
                    return obj.length === 1 && hasSelectedIndexProperty(obj[0]);
                },
                getValue: function (obj) {
                    return obj[0].selectedIndex;
                },
                generateGetter: function (selector) {
                    return selector + '[0].selectedIndex';
                }
            },
            {
                name: 'checked',
                isSuitable: function (obj) {
                    return obj.length === 1 && hasCheckedProperty(obj[0]);
                },
                getValue: function (obj) {
                    return obj[0].checked;
                },
                generateGetter: function (selector) {
                    return selector + '[0].checked';
                }
            },
            {
                name: 'visible',
                isSuitable: function () {
                    return true;
                },
                getValue: function (obj) {
                    return obj.is(':visible');
                },
                generateGetter: function (selector) {
                    return selector + '.is(":visible")';
                }
            },
            {
                name: 'class',
                isSuitable: function (obj) {
                    return obj.length === 1;
                },
                getValue: function (obj) {
                    return obj.attr('class');
                },
                generateGetter: function (selector) {
                    return selector + '.attr("class")';
                }
            },
            {
                name: 'matching element count',
                isSuitable: function () {
                    return true;
                },
                getValue: function (obj) {
                    return obj.length;
                },
                generateGetter: function (selector) {
                    return selector + '.length';
                }
            },
            {
                name: 'child element count',
                isSuitable: function (obj) {
                    return !!obj.length;
                },
                getValue: function (obj) {
                    return obj.children().length;
                },
                generateGetter: function (selector) {
                    return selector + '.children().length';
                }
            },
            {
                name: 'width',
                isSuitable: function (obj) {
                    return obj.length === 1;
                },
                getValue: function (obj) {
                    return obj.outerWidth();
                },
                generateGetter: function (selector) {
                    return selector + '.outerWidth()';
                }
            },
            {
                name: 'height',
                isSuitable: function (obj) {
                    return obj.length === 1;
                },
                getValue: function (obj) {
                    return obj.outerHeight();
                },
                generateGetter: function (selector) {
                    return selector + '.outerHeight()';
                }
            },
            {
                name: 'left',
                isSuitable: function (obj) {
                    return obj.length === 1;
                },
                getValue: function (obj) {
                    return obj.offset().left;
                },
                generateGetter: function (selector) {
                    return selector + '.offset().left;';
                }
            },
            {
                name: 'top',
                isSuitable: function (obj) {
                    return obj.length === 1;
                },
                getValue: function (obj) {
                    return obj.offset().top;
                },
                generateGetter: function (selector) {
                    return selector + '.offset().top';
                }
            }
        ];
    }

    function getGeneralPropertyConstructorsForDomElement() {
        return [
            {
                name: 'exists',
                isSuitable: function () {
                    return true;
                },
                getValue: function () {
                    return true;
                },
                generateGetter: function (selector) {
                    return '!!(' + selector + ')';
                },
                isDefault: true
            },
            {
                name: 'text',
                isSuitable: function () {
                    return true;
                },
                getValue: function (obj) {
                    return obj.textContent;
                },
                generateGetter: function (selector) {
                    return selector + '.textContent';
                }
            },
            {
                name: 'value',
                isSuitable: function (obj) {
                    return hasValueProperty(obj);
                },
                getValue: function (obj) {
                    return obj.value;
                },
                generateGetter: function (selector) {
                    return selector + '.value';
                }
            },
            {
                name: 'selectedIndex',
                isSuitable: function (obj) {
                    return hasSelectedIndexProperty(obj);
                },
                getValue: function (obj) {
                    return obj.selectedIndex;
                },
                generateGetter: function (selector) {
                    return selector + '.selectedIndex';
                }
            },
            {
                name: 'checked',
                isSuitable: function (obj) {
                    return hasCheckedProperty(obj);
                },
                getValue: function (obj) {
                    return obj.checked;
                },
                generateGetter: function (selector) {
                    return selector + '.checked';
                }
            },
            {
                name: 'visible',
                isSuitable: function () {
                    return true;
                },
                getValue: function (obj) {
                    return $(obj).is(':visible');
                },
                generateGetter: function (selector) {
                    return '$(' + selector + ').is(":visible")';
                }
            },
            {
                name: 'class',
                isSuitable: function () {
                    return true;
                },
                getValue: function (obj) {
                    return obj.className;
                },
                generateGetter: function (selector) {
                    return selector + '.className';
                }
            },
            {
                name: 'child element count',
                isSuitable: function () {
                    return true;
                },
                getValue: function (obj) {
                    return obj.childElementCount;
                },
                generateGetter: function (selector) {
                    return selector + '.childElementCount';
                }
            },
            {
                name: 'width',
                isSuitable: function () {
                    return true;
                },
                getValue: function (obj) {
                    return obj.offsetWidth;
                },
                generateGetter: function (selector) {
                    return selector + '.offsetWidth';
                }
            },
            {
                name: 'height',
                isSuitable: function () {
                    return true;
                },
                getValue: function (obj) {
                    return obj.offsetHeight;
                },
                generateGetter: function (selector) {
                    return selector + '.offsetHeight';
                }
            },
            {
                name: 'left',
                isSuitable: function () {
                    return true;
                },
                getValue: function (obj) {
                    return $(obj).offset().left;
                },
                generateGetter: function (selector) {
                    return '$(' + selector + ').offset().left';
                }
            },
            {
                name: 'top',
                isSuitable: function () {
                    return true;
                },
                getValue: function (obj) {
                    return $(obj).offset().top;
                },
                generateGetter: function (selector) {
                    return '$(' + selector + ').offset().top';
                }
            }
        ];
    }

    function getGeneralPropertyConstructorsForCollection() {
        return [
            {
                name: 'length',
                isSuitable: function () {
                    return true;
                },
                getValue: function (obj) {
                    return obj.length;
                },
                generateGetter: function (selector) {
                    return selector + '.length';
                },
                isDefault: true
            }
        ];
    }

    function hasValueProperty(el) {
        return el.tagName.toLowerCase() === 'input' || el.tagName.toLowerCase() === 'button' || el.tagName.toLowerCase() === 'select';
    }

    function hasCheckedProperty(el) {
        return el.tagName.toLowerCase() === 'input' && el.getAttribute('type') === 'checkbox';
    }

    function hasSelectedIndexProperty(el) {
        return el.tagName.toLowerCase() === 'select';
    }

    function getGeneralProperties(obj, selector, callback, context) {
        function onMessage(e) {
            if (e.message.cmd === IFrameMessages.GENERATE_GENERAL_PROPERTIES_IN_IFRAME_RESPONSE_CMD) {
                callback(e.message.properties);
                MessageSandbox.off(MessageSandbox.SERVICE_MSG_RECEIVED, onMessage);
            }
        }

        if (context) {
            MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, onMessage);

            var msg = {
                cmd: IFrameMessages.GENERATE_GENERAL_PROPERTIES_IN_IFRAME_REQUEST_CMD,
                selector: selector
            };

            MessageSandbox.sendServiceMsg(msg, context);

            return;
        }

        if (!obj)
            obj = JavascriptExecutor.parseSelectorSync(selector).evalResults;

        var propertyConstructors,
            generalProperties = [];

        if (Util.isJQueryObj(obj))
            propertyConstructors = getGeneralPropertyConstructorsForJQueryObject();
        else if (Util.isDomElement(obj))
            propertyConstructors = getGeneralPropertyConstructorsForDomElement();
        else if ($.isArray(obj) || obj instanceof NodeList || obj instanceof HTMLCollection)
            propertyConstructors = getGeneralPropertyConstructorsForCollection();
        else {
            callback(generalProperties);
            return;
        }

        for (var i = 0; i < propertyConstructors.length; i++)
            if (propertyConstructors[i].isSuitable(obj))
                generalProperties.push({
                    name: propertyConstructors[i].name,
                    value: propertyConstructors[i].getValue(obj),
                    getter: propertyConstructors[i].generateGetter(selector),
                    isDefault: propertyConstructors[i].isDefault || false
                });

        callback(generalProperties);
    }

    function getCssProperties(obj, selector, callback, context) {
        function onMessage(e) {
            if (e.message.cmd === IFrameMessages.GENERATE_CSS_PROPERTIES_IN_IFRAME_RESPONSE_CMD) {
                callback(e.message.properties);
                MessageSandbox.off(MessageSandbox.SERVICE_MSG_RECEIVED, onMessage);
            }
        }

        if (context) {
            MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, onMessage);

            var msg = {
                cmd: IFrameMessages.GENERATE_CSS_PROPERTIES_IN_IFRAME_REQUEST_CMD,
                selector: selector
            };

            MessageSandbox.sendServiceMsg(msg, context);

            return;
        }

        if (!obj)
            obj = JavascriptExecutor.parseSelectorSync(selector).evalResults;

        var getterGenerator,
            $el,
            propertyNames = [],
            cssProperties = [];

        if (Util.isJQueryObj(obj) && obj.length) {
            $el = obj;
            getterGenerator = function (selector, propertyName) {
                return selector + '.css("' + propertyName + '")';
            };
        }
        else if (Util.isDomElement(obj)) {
            $el = $(obj);
            getterGenerator = function (selector, propertyName) {
                return '$(' + selector + ').css("' + propertyName + '")';
            };
        }
        else {
            callback(cssProperties);
            return;
        }

        /* jshint ignore:start */
        for (var propertyName in $el[0].style) {
            if (!/[^a-z\-]/i.test(propertyName))
                propertyNames.push(propertyName);
        }
        /* jshint ignore:end */

        propertyNames.sort(function (name1, name2) {
            return name1.toLowerCase().localeCompare(name2.toLowerCase());
        });

        for (var i = 0; i < propertyNames.length; i++) {
            var propertyValue = $el.css(propertyNames[i]);
            if (typeof propertyValue !== 'object')
                cssProperties.push({
                    name: propertyNames[i],
                    value: propertyValue,
                    getter: getterGenerator(selector, propertyNames[i])
                });
        }

        callback(cssProperties);
    }

    function getAttributes(obj, selector, callback, context) {
        function onMessage(e) {
            if (e.message.cmd === IFrameMessages.GENERATE_ATTRIBUTES_IN_IFRAME_RESPONSE_CMD) {
                callback(e.message.properties);
                MessageSandbox.off(MessageSandbox.SERVICE_MSG_RECEIVED, onMessage);
            }
        }

        if (context) {
            MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, onMessage);

            var msg = {
                cmd: IFrameMessages.GENERATE_ATTRIBUTES_IN_IFRAME_REQUEST_CMD,
                selector: selector
            };

            MessageSandbox.sendServiceMsg(msg, context);

            return;
        }

        if (!obj)
            obj = JavascriptExecutor.parseSelectorSync(selector).evalResults;

        var res = [],
            attributes = Hammerhead.getOriginElementAttributes(obj[0]);

        if (!attributes)
            callback(null);
        else {
            for (var i = 0; i < attributes.length; i++) {
                res.push({
                    name: attributes[i].name,
                    value: obj[0].getAttribute(attributes[i].name),
                    getter: selector + '.attr("' + attributes[i].name + '")'
                });
            }

            callback(res);
        }
    }

    exports.getGeneralProperties = getGeneralProperties;
    exports.getCssProperties = getCssProperties;
    exports.getAttributes = getAttributes;
});