TestCafeClient.define('Recorder.Recorder', function (require) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        MessageSandbox = Hammerhead.MessageSandbox,
        Util = Hammerhead.Util,
        NativeMethods = Hammerhead.NativeMethods,

        Automation = require('Automation'),
        ElementPicker = require('Recorder.ElementPicker'),
        EventListener = require('Recorder.EventListener'),
        JavascriptExecutor = require('Base.JavascriptExecutor'),
        ModalBackground = require('UI.ModalBackground'),
        RecorderBase = require('Recorder.RecorderBase'),
        RecorderUI = require('UI.Recorder'),
        RecorderUtil = require('Recorder.Util'),
        SelectorGenerator = require('Recorder.SelectorGenerator'),
        ServiceCommands = require('Shared.ServiceCommands'),
        Settings = require('Settings'),
        SharedConst = require('Shared.Const'),
        Tooltip = require('UI.RecorderWidgets.Tooltip'),
        Transport = require('Base.Transport'),

        UXLog = require('UI.UXLog');


    var ACTION_WAIT_DEFAULT_VALUE = 1000,

        UPLOAD_CLEARED_MESSAGE = 'The file input was cleared.',
        UPLOAD_ERROR_MESSAGE = 'Cannot copy a file to the test folder. Probably, TestCafe does not have rights to do this or the hard disk is full.',
        UPLOAD_LOADING_PANEL_DELAY = 750,
        UPLOAD_MESSAGE_TEMPLATE = 'The following files have been copied to the test folder:<br><br>%s';

    //Util
    function createExecutableShortcutHandler(shortcut) {
        return {
            start: function () {
                EventListener.setPreventingMode(true);
            },
            end: function () {
                RecorderUI.Shortcuts[shortcut]();
                EventListener.setPreventingMode(false);
            },
            multiple: false
        };
    }

    //Recorder
    var Recorder = this.exports = function (storedNativeDialogs) {
        var recorder = this;

        RecorderBase.apply(this, [storedNativeDialogs]);

        this.stepsInfo = [];
        this.hasUnsavedChanges = false;
        this.testSaved = false;

        this.uploadingInProcess = false;
        this.elementPickingInProcess = false;
        this.paused = false;
        this.nativeDialogQueue = Settings.NATIVE_DIALOGS_QUEUE || [];

        if (!RecorderUI.isInitialized()) {
            RecorderUI.init(function () {
                    return recorder.stepsInfo;
                }, function () {
                    return recorder.hasUnsavedChanges;
                },
                function () {
                    return recorder.testSaved;
                }, Settings.SHOW_STEPS);
        }

        RecorderUI.events.on(RecorderUI.START_PICKING_ELEMENT_EVENT, function (e) {
            recorder._startPickingElement(function (elementSelectors, iFrameSelectors) {
                e.callback(elementSelectors, iFrameSelectors);
            }, e.iFrameContext);
        });

        RecorderUI.events.on(RecorderUI.STOP_PICKING_ELEMENT_EVENT, function () {
            RecorderUI.onStopPickElement();
        });
    };

    Util.inherit(Recorder, RecorderBase);

    Recorder.prototype.loadStepsInfo = function (callback) {
        var recorder = this;

        Transport.asyncServiceMsg({cmd: ServiceCommands.STEPS_INFO_GET}, function (res) {
            recorder.stepsInfo = [];

            var resStepsInfo = res.stepsInfo || [];

            for (var i = 0; i < resStepsInfo.length; i++)
                recorder.stepsInfo.push(RecorderUtil.parseServerCompatibleStepInfo(resStepsInfo[i]));

            recorder.hasUnsavedChanges = res.hasUnsavedChanges || false;
            recorder.testSaved = res.testSaved || false;

            //NOTE: we set actual step on recorder init
            callback(recorder.stepsInfo, recorder.hasUnsavedChanges);
        });
    };

    Recorder.prototype.start = function (initIFrameNativeDialogs) {
        UXLog.write('Recording started');
        Recorder.base.start.call(this, initIFrameNativeDialogs);
        this._checkNativeDialogsQueue();
    };

    Recorder.prototype._parseAndSendStepsInfo = function (callback) {
        $.each(this.stepsInfo, function (index, value) {
            RecorderUtil.parseStepArguments(value);
        });

        RecorderUtil.sendStepsInfo(this.stepsInfo, this.hasUnsavedChanges, callback);
    };

    Recorder.prototype._setDialogHandler = function (stepInfo, dialog, retValue) {
        this._setHasUnsavedChanges(true);

        if (!stepInfo.nativeDialogHandlers)
            stepInfo.nativeDialogHandlers = [];

        var dialogIndex = RecorderUtil.BROWSER_DIALOG_INDEXES[dialog],
            dialogHandler = {
                dialog: dialog,
                retValue: retValue
            };

        if (stepInfo.nativeDialogHandlers[dialogIndex] && stepInfo.nativeDialogHandlers[dialogIndex].length)
            stepInfo.nativeDialogHandlers[dialogIndex].push(dialogHandler);
        else
            stepInfo.nativeDialogHandlers[dialogIndex] = [dialogHandler];


        RecorderUtil.sendStepsInfo(this.stepsInfo, true);
        RecorderUI.handlerAdded(dialog);
    };

    Recorder.prototype._saveDialogHandler = function (dialog, retValue) {
        var recorder = this;

        //NOTE: the following dialogs are not raised in browsers after before unload event
        if (/alert|confirm|prompt/.test(dialog) && this.beforeUnloadEventWasRaised)
            return;

        function saveDialogHandler(stepNum, dialog, retValue) {
            var stepInfo = recorder.stepsInfo[stepNum];

            if (stepInfo)
                recorder._setDialogHandler(stepInfo, dialog, retValue);
            else {
                //NOTE: for dialogs before other test actions and before start recording
                var dialogIndex = RecorderUtil.BROWSER_DIALOG_INDEXES[dialog],
                    dialogHandler = {
                        dialog: dialog,
                        retValue: retValue
                    };

                if (recorder.nativeDialogQueue[dialogIndex] && recorder.nativeDialogQueue[dialogIndex].length)
                    recorder.nativeDialogQueue[dialogIndex].push(dialogHandler);
                else
                    recorder.nativeDialogQueue[dialogIndex] = [dialogHandler];


                Transport.asyncServiceMsg({
                    cmd: ServiceCommands.SET_NATIVE_DIALOGS_QUEUE,
                    queue: recorder.nativeDialogQueue
                });
            }
        }

        function getStepNumberForCurrentNativeDialogHandler() {
            var stepsLength = recorder.stepsInfo.length;

            if (recorder.uploadingInProcess || recorder.elementPickingInProcess)
                return stepsLength;

            //NOTE: action saved before the its handlers are raised (B251638)
            return stepsLength ? stepsLength - 1 : 0;
        }

        //NOTE: for Mozilla we should save native dialog immediately (before reload the page)
        if (Util.isMozilla && dialog === 'beforeUnload')
            recorder._setDialogHandler(recorder.stepsInfo[recorder.stepsInfo.length - 1], 'beforeUnload');
        else
            saveDialogHandler(getStepNumberForCurrentNativeDialogHandler(), dialog, retValue);
    };

    Recorder.prototype._checkNativeDialogsQueue = function () {
        var recorder = this;

        $.each(this.nativeDialogQueue, function (index, dialogs) {
            if (dialogs) {
                $.each(dialogs, function (i, dialog) {
                    var stepInfo = recorder.stepsInfo[recorder.stepsInfo.length - 1];

                    if (stepInfo)
                        recorder._setDialogHandler(stepInfo, dialog.dialog, dialog.retValue);
                });
            }
        });

        this.nativeDialogQueue = [];
    };

    Recorder.prototype._setHasUnsavedChanges = function (hasUnsavedChanges, sendServiceMessage) {
        this.hasUnsavedChanges = hasUnsavedChanges;
        RecorderUI.updateToolbarButtons();

        if (sendServiceMessage) {
            var unsavedChangesMsg = {
                cmd: ServiceCommands.SET_HAS_UNSAVED_CHANGES,
                hasUnsavedChanges: hasUnsavedChanges
            };

            Transport.asyncServiceMsg(unsavedChangesMsg, function () {
                RecorderUI.updateToolbarButtons();
            });
        }
    };

    Recorder.prototype._saveStep = function (stepInfo, overwriteLastStep) {
        RecorderUtil.parseStepArguments(stepInfo);

        this._checkNativeDialogsQueue();

        if (overwriteLastStep)
            this.stepsInfo[this.stepsInfo.length - 1] = stepInfo;
        else
            this.stepsInfo.splice(this.stepsInfo.length, 0, stepInfo);

        this._setHasUnsavedChanges(true);
        RecorderUtil.sendStepsInfo(this.stepsInfo, this.hasUnsavedChanges);

        if (!overwriteLastStep)
            RecorderUI.stepAdded(stepInfo);

        if (stepInfo.isAssertion)
            UXLog.write('Assertion is saved');
        else
            UXLog.write('Action is saved: ' + stepInfo.actionDescriptor.type);
    };

    Recorder.prototype._startPickingElement = function (callback, iframeContext) {
        var recorder = this;

        this.elementPickingInProcess = true;
        Recorder.base._startPickingElement.call(this);

        this.elementPickCallback = function (elementSelectors, iFrameSelectors, options) {
            recorder._stopPickingElement();
            callback(elementSelectors, iFrameSelectors, options);
        };

        ElementPicker.start(this.elementPickCallback, iframeContext);

        RecorderUI.onStartPickElement();

        var msg = {
            cmd: RecorderBase.IFRAME_START_PICKING_ELEMENT_CMD
        };

        if (iframeContext)
            MessageSandbox.sendServiceMsg(msg, iframeContext);
        else
            this._sendMessageToAllIFrame(msg);
    };

    Recorder.prototype._stopPickingElement = function () {
        this.elementPickingInProcess = false;
        Recorder.base._stopPickingElement.call(this);

        this._sendMessageToAllIFrame({
            cmd: RecorderBase.IFRAME_STOP_PICKING_ELEMENT_CMD
        });
    };

    Recorder.prototype._exitRecording = function () {
        var recorder = this;

        this.recording = false;

        Transport.asyncServiceMsg({
            cmd: ServiceCommands.EXIT_RECORDING
        }, function (res) {
            var fixtureFilename = res.fixtureFilename;

            if (!res.err) {
                recorder.recording = false;
                window.location.href = Settings.RETURN_URL + Settings.FIXTURE_FILE_NAME_PARAM + fixtureFilename;
            }
        });
    };

    //Ui
    Recorder.prototype._initUi = function () {
        RecorderUI.activateRecordingMode();

        this.recordingManagerOptions.typingStateChangedCallback = RecorderUI.onTypingStateChanged;
        this.recordingManagerOptions.clickEditorStateChangedCallback = RecorderUI.onClickEditorStateChanged;

        if (!RecorderUI.Shortcuts)
            this.recordingManagerOptions.executableShortcuts = {};
        else {
            for (var executableShortcut in RecorderUI.Shortcuts) {
                if (RecorderUI.Shortcuts.hasOwnProperty(executableShortcut))
                    this.recordingManagerOptions.executableShortcuts[executableShortcut] = createExecutableShortcutHandler(executableShortcut);
            }
        }
    };

    Recorder.prototype._initRecordingUiEventHandling = function () {
        var recorder = this,
            paused = false;

        function pauseEventListening() {
            if (!paused) {
                paused = true;
                EventListener.pause();
            }
        }

        function resumeEventListening() {
            if (paused) {
                paused = false;
                EventListener.resume();
            }
        }

        RecorderUI.events.on(RecorderUI.ADD_HOVER_ACTION_EVENT, function () {
            //NOTE: we set this flag here again only for 'toolbar button press' handler
            EventListener.setPreventingMode(true);

            recorder._createAllDeferred();

            recorder._sendMessageToAllIFrame({
                cmd: RecorderBase.IFRAME_CREATE_ALL_DEFERRED_ACTIONS_CMD
            });

            EventListener.pause();
            recorder._startPickingElement(function (elementSelectors, iFrameSelectors, options) {
                RecorderUI.onStopPickElement();
                EventListener.resume();

                if (!elementSelectors)
                    return;

                if (iFrameSelectors) {
                    var iFrame = JavascriptExecutor.parseSelectorSync(iFrameSelectors[0].selector).$elements[0];

                    MessageSandbox.sendServiceMsg({
                        cmd: RecorderBase.IFRAME_HOVER_ACTION_RECORDED_CMD,
                        elementSelectors: elementSelectors,
                        iFrameSelectors: iFrameSelectors,
                        options: options
                    }, iFrame.contentWindow);
                }
                else {
                    var actionDescriptor = $.extend(Automation.defaultMouseActionDescriptor, {
                        type: 'hover',
                        element: JavascriptExecutor.parseSelectorSync(elementSelectors[0].selector).$elements[0],
                        apiArguments: {
                            options: options
                        },
                        selector: elementSelectors[0].selector,
                        serviceInfo: {
                            selectors: elementSelectors
                        }
                    });

                    recorder._onActionParsed(actionDescriptor);
                }
            });

            EventListener.setPreventingMode(false);
        });

        RecorderUI.events.on(RecorderUI.ADD_WAIT_ACTION_EVENT, function () {
            //NOTE: we set this flag here again only for 'toolbar button press' handler (to force create deferred action)
            EventListener.setPreventingMode(true);

            recorder._createAllDeferred();

            //NOTE: deferred actions in an iframe don't have time to be created
            recorder._sendMessageToAllIFrame({
                cmd: RecorderBase.IFRAME_CREATE_ALL_DEFERRED_ACTIONS_CMD
            });

            var actionDescriptor = $.extend(Automation.defaultWaitActionDescriptor, {
                apiArguments: { ms: ACTION_WAIT_DEFAULT_VALUE }
            });

            //NOTE: for wait action we should always use confirm dialog
            recorder._onActionParsed(actionDescriptor, true);


            EventListener.setPreventingMode(false);
        });

        RecorderUI.events.on(RecorderUI.ADD_SCREENSHOT_ACTION_EVENT, function () {
            //NOTE: we set this flag here again only for 'toolbar button press' handler (to force create deferred action)
            EventListener.setPreventingMode(true);

            recorder._createAllDeferred();

            //NOTE: deferred actions in an iframe don't have time to be created
            recorder._sendMessageToAllIFrame({
                cmd: RecorderBase.IFRAME_CREATE_ALL_DEFERRED_ACTIONS_CMD
            });

            //NOTE: for wait action we should always use confirm dialog
            recorder._onActionParsed(Automation.defaultScreenshotActionDescriptor);

            EventListener.setPreventingMode(false);
        });

        RecorderUI.events.on(RecorderUI.ADD_ASSERTIONS_STEP_EVENT, function () {
            var isPrevStepAssertion = recorder.stepsInfo.length && recorder.stepsInfo[recorder.stepsInfo.length - 1].isAssertion,
                stepNum = isPrevStepAssertion ? recorder.stepsInfo.length : recorder.stepsInfo.length + 1,
                stepInfo = isPrevStepAssertion ? recorder.stepsInfo[recorder.stepsInfo.length - 1] : RecorderUtil.getAssertionsStepInfo();

            RecorderUI.confirmAction(stepNum, stepInfo, function (confirmed, stepInfo) {
                if (confirmed)
                    recorder._saveStep(stepInfo, isPrevStepAssertion);
            });
        });

        RecorderUI.events.on(RecorderUI.RUN_PLAYBACK_EVENT, function () {

            RecorderUI.hideToolbar();
            ModalBackground.show();
            ModalBackground.showLoadingIcon();

            recorder.recording = false;
            Transport.asyncServiceMsg({cmd: ServiceCommands.START_PLAYBACK}, function (res) {
                if (res.err)
                    window.location.href = Settings.RETURN_URL;
                else
                    Transport.switchToStartRecordingUrl(res.recordingUrl);
            });
        });

        RecorderUI.events.on(RecorderUI.SAVE_TEST_EVENT, function (e) {
            recorder.recording = false;

            Transport.asyncServiceMsg({
                cmd: ServiceCommands.SAVE_TEST,
                testName: e.testName
            }, function (res) {
                var err = res.err;

                if (!err)
                    recorder._setHasUnsavedChanges(false, true);

                e.callback(err);

                if (!err) {
                    recorder.recording = true;
                    recorder.testSaved = true;

                    if (e.complete)
                        recorder._exitRecording();
                }
            });
        });

        RecorderUI.events.on(RecorderUI.EXIT_RECORDING_EVENT, function (e) {
            if (e && e.skipCheckChanges) {
                recorder._exitRecording();
                return;
            }

            //T183278 - Recorder stops work after click 'Cancel' button in 'Unsaved changes dialog'
            //We should not stop listener, we just need create deferred actions
            EventListener.createAllDeferred();

            if (!recorder.hasUnsavedChanges)
                recorder._exitRecording();
            else
                RecorderUI.openChangesWarningDialog(e && e.hasErrors);
        });

        RecorderUI.events.on(RecorderUI.RESTART_RECORDING_EVENT, function () {
            recorder.recording = false;
            Transport.asyncServiceMsg({cmd: ServiceCommands.RESTART_RECORDING}, function (res) {
                if (res.err)
                    window.location.href = Settings.RETURN_URL;
                else
                    Transport.switchToStartRecordingUrl(res.recordingUrl);
            });
        });

        RecorderUI.events.on(RecorderUI.SEND_STEPS_INFO_EVENT, function () {
            recorder._setHasUnsavedChanges(!!recorder.stepsInfo.length);
            recorder._parseAndSendStepsInfo();
        });

        RecorderUI.events.on(RecorderUI.POPUP_DIALOG_OPENING_EVENT, pauseEventListening);

        RecorderUI.events.on(RecorderUI.POPUP_DIALOG_CLOSED_EVENT, function (e) {
            if (e.needSendStepsInfo)
                recorder._parseAndSendStepsInfo();

            resumeEventListening();
        });

        RecorderUI.events.on(RecorderUI.STEP_EDITING_STARTED, pauseEventListening);

        RecorderUI.events.on(RecorderUI.STEP_EDITING_FINISHED, resumeEventListening);
    };

    // Upload
    Recorder.prototype._startUpload = function (completed, callback) {
        var uploaded = false,
            upload_errs = null;

        ModalBackground.show();
        ModalBackground.showLoadingIcon();
        RecorderUI.hideToolbar();

        var loadingPanelTimeout = NativeMethods.setTimeout.call(window, function () {
            loadingPanelTimeout = null;

            if (uploaded)
                completed(upload_errs);
        }, UPLOAD_LOADING_PANEL_DELAY);

        callback(function (errs) {
            uploaded = true;
            upload_errs = !!errs.length;

            if (!loadingPanelTimeout)
                completed(upload_errs);
        });
    };

    Recorder.prototype._uploadCompleted = function (fileNames, filePaths, errs) {
        var message = errs ? UPLOAD_ERROR_MESSAGE : UPLOAD_MESSAGE_TEMPLATE.replace('%s', filePaths.join('<br>'));

        if (errs)
            message = UPLOAD_ERROR_MESSAGE;
        else {
            if (filePaths.join(''))
                message = UPLOAD_MESSAGE_TEMPLATE.replace('%s', filePaths.join('<br>'));
            else
                message = UPLOAD_CLEARED_MESSAGE;
        }

        RecorderUI.showToolbar();
        ModalBackground.hideLoadingIcon();
        ModalBackground.hide();

        Tooltip.show(message);
    };

    Recorder.prototype._initHammerheadEventHandling = function () {
        var recorder = this;

        Hammerhead.on(Hammerhead.FILE_UPLOADING_EVENT, function (fileNames, input, callback) {
            recorder.uploadingInProcess = true;

            var selectors = SelectorGenerator.generate($(input));

            recorder._startUpload(function (errs) {
                recorder.uploadingInProcess = false;
                var filePaths = fileNames.map(function (name) {
                    return SharedConst.UPLOADED_FILES_PATH + name;
                });

                recorder._uploadCompleted(fileNames, filePaths, errs);

                recorder._onActionParsed({
                    type: 'upload',
                    apiArguments: {
                        files: filePaths
                    },
                    serviceInfo: {
                        selectors: selectors,
                        fileNames: fileNames
                    }
                });
            }, callback);
        });
    };

    Recorder.prototype._onUploadRequested = function (fileNames, selectors, owner) {
        var recorder = this,
            filePaths = fileNames.map(function (name) {
                return SharedConst.UPLOADED_FILES_PATH + name;
            });

        recorder._startUpload(function (errs) {
            recorder._uploadCompleted(fileNames, filePaths, errs);
        }, function (completed) {
            MessageSandbox.sendServiceMsg({
                cmd: RecorderBase.IFRAME_UPLOAD_ALLOWED_CMD
            }, owner);

            var onUploadCompleted = function (e) {
                if (e.message.cmd === RecorderBase.IFRAME_UPLOAD_COMPLETED_CMD) {
                    MessageSandbox.off(MessageSandbox.SERVICE_MSG_RECEIVED, onUploadCompleted);
                    completed(e.message.errs);
                }
            };

            MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, onUploadCompleted);
        });
    };

    //Recording management
    Recorder.prototype._confirmAction = function (parsedStepInfo, callback) {
        RecorderUI.confirmAction(this.stepsInfo.length + 1, parsedStepInfo, callback);
    };
});
