TestCafeClient.define('Recorder.RecorderBase', function (require) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        Util = Hammerhead.Util,
        MessageSandbox = Hammerhead.MessageSandbox,
        Settings = require('Settings'),
        EventListener = require('Recorder.EventListener'),
        Automation = require('Automation'),
        AutomationIFrameBehavior = require('Automation.IFrameBehavior'),
        JavaScriptExecutor = require('Base.JavascriptExecutor'),
        RecorderUtil = require('Recorder.Util'),
        SelectorGenerator = require('Recorder.SelectorGenerator'),
        CursorWidget = require('UI.Cursor'),
        ElementPicker = require('Recorder.ElementPicker'),
        IFrameMessages = require('Base.CrossDomainMessages'),
        PropertyListGenerator = require('Recorder.PropertyListGenerator'),
        Assertions = require('TestRunner.API.Assertions'),
        ObjectViewer = require('UI.RecorderWidgets.ObjectViewer');

    var RecorderBase = this.exports = function (storedNativeDialogs) {
        this.eventEmitter = new Util.EventEmitter();
        this.waitingIframeRecorders = [];
        this.iFrameListeners = [];

        this.recordingManagerOptions = {
            recordableShortcutList: Automation.SUPPORTED_SHORTCUTS,
            executableShortcuts: {}
        };
        this.elementPickCallback = null;
        this.beforeUnloadEventWasRaised = false;
        this.nativeDialogs = storedNativeDialogs;

        this._initIFrameBehavior();

        if (!Settings.PLAYBACK)
            this.initNativeDialogsHandling();
    };

    RecorderBase.IFRAME_RECORDER_INITIALIZED_CMD = 'iframeRecorderInitialized';
    RecorderBase.START_IFRAME_RECORDER_CMD = 'startIframeRecorder';
    RecorderBase.IFRAME_SAVE_STEP_REQUEST_CMD = 'saveStepRequest';
    RecorderBase.IFRAME_SAVE_STEP_RESPONSE_CMD = 'saveStepResponse';
    RecorderBase.IFRAME_UPLOAD_ALLOWED_CMD = 'iframeUploadAllowed';
    RecorderBase.IFRAME_UPLOAD_COMPLETED_CMD = 'iframeUploadCompleted';
    RecorderBase.IFRAME_UPLOAD_REQUEST_TO_UPLOAD_CMD = 'iframeRequestToUpload';
    RecorderBase.IFRAME_CONFIRM_ACTION_REQUEST_CMD = 'confirmActionRequest';
    RecorderBase.IFRAME_CONFIRM_ACTION_RESPONSE_CMD = 'confirmActionResponse';
    RecorderBase.IFRAME_TYPING_STATE_CHANGED_REQUEST_CMD = 'typingStateChangedRequest';
    RecorderBase.IFRAME_TYPING_STATE_CHANGED_RESPONSE_CMD = 'typingStateChangedResponse';
    RecorderBase.IFRAME_CLICK_EDITOR_STATE_CHANGED_REQUEST_CMD = 'clickEditorStateChangedRequest';
    RecorderBase.IFRAME_CLICK_EDITOR_STATE_CHANGED_RESPONSE_CMD = 'clickEditorStateChangedResponse';
    RecorderBase.IFRAME_SHORTCUT_PRESSED_START_CMD = 'shortcutPressedStart';
    RecorderBase.IFRAME_SHORTCUT_PRESSED_END_CMD = 'shortcutPressedEnd';
    RecorderBase.IFRAME_SAVE_DIALOG_HANDLER_CMD = 'saveDialogHandler';
    RecorderBase.IFRAME_START_PICKING_ELEMENT_CMD = 'iFrameStartPickingElement';
    RecorderBase.IFRAME_STOP_PICKING_ELEMENT_CMD = 'iFrameStopPickingElement';
    RecorderBase.IFRAME_ELEMENT_PICKED_CMD = 'iFrameElementPicked';
    RecorderBase.IFRAME_HOVER_ACTION_RECORDED_CMD = 'iFrameHoverActionRecorded';
    RecorderBase.IFRAME_CREATE_ALL_DEFERRED_ACTIONS_CMD = 'iFrameCreateAllDeferredActions';

    RecorderBase.prototype.on = function () {
        return this.eventEmitter.on.apply(this.eventEmitter, arguments);
    };

    function addIFrameSelectorsToStepInfo(windowObj, stepInfo) {
        var iFrame = Util.getIFrameByWindow(windowObj);

        stepInfo = RecorderUtil.parseStepInfo(stepInfo);
        stepInfo.iFrameSelectors = SelectorGenerator.generate($(iFrame));

        return stepInfo;
    }

    RecorderBase.prototype._initIFrameBehavior = function () {
        var recorder = this;

        AutomationIFrameBehavior.init();

        MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, function (e) {
            var message = e.message,
                source = e.source,
                responseMessage = null,
                pos = null,
                rect = null;

            switch (message.cmd) {
                case RecorderBase.IFRAME_UPLOAD_REQUEST_TO_UPLOAD_CMD:
                    recorder._onUploadRequested(message.fileNames, message.selectors, source);
                    break;

                case RecorderBase.IFRAME_RECORDER_INITIALIZED_CMD:
                    if (recorder.started)
                        MessageSandbox.sendServiceMsg({
                            cmd: RecorderBase.START_IFRAME_RECORDER_CMD
                        }, source);
                    else
                        recorder.waitingIframeRecorders.push(source);

                    if ($.inArray(source, recorder.iFrameListeners) < 0)
                        recorder.iFrameListeners.push(source);
                    break;

                case RecorderBase.IFRAME_SAVE_STEP_REQUEST_CMD:
                    recorder._saveStep(addIFrameSelectorsToStepInfo(source, message.stepInfo));

                    MessageSandbox.sendServiceMsg({
                        cmd: RecorderBase.IFRAME_SAVE_STEP_RESPONSE_CMD
                    }, source);
                    break;

                case RecorderBase.IFRAME_CONFIRM_ACTION_REQUEST_CMD:
                    recorder._confirmAction(addIFrameSelectorsToStepInfo(source, message.stepInfo), function (confirmed, stepInfo) {
                        responseMessage = {
                            cmd: RecorderBase.IFRAME_CONFIRM_ACTION_RESPONSE_CMD,
                            confirmed: confirmed,
                            stepInfo: stepInfo ? RecorderUtil.serializeStepInfo(stepInfo) : null
                        };

                        MessageSandbox.sendServiceMsg(responseMessage, source);
                    });
                    break;

                case RecorderBase.IFRAME_TYPING_STATE_CHANGED_REQUEST_CMD:
                    if (message.isTypingStarted) {
                        pos = Util.getFixedPosition({ x: message.rect.left, y: message.rect.top }, source);
                        rect = {
                            left: pos.x,
                            top: pos.y,
                            width: message.rect.width,
                            height: message.rect.height
                        };
                    }

                    recorder.recordingManagerOptions.typingStateChangedCallback(message.isTypingStarted, function () {
                        MessageSandbox.sendServiceMsg({ cmd: RecorderBase.IFRAME_TYPING_STATE_CHANGED_RESPONSE_CMD }, source);
                    }, rect);

                    break;

                case RecorderBase.IFRAME_CLICK_EDITOR_STATE_CHANGED_REQUEST_CMD:
                    if (message.isClickOnEditorStarted) {
                        pos = Util.getFixedPosition({ x: message.rect.left, y: message.rect.top }, source);
                        rect = {
                            left: pos.x,
                            top: pos.y,
                            width: message.rect.width,
                            height: message.rect.height
                        };
                    }

                    recorder.recordingManagerOptions.clickEditorStateChangedCallback(message.isClickOnEditorStarted, function () {
                        MessageSandbox.sendServiceMsg({ cmd: RecorderBase.IFRAME_CLICK_EDITOR_STATE_CHANGED_RESPONSE_CMD }, source);
                    }, rect);

                    break;

                case RecorderBase.IFRAME_SHORTCUT_PRESSED_START_CMD:
                    if (typeof recorder.recordingManagerOptions.executableShortcuts[message.shortcut].start === 'function')
                        recorder.recordingManagerOptions.executableShortcuts[message.shortcut].start();

                    else if (typeof recorder.recordingManagerOptions.executableShortcuts[message.shortcut] === 'function')
                        recorder.recordingManagerOptions.executableShortcuts[message.shortcut]();

                    break;

                case RecorderBase.IFRAME_SHORTCUT_PRESSED_END_CMD:
                    if (typeof recorder.recordingManagerOptions.executableShortcuts[message.shortcut].end === 'function')
                        recorder.recordingManagerOptions.executableShortcuts[message.shortcut].end();

                    break;

                case RecorderBase.IFRAME_SAVE_DIALOG_HANDLER_CMD:
                    recorder._saveDialogHandler(message.dialog, message.retValue);
                    break;

                case RecorderBase.IFRAME_ELEMENT_PICKED_CMD:
                    recorder.elementPickCallback(message.elementSelector, message.iFrameSelectors, message.options);
                    break;

                case IFrameMessages.GENERATE_GENERAL_PROPERTIES_IN_IFRAME_REQUEST_CMD:
                    PropertyListGenerator.getGeneralProperties(null, message.selector, function (properties) {
                        responseMessage = {
                            cmd: IFrameMessages.GENERATE_GENERAL_PROPERTIES_IN_IFRAME_RESPONSE_CMD,
                            properties: properties
                        };

                        MessageSandbox.sendServiceMsg(responseMessage, source);
                    });
                    break;

                case IFrameMessages.GENERATE_CSS_PROPERTIES_IN_IFRAME_REQUEST_CMD:
                    PropertyListGenerator.getCssProperties(null, message.selector, function (properties) {
                        responseMessage = {
                            cmd: IFrameMessages.GENERATE_CSS_PROPERTIES_IN_IFRAME_RESPONSE_CMD,
                            properties: properties
                        };

                        MessageSandbox.sendServiceMsg(responseMessage, source);
                    });
                    break;

                case IFrameMessages.GENERATE_ATTRIBUTES_IN_IFRAME_REQUEST_CMD:
                    PropertyListGenerator.getAttributes(null, message.selector, function (properties) {
                        responseMessage = {
                            cmd: IFrameMessages.GENERATE_ATTRIBUTES_IN_IFRAME_RESPONSE_CMD,
                            properties: properties
                        };

                        MessageSandbox.sendServiceMsg(responseMessage, source);
                    });
                    break;

                case IFrameMessages.ASSERT_REQUEST_CMD:
                    var args = message.args,
                        operator = message.operator;

                    args[0] = JavaScriptExecutor.eval(args[0]);

                    if (operator === 'eq' || operator === 'notEq')
                        args[1] = JavaScriptExecutor.eval(args[1]);

                    Assertions.assert(operator, args, function (err) {
                        MessageSandbox.sendServiceMsg({
                            cmd: IFrameMessages.ASSERT_RESPONSE_CMD,
                            err: err
                        }, source);
                    });
                    break;

                case IFrameMessages.OBJECT_VIEWER_GET_OBJECT_VIEW_REQUEST_CMD:
                    ObjectViewer.getObjectView(message.stringValue, null, function (objectView) {

                        delete objectView.parsedValue;

                        MessageSandbox.sendServiceMsg({
                            cmd: IFrameMessages.OBJECT_VIEWER_GET_OBJECT_VIEW_RESPONSE_CMD,
                            objectView: objectView,
                            stringValue: message.stringValue
                        }, source);
                    });

                    break;

                case IFrameMessages.OBJECT_VIEWER_GET_OBJECT_PROPERTIES_ROWS_REQUEST_CMD:
                    ObjectViewer.getObjectPropertiesRows(message.stringValue, message.getInheritedProperties, null, function (rows) {
                        MessageSandbox.sendServiceMsg({
                            cmd: IFrameMessages.OBJECT_VIEWER_GET_OBJECT_PROPERTIES_ROWS_RESPONSE_CMD,
                            rows: rows,
                            stringValue: message.stringValue
                        }, source);
                    });

                    break;
            }
        });
    };

    RecorderBase.prototype.loadStepsInfo = function () {
    };

    RecorderBase.prototype.start = function (initIFrameNativeDialogs) {
        var recorder = this;

        this.recording = true;
        this.started = true;

        CursorWidget.init();
        this._initUi();
        this._initRecordingUiEventHandling();
        this._initHammerheadEventHandling();

        Automation.init();
        JavaScriptExecutor.init();

        EventListener.start(function (actionDescriptor) {
            recorder._onActionParsed(actionDescriptor);
        }, this.recordingManagerOptions);

        for (var i = 0; i < this.waitingIframeRecorders.length; i++)
            MessageSandbox.sendServiceMsg({
                cmd: RecorderBase.START_IFRAME_RECORDER_CMD,
                initNativeDialogs: initIFrameNativeDialogs
            }, this.waitingIframeRecorders[i]);
    };

    RecorderBase.prototype._saveDialogHandler = function () {
    };

    RecorderBase.prototype.initNativeDialogsHandling = function () {
        var recorder = this;

        var nativeDialogs = this.nativeDialogs || RecorderUtil.getNativeDialogs();

        window.alert = function (message) {
            EventListener.checkActionsBeforeAdditionNativeDialogHandler();
            recorder._saveDialogHandler('alert');
            return nativeDialogs.alert(message);
        };

        window.confirm = function (message) {
            EventListener.checkActionsBeforeAdditionNativeDialogHandler();
            var ret = nativeDialogs.confirm(message);
            recorder._saveDialogHandler('confirm', ret);
            return ret;
        };

        window.prompt = function (message) {
            EventListener.checkActionsBeforeAdditionNativeDialogHandler();
            var ret = nativeDialogs.prompt(message);
            recorder._saveDialogHandler('prompt', ret);
            return ret;
        };

        Hammerhead.on(Hammerhead.BEFORE_BEFORE_UNLOAD_EVENT, function (e) {
            if (!e.isFakeIEEvent) {
                EventListener.createAllDeferred();
                recorder.beforeUnloadEventWasRaised = true;
            }
        });

        Hammerhead.on(Hammerhead.BEFORE_UNLOAD_EVENT, function (e) {
            if (recorder.recording && !Settings.PLAYBACK && e.prevented && !e.isFakeIEEvent)
                recorder._saveDialogHandler('beforeUnload');

            if (!recorder.recording)
                recorder.recording = true;
        });

        Hammerhead.on(Hammerhead.UNLOAD_EVENT, function () {
            EventListener.stop();
        });
    };

    RecorderBase.prototype._sendMessageToAllIFrame = function (msg) {
        for (var i = 0; i < this.iFrameListeners.length; i++){
            try {
                MessageSandbox.sendServiceMsg(msg, this.iFrameListeners[i]);
            } catch (e) {
                // Hack for MS Edge: after an iframe was removed from DOM, some content windows preventing access
                // to all their properties, throwing 'Permission denied' error
                if (e.message && e.message.indexOf('Permission denied') > -1)
                    this.iFrameListeners.splice(i, 1);
                else
                    throw e;
            }
        }

    };

    RecorderBase.prototype._startPickingElement = function () {
        this.isPlayback = Settings.PLAYBACK;

        Settings.PLAYBACK = false;
    };

    RecorderBase.prototype._stopPickingElement = function () {
        Settings.PLAYBACK = this.isPlayback;

        this.elementPickCallback = null;

        ElementPicker.stop();
    };

    RecorderBase.prototype._saveStep = function () {
    };

    RecorderBase.prototype._createAllDeferred = function () {
        EventListener.createAllDeferred();
    };

    //Ui
    RecorderBase.prototype._initUi = function () {
    };

    RecorderBase.prototype._initRecordingUiEventHandling = function () {
    };

    // Upload
    RecorderBase.prototype._initHammerheadEventHandling = function () {
    };

    RecorderBase.prototype._onUploadRequested = function () {
    };

    //Recording management
    RecorderBase.prototype._confirmAction = function () {
    };

    RecorderBase.prototype._onActionParsed = function (actionDescriptor, dialogRequired) {
        var recorder = this,
            parsedStepInfo = RecorderUtil.getActionStepInfo(actionDescriptor);

        if (!dialogRequired)
            this._saveStep(parsedStepInfo);
        else {
            this._confirmAction(parsedStepInfo, function (confirmed, stepInfo) {
                if (confirmed)
                    recorder._saveStep(stepInfo);
            });
        }
    };
});