TestCafeClient.define('Recorder.SelectorGenerator', function (require, exports) {
        var Hammerhead = HammerheadClient.get('Hammerhead'),
            $ = Hammerhead.$,
            Util = Hammerhead.Util,
            GeneratorRules = require('Recorder.SelectorGenerator.Rules'),
            JavascriptExecutor = require('Base.JavascriptExecutor');

        var generatingRules = [],
            rulePriorities = {},
            initialized = false,

        //selector by tags tree becomes more of a priority when number of elements found by other selectors more than:
            ANCESTOR_AND_TAGS_TREE_RULE_VALIDITY_BOUND = 10,
            TAGS_TREE_RULE_VALIDITY_BOUND = 20;

        var setRulePriority = function (ruleID, priority) {
            if (priority > 0) {
                for (var i = 0; i < generatingRules.length; i++)
                    if (rulePriorities[generatingRules[i].id] === priority && generatingRules[i] !== this) {
                        setRulePriority(generatingRules[i].id, priority + 1);
                        break;
                    }
                rulePriorities[ruleID] = priority;
            }
            else
                rulePriorities[ruleID] = 0;
        };

        var getRulePriority = function (ruleID) {
            return rulePriorities[ruleID] || 0;
        };

        exports.getRulePriorities = function () {
            return rulePriorities;
        };

        //priority is incremented by this value for each element found by selector except the desired
        var getPriorityShift = function () {
            return Math.pow(10, generatingRules.length.toString().length);
        };

        var initPriorities = function () {
            rulePriorities = {};
            setRulePriority(GeneratorRules.ruleIDs.FOR_HTML, 1);
            setRulePriority(GeneratorRules.ruleIDs.FOR_BODY, 2);
            setRulePriority(GeneratorRules.ruleIDs.BY_ID, 3);
            setRulePriority(GeneratorRules.ruleIDs.BY_INPUT_NAME_AND_FORM, 4);
            setRulePriority(GeneratorRules.ruleIDs.BY_TEXT, 5);
            setRulePriority(GeneratorRules.ruleIDs.BY_ANCESTOR_AND_TEXT, 6);
            setRulePriority(GeneratorRules.ruleIDs.BY_ATTR, 7);
            setRulePriority(GeneratorRules.ruleIDs.BY_ANCESTOR_AND_ATTR, 8);

            setRulePriority(GeneratorRules.ruleIDs.BY_ANCESTOR_WITH_ATTR_AND_TAGS_TREE, getPriorityShift() * ANCESTOR_AND_TAGS_TREE_RULE_VALIDITY_BOUND);
            setRulePriority(GeneratorRules.ruleIDs.BY_ANCESTOR_WITH_TEXT_AND_TAGS_TREE, getRulePriority(GeneratorRules.ruleIDs.BY_ANCESTOR_WITH_ATTR_AND_TAGS_TREE) + 1);
            setRulePriority(GeneratorRules.ruleIDs.BY_TAGS_TREE, getPriorityShift() * TAGS_TREE_RULE_VALIDITY_BOUND);
        };

        var getSelectorPriority = function ($el, $selectorResult, rule) {
            var wasElementChanged = rule.wasElementChanged($el);

            if ($selectorResult.length === 1 && !wasElementChanged && $selectorResult[0] === $el[0])
                return getRulePriority(rule.id);
            else if ($selectorResult.length === 0 && wasElementChanged)
                return getRulePriority(rule.id);
            else if ($selectorResult.length && (wasElementChanged || $selectorResult.index($el) !== -1)) {
                return getRulePriority(rule.id) + getPriorityShift() * ($selectorResult.length + (wasElementChanged ? 1 : 0));
            }
            else return 0;
        };

        var sortSelectorsByPriority = function (selectorObjects, selectorPriorities) {
            selectorObjects.sort(function (obj1, obj2) {
                if (selectorPriorities[obj1.selector] < selectorPriorities[obj2.selector])
                    return -1;
                else if (selectorPriorities[obj1.selector] > selectorPriorities[obj2.selector])
                    return 1;
                else return 0;
            });
        };

        var getContextString = function ($el) {
            if (Util.isElementInIframe($el[0])) {
                var iframe = Util.getIFrameByElement($el[0]);
                if (iframe) {
                    var iFrameSelectors = exports.generate($(iframe));

                    if (iFrameSelectors.length)
                        return iFrameSelectors[0].selector + '.contents()';
                }
            }
            return '';
        };

        var buildJQueryExpression = function ($el, elementSelector, ancestorExpression) {
            if (ancestorExpression)
                return [
                    ancestorExpression, '.find("', elementSelector, '")'
                ].join('');
            else {
                var contextString = getContextString($el);
                if (contextString)
                    return [
                        '$("', elementSelector, '", ', contextString, ')'
                    ].join('');
                else return [
                    '$("', elementSelector, '")'
                ].join('');
            }
        };

        var generateJQueryExpressionByRule = function ($el, generatingRule) {
            var generatingRulesResult = generatingRule.generate($el),
                ancestorExpression = '';

            if (typeof generatingRulesResult === 'string')
                return generatingRulesResult;

            if (generatingRulesResult.ancestor)
                ancestorExpression = generateJQueryExpressionByRule(generatingRulesResult.ancestor.$el, generatingRulesResult.ancestor.rule);

            return buildJQueryExpression($el, generatingRulesResult.selector, ancestorExpression);
        };

        var init = function () {
            generatingRules = GeneratorRules.getRules();
            initPriorities();
            initialized = true;
        };

        exports.generate = function ($element) {
            if (!initialized)
                init();

            var selectorObjects = [];
            var selectorPriorities = {};

            if (!$element.length)
                return selectorObjects;

            for (var i = 0; i < generatingRules.length; i++) {
                if (generatingRules[i].check($element)) {
                    var selector = generateJQueryExpressionByRule($element, generatingRules[i]),
                        $selectorResult = JavascriptExecutor.parseSelectorSync(selector).$elements;

                    if (!selector || !$selectorResult)
                        continue;
                    var priority = getSelectorPriority($element, $selectorResult, generatingRules[i]);
                    var description = generatingRules[i].getDescription(selector);
                    var selectorObject = {id: generatingRules[i].id, selector: selector, description: description};

                    GeneratorRules.makeSelectorUnique(selectorObject, $element, $selectorResult);

                    if (priority > 0) {
                        if (!selectorPriorities[selectorObject.selector]) {
                            selectorObjects.push(selectorObject);
                            selectorPriorities[selectorObject.selector] = priority;
                        }
                        else if (priority < selectorPriorities[selectorObject.selector]) {
                            selectorPriorities[selectorObject.selector] = priority;
                            for (var j = 0; j < selectorObjects.length; j++)
                                selectorObjects[j] = selectorObject;
                        }
                    }
                }
            }

            sortSelectorsByPriority(selectorObjects, selectorPriorities);

            return selectorObjects;
        };
    }
)
;