TestCafeClient.define('Recorder.Util', function (require, exports) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        JSON = Hammerhead.JSON,
        StepNameGenerator = require('Recorder.StepNameGenerator'),
        ServiceCommands = require('Shared.ServiceCommands'),
        Transport = require('Base.Transport');

    exports.ELEMENT_PICKING_CLASSES = {
        elementFrame: 'element-frame',
        elementFramePart: 'element-frame-part'
    };

    exports.ELEMENTS_MARKER_CLASS = 'elements-marker';

    exports.BROWSER_DIALOG_INDEXES = {
        alert: 0,
        confirm: 1,
        prompt: 2,
        beforeUnload: 3
    };

    exports.getActionStepInfo = function (actionDescriptor) {
        var element = actionDescriptor.element;

        if (element)
            actionDescriptor.element = null;    //B253740

        var actionDescriptorClone = $.extend(true, {}, actionDescriptor);

        if (element) {
            actionDescriptor.element = element;
            actionDescriptorClone.element = element;
        }

        return {
            name: StepNameGenerator.generateStepName(actionDescriptor),
            actionDescriptor: actionDescriptorClone,
            selectors: actionDescriptor.serviceInfo && actionDescriptor.serviceInfo.selectors,
            iFrameSelectors: null,
            currentSelectorIndex: 0,
            currentIFrameSelectorIndex: 0,
            useOffsets: actionDescriptor.serviceInfo && actionDescriptor.serviceInfo.useOffsets
        };
    };

    exports.getAssertionsStepInfo = function () {
        return {
            isAssertion: true,
            name: 'Assert',
            blocks: [
                /*{
                 assertions: [],
                 iFrameContext: 'context' || null
                 }, ...*/
            ]
        };
    };

    //NOTE: we should change string values in action arguments to integer values
    exports.parseStepArguments = function (stepInfo) {
        if (stepInfo.isAssertion)
            return;

        var actionDescr = stepInfo.actionDescriptor;

        if (actionDescr.type === 'drag') {
            if (actionDescr.apiArguments.dragOffsetX === '')
                actionDescr.apiArguments.dragOffsetX = 0;
            if (actionDescr.apiArguments.dragOffsetX === '')
                actionDescr.apiArguments.dragOffsetX = 0;
        }

        if (actionDescr.type === 'wait') {
            if (actionDescr.apiArguments.ms === '')
                actionDescr.apiArguments.ms = 0;
        }
    };

    //NOTE: modify stepInfo to parsing on server
    exports.getServerCompatibleStepInfo = function (stepInfo) {
        if (stepInfo.isAssertion)
            return stepInfo;
        else
            return exports.getServerCompatibleActionStepInfo(stepInfo);
    };

    exports.sendStepsInfo = function (stepsInfo, hasUnsavedChanges, callback, getStepsScript) {
        var res = [];

        for (var i = 0; i < stepsInfo.length; i++)
            res.push(exports.getServerCompatibleStepInfo(stepsInfo[i]));

        var setStepsInfoMsg = {
            cmd: ServiceCommands.STEPS_INFO_SET,
            stepsInfo: res,
            hasUnsavedChanges: hasUnsavedChanges,
            getStepsScript: getStepsScript
        };

        Transport.queuedAsyncServiceMsg(setStepsInfoMsg, callback);
    };

    exports.getServerCompatibleActionStepInfo = function (stepInfo) {
        var res = {
            action: stepInfo.actionDescriptor.type,
            actionArgs: [],
            failed: stepInfo.failed || false,
            error: stepInfo.error || null,
            dialogError: stepInfo.dialogError || null
        };

        //NOTE: We can not use jquery extend method because it causes an 'Permission denied' error
        // if the action contains an element of iframe that has been deleted.
        // Therefore, we should copy in an empty object all the necessary properties of stepInfo object (B253740)
        for (var option in stepInfo) {
            if (stepInfo.hasOwnProperty(option) && !/actiondescriptor|actionargs/.test(option.toLowerCase())) {
                //NOTE: all value passed by reference must be copied separately (such as selectors, nativeDialogHandlers)
                if (/selectors/.test(option.toLowerCase()) && typeof stepInfo[option] !== 'undefined')
                    res[option] = (stepInfo[option] && stepInfo[option].length) ? $.extend(true, [], stepInfo[option]) : null;
                if (/nativedialoghandlers/.test(option.toLowerCase()) && typeof stepInfo[option] !== 'undefined')
                    res[option] = (stepInfo[option] && stepInfo[option].length) ? $.extend(true, [], stepInfo[option]) : null;
                else
                    res[option] = stepInfo[option];
            }
        }

        var apiArguments = stepInfo.actionDescriptor.apiArguments,
            actionOptions = null;

        if (stepInfo.actionDescriptor.type === 'drag') {
            res.actionArgs.push({ dragOffsetX: apiArguments.dragOffsetX});
            res.actionArgs.push({ dragOffsetY: apiArguments.dragOffsetY });
        }

        if (stepInfo.actionDescriptor.type === 'type')
            res.actionArgs.push({ text: apiArguments.text });

        if (stepInfo.actionDescriptor.type === 'press')
            res.actionArgs.push({ keysCommand: apiArguments.keysCommand });

        if (stepInfo.actionDescriptor.type === 'wait')
            res.actionArgs.push({ ms: apiArguments.ms });

        if (stepInfo.actionDescriptor.type === 'select') {
            res.actionArgs.push({ startPos: apiArguments.startPos });
            res.actionArgs.push({ endPos: apiArguments.endPos });
        }

        if (stepInfo.actionDescriptor.type === 'upload')
            res.actionArgs.push({ files: apiArguments.files});

        if (/click|drag|type|hover/.test(stepInfo.actionDescriptor.type) && apiArguments.options) {
            actionOptions = $.extend(true, {}, apiArguments.options);

            if (actionOptions)
                res.actionArgs.push({ options: actionOptions});
        }

        return res;
    };

    exports.parseServerCompatibleStepInfo = function (stepInfo) {
        if (stepInfo.isAssertion)
            return stepInfo;
        else
            return exports.parseServerCompatibleActionStepInfo(stepInfo);
    };

    exports.parseServerCompatibleActionStepInfo = function (stepInfo) {
        var res = $.extend(true, {}, stepInfo, {
            actionDescriptor: {
                type: stepInfo.action,
                apiArguments: {},
                failed: stepInfo.failed,
                error: null
            }
        });

        for (var i = 0; i < stepInfo.actionArgs.length; i++) {
            var arg = stepInfo.actionArgs[i];

            for (var field in arg) {
                if (arg.hasOwnProperty(field))
                    res.actionDescriptor.apiArguments[field] = arg[field];
            }
        }

        return res;
    };

    var storedStates = {};

    exports.serializeStepInfo = function (stepInfo) {
        if (!stepInfo.id)
            stepInfo.id = Date.now();

        if (stepInfo.actionDescriptor) {
            if (stepInfo.actionDescriptor.serviceInfo && stepInfo.actionDescriptor.serviceInfo.prevPageState) {
                storedStates[stepInfo.id] = stepInfo.actionDescriptor.serviceInfo.prevPageState;
                stepInfo.actionDescriptor.serviceInfo.prevPageState = null;
            }

            stepInfo.actionDescriptor.element = null;
        }

        return JSON.stringify(stepInfo);
    };

    exports.parseStepInfo = function (serializedStepInfo) {
        var stepInfo = JSON.parse(serializedStepInfo);

        if (stepInfo.id && storedStates[stepInfo.id])
            stepInfo.actionDescriptor.serviceInfo.prevPageState = storedStates[stepInfo.id];

        return stepInfo;
    };

    exports.getNativeDialogs = function () {
        return (function (_alert, _confirm, _prompt) {
            return {
                alert: function (message) {
                    return _alert.call(window, message);
                },
                confirm: function (message) {
                    return _confirm.call(window, message);
                },
                prompt: function (message) {
                    return _prompt.call(window, message);
                }
            };
        })(window.alert, window.confirm, window.prompt);
    };
});