TestCafeClient.define('UI.RecorderWidgets.ActionStep', function (require) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        Util = Hammerhead.Util,
        ShadowUI = Hammerhead.ShadowUI,

        SharedErrors = require('Shared.Errors'),
        SharedConst = require('Shared.Const'),

        BaseStep = require('UI.RecorderWidgets.BaseStep'),
        TextSelection = Hammerhead.TextSelection,
        ComboboxWidget = require('UI.RecorderWidgets.Combobox'),
        DialogPropertyWidget = require('UI.RecorderWidgets.DialogProperty'),
        ActionOptionsEditorWidget = require('UI.RecorderWidgets.ActionOptionsEditor'),
        ActionDialogsEditorWidget = require('UI.RecorderWidgets.ActionDialogsEditor'),

        SelectorEditorWidget = require('UI.RecorderWidgets.SelectorEditor'),
        ElementsMarkerWidget = require('UI.RecorderWidgets.ElementsMarker');

    //Const
    var API_ARGUMENTS_ORDERS = {
            click: ['options'],
            rclick: ['options'],
            dblclick: ['options'],
            drag: ['options', 'dragOffsetX', 'dragOffsetY'],
            hover: ['options'],
            type: ['options', 'text'],
            press: ['keysCommand'],
            wait: ['ms'],
            select: ['startPos', 'endPos'],
            upload: ['files'],
            screenshot: []
        },

        ACTION_STEP_CLASS = 'action-step',
        SELECTOR_EDITOR_CONTAINER_CLASS = 'selector-editor-container',
        ARGUMENTS_CLASS = 'arguments',
        OPTIONS_CLASS = 'options',
        ACTION_TYPE_CLASS = 'action-type',
        ACTION_ICON_CLASS = 'action-icon',
        PROPERTY_CLASS = 'dialog-property',
        SELECTOR_AREA_CLASS = 'selector-area',
        ARGUMENTS_CONTAINER_CLASS = 'arguments-container',

        IFRAME_SELECTOR_LABEL_TEXT = 'IFrame selector:',
        ELEMENT_SELECTOR_LABEL_TEXT = 'Target selector:',

        ELEMENT_SELECTOR_HEIGHT = 100,
        IFRAME_SELECTOR_HEIGHT = 100;

    //ActionStep
    var ActionStep = this.exports = function ($container, stepNum, stepInfo, options) {
        options = options || {};
        this.options = {
            enableFloatMode: options.enableFloatMode || false,
            $floatingParent: options.$floatingParent || null,
            enableSelectorValidation: options.enableSelectorValidation || false,
            enableElementsMarking: options.enableElementsMarking || false,
            stretchSelectorEditor: options.stretchSelectorEditor || false,
            allowVisibleElementsOnly: options.allowVisibleElementsOnly || false,
            dialogError: options.dialogError || stepInfo.dialogError
        };

        BaseStep.apply(this, [$container, stepNum, stepInfo, this.options]);

        this.actionDescriptor = stepInfo.actionDescriptor;

        this.argumentPropertyEditors = [];
        this.optionsEditor = null;
        this.dialogsEditor = null;
        this.selectorEditor = null;
        this.iFrameSelectorEditor = null;

        this.$argumentsContainer = null;

        this._createMarkup();
        this._init();
    };

    Util.inherit(ActionStep, BaseStep);

    //Events
    ActionStep.STEP_NAME_CHANGED_EVENT = BaseStep.STEP_NAME_CHANGED_EVENT;
    ActionStep.STEP_INFO_CHANGED_EVENT = BaseStep.STEP_INFO_CHANGED_EVENT;
    ActionStep.SELECTOR_EDITOR_FOCUSED_EVENT = 'selectorEditorFocused';
    ActionStep.SELECTOR_EDITOR_BLURED_EVENT = 'selectorEditorBlured';

    //Util
    function getPropertyOptions(propertyName) {
        return {
            isReadOnly: /^files/i.test(propertyName),
            isKeysProperty: /^key/.test(propertyName.toLowerCase()),
            isNumericProperty: !(/text|^key/.test(propertyName.toLowerCase())),
            isPositiveNumericProperty: (/^ms$/.test(propertyName.toLowerCase())),
            maximizeInput: propertyName.toLowerCase() === 'text',
            smallInput: /startPos|endPos|ms|keys/i.test(propertyName)
        };
    }

    //Markup
    ActionStep.prototype._createMarkup = function () {
        ShadowUI.addClass(this.$step, ACTION_STEP_CLASS);
        this._createBackground();
        this._createArguments(this.$step);

        if (typeof this.stepInfo.selectors !== 'undefined')
            this._createSelectorArea(this.$step);
    };

    ActionStep.prototype._createComboBoxProperty = function ($container, name, value) {
        var $property = $('<div></div>').appendTo($container);
        ShadowUI.addClass($property, PROPERTY_CLASS);

        var inputId = SharedConst.PROPERTY_PREFIX + name + Math.random().toString().replace('0.', '-');

        this._createLabel($property, name, inputId);
        return new ComboboxWidget($property, ['true', 'false'], value, inputId);
    };

    ActionStep.prototype._createArgumentsProperty = function ($container, name, value, obj) {
        var step = this;

        var property = new DialogPropertyWidget($container, this._getArgumentPropertyLabelText(name), value, getPropertyOptions(name));

        this.argumentPropertyEditors.push(property);

        property.on(DialogPropertyWidget.VALUE_CHANGED_EVENT, function (e) {
            obj[name] = e.value;

            step._onChangePropertyValue();
        });
    };

    ActionStep.prototype._getArgumentPropertyLabelText = function (propertyName) {
        if (/keyscommand/i.test(propertyName))
            return 'Keys:';
        else
            return propertyName + ':';
    };

    ActionStep.prototype._createArgumentsContainer = function ($container) {
        this.$argumentsContainer = $('<div></div>').appendTo($container);
        ShadowUI.addClass(this.$argumentsContainer, ARGUMENTS_CONTAINER_CLASS);
    };


    ActionStep.prototype._createOptionsEditor = function ($container) {
        var step = this;

        if (!this.$argumentsContainer)
            this._createArgumentsContainer($container);

        this.optionsEditor = new ActionOptionsEditorWidget(this.$argumentsContainer, this.stepInfo);

        this.optionsEditor.on(ActionOptionsEditorWidget.OPTION_CHANGED_EVENT, function () {
            step._onChangePropertyValue();
        });
    };

    ActionStep.prototype._createBrowserDialogsEditor = function ($container) {
        var step = this;

        if (!this.$argumentsContainer)
            this._createArgumentsContainer($container);

        this.dialogsEditor = new ActionDialogsEditorWidget(this.$argumentsContainer, this.stepInfo);

        this.dialogsEditor.on(ActionDialogsEditorWidget.OPTION_CHANGED_EVENT, function (e) {
            if (step._hasNativeDialogError() && e.dialog) {
                if ((step.options.dialogError.code === SharedErrors.API_EXPECTED_DIALOG_DOESNT_APPEAR && step.options.dialogError.dialog === e.dialog && !e.state) ||
                    (step.options.dialogError.code === SharedErrors.API_UNEXPECTED_DIALOG && step.options.dialogError.dialog === e.dialog && e.state)) {
                    step.options.dialogError = null;
                    step.stepInfo.dialogError = null;
                    step.dialogsEditor.removeError();
                }
            }

            step._onChangePropertyValue();
        });
    };

    ActionStep.prototype._createArguments = function ($container) {
        var step = this;

        var $arguments = $('<div></div>').appendTo($container);
        ShadowUI.addClass($arguments, ARGUMENTS_CLASS);

        var $actionType = $('<div></div>').appendTo($arguments);
        ShadowUI.addClass($actionType, ACTION_TYPE_CLASS);

        var actionName = this.actionDescriptor.type.toLowerCase();
        $('<span></span>').text(actionName).appendTo($actionType);
        var $icon = $('<div></div>').appendTo($actionType);
        ShadowUI.addClass($icon, ACTION_ICON_CLASS);
        ShadowUI.addClass($icon, actionName);

        var $options = null,
            apiArgumentOrder = API_ARGUMENTS_ORDERS[step.actionDescriptor.type],
            apiArguments = step.actionDescriptor.apiArguments;

        if (!apiArgumentOrder.length)
            step._createBrowserDialogsEditor($arguments);

        for (var i = 0; i < apiArgumentOrder.length; i++) {
            var argName = apiArgumentOrder[i],
                argOptions = getPropertyOptions(argName),
                argValue = apiArguments[argName],
                isArgValueEmpty = argValue ? (argValue.join ? !argValue.join('') : !argValue) : true;

            if (argOptions.isReadOnly && isArgValueEmpty)
                continue;

            if (argValue !== null) {
                if (argName === 'options') {
                    step._createOptionsEditor($arguments);
                    step._createBrowserDialogsEditor($arguments);
                }
                else {
                    if (!this.dialogsEditor)
                        step._createBrowserDialogsEditor($arguments);

                    if (!$options) {
                        $options = $('<div>').appendTo($arguments);
                        ShadowUI.addClass($options, OPTIONS_CLASS);
                    }
                    step._createArgumentsProperty($options, argName, argValue, apiArguments);
                }
            }
        }
    };

    ActionStep.prototype._createSelectorArea = function ($container) {
        var $selectorArea = $('<div></div>').appendTo($container);

        ShadowUI.addClass($selectorArea, SELECTOR_AREA_CLASS);

        if (this.stepInfo.iFrameSelectors)
            this._createIFrameSelectorEditor($selectorArea);

        this._createElementSelectorEditor($selectorArea);
    };

    ActionStep.prototype._createSelectorProperty = function ($container, text) {
        var $property = $('<div></div>').appendTo($container);
        ShadowUI.addClass($property, PROPERTY_CLASS);

        var $label = this._createLabel($property, text);
        $label.css('textTransform', 'none');
    };

    ActionStep.prototype._createIFrameSelectorEditor = function ($container) {
        this._createSelectorProperty($container, IFRAME_SELECTOR_LABEL_TEXT);

        var $iFrameSelectorEditorContainer = $('<div></div>').appendTo($container);
        ShadowUI.addClass($iFrameSelectorEditorContainer, SELECTOR_EDITOR_CONTAINER_CLASS);

        this.iFrameSelectorEditor = new SelectorEditorWidget($iFrameSelectorEditorContainer, {
            width: '100%',
            height: IFRAME_SELECTOR_HEIGHT,
            selectors: this.stepInfo.iFrameSelectors,
            currentSelectorIndex: this.stepInfo.currentIFrameSelectorIndex,
            allowEdit: true,
            enableFloatMode: false,
            enableValidation: false,
            enableElementsMarking: false
        });
    };

    ActionStep.prototype._createElementSelectorEditor = function ($container) {
        this._createSelectorProperty($container, ELEMENT_SELECTOR_LABEL_TEXT);

        var $editorContainer = $('<div></div>').appendTo($container);
        ShadowUI.addClass($editorContainer, SELECTOR_EDITOR_CONTAINER_CLASS);

        var context = (this.iFrameSelectorEditor && this.iFrameSelectorEditor.getParsedSelector().length) ?
            this.iFrameSelectorEditor.getParsedSelector().$elements[0].contentWindow :
            null;

        this.selectorEditor = new SelectorEditorWidget($editorContainer, {
            width: '100%',
            height: ELEMENT_SELECTOR_HEIGHT,
            selectors: this.stepInfo.selectors,
            currentSelectorIndex: this.stepInfo.currentSelectorIndex,
            allowEdit: true,
            enableFloatMode: this.options.enableFloatMode,
            $floatingParent: this.options.$floatingParent,
            enableValidation: this.options.enableSelectorValidation,
            enableElementsMarking: this.options.enableElementsMarking,
            allowVisibleElementsOnly: this.options.allowVisibleElementsOnly,
            context: context
        });
    };

    ActionStep.prototype._createBackground = function () {
        if (this.options.enableElementsMarking)
            ElementsMarkerWidget.mark(this.options.$floatingParent || this.$step, $());
    };

    //Behavior
    ActionStep.prototype._init = function () {
        var step = this;

        if (this._hasNativeDialogError())
            this.dialogsEditor.showPopup();

        if (typeof step.stepInfo.selectors !== 'undefined')
            this._initSelectorEditor();

        if (this.iFrameSelectorEditor)
            this._initIFrameSelectorEditor();

        if (this.selectorEditor) {
            this.selectorEditor.on(SelectorEditorWidget.FOCUS_EVENT, function () {
                step.eventEmitter.emit(ActionStep.SELECTOR_EDITOR_FOCUSED_EVENT);
            });

            this.selectorEditor.on(SelectorEditorWidget.BLUR_EVENT, function () {
                step.eventEmitter.emit(ActionStep.SELECTOR_EDITOR_BLURED_EVENT);
            });
        }

        if (this.actionDescriptor.type === 'wait')
            this._initWaitActionDialogSpecificBehavior();
    };

    ActionStep.prototype._initWaitActionDialogSpecificBehavior = function () {
        var step = this,
            msInput = null;

        function getAutoGeneratedStepName(ms) {
            return 'Wait ' + ms + ' milliseconds';
        }

        function isStepNameAutoGenerated(ms, message) {
            return getAutoGeneratedStepName(ms) === message;
        }

        for (var i = 0; i < this.argumentPropertyEditors.length; i++)
            if (/^ms:$/i.test(this.argumentPropertyEditors[i].getLabelText()))
                msInput = this.argumentPropertyEditors[i].getInputElement();

        msInput.focus();
        TextSelection.select(msInput);

        var savedValue = msInput.value;

        $(msInput).bind('input', function () {
            if (isStepNameAutoGenerated(savedValue, step.stepNamePropertyEditor.getEditorText())) {
                step.stepNamePropertyEditor.setEditorText(getAutoGeneratedStepName(msInput.value));
            }

            savedValue = msInput.value;
        });
    };

    ActionStep.prototype._initIFrameSelectorEditor = function () {
        var step = this;

        this.iFrameSelectorEditor.on(SelectorEditorWidget.SELECTOR_CHANGED_EVENT, function (e) {
            step.stepInfo.currentIFrameSelectorIndex = e.index;

            step.eventEmitter.emit(ActionStep.STEP_INFO_CHANGED_EVENT, {
                stepInfo: step.stepInfo
            });
        });
    };

    ActionStep.prototype._initSelectorEditor = function () {
        var step = this;

        this.selectorEditor.on(SelectorEditorWidget.SELECTOR_CHANGED_EVENT, function (e) {
            step.actionDescriptor.selector = e.parsedSelector.selector;
            //NOTE: we store element in descriptor before it's changed only
            step.actionDescriptor.element = null;

            step.stepInfo.currentSelectorIndex = e.index;

            step.eventEmitter.emit(ActionStep.STEP_INFO_CHANGED_EVENT, {
                selector: e.text
            });
        });
    };

    ActionStep.prototype._onChangePropertyValue = function () {
        this.eventEmitter.emit(ActionStep.STEP_INFO_CHANGED_EVENT, {
            stepInfo: this.stepInfo
        });
    };

    //Validation
    ActionStep.prototype._hasPropertyError = function () {
        for (var i = 0; i < this.argumentPropertyEditors.length; i++)
            if (!this.argumentPropertyEditors[i].isValid())
                return true;
        return false;
    };

    ActionStep.prototype._hasNativeDialogError = function () {
        return  this.options.dialogError && this.options.dialogError.dialog &&
            (this.options.dialogError.code === SharedErrors.API_EXPECTED_DIALOG_DOESNT_APPEAR || this.options.dialogError.code === SharedErrors.API_UNEXPECTED_DIALOG);
    };

    //API
    ActionStep.prototype.getParsedSelector = function () {
        return this.selectorEditor ? this.selectorEditor.getParsedSelector() : null;
    };

    ActionStep.prototype.isValid = function (skipPageStateDependentErrors) {
        if (this._hasPropertyError())
            return false;
        else if (!skipPageStateDependentErrors && this._hasNativeDialogError())
            return false;
        else if (this.optionsEditor && !this.optionsEditor.isValid())
            return false;
        else if (this.selectorEditor) {
            if (this.iFrameSelectorEditor && !this.iFrameSelectorEditor.isValid(skipPageStateDependentErrors))
                return false;

            return this.selectorEditor.isValid(skipPageStateDependentErrors);
        }
        else
            return true;
    };

    ActionStep.prototype.destroy = function () {
        if (this.selectorEditor)
            this.selectorEditor.destroy();

        ElementsMarkerWidget.clear();
    };
});