TestCafeClient.define('UI.RecorderWidgets.Assertion', function (require) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        Util = Hammerhead.Util,
        ShadowUI = Hammerhead.ShadowUI,
        MessageSandbox = Hammerhead.MessageSandbox,
        IFrameMessages = require('Base.CrossDomainMessages'),
        JavascriptExecutor = require('Base.JavascriptExecutor'),
        AssertionsAPI = require('TestRunner.API.Assertions'),
        ButtonWidget = require('UI.RecorderWidgets.Button'),

        OPERATOR_ICON_CLASSES = {
            eq: 'eq-icon',
            notEq: 'not-eq-icon',
            ok: 'ok-icon',
            notOk: 'not-ok-icon'
        },

        ASSERTION_CLASS = 'assertion',
        OPERATOR_AREA_CLASS = 'operator-area',
        OPERATOR_ICON_CLASS = 'operator-icon',
        SUCCESSFUL_CLASS = 'successful',
        FAILED_CLASS = 'failed',
        MESSAGE_AREA_CLASS = 'message-area',
        REMOVE_ASSERTION_BUTTON_AREA_CLASS = 'remove-assertion-button-area';

    //Utils
    function addEventEmitter(target) {
        target.eventEmitter = new Util.EventEmitter();
        target.on = function (ev, listener) {
            target.eventEmitter.on(ev, listener);
        };
    }

    function getAssertionAutoMessage(assertionInfo) {
        var actual = assertionInfo.arguments[0],
            expected = assertionInfo.arguments[1];

        function generate(op, args) {
            var res = [op + '('];

            for (var i = 0; i < args.length; i++) {
                if (i)
                    res.push(', ');

                res.push(args[i] || '...');
            }

            res.push([')']);
            return res.join('');
        }

        return generate(assertionInfo.operator, expected ? [actual, expected] : [actual]);
    }

    var Assertion = this.exports = function ($assertion, assertionInfo, options) {
        var assertion = this;

        this.options = {
            enableValidation: options.enableValidation
        };

        this.assertionInfo = assertionInfo;
        this.$assertion = $assertion;
        this.$operatorArea = null;
        this.$messageArea = null;
        this.$removeButtonArea = null;
        this.isValid = options.isValid;
        this.iFrameSelector = options.iFrameSelector || null;
        this.curError = null;
        this.failed = assertionInfo.failed;

        ShadowUI.addClass(this.$assertion, ASSERTION_CLASS);

        addEventEmitter(this);

        this._createOperator(assertionInfo.operator);
        this._createMessage(assertionInfo.message || getAssertionAutoMessage(assertionInfo));
        this._createRemoveButton();

        this.assertionsAPI = new AssertionsAPI(function (err) {
            assertion.curError = err;
        });

        this.validate();

        this.$assertion.click(function () {
            assertion.eventEmitter.emit(Assertion.ASSERTION_CLICKED_EVENT, {});
        });
    };

//Assertion events
    Assertion.ASSERTION_CLICKED_EVENT = 'assertionClicked';
    Assertion.ASSERTION_REMOVED_EVENT = 'assertionRemoved';
    Assertion.VALIDITY_CHANGED_EVENT = 'validityChanged';

//Assertion public methods
    Assertion.prototype.update = function (assertionInfo, isValid, iFrameSelector) {
        if (this.assertionInfo.message !== assertionInfo.message)
            this._setMessage(assertionInfo.message || getAssertionAutoMessage(assertionInfo));

        this.assertionInfo.arguments = assertionInfo.arguments;
        this.assertionInfo.message = assertionInfo.message;
        this.assertionInfo.operator = assertionInfo.operator;

        this.assertionInfo = assertionInfo;
        this.isValid = isValid;
        this.iFrameSelector = iFrameSelector;
        this.failed = assertionInfo.failed;
        this._updateAssertionState();
    };

    Assertion.prototype.getContainer = function () {
        return this.$assertion;
    };

    Assertion.prototype.setIFrameSelector = function (iFrameSelector) {
        this.iFrameSelector = iFrameSelector;
    };

//Assertion private methods
    Assertion.prototype._createOperator = function (operator) {
        this.$operatorArea = $('<div></div>').appendTo(this.$assertion);
        ShadowUI.addClass(this.$operatorArea, OPERATOR_AREA_CLASS);

        var $operatorIcon = $('<div></div>').appendTo(this.$operatorArea);
        ShadowUI.addClass($operatorIcon, OPERATOR_ICON_CLASS);
        ShadowUI.addClass($operatorIcon, OPERATOR_ICON_CLASSES[operator]);
    };

    Assertion.prototype._createMessage = function (message) {
        this.$messageArea = $('<div></div>').appendTo(this.$assertion);
        ShadowUI.addClass(this.$messageArea, MESSAGE_AREA_CLASS);

        this._setMessage(message);
    };

    Assertion.prototype._createRemoveButton = function () {
        var assertion = this;

        this.$removeButtonArea = $('<div></div>').appendTo(this.$assertion);
        ShadowUI.addClass(this.$removeButtonArea, REMOVE_ASSERTION_BUTTON_AREA_CLASS);

        var $removeButton = ButtonWidget.create(this.$removeButtonArea, '', true);

        $removeButton.click(function () {
            assertion.eventEmitter.emit(Assertion.ASSERTION_REMOVED_EVENT, { assertion: assertion });
            return false;
        });
    };

    Assertion.prototype._setValidity = function (value) {
        var changed = value !== this.isValid;

        this.isValid = value;

        if (changed)
            this.eventEmitter.emit(Assertion.VALIDITY_CHANGED_EVENT, {
                isValid: value
            });
    };

    Assertion.prototype.validate = function () {
        var assertion = this;

        function onMessage(e) {
            if (e.message.cmd === IFrameMessages.ASSERT_RESPONSE_CMD) {
                assertion._setValidity(!e.message.err);
                assertion._updateAssertionState();

                MessageSandbox.off(MessageSandbox.SERVICE_MSG_RECEIVED, onMessage);
            }
        }

        if (!this.options.enableValidation)
            this._updateAssertionState();
        else {
            this.curError = null;

            if (this.iFrameSelector) {
                var $iFrame = JavascriptExecutor.parseSelectorSync(this.iFrameSelector).$elements,
                    iFrameContext = ($iFrame && $iFrame.length) ? $iFrame[0].contentWindow : null;

                if (iFrameContext) {
                    var msg = {
                        cmd: IFrameMessages.ASSERT_REQUEST_CMD,
                        args: this.assertionInfo.arguments,
                        operator: this.assertionInfo.operator
                    };

                    MessageSandbox.on(MessageSandbox.SERVICE_MSG_RECEIVED, onMessage);
                    MessageSandbox.sendServiceMsg(msg, iFrameContext);
                }
            }
            else {
                var assertionArguments = $.map(this.assertionInfo.arguments, function (argument) {
                    return argument ? JavascriptExecutor.eval(argument, function (err) {
                        assertion.curError = err;
                    }) : argument;
                });

                assertionArguments.push(this.assertionInfo.message);

                if (!this.curError)
                    this.assertionsAPI[this.assertionInfo.operator].apply(this.assertionsAPI, assertionArguments);

                assertion._setValidity(!this.curError);
                this._updateAssertionState();
            }
        }
    };

    Assertion.prototype._updateAssertionState = function () {
        if (this.failed) {
            ShadowUI.removeClass(this.$operatorArea, SUCCESSFUL_CLASS);
            ShadowUI.addClass(this.$operatorArea, FAILED_CLASS);
        }
        else {
            ShadowUI.removeClass(this.$operatorArea, SUCCESSFUL_CLASS);
            ShadowUI.removeClass(this.$operatorArea, FAILED_CLASS);
        }

        if (!this.options.enableValidation)
            return;

        if (this.isValid) {
            ShadowUI.removeClass(this.$operatorArea, FAILED_CLASS);
            ShadowUI.addClass(this.$operatorArea, SUCCESSFUL_CLASS);
        }
        else {
            ShadowUI.removeClass(this.$operatorArea, SUCCESSFUL_CLASS);
            ShadowUI.addClass(this.$operatorArea, FAILED_CLASS);
        }
    };

    Assertion.prototype._setMessage = function (message) {
        this.$messageArea.text(message || '');
    };
});