TestCafeClient.define('UI.RecorderWidgets.EditAssertionDialog', function (require) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        Util = Hammerhead.Util,
        ShadowUI = Hammerhead.ShadowUI,
        PopupWidget = require('UI.RecorderWidgets.Popup'),
        ButtonWidget = require('UI.RecorderWidgets.Button'),
        AssertionsAPI = require('TestRunner.API.Assertions'),
        AssertionArgumentEditor = require('UI.RecorderWidgets.AssertionArgument'),
        DialogPropertyWidget = require('UI.RecorderWidgets.DialogProperty'),
        IFrameContextLabel = require('Recorder.IFrameContextLabel'),
        JavascriptExecutor = require('Base.JavascriptExecutor'),

        DOUBLE_ARGUMENTS_DIALOG_WIDTH = 931,
        SINGLE_ARGUMENT_DIALOG_WIDTH = 447,

        START_PICKING_ELEMENT = 'startPickingElement',
        ACTUAL_ARGUMENT_NAME = 'Actual',
        EXPECTED_ARGUMENT_NAME = 'Expected',
        EXPRESSION_ARGUMENT_NAME = 'Expression',

        OPERATOR_ICON_CLASSES = {
            eq: 'eq-icon',
            notEq: 'not-eq-icon',
            ok: 'ok-icon',
            notOk: 'not-ok-icon'
        },

        RECORDER_CLASS = 'recorder',
        BUTTONS_CLASS = 'buttons',
        EDIT_ASSERTION_DIALOG = 'edit-assertion-dialog',
        SEPARATOR_CLASS = 'separator',
        MIDDLE_AREA_CLASS = 'middle-area',
        COPY_BUTTON_CLASS = 'copy-button',
        POPUP_HEADER_ICON_CLASS = 'header-operator-icon',
        SUCCESSFUL_CLASS = 'successful',
        FAILED_CLASS = 'failed',
        ARGUMENTS_AREA_CLASS = 'arguments-area',
        ARGUMENT_AREA_CLASS = 'argument-area',
        STATE_INDICATOR_CLASS = 'state-indicator',
        STATE_TEXT_CLASS = 'state-text',
        IFRAME_CONTEXT_CLASS = 'iframe-context';

    var EditAssertionDialog = this.exports = function (assertionInfo, options) {
        var dialog = this;

        options = options || {};
        this.options = {
            enableAssertionValidation: options.enableAssertionValidation
        };

        this.popup = null;
        this.$confirmButton = null;
        this.$cancelButton = null;
        this.$dialog = null;
        this.$middleArea = null;
        this.$copyButton = null;
        this.$stateIndicator = null;
        this.$stateText = null;
        this.$argumentsArea = null;
        this.messageEditor = null;
        this.expected = null;
        this.actual = null;
        this.operator = assertionInfo.operator;
        this.error = null;

        this.iFrameSelector = options.iFrameSelector || null;

        var $iFrame = this.iFrameSelector ? JavascriptExecutor.parseSelectorSync(this.iFrameSelector).$elements : null;
        this.iFrameContext = ($iFrame && $iFrame.length) ? $iFrame[0].contentWindow : null;


        this.eventEmitter = new Util.EventEmitter();
        this.on = function (ev, listener) {
            dialog.eventEmitter.on(ev, listener);
        };

        this._createDialog(assertionInfo);
        this._init();
        this._validate();

        if (this.iFrameSelector)
            this._createIFrameContextLabel(this.iFrameSelector);
    };

    //Assertion events
    EditAssertionDialog.CONFIRM_BUTTON_CLICK_EVENT = 'confirmButtonClick';
    EditAssertionDialog.CANCEL_BUTTON_CLICK_EVENT = 'cancelButtonClick';
    EditAssertionDialog.START_PICKING_ELEMENT_EVENT = START_PICKING_ELEMENT;

    EditAssertionDialog.prototype._createDialog = function (assertionInfo) {
        this.$dialog = $('<div></div>');
        ShadowUI.addClass(this.$dialog, EDIT_ASSERTION_DIALOG);

        var popupOptions = {
            width: assertionInfo.arguments.length === 2 ? DOUBLE_ARGUMENTS_DIALOG_WIDTH : SINGLE_ARGUMENT_DIALOG_WIDTH,
            headerText: 'Edit assertion',
            headerIconClass: POPUP_HEADER_ICON_CLASS + ' ' + OPERATOR_ICON_CLASSES[assertionInfo.operator],
            content: this.$dialog,
            footerContent: this._createButtons(),
            backgroundOpacity: true,
            showAtWindowCenter: true,
            hasParentPopup: true
        };

        var $popupParent = ShadowUI.select('.' + RECORDER_CLASS, ShadowUI.getRoot());

        this.popup = new PopupWidget($popupParent.length ? $popupParent : ShadowUI.getRoot(), popupOptions);

        this._createMessage(assertionInfo.message);
        this._createArgumentsArea(assertionInfo.arguments);

        if (this.options.enableAssertionValidation)
            this._createStateIndicator();

        this.popup.showAtWindowCenter();
    };

    EditAssertionDialog.prototype._createButtons = function () {
        var $buttons = $('<div></div>');
        ShadowUI.addClass($buttons, BUTTONS_CLASS);

        this.$confirmButton = ButtonWidget.create($buttons, 'OK');
        this.$cancelButton = ButtonWidget.create($buttons, 'Cancel');

        return $buttons;
    };

    EditAssertionDialog.prototype._createMessage = function (value) {
        this.messageEditor = new DialogPropertyWidget(this.$dialog, 'message', value, {maximizeInput: true});

        var $separator = $('<div></div>').appendTo(this.$dialog);
        ShadowUI.addClass($separator, SEPARATOR_CLASS);
    };

    EditAssertionDialog.prototype._createIFrameContextLabel = function (value) {
        var dialog = this,
            $labelContainer = $('<div></div>').prependTo(this.$argumentsArea);

        ShadowUI.addClass($labelContainer, ARGUMENT_AREA_CLASS);
        ShadowUI.addClass($labelContainer, IFRAME_CONTEXT_CLASS);

        IFrameContextLabel.create($labelContainer, value, function () {
            dialog.iFrameContext = dialog.iFrameSelector = null;
            dialog.actual.removeIFrameContext();

            if (dialog.expected)
                dialog.expected.removeIFrameContext();
        }, false);
    };

    EditAssertionDialog.prototype._createArgumentsArea = function (argumentValues) {
        var $argumentsArea = this.$argumentsArea = $('<div></div>').appendTo(this.$dialog);
        ShadowUI.addClass($argumentsArea, ARGUMENTS_AREA_CLASS);

        var $actualValueArea = $('<div></div>').appendTo($argumentsArea);
        ShadowUI.addClass($actualValueArea, ARGUMENT_AREA_CLASS);

        this.actual = this._createArgument($actualValueArea, argumentValues.length === 1 ? EXPRESSION_ARGUMENT_NAME : ACTUAL_ARGUMENT_NAME, argumentValues[0]);

        this.actual.editor.focus();
        this.actual.editor.setCursorToEnd();

        if (this.operator === 'eq' || this.operator === 'notEq')
            this._createMiddleArea($argumentsArea);

        if (argumentValues.length > 1) {
            var $expectedValueArea = $('<div></div>').appendTo($argumentsArea);
            ShadowUI.addClass($expectedValueArea, ARGUMENT_AREA_CLASS);
            this.expected = this._createArgument($expectedValueArea, EXPECTED_ARGUMENT_NAME, argumentValues[1]);

            this._monitorActualValueStateUpdate();
        }
    };

    EditAssertionDialog.prototype._createStateIndicator = function () {
        this.$stateIndicator = $('<div></div>').appendTo(this.$dialog);
        ShadowUI.addClass(this.$stateIndicator, STATE_INDICATOR_CLASS);

        this.$stateText = $('<div></div>').appendTo(this.$stateIndicator);
        ShadowUI.addClass(this.$stateText, STATE_TEXT_CLASS);
    };

    EditAssertionDialog.prototype._createArgument = function ($container, argumentName, value) {
        var assertionDialog = this;

        var argumentEditor = new AssertionArgumentEditor($container, {
            argumentName: argumentName,
            value: value,
            iFrameSelector: assertionDialog.iFrameSelector,
            iFrameContext: assertionDialog.iFrameContext,
            enableValidation: assertionDialog.options.enableAssertionValidation,
            parentPopup: assertionDialog.popup
        });

        argumentEditor.on(AssertionArgumentEditor.ARGUMENT_CHANGED_EVENT, function (e) {
            if (e.iFrameContext !== assertionDialog.iFrameContext) {
                assertionDialog.iFrameContext = e.iFrameContext;
                assertionDialog.iFrameSelector = e.iFrameSelector;

                if (e.iFrameContext)
                    assertionDialog._createIFrameContextLabel(e.iFrameSelector);

                if (argumentEditor !== assertionDialog.actual)
                    assertionDialog.actual.setIFrameContext(e.iFrameContext);
                else if (assertionDialog.expected)
                    assertionDialog.expected.setIFrameContext(e.iFrameContext);
            }

            assertionDialog._onChange();
        });

        argumentEditor.on(AssertionArgumentEditor.START_PICKING_ELEMENT_EVENT, function (e) {
            assertionDialog.eventEmitter.emit(EditAssertionDialog.START_PICKING_ELEMENT_EVENT, e);
        });

        return argumentEditor;
    };

    EditAssertionDialog.prototype._createMiddleArea = function ($container) {
        this.$middleArea = $('<div></div>').appendTo($container);
        ShadowUI.addClass(this.$middleArea, MIDDLE_AREA_CLASS);

        var $operatorIcon = $('<div></div>').appendTo(this.$middleArea);
        ShadowUI.addClass($operatorIcon, OPERATOR_ICON_CLASSES[this.operator]);

        if (this.operator === 'eq' && this.options.enableAssertionValidation) {
            this.$copyButton = ButtonWidget.create(this.$middleArea, '', true);
            ShadowUI.addClass(this.$copyButton, COPY_BUTTON_CLASS);

            this._setCopyButtonState();
        }
    };

    EditAssertionDialog.prototype._setCopyButtonState = function () {
        if (this.$copyButton) {
            if (this.actual.value && this.actual.isReproducibleValue)
                this.$copyButton.removeAttr('disabled');
            else
                this.$copyButton.attr('disabled', 'disabled');
        }
    };

    EditAssertionDialog.prototype._monitorActualValueStateUpdate = function () {
        var assertionDialog = this;

        this.actual.on(AssertionArgumentEditor.ARGUMENT_REPRODUCIBLE_STATE_UPDATED, function () {
            assertionDialog._setCopyButtonState();
        });
    };

    EditAssertionDialog.prototype._onChange = function () {
        this._validate();
    };

    EditAssertionDialog.prototype._validate = function () {
        var dialog = this,
            areArgumentsValid = this._argumentsFilled() && this.actual.isValid() && (!this.expected || this.expected.isValid());

        if (!areArgumentsValid) {
            this._onError();
            return;
        }

        this._onSuccess();

        if (!this.options.enableAssertionValidation)
            return;

        this.error = null;

        var iFrameContext = this.actual.iFrameContext,
            assertionArguments = [iFrameContext ? this.actual.value : this.actual.getLastParsedValue()];

        if (this.expected)
            assertionArguments.push(iFrameContext ? this.expected.value : this.expected.getLastParsedValue());

        assertionArguments.push(this.messageEditor.getEditorText());

        AssertionsAPI.assert(this.operator, assertionArguments, function (err) {
            dialog.error = err;
            dialog._setState(!dialog.error);
        }, iFrameContext);
    };

    EditAssertionDialog.prototype._onError = function () {
        if (this.$stateIndicator)
            this.$stateIndicator.css('visibility', 'hidden');

        this.$confirmButton.attr('disabled', 'disabled');
    };

    EditAssertionDialog.prototype._onSuccess = function () {
        this.$confirmButton.removeAttr('disabled');
    };

    EditAssertionDialog.prototype._setState = function (successful) {
        if (successful) {
            ShadowUI.removeClass(this.$stateIndicator, FAILED_CLASS);
            ShadowUI.addClass(this.$stateIndicator, SUCCESSFUL_CLASS);
            this.$stateText.text('passed');
        }
        else {
            ShadowUI.removeClass(this.$stateIndicator, SUCCESSFUL_CLASS);
            ShadowUI.addClass(this.$stateIndicator, FAILED_CLASS);
            this.$stateText.text('failed');
        }

        this.$stateIndicator.css('visibility', this._argumentsFilled() ? '' : 'hidden');
    };

    EditAssertionDialog.prototype._argumentsFilled = function () {
        return this.actual.value && !(this.expected && !this.expected.value);
    };

    EditAssertionDialog.prototype._init = function () {
        var dialog = this;

        if (this.$copyButton) {
            ShadowUI.bind(this.$copyButton, 'click', function () {
                if (dialog.expected)
                    dialog.expected.setValue(dialog.actual.objectViewer.getText());
            });
        }

        ShadowUI.bind(this.$confirmButton, 'click', function () {
            dialog._close(function () {
                dialog.eventEmitter.emit(EditAssertionDialog.CONFIRM_BUTTON_CLICK_EVENT, {
                    assertionInfo: {
                        operator: dialog.operator,
                        message: dialog.messageEditor.getEditorText(),
                        arguments: dialog.expected ? [dialog.actual.value, dialog.expected.value] : [dialog.actual.value]
                    },
                    iFrameSelector: dialog.iFrameSelector,
                    isValid: !dialog.error
                });
            });
        });

        ShadowUI.bind(this.$cancelButton, 'click', function () {
            dialog._close(function () {
                dialog.eventEmitter.emit(EditAssertionDialog.CANCEL_BUTTON_CLICK_EVENT, {});
            });
        });

        this.popup.onkeydown(function (e) {
            if (e.keyCode === Util.KEYS_MAPS.SPECIAL_KEYS.enter) {
                if (dialog.$confirmButton && dialog.$confirmButton.css('visibility') !== 'hidden')
                    dialog.$confirmButton.trigger('click');
                Util.preventDefault(e);
            }

            if (e.keyCode === Util.KEYS_MAPS.SPECIAL_KEYS.esc) {
                dialog.$cancelButton.trigger('click');
                Util.preventDefault(e);
            }

            //NOTE: we should prevent page scrolling
            if (e.target === dialog.popup.$popup[0] && Util.isArrowKey(e.keyCode))
                Util.preventDefault(e);
        }, true);
    };

    EditAssertionDialog.prototype._close = function (callback) {
        this.popup.close(callback);
    };
});