TestCafeClient.define('UI.RecorderWidgets.CodeEditor', function (require) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        Util = Hammerhead.Util,
        ShadowUI = Hammerhead.ShadowUI,
        NativeMethods = Hammerhead.NativeMethods,
        KeyEventParser = require('Recorder.KeyEventParser'),
        CodeFormatter = require('UI.RecorderWidgets.CodeFormatter');

    //Consts
    var EDITOR_CLASS = 'codeEditor',
        EDIT_LAYOUT_CLASS = 'editable',
        SCROLL_LAYOUT_CLASS = 'scrollable',
        LINE_CLASS = 'line',
        CURSOR_CLASS = 'editorCursor',
        TEXT_AREA_CLASS = 'editorTextarea',
        LETTER_WIDTH_FINDER = 'letter-width-finder',
        SELECTION_CLASS = 'selection',
        SELECTION_START_CLASS = 'start',
        SELECTION_END_CLASS = 'end',
        DISABLED_CLASS = 'disabled',

        EDITOR_WIDTH_DEFAULT = 400,
        EDITOR_HEIGHT_DEFAULT = 150,

        EDITOR_MAX_EXPANDED_SYMBOLS_COUNT = 80,

        EDITOR_PADDING_RIGHT = 4,
        EDITOR_PADDING_BOTTOM = 4,

        CURSOR_BLINKING_INTERVAL = 500,
        CURSOR_BLINKING_DELAY = 500,

        TAB_INDENT = '    ',

        STATE_STACK_SIZE = 100,

        OPENING_BRACKET_CHAR_CODES = [40, 91, 123],
        CLOSING_BRACKET_CHAR_CODES = [41, 93, 125],

        SHORTCUT_MODIFIER = Util.isMacPlatform ? 'meta' : 'ctrl';

    //Globals
    var CodeEditor = this.exports = function ($editorArea, options) {
        this.$editor = null;
        this.$scrollLayout = null;
        this.$editLayout = null;
        this.$cursor = null;
        this.$textarea = null;
        this.$lines = [];
        this.$selectionStart = null;
        this.$selectionCenter = null;
        this.$selectionEnd = null;

        this.editorWidth = 0;
        this.editorHeight = 0;
        this.fixedHeight = true;
        this.floatingWidth = false;
        this.expandDirection = null;
        this.holdFocus = false;
        this.inPaste = false;

        this.state = {
            lineIndex: 0,
            position: 0,
            selectionRegion: {
                startPosition: 0,
                startLine: 0,
                endPosition: 0,
                endLine: 0,
                inverse: false
            }
        };
        this.stateStack = [];
        this.stateIndex = 0;

        this.mouseLeftButtonPressed = false;
        this.mouseRightButtonPressed = false;
        this.allowEdit = true;
        this.expanded = false;
        this.buffer = [];
        this.changes = {
            modified: [],
            added: [],
            removed: []
        };

        this.lineHeight = null;
        this.letterWidth = null;

        this.codeFormatter = new CodeFormatter();

        this._init($editorArea, options);
    };

//Events
    CodeEditor.CHANGE_EVENT = 'change';
    CodeEditor.FOCUS_EVENT = 'focus';
    CodeEditor.BLUR_EVENT = 'blur';

//Markup
    CodeEditor.prototype._createEditor = function ($editorArea) {
        this.$editor = $editorArea;
        ShadowUI.addClass(this.$editor, EDITOR_CLASS);

        this.$scrollLayout = $('<div>').appendTo(this.$editor);
        ShadowUI.addClass(this.$scrollLayout, SCROLL_LAYOUT_CLASS);
        this.$editLayout = $('<div>').appendTo(this.$scrollLayout);
        ShadowUI.addClass(this.$editLayout, EDIT_LAYOUT_CLASS);
    };

    CodeEditor.prototype._createCursor = function () {
        this.$cursor = $('<div>').appendTo(this.$scrollLayout);
        ShadowUI.addClass(this.$cursor, CURSOR_CLASS);
        this.$cursor.css({
            height: this._getLineHeight(),
            opacity: 0
        });
    };

    CodeEditor.prototype._createTextArea = function () {
        this.$textarea = $('<textarea>').appendTo(this.$editor);
        //NOTE: letterWidth = 7
        this.$textarea.css('width', 3 * 7);
        //NOTE: for IE
        if (Util.isIE) {
            this.$textarea.css('background-color', 'red');
            this.$textarea.css('font-size', '0em');
        }
        ShadowUI.addClass(this.$textarea, TEXT_AREA_CLASS);
    };

    CodeEditor.prototype._createSelectionAreas = function () {
        this.$selectionStart = $('<div>').appendTo(this.$scrollLayout);
        ShadowUI.addClass(this.$selectionStart, SELECTION_CLASS);
        ShadowUI.addClass(this.$selectionStart, SELECTION_START_CLASS);
        this.$selectionStart.css('height', this._getLineHeight());

        this.$selectionCenter = $('<div>').appendTo(this.$scrollLayout);
        ShadowUI.addClass(this.$selectionCenter, SELECTION_CLASS);
        this.$selectionCenter.css('width', this.$scrollLayout[0].clientWidth);
        this.$selectionCenter.css('height', 0);

        this.$selectionEnd = $('<div>').appendTo(this.$scrollLayout);
        ShadowUI.addClass(this.$selectionEnd, SELECTION_CLASS);
        ShadowUI.addClass(this.$selectionEnd, SELECTION_END_CLASS);
        this.$selectionEnd.css('height', this._getLineHeight());
    };

    CodeEditor.prototype._createLines = function (startIndex, lineTexts) {
        if (!lineTexts.length)
            return [];
        var $lines = null,
            lineTemplate = '<div><span></span></div>';

        $.each(lineTexts, function (index, text) {
            var $line = $(lineTemplate);
            $line.children('span').text(text || '');
            ShadowUI.addClass($line, LINE_CLASS);
            if (!$lines)
                $lines = $line;
            else
                $lines = $lines.add($line);
        });

        var prevChild = this.$editLayout.children(':nth-child(' + startIndex + ')');

        if (startIndex === 0)
            $lines.prependTo(this.$editLayout);
        else if (prevChild.length)
            $lines.insertAfter(prevChild);
        else
            $lines.appendTo(this.$editLayout);

        return $lines;
    };

//highlight
    CodeEditor.prototype._highlight = function (modifiedLines) {
        if (this.codeFormatter)
            this.codeFormatter.highlight(this.$lines, modifiedLines, this.state.lineIndex);
    };

//Behavior

//initialization
    CodeEditor.prototype._init = function ($editorArea, options) {
        var codeEditor = this;
        options = options || {};

        options = {
            width: options.width || EDITOR_WIDTH_DEFAULT,
            height: options.height || EDITOR_HEIGHT_DEFAULT,
            fixedHeight: typeof options.fixedHeight === 'undefined' ? true : options.fixedHeight,
            text: options.text || '',
            allowEdit: typeof options.allowEdit === 'undefined' ? true : options.allowEdit,
            floatingWidth: options.floatingWidth || false,
            expandDirection: options.expandDirection || 'right'
        };

        codeEditor.eventEmitter = new Util.EventEmitter();
        codeEditor.events = {
            on: function (ev, listener) {
                codeEditor.eventEmitter.on(ev, listener);
            }
        };

        codeEditor._createEditor($editorArea);
        codeEditor._initEditor(options);

        if (options.text) {
            codeEditor._setEditorText(options.text);
        }
        else
            codeEditor._addLine(undefined, '');

        codeEditor._createCursor();
        codeEditor._createTextArea();
        codeEditor._initTextArea();
        codeEditor._createSelectionAreas();
        codeEditor._initSelectionAreas();
        codeEditor._setCursor(0, 0);
        if (Util.isIE)
            Util.setUnselectableAttributeRecursive(codeEditor.$scrollLayout[0]);
        codeEditor._clearChanges();

        codeEditor._textChanged();
    };

    CodeEditor.prototype._initEditor = function (options) {
        var codeEditor = this;

        if (typeof options.width === 'string' && options.width.indexOf('%') !== -1)
            codeEditor.editorWidth = Math.floor(codeEditor.$editor.parent().width() *
                options.width.substring(0, options.width.indexOf('%')) / 100);
        else
            codeEditor.editorWidth = options.width;

        if (typeof options.height === 'string' && options.height.indexOf('%') !== -1)
            codeEditor.editorHeight = Math.floor(codeEditor.$editor.parent().height() *
                options.height.substring(0, options.height.indexOf('%')) / 100);
        else
            codeEditor.editorHeight = options.height;
        codeEditor.fixedHeight = options.fixedHeight;
        codeEditor.allowEdit = options.allowEdit;
        codeEditor.floatingWidth = options.floatingWidth;
        codeEditor.expandDirection = options.expandDirection;

        if (!codeEditor.allowEdit)
            ShadowUI.addClass(this.$editor, DISABLED_CLASS);

        codeEditor.$scrollLayout.css('width', codeEditor.editorWidth);
        codeEditor.$editor.css('height', codeEditor.editorHeight);
        codeEditor.$scrollLayout.css('height', codeEditor.editorHeight);

        if (!codeEditor.fixedHeight)
            codeEditor.$scrollLayout.css('overflowY', 'visible');

        codeEditor.$scrollLayout.mousedown(function (e) {
            codeEditor._mousedown(e);
            return false;
        });

        codeEditor.$scrollLayout.dblclick(function (e) {
            codeEditor._dblclick(e);
            return false;
        });

        ShadowUI.bind($(document), 'mouseup', function () {
            codeEditor.stopSelectionProcess();
        });

        codeEditor.$scrollLayout.mouseup(function () {
            codeEditor.mouseRightButtonPressed = false;
            codeEditor.mouseLeftButtonPressed = false;
            if (!Util.isMozilla)
                return false;
        });

        ShadowUI.bind($(document), 'mousemove', function (e) {
            codeEditor._mousemove(e);
        });

        ShadowUI.bind(codeEditor.$scrollLayout, typeof document.onwheel !== 'undefined' ? 'wheel' : 'mousewheel', function (e) {
            var delta = -e.wheelDelta || (e.originalEvent ? (e.originalEvent.deltaY * 8) : e.deltaY * 100),
                $el = $(this),
                scrollTop = $el.scrollTop(),
                bordersWidth = Util.getBordersWidth($el),
                sizeDifference = this.scrollHeight - this.clientHeight,
                newTop = Math.min(Math.max(scrollTop + delta, 0), sizeDifference);

            if (this.scrollHeight !== this.clientHeight + bordersWidth.top + bordersWidth.bottom)
                $el.scrollTop(newTop);

            Util.preventDefault(e);
            return false;
        });
    };

    CodeEditor.prototype._expandEditor = function () {
        this.expanded = true;
        ShadowUI.addClass(this.$editor, 'expanded');

        if (this.$editLayout.width() <= this.$scrollLayout.width())
            return;

        this._updateEditorSize();
    };

    CodeEditor.prototype._collapseEditor = function () {
        this.expanded = false;
        ShadowUI.removeClass(this.$editor, 'expanded');
        if (this.expandDirection === 'left') {
            this.$scrollLayout.css('marginLeft', 0);
            this.$scrollLayout.css('position', 'relative');
        }
        this.$scrollLayout.width(this.editorWidth);
        this._updateEditorSize();
    };

    CodeEditor.prototype._initTextArea = function () {
        var codeEditor = this;

        codeEditor._initEventParser();

        ShadowUI.bind(codeEditor.$textarea, 'blur', function () {
            codeEditor._blur();
        });

        ShadowUI.bind(codeEditor.$textarea, 'focus', function () {
            codeEditor._focus();
        });

        ShadowUI.bind(codeEditor.$textarea, 'contextmenu', function () {
            NativeMethods.setTimeout.call(window, function () {
                codeEditor.$textarea.css('z-index', 0);
                if (Util.isIE)
                    codeEditor.$textarea.css('font-size', '0em');
                if (!Util.isMozilla && !Util.isIE)
                    codeEditor.$textarea[0].value = '';
            }, 0);
        });

        ShadowUI.bind(codeEditor.$textarea, 'copy', function () {
            codeEditor._onCopyCodeEditorText();
        });

        ShadowUI.bind(codeEditor.$textarea, 'paste', function () {
            codeEditor._onPasteCodeEditor();
        });

        ShadowUI.bind(codeEditor.$textarea, 'cut', function () {
            codeEditor._onCutCodeEditor();
        });
    };

    CodeEditor.prototype._getShortcutHandlers = function () {
        var codeEditor       = this;
        var shortcutHandlers = {};

        shortcutHandlers['left'] = function () {
            codeEditor._shortcutHandlers().left();
        };
        shortcutHandlers['right'] = function () {
            codeEditor._shortcutHandlers().right();
        };
        shortcutHandlers['up'] = function () {
            codeEditor._shortcutHandlers().up();
        };
        shortcutHandlers['down'] = function () {
            codeEditor._shortcutHandlers().down();
        };
        shortcutHandlers['enter'] = function () {
            codeEditor._shortcutHandlers().enter();
        };
        shortcutHandlers['backspace'] = function () {
            codeEditor._shortcutHandlers().backspace();
        };
        shortcutHandlers['delete'] = function () {
            codeEditor._shortcutHandlers()['delete']();
        };
        shortcutHandlers['home'] = function () {
            codeEditor._shortcutHandlers().home();
        };
        shortcutHandlers['end'] = function () {
            codeEditor._shortcutHandlers().end();
        };
        shortcutHandlers['tab'] = function () {
            codeEditor._shortcutHandlers().tab();
        };
        shortcutHandlers['shift+tab'] = function () {
            codeEditor._shortcutHandlers().shiftTab();
        };
        shortcutHandlers['shift+left'] = function () {
            codeEditor._shortcutHandlers().shiftLeft();
        };
        shortcutHandlers['shift+right'] = function () {
            codeEditor._shortcutHandlers().shiftRight();
        };
        shortcutHandlers['shift+up'] = function () {
            codeEditor._shortcutHandlers().shiftUp();
        };
        shortcutHandlers['shift+down'] = function () {
            codeEditor._shortcutHandlers().shiftDown();
        };
        shortcutHandlers['shift+home'] = function () {
            codeEditor._shortcutHandlers().shiftHome();
        };
        shortcutHandlers['shift+end'] = function () {
            codeEditor._shortcutHandlers().shiftEnd();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+left'] = function () {
            codeEditor._shortcutHandlers().ctrlLeft();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+right'] = function () {
            codeEditor._shortcutHandlers().ctrlRight();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+home'] = function () {
            codeEditor._shortcutHandlers().ctrlHome();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+end'] = function () {
            codeEditor._shortcutHandlers().ctrlEnd();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+backspace'] = function () {
            codeEditor._shortcutHandlers().ctrlBackspace();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+delete'] = function () {
            codeEditor._shortcutHandlers().ctrlDelete();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+shift+left'] = function () {
            codeEditor._shortcutHandlers().ctrlShiftLeft();
        };
        shortcutHandlers['shift+' + SHORTCUT_MODIFIER + '+left'] = function () {
            codeEditor._shortcutHandlers().ctrlShiftLeft();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+shift+right'] = function () {
            codeEditor._shortcutHandlers().ctrlShiftRight();
        };
        shortcutHandlers['shift+' + SHORTCUT_MODIFIER + '+right'] = function () {
            codeEditor._shortcutHandlers().ctrlShiftRight();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+shift+up'] = function () {
            codeEditor._shortcutHandlers().ctrlShiftUp();
        };
        shortcutHandlers['shift+' + SHORTCUT_MODIFIER + '+up'] = function () {
            codeEditor._shortcutHandlers().ctrlShiftUp();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+shift+down'] = function () {
            codeEditor._shortcutHandlers().ctrlShiftDown();
        };
        shortcutHandlers['shift+' + SHORTCUT_MODIFIER + '+down'] = function () {
            codeEditor._shortcutHandlers().ctrlShiftDown();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+x'] = function () {
            codeEditor._shortcutHandlers().ctrlX();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+c'] = function () {
            codeEditor._shortcutHandlers().ctrlC();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+v'] = function () {
            codeEditor._shortcutHandlers().ctrlV();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+a'] = function () {
            codeEditor._shortcutHandlers().ctrlA();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+z'] = function () {
            codeEditor._shortcutHandlers().ctrlZ();
        };
        shortcutHandlers[SHORTCUT_MODIFIER + '+shift+z'] = function () {
            codeEditor._shortcutHandlers().ctrlShiftZ();
        };
        shortcutHandlers['shift+' + SHORTCUT_MODIFIER + '+z'] = function () {
            codeEditor._shortcutHandlers().ctrlShiftZ();
        };
        return shortcutHandlers;
    };

    CodeEditor.prototype._initEventParser = function () {
        var codeEditor = this;
        codeEditor.keyEventParser = new KeyEventParser();
        codeEditor.keyEventParser.init({
            symbolPressed: function (charCode) {
                codeEditor._letterPressed(charCode);
            },
            shortcutHandlers: codeEditor._getShortcutHandlers()
        });

        var listenedEvents = ['keydown', 'keypress', 'keyup'];

        $.each(listenedEvents, function () {
            codeEditor.$textarea[this](function (e) {
                //NOTE: allow ctrl+c, ctrl+a always
                if (codeEditor.allowEdit ||
                    (/key(down|up)/.test(e.type) && (e.keyCode === 17 || e.keyCode === 65 || e.keyCode === 67)))
                    codeEditor.keyEventParser.pushEvent(e);

                //NOTE: we need to prevent page scrolling in Chrome
                if (/key(down|up)/.test(e.type) && Util.isWebKit && (e.keyCode === 37 || e.keyCode === 39))
                    e.preventDefault();


                //NOTE: we're should raise textarea blur event to code editor lost focus
                if (e.type === 'keydown' && Util.KEYS_MAPS.REVERSED_SPECIAL_KEYS[e.keyCode] === 'esc')
                    NativeMethods.blur.call(codeEditor.$textarea[0]);

                e.stopPropagation();

                if ((e.type === 'keypress' && !((e.charCode === 3 || (Util.isMozilla && e.charCode === 118) ||
                    ((e.charCode === 22 || e.charCode === 24))) && e.ctrlKey)) ||
                    (e.type === 'keydown' && (e.keyCode === 9 || e.keyCode === 8 || e.keyCode === 16))) {
                    return false;
                }
            });
        });
    };

    CodeEditor.prototype._initSelectionAreas = function () {
        var codeEditor = this;
        this.$selectionStart.mousedown(function (e) {
            codeEditor._mousedown(e);
            return false;
        });
        this.$selectionCenter.mousedown(function (e) {
            codeEditor._mousedown(e);
            return false;
        });
        this.$selectionEnd.mousedown(function (e) {
            codeEditor._mousedown(e);
            return false;
        });
    };

    CodeEditor.prototype._updateScroll = function () {
        var cursor = this.$cursor[0],
            cursorOffset = Util.getOffsetPosition(cursor),
            cursorClientPosition = Util.offsetToClientCoords({x: cursorOffset.left, y: cursorOffset.top});
        if (cursorClientPosition.top < 0)
            cursor.scrollIntoView();
        else if ((cursorClientPosition.top + this._getLineHeight()) >= $(window).height())
            cursor.scrollIntoView(false);

        var cursorPosition = {
            left: parseInt(this.$cursor.css('left').replace('px', '')),
            top: parseInt(this.$cursor.css('top').replace('px', ''))
        };
        //need scroll horizontal
        if (cursorPosition.left < this.$scrollLayout.scrollLeft())
            this.$scrollLayout.scrollLeft(cursorPosition.left);
        else if (cursorPosition.left >= this.$scrollLayout.scrollLeft() + this.$scrollLayout[0].clientWidth - this._getVerticalScrollWidth())
            this.$scrollLayout.scrollLeft(cursorPosition.left - this.$scrollLayout[0].clientWidth + this._getVerticalScrollWidth() + EDITOR_PADDING_RIGHT);
        //need scroll vertical
        if (!this.fixedHeight)
            return;
        if (cursorPosition.top < this.$scrollLayout.scrollTop())
            this.$scrollLayout.scrollTop(cursorPosition.top);
        else if (cursorPosition.top + this._getLineHeight() >= this.$scrollLayout.scrollTop() + this.editorHeight - this._getHorizontalScrollHeight()) {
            this.$scrollLayout.scrollTop(cursorPosition.top + this._getLineHeight() - this.editorHeight + this._getHorizontalScrollHeight() + EDITOR_PADDING_BOTTOM);
        }
    };

//events
    CodeEditor.prototype._textChanged = function () {
        this._updateEditorSize();

        if (this.changes.added.length || this.changes.removed.length)
            this._highlight();
        else
            this._highlight($.map(this.changes.modified, function (item) {
                return item.index;
            }));
        this._saveState(false);
        this.eventEmitter.emit(CodeEditor.CHANGE_EVENT, {text: this.getText()});
    };

    CodeEditor.prototype._focus = function () {
        if (this.holdFocus)
            return;

        this._showSelection();

        if (this.floatingWidth)
            this._expandEditor();
        this._enableCursor();
        this.eventEmitter.emit(CodeEditor.FOCUS_EVENT);
    };

    CodeEditor.prototype._blur = function () {
        if (this.holdFocus)
            return;

        this._hideSelection();

        this._disableCursor();

        if (this.floatingWidth)
            this._collapseEditor();
        this.eventEmitter.emit(CodeEditor.BLUR_EVENT);
    };

//cursor
    CodeEditor.prototype._enableCursorBlinking = function () {
        var codeEditor = this;
        codeEditor.cursorInterval = NativeMethods.setInterval.call(window, function () {
            codeEditor.$cursor.css('opacity', (codeEditor.$cursor.css('opacity') === '1' ? 0 : 1));
        }, CURSOR_BLINKING_INTERVAL);
    };

    CodeEditor.prototype._enableCursor = function () {
        if (!this.allowEdit || this.$cursor.data('enabled'))
            return;

        this._enableCursorBlinking();
        this.$cursor.data('enabled', true);
        this.$cursor.css('opacity', 1);
    };

    CodeEditor.prototype._disableCursor = function () {
        if (!this.allowEdit)
            return;
        var codeEditor = this;

        if (codeEditor.cursorInterval)
            window.clearInterval(codeEditor.cursorInterval);
        if (codeEditor.cursorTimeout)
            window.clearTimeout(codeEditor.cursorTimeout);

        codeEditor.cursorInterval = null;
        codeEditor.cursorTimeout = null;
        codeEditor.$cursor.css('opacity', 0);
        codeEditor.$cursor.data('enabled', false);
    };

    CodeEditor.prototype._suspendCursorBlinking = function () {
        if (!this.allowEdit || !this.$cursor.data('enabled'))
            return;

        var codeEditor = this;

        if (codeEditor.cursorInterval)
            window.clearInterval(codeEditor.cursorInterval);
        if (codeEditor.cursorTimeout)
            window.clearTimeout(codeEditor.cursorTimeout);

        codeEditor.$cursor.css('opacity', 1);

        codeEditor.cursorTimeout = NativeMethods.setTimeout.call(window, function () {
            codeEditor._enableCursorBlinking();
        }, CURSOR_BLINKING_DELAY);
    };

    CodeEditor.prototype._setCursor = function (lineIndex, position, saveState, withSelection) {
        if (withSelection) {
            this._setSelection(lineIndex, position);
            lineIndex = this.state.selectionRegion.inverse ? this.state.selectionRegion.startLine : this.state.selectionRegion.endLine;
            position = this.state.selectionRegion.inverse ? this.state.selectionRegion.startPosition : this.state.selectionRegion.endPosition;
        }
        else
            this._clearSelection(lineIndex, position);

        this.state.lineIndex = lineIndex;
        this.state.position = position;
        this._setElementPosition(this.$cursor, lineIndex, position);
        this._updateScroll();

        this.$textarea.css({
            left: this.$cursor[0].offsetLeft - this.$scrollLayout[0].scrollLeft + this.$scrollLayout[0].offsetLeft - this.letterWidth,
            top: this.$cursor[0].offsetTop - this.$scrollLayout[0].scrollTop + this.$scrollLayout[0].offsetTop
        });

        this._suspendCursorBlinking();

        if (saveState)
            this._saveState(true);
    };

    CodeEditor.prototype._setTextarea = function (lineIndex, position) {
        var left = this._getElementPosition(lineIndex, position).left - this.$scrollLayout[0].scrollLeft + this.$scrollLayout[0].offsetLeft - this.letterWidth,
            top = this._getElementPosition(lineIndex, position).top - this.$scrollLayout[0].scrollTop + this.$scrollLayout[0].offsetTop;

        this.$textarea.css({
            left: left > this.$scrollLayout.width() + this.$scrollLayout[0].offsetLeft ?
                this.$scrollLayout.width() + this.$scrollLayout[0].offsetLeft : left,
            top: top > this.$scrollLayout.height() + this.$scrollLayout[0].offsetTop ?
                this.$scrollLayout.height() + this.$scrollLayout[0].offsetTop : top
        });
    };

//lines
    CodeEditor.prototype._addLine = function (index, text, withIndent) {
        text = withIndent ? TAB_INDENT + text : text;

        this._addLines(index, [text]);
    };

    CodeEditor.prototype._addLines = function (startIndex, linesArray) {
        var codeEditor = this,
            height = null,
            $lines = codeEditor._createLines(startIndex, linesArray),
            newIndex,
            newValue,
            remIndex;
        $.each($lines, function (index, item) {
            var $item = $(item);
            if (typeof startIndex !== 'undefined') {
                codeEditor.$lines.splice(startIndex, 0, $item);
                newIndex = startIndex;
                newValue = codeEditor._getLineText(newIndex);
                remIndex = codeEditor._getStateIndex(codeEditor.changes.removed, newIndex);
                startIndex++;
            }
            else {
                codeEditor.$lines.push($item);
                newIndex = codeEditor.$lines.length - 1;
                newValue = codeEditor._getLineText(newIndex);
                remIndex = codeEditor._getStateIndex(codeEditor.changes.removed, newIndex);
            }
            if (typeof remIndex === 'undefined')
                codeEditor.changes.added.push({
                    index: newIndex,
                    value: newValue
                });
            else {
                codeEditor.changes.modified.push({
                    index: newIndex,
                    oldValue: codeEditor.changes.removed[remIndex].value,
                    newValue: newValue
                });
                codeEditor.changes.removed.splice(remIndex, 1);
            }
            if (!height)
                height = codeEditor._getLineHeight();
            $item.css('height', height);
            if (Util.isIE)
                Util.setUnselectableAttributeRecursive(item);
        });
    };

    CodeEditor.prototype._removeLine = function (lineIndex) {
        var value = this._getLineText(lineIndex),
            addIndex = this._getStateIndex(this.changes.added, lineIndex),
            modIndex = this._getStateIndex(this.changes.modified, lineIndex);

        if (typeof addIndex !== 'undefined')
            this.changes.added.splice(addIndex, 1);
        else {
            if (typeof modIndex !== 'undefined')
                this.changes.modified.splice(modIndex, 1);
            this.changes.removed.push({
                index: lineIndex,
                value: value
            });
        }
        this.$lines[lineIndex].remove();
        this.$lines.splice(lineIndex, 1);
    };

    CodeEditor.prototype._setLineText = function (lineIndex, text) {
        var oldValue = this._getLineText(lineIndex),
            modIndex = this._getStateIndex(this.changes.modified, lineIndex),
            addIndex = this._getStateIndex(this.changes.added, lineIndex);

        if (this.$lines[lineIndex])
            this.$lines[lineIndex].children('span:first').text(text);

        if (typeof modIndex !== 'undefined')
            this.changes.modified[modIndex].newValue = text;
        else if (typeof addIndex !== 'undefined')
            this.changes.added[addIndex].value = text;
        else
            this.changes.modified.push({
                index: lineIndex,
                oldValue: oldValue,
                newValue: text
            });
    };

    CodeEditor.prototype._getLineText = function (lineIndex) {
        if (this.$lines[lineIndex])
            return this.$lines[lineIndex].text();
        return '';
    };

//selection
    CodeEditor.prototype._setSelection = function (line, position) {
        var startSelectionLineIndex = this.state.selectionRegion.inverse ?
                this.state.selectionRegion.endLine :
                this.state.selectionRegion.startLine,

            startSelectionPosition = this.state.selectionRegion.inverse ?
                this.state.selectionRegion.endPosition :
                this.state.selectionRegion.startPosition;

        if (line === startSelectionLineIndex && position === startSelectionPosition)
            this._clearSelection(line, position);

        else if (line < startSelectionLineIndex || (line === startSelectionLineIndex && position < startSelectionPosition)) {
            this.state.selectionRegion.endLine = startSelectionLineIndex;
            this.state.selectionRegion.endPosition = startSelectionPosition;
            this.state.selectionRegion.startLine = line;
            this.state.selectionRegion.startPosition = position;
            this.state.selectionRegion.inverse = true;
        }
        else {
            this.state.selectionRegion.startLine = startSelectionLineIndex;
            this.state.selectionRegion.startPosition = startSelectionPosition;
            this.state.selectionRegion.endLine = line;
            this.state.selectionRegion.endPosition = position;
            this.state.selectionRegion.inverse = false;
        }

        this._updateSelectionAreas();
    };

    CodeEditor.prototype._shiftSelection = function (lineOffset, positionOffset) {
        if (lineOffset) {
            this.state.selectionRegion.startLine += lineOffset;
            this.state.selectionRegion.endLine += lineOffset;
        }
        if (positionOffset) {
            this.state.selectionRegion.startPosition += positionOffset;
            this.state.selectionRegion.endPosition += positionOffset;
        }
        this._updateSelectionAreas();
    };

    CodeEditor.prototype._moveLines = function (offset) {
        var changingLineIndex = null;

        if (!this._hasSelection(this.state))
            changingLineIndex = this.state.lineIndex + offset;
        else if (this._hasSelection(this.state) && offset === -1)
            changingLineIndex = this.state.selectionRegion.startLine - 1;
        else if (this._hasSelection(this.state) && offset === 1)
            changingLineIndex = this.state.selectionRegion.endLine + 1;

        if (!this.$lines[changingLineIndex])
            return;

        var lineText = this._getLineText(changingLineIndex);

        if (!this._hasSelection(this.state) || this.state.selectionRegion.startLine === this.state.selectionRegion.endLine) {
            this._setLineText(this.state.lineIndex + offset, this._getLineText(this.state.lineIndex));
            this._setLineText(this.state.lineIndex, lineText);
        }
        else {
            if (offset === -1) {
                for (var i = this.state.selectionRegion.startLine - 1; i < this.state.selectionRegion.endLine; i++)
                    this._setLineText(i, this._getLineText(i + 1));

                this._setLineText(this.state.selectionRegion.endLine, lineText);
            }
            else {
                for (var j = this.state.selectionRegion.endLine + 1; j > this.state.selectionRegion.startLine; j--)
                    this._setLineText(j, this._getLineText(j - 1));

                this._setLineText(this.state.selectionRegion.startLine, lineText);
            }
        }
        this._shiftSelection(offset);
        this._setCursor(this.state.lineIndex + offset, this.state.position, false, true);
        this._textChanged();
    };

    CodeEditor.prototype._updateSelectionAreas = function () {
        var startLine = this.state.selectionRegion.startLine,
            endLine = this.state.selectionRegion.endLine,
            startPosition = this.state.selectionRegion.startPosition,
            endPosition = this.state.selectionRegion.endPosition;

        if (startLine === endLine) {
            this._clearSelectionArea(this.$selectionCenter);
            this._clearSelectionArea(this.$selectionEnd);

            var width = Math.abs(
                // Browser rounds *.5 position to the smallest integer.
                Math.floor(endPosition * this._getLetterWidth(startLine)) -
                    Math.floor(startPosition * this._getLetterWidth(startLine))
            );

            this._setElementPosition(this.$selectionStart, startLine, Math.min(endPosition, startPosition));
            this.$selectionStart.css('width', width);
        }
        else {
            var startWidth = Math.floor(this.$editLayout.width() - (startPosition * this._getLetterWidth(startLine))),
                endWidth = Math.floor(endPosition * this._getLetterWidth(endLine));

            if (Math.abs(endLine - startLine) === 1)
                this._clearSelectionArea(this.$selectionCenter);
            else {
                var selectionCenterHeight = (Math.abs(endLine - startLine) - 1) * this._getLineHeight();
                this._setElementPosition(this.$selectionCenter, startLine + 1, 0);
                this.$selectionCenter.css('height', selectionCenterHeight + 'px');
            }

            this._setElementPosition(this.$selectionStart, startLine, startPosition);
            this._setElementPosition(this.$selectionEnd, endLine, 0);
            this.$selectionStart.css('width', startWidth);
            this.$selectionEnd.css('width', endWidth);
        }
    };

    CodeEditor.prototype._removeSelectionContent = function () {
        var startLineText = this._getLineText(this.state.selectionRegion.startLine).substring(0, this.state.selectionRegion.startPosition),
            endLineText = this._getLineText(this.state.selectionRegion.endLine).substring(this.state.selectionRegion.endPosition, this._getLineLength(this.state.selectionRegion.endLine));

        this._setLineText(this.state.selectionRegion.startLine, startLineText + endLineText);

        for (var i = this.state.selectionRegion.endLine; i > this.state.selectionRegion.startLine; i--)
            this._removeLine(i);

        this._setCursor(this.state.selectionRegion.startLine, this.state.selectionRegion.startPosition);
    };

    CodeEditor.prototype._getSelectionTextAsArray = function () {
        var selectionText = [];
        if (this._hasSelection(this.state)) {
            selectionText
                .push(this._getLineText(this.state.selectionRegion.startLine)
                    .substring(this.state.selectionRegion.startPosition,
                        this.state.selectionRegion.startLine === this.state.selectionRegion.endLine ?
                            this.state.selectionRegion.endPosition :
                            this._getLineLength(this.state.selectionRegion.startLine)));

            for (var i = this.state.selectionRegion.startLine + 1; i < this.state.selectionRegion.endLine; i++)
                selectionText.push(this._getLineText(i));

            if (this.state.selectionRegion.startLine !== this.state.selectionRegion.endLine)
                selectionText.push(this._getLineText(this.state.selectionRegion.endLine).substring(0, this.state.selectionRegion.endPosition));
        }
        return selectionText;
    };

    CodeEditor.prototype._clearSelection = function (line, position) {
        this.state.selectionRegion = {
            startLine: line,
            startPosition: position,
            endLine: line,
            endPosition: position,
            inverse: false
        };

        this._clearSelectionArea(this.$selectionStart);
        this._clearSelectionArea(this.$selectionCenter);
        this._clearSelectionArea(this.$selectionEnd);
    };

    CodeEditor.prototype._clearSelectionArea = function ($area) {
        $area.css('left', 0);

        if ($area === this.$selectionCenter)
            $area.css('height', 0);
        else
            $area.css('width', 0);

        $area.css('top', 0);
    };

    CodeEditor.prototype._hideSelection = function () {
        this.$selectionStart.css('visibility', 'hidden');
        this.$selectionCenter.css('visibility', 'hidden');
        this.$selectionEnd.css('visibility', 'hidden');
    };

    CodeEditor.prototype._showSelection = function () {
        this.$selectionStart.css('visibility', '');
        this.$selectionCenter.css('visibility', '');
        this.$selectionEnd.css('visibility', '');
    };

//moving position
    CodeEditor.prototype._moveRight = function (saveState, withSelection) {
        var line = this.state.lineIndex,
            position = this.state.position;

        if (this._getLineLength(this.state.lineIndex) >= this.state.position + 1)
            position++;
        else if (this.$lines[this.state.lineIndex + 1]) {
            line++;
            position = 0;
        }
        else {
            if (this._hasSelection(this.state))
                this._setCursor(line, position, saveState, withSelection);
            return;
        }

        this._setCursor(line, position, saveState, withSelection);
    };

    CodeEditor.prototype._moveLeft = function (withSelection) {
        var line = this.state.lineIndex,
            position = this.state.position;

        if ((this.state.position - 1) === -1) {
            if ((this.state.lineIndex - 1) === -1) {
                if (this._hasSelection(this.state))
                    this._setCursor(line, position, true, withSelection);
                return;
            }
            line--;
            position = this._getLineLength(line);
        }
        else
            position--;

        this._setCursor(line, position, true, withSelection);
    };

    CodeEditor.prototype._moveUp = function (withSelection) {
        var line = this.state.lineIndex,
            position = this.state.position;

        if ((this.state.lineIndex - 1) === -1)
            if ((this.state.position - 1) === -1) {
                if (this._hasSelection(this.state))
                    this._setCursor(line, position, true, withSelection);
                return;
            }
            else
                position = 0;
        else if (this._getLineLength(this.state.lineIndex - 1) <= this.state.position) {
            line--;
            position = this._getLineLength(line);
        }
        else
            line--;

        this._setCursor(line, position, true, withSelection);
    };

    CodeEditor.prototype._moveDown = function (withSelection) {
        var line = this.state.lineIndex,
            position = this.state.position;

        if (this.$lines[this.state.lineIndex + 1])
            if (this._getLineLength(this.state.lineIndex + 1) <= this.state.position) {
                line++;
                position = this._getLineLength(line);
            }
            else
                line++;
        else if (this.state.position !== this._getLineLength(this.$lines.length - 1))
            position = this._getLineLength(this.$lines.length - 1);
        else {
            if (this._hasSelection(this.state))
                this._setCursor(line, position, true, withSelection);
            return;
        }
        this._setCursor(line, position, true, withSelection);
    };

//handlers
    CodeEditor.prototype._mousedown = function (e) {
        var codeEditor = this,

            x = e.pageX,
            y = e.pageY,

            scrollLeft = this.$scrollLayout[0].scrollLeft,
            scrollTop = this.$scrollLayout[0].scrollTop,

            isTextareaActive = document.activeElement === this.$textarea[0];

        x = x - this._getEditorOffset().left + scrollLeft;
        y = y - this._getEditorOffset().top + scrollTop;

        //return if click on scroll
        if (y > this.$scrollLayout[0].offsetHeight + scrollTop - this._getHorizontalScrollHeight() ||
            x > this.$scrollLayout[0].offsetWidth + scrollLeft - this._getVerticalScrollWidth()) {
            return;
        }

        var newCursorPosition = this._getCodeEditorRelativePosition(x, y),
            contained = this._selectionContains({lineIndex: newCursorPosition.lineIndex, position: newCursorPosition.position});

        var mousedownHandler = function () {
            //NOTE: we should restore scrollContainer scroll values after focus textarea
            //because its can be modified (B237719)
            if (!isTextareaActive) {
                codeEditor.$scrollLayout[0].scrollLeft = scrollLeft;
                codeEditor.$scrollLayout[0].scrollTop = scrollTop;
            }

            if (Util.BUTTON.LEFT === e.button)
                codeEditor.mouseLeftButtonPressed = true;
            else if (Util.BUTTON.RIGHT === e.button) {
                if (Util.isIE)
                    codeEditor.$textarea.css('font-size', '1em');
                codeEditor.$textarea.css('z-index', 10);
                codeEditor.mouseRightButtonPressed = true;
                if (codeEditor._hasSelection(codeEditor.state) && contained)
                    codeEditor._createSelectionObject();
            }

            if (!codeEditor.mouseRightButtonPressed)
                codeEditor._setCursor(newCursorPosition.cursorLineIndex, newCursorPosition.cursorPosition, true, e.shiftKey);
            else if (!codeEditor._hasSelection(codeEditor.state)) {
                codeEditor._setCursor(newCursorPosition.cursorLineIndex, newCursorPosition.cursorPosition, true, e.shiftKey);
                if (newCursorPosition.cursorLineIndex !== newCursorPosition.lineIndex ||
                    newCursorPosition.cursorPosition !== newCursorPosition.position)
                    codeEditor._setTextarea(newCursorPosition.lineIndex, newCursorPosition.position);
            }
            else {
                if (!contained) {
                    codeEditor._clearSelection(newCursorPosition.cursorLineIndex, newCursorPosition.cursorPosition);
                    codeEditor.$textarea[0].value = '';
                    codeEditor._setCursor(newCursorPosition.cursorLineIndex, newCursorPosition.cursorPosition, true, e.shiftKey);
                }
                codeEditor._setTextarea(newCursorPosition.lineIndex, newCursorPosition.position);
            }

            if (!isTextareaActive)
                NativeMethods.focus.call(codeEditor.$textarea[0]);

            Util.preventDefault(e);
        };

        //B237719 - TD6 - Unnecessary text selection and scroll codeEditor in IE
        if (!Util.isIE)
            mousedownHandler();
        else
            NativeMethods.setTimeout.call(window, mousedownHandler, 0);
    };

    CodeEditor.prototype._dblclick = function (e) {
        var ev = e.originalEvent || e,

            x = ev.pageX || ev.clientX,
            y = ev.pageY || ev.clientY;

        x = x - this._getEditorOffset().left + this.$scrollLayout[0].scrollLeft;
        y = y - this._getEditorOffset().top + this.$scrollLayout[0].scrollTop;

        //return if click on scroll
        if (y > this.$scrollLayout[0].offsetHeight + this.$scrollLayout[0].scrollTop - this._getHorizontalScrollHeight() ||
            x > this.$scrollLayout[0].offsetWidth + this.$scrollLayout[0].scrollLeft - this._getVerticalScrollWidth()) {
            return;
        }

        var positionLeft = this._findCurrentWordPosition(this.state.lineIndex, this.state.position),
            positionRight = this._findCurrentWordEndPosition(this.state.lineIndex, this.state.position),
            checkPositionLeft = this._findCurrentWordPosition(this.state.lineIndex, positionRight),
            checkPositionRight = this._findCurrentWordEndPosition(this.state.lineIndex, positionLeft);

        if (this.state.position === checkPositionLeft)
            positionLeft = this.state.position;
        else if (this.state.position === checkPositionRight) {
            positionRight = this.state.position;
        }
        this._setCursor(this.state.lineIndex, positionLeft, false);
        this._setCursor(this.state.lineIndex, positionLeft, true, false);
        this._setCursor(this.state.lineIndex, positionRight, true, true);

        e.preventDefault();
    };

    CodeEditor.prototype._mousemove = function (e) {
        if (!this.mouseLeftButtonPressed && !this.mouseRightButtonPressed)
            return;

        var target = e.target,
            $target = $(target);

        var x = e.pageX,
            y = e.pageY,
            cursorLineIndex = 0,
            cursorPosition = 0,
            newPosition = null;

        if (this.$editor.has($target).length || $target.is(this.$editor)) {
            x = x - this._getEditorOffset().left + this.$scrollLayout[0].scrollLeft;
            y = y - this._getEditorOffset().top + this.$scrollLayout[0].scrollTop;

            newPosition = this._getCodeEditorRelativePosition(x, y);
            cursorLineIndex = newPosition.cursorLineIndex;
            cursorPosition = newPosition.cursorPosition;
        }
        else {
            if (y > this._getEditorOffset().top + this.$editor.height()) {
                cursorLineIndex = this.$lines.length - 1;
                cursorPosition = this._getLineLength(cursorLineIndex);
            }
            else if (y > this._getEditorOffset().top) {
                cursorLineIndex = Math.min(
                    Math.floor((y + this.$scrollLayout[0].scrollTop - this._getEditorOffset().top) / this._getLineHeight()),
                    this.$lines.length - 1
                );
                x = x - this._getEditorOffset().left + this.$scrollLayout[0].scrollLeft;
                cursorPosition = Math.min(this._getLineLength(cursorLineIndex), Math.round(x / this._getLetterWidth(cursorLineIndex)));
            }

            //B239062 - Speedy selection in addAction dialog's codeEditor doesn't work
            if (e.pageX < this._getEditorOffset().left)
                cursorPosition = 0;
        }

        if (cursorLineIndex === this.state.lineIndex && cursorPosition === this.state.position)
            return;

        if (this.mouseLeftButtonPressed)
            this._setCursor(cursorLineIndex, cursorPosition, true, true);
        else if (newPosition)
            this._setTextarea(newPosition.lineIndex, newPosition.position);
    };

    CodeEditor.prototype._letterPressed = function (charCode) {
        if (this._hasSelection(this.state))
            this._removeSelectionContent();

        var newText = String.fromCharCode(charCode),
            text = this._getLineText(this.state.lineIndex),
            nextSymbolCharCode = text.charCodeAt(this.state.position);

        if ($.inArray(charCode, OPENING_BRACKET_CHAR_CODES) !== -1 || charCode === 39 || charCode === 34) {
            //NOTE: if empty line, space, one of the closing bracket or ";"
            var needPastePairChar = (text.length === this.state.position ||
                nextSymbolCharCode === 32 || nextSymbolCharCode === 59 ||
                ($.inArray(nextSymbolCharCode, CLOSING_BRACKET_CHAR_CODES) !== -1));
            if (needPastePairChar) {
                switch (charCode) {
                    case 40:
                        newText += String.fromCharCode(41);
                        break;
                    case 91:
                        newText += String.fromCharCode(93);
                        break;
                    case 123:
                        newText += String.fromCharCode(125);
                        break;
                    case 39:
                        newText += String.fromCharCode(39);
                        break;
                    case 34:
                        newText += String.fromCharCode(34);
                        break;
                }
            }
        }
        if (charCode === nextSymbolCharCode &&
            (($.inArray(nextSymbolCharCode, CLOSING_BRACKET_CHAR_CODES) !== -1) ||
                nextSymbolCharCode === 39 || nextSymbolCharCode === 34)) {
            this._moveRight();
            return;
        }
        this._setLineText(this.state.lineIndex, text.substring(0, this.state.position) + newText + text.substring(this.state.position, text.length));
        this._moveRight();
        this._textChanged();
    };

    CodeEditor.prototype._shortcutHandlers = function () {
        var codeEditor = this;
        return {
            left: function () {
                if (codeEditor._hasSelection(codeEditor.state) && !codeEditor.state.selectionRegion.inverse) {
                    codeEditor._setCursor(codeEditor.state.selectionRegion.startLine, codeEditor.state.selectionRegion.startPosition, true);
                }
                else
                    codeEditor._moveLeft();
            },

            up: function () {
                codeEditor._moveUp();
            },

            right: function () {
                if (codeEditor._hasSelection(codeEditor.state) && codeEditor.state.selectionRegion.inverse) {
                    codeEditor._setCursor(codeEditor.state.selectionRegion.endLine, codeEditor.state.selectionRegion.endPosition, true);
                }
                else
                    codeEditor._moveRight(true);
            },

            down: function () {
                codeEditor._moveDown();
            },

            enter: function () {
                if (codeEditor._hasSelection(codeEditor.state))
                    codeEditor._removeSelectionContent();
                var text = codeEditor._getLineText(codeEditor.state.lineIndex),
                    nextLineText = text.substring(codeEditor.state.position, text.length),
                    curLineNewText = text.substring(0, codeEditor.state.position),

                //auto-indent
                    nonSpaceRegExp = /\S/g,
                    firstSymbol = nonSpaceRegExp.exec(curLineNewText),
                    indent = firstSymbol ? nonSpaceRegExp.lastIndex - 1 : curLineNewText.length,
                    indentString = [''];
                indentString.length += indent;
                indentString = indentString.join(' ');

                codeEditor._setLineText(codeEditor.state.lineIndex, curLineNewText);

                if (codeEditor._checkBracketPairCharCodes(text.charCodeAt(codeEditor.state.position - 1), nextLineText.charCodeAt(0))) {
                    indent += 4;
                    codeEditor._addLine(codeEditor.state.lineIndex + 1, TAB_INDENT + indentString);
                    codeEditor._addLine(codeEditor.state.lineIndex + 2, indentString + nextLineText);
                }
                else {
                    codeEditor._addLine(codeEditor.state.lineIndex + 1, indentString + nextLineText);
                }
                codeEditor._setCursor(codeEditor.state.lineIndex + 1, indent);
                codeEditor._textChanged();
            },

            backspace: function () {
                if (codeEditor._hasSelection(codeEditor.state))
                    codeEditor._removeSelectionContent();
                else {
                    var text = codeEditor._getLineText(codeEditor.state.lineIndex);
                    if (codeEditor.state.position === 0) {
                        if (!codeEditor.$lines[codeEditor.state.lineIndex - 1])
                            return;
                        var newPosition = codeEditor._getLineLength(codeEditor.state.lineIndex - 1);
                        codeEditor._setLineText(codeEditor.state.lineIndex - 1, codeEditor._getLineText(codeEditor.state.lineIndex - 1) + text);
                        codeEditor._removeLine(codeEditor.state.lineIndex);
                        codeEditor._setCursor(codeEditor.state.lineIndex - 1, newPosition);
                    }
                    else {
                        if (codeEditor._checkBracketPairCharCodes(text.charCodeAt(codeEditor.state.position - 1), text.charCodeAt(codeEditor.state.position)) ||
                            codeEditor._checkQuotesPairCharCodes(text.charCodeAt(codeEditor.state.position - 1), text.charCodeAt(codeEditor.state.position)))
                            codeEditor._setLineText(codeEditor.state.lineIndex,
                                text.substring(0, codeEditor.state.position - 1) + text.substring(codeEditor.state.position + 1, text.length));
                        else
                            codeEditor._setLineText(codeEditor.state.lineIndex,
                                text.substring(0, codeEditor.state.position - 1) + text.substring(codeEditor.state.position, text.length));
                        codeEditor._setCursor(codeEditor.state.lineIndex, codeEditor.state.position - 1);
                    }
                }
                codeEditor._textChanged();
            },

            'delete': function () {
                if (codeEditor._hasSelection(codeEditor.state))
                    codeEditor._removeSelectionContent();
                else {
                    var text = codeEditor._getLineText(codeEditor.state.lineIndex);
                    if (codeEditor.state.position === codeEditor._getLineLength(codeEditor.state.lineIndex)) {
                        if (!codeEditor.$lines[codeEditor.state.lineIndex + 1])
                            return;
                        codeEditor._setLineText(codeEditor.state.lineIndex, text + codeEditor._getLineText(codeEditor.state.lineIndex + 1));
                        codeEditor._removeLine(codeEditor.state.lineIndex + 1);
                    }
                    else {
                        codeEditor._setLineText(codeEditor.state.lineIndex,
                            text.substring(0, codeEditor.state.position) + text.substring(codeEditor.state.position + 1, text.length));
                    }
                }
                codeEditor._textChanged();
            },

            home: function () {
                var lineSpaceCount = codeEditor._getLineSpaceCount(codeEditor.state.lineIndex),
                    position = 0;
                if (codeEditor.state.position === 0 || codeEditor.state.position > lineSpaceCount)
                    position = lineSpaceCount;
                codeEditor._setCursor(codeEditor.state.lineIndex, position, true);
            },

            end: function () {
                if (codeEditor.state.position !== codeEditor._getLineLength(codeEditor.state.lineIndex))
                    codeEditor._setCursor(codeEditor.state.lineIndex, codeEditor._getLineLength(codeEditor.state.lineIndex), true);
            },

            tab: function () {
                var text = codeEditor._getLineText(codeEditor.state.lineIndex);
                if (!codeEditor._hasSelection(codeEditor.state)) {
                    var spacesCount = 4 - (codeEditor.state.position % 4),
                        indent = '';
                    for (var i = 0; i < spacesCount; i++)
                        indent += ' ';
                    codeEditor._setLineText(codeEditor.state.lineIndex, text.substring(0, codeEditor.state.position) + indent + text.substring(codeEditor.state.position, text.length));
                    codeEditor._setCursor(codeEditor.state.lineIndex, codeEditor.state.position + spacesCount, false);
                }
                else {
                    for (var j = codeEditor.state.selectionRegion.startLine; j <= codeEditor.state.selectionRegion.endLine; j++)
                        if (codeEditor._getLineLength(j) !== 0)
                            codeEditor._setLineText(j, TAB_INDENT + codeEditor._getLineText(j));
                    codeEditor._shiftSelection(0, TAB_INDENT.length);

                    codeEditor._setCursor(codeEditor.state.lineIndex, codeEditor.state.position + TAB_INDENT.length > text.length ?
                        text.length : codeEditor.state.position + TAB_INDENT.length, false, true);
                }
                codeEditor._textChanged();
            },

            shiftTab: function () {
                var tabSpaceCount = 4,
                    lineSpaceCount = codeEditor._getLineSpaceCount(codeEditor.state.lineIndex),
                    removingSpaceCount = lineSpaceCount >= tabSpaceCount ?
                        tabSpaceCount :
                        lineSpaceCount,
                    saveState = false;

                if (!codeEditor._hasSelection(codeEditor.state))
                    saveState = codeEditor._removeLineTabIndent(codeEditor.state.lineIndex);
                else {
                    for (var i = codeEditor.state.selectionRegion.startLine; i <= codeEditor.state.selectionRegion.endLine; i++)
                        if (codeEditor._removeLineTabIndent(i))
                            saveState = true;
                    codeEditor._shiftSelection(0, -removingSpaceCount);
                }

                codeEditor._setCursor(codeEditor.state.lineIndex,
                    codeEditor.state.position - removingSpaceCount > 0 ?
                        codeEditor.state.position - removingSpaceCount : 0, false, codeEditor._hasSelection(codeEditor.state));

                if (saveState)
                    codeEditor._textChanged();
            },

            shiftLeft: function () {
                codeEditor._moveLeft(true);
            },

            shiftRight: function () {
                codeEditor._moveRight(true, true);
            },

            shiftUp: function () {
                codeEditor._moveUp(true);
            },

            shiftDown: function () {
                codeEditor._moveDown(true);
            },

            shiftHome: function () {
                var lineSpaceCount = codeEditor._getLineSpaceCount(codeEditor.state.lineIndex),
                    position = 0;
                if (codeEditor.state.position === 0 || codeEditor.state.position > lineSpaceCount)
                    position = lineSpaceCount;
                codeEditor._setCursor(codeEditor.state.lineIndex, position, true, true);
            },

            shiftEnd: function () {
                codeEditor._setCursor(codeEditor.state.lineIndex, codeEditor._getLineLength(codeEditor.state.lineIndex), true, true);
            },

            ctrlLeft: function () {
                var newPosition = codeEditor._findWordBoundary('left');
                if (codeEditor.state.lineIndex !== newPosition.lineIndex || codeEditor.state.position !== newPosition.position)
                    codeEditor._setCursor(newPosition.lineIndex, newPosition.position, true);
            },

            ctrlRight: function () {
                var newPosition = codeEditor._findWordBoundary('right');
                if (codeEditor.state.lineIndex !== newPosition.lineIndex || codeEditor.state.position !== newPosition.position)
                    codeEditor._setCursor(newPosition.lineIndex, newPosition.position, true);
            },

            ctrlBackspace: function () {
                if (codeEditor._hasSelection(codeEditor.state))
                    codeEditor._removeSelectionContent();
                else {
                    var lineIndex = codeEditor.state.lineIndex,
                        position = codeEditor.state.position;

                    if (lineIndex === 0 && position === 0)
                        return;

                    var nextPosition = codeEditor._findWordBoundary('left'),
                        lineText = codeEditor._getLineText(lineIndex);

                    codeEditor._setCursor(nextPosition.lineIndex, nextPosition.position);

                    if (nextPosition.lineIndex === lineIndex) {
                        codeEditor._setLineText(lineIndex, lineText.substring(0, nextPosition.position) + lineText.substring(position, codeEditor._getLineLength(lineIndex)));
                    }
                    else {
                        codeEditor._setLineText(nextPosition.lineIndex, codeEditor._getLineText(nextPosition.lineIndex) + lineText.substring(position, codeEditor._getLineLength(lineIndex)));
                        codeEditor._removeLine(lineIndex);
                    }
                }
                codeEditor._textChanged();
            },

            ctrlDelete: function () {
                if (codeEditor._hasSelection(codeEditor.state))
                    codeEditor._removeSelectionContent();
                else {
                    var lineIndex = codeEditor.state.lineIndex,
                        position = codeEditor.state.position,
                        lineText = codeEditor._getLineText(lineIndex);

                    if (lineIndex === codeEditor.$lines.length - 1 && position === lineText.length)
                        return;

                    var nextPosition = codeEditor._findWordBoundary('right');

                    if (nextPosition.lineIndex === lineIndex)
                        codeEditor._setLineText(lineIndex, lineText.substring(0, position) +
                            lineText.substring(nextPosition.position, codeEditor._getLineLength(lineIndex)));
                    else {
                        codeEditor._setLineText(lineIndex, codeEditor._getLineText(lineIndex) +
                            codeEditor._getLineText(nextPosition.lineIndex).substring(nextPosition.position, codeEditor._getLineLength(nextPosition.lineIndex)));
                        codeEditor._removeLine(nextPosition.lineIndex);
                    }
                }
                codeEditor._textChanged();
            },

            ctrlHome: function () {
                if (codeEditor.state.lineIndex !== 0 || codeEditor.state.position !== 0)
                    codeEditor._setCursor(0, 0, true);
            },

            ctrlEnd: function () {
                if (codeEditor.state.lineIndex !== codeEditor.$lines.length - 1 || codeEditor.state.position !== codeEditor._getLineLength(codeEditor.$lines.length - 1))
                    codeEditor._setCursor(codeEditor.$lines.length - 1, codeEditor._getLineLength(codeEditor.$lines.length - 1), true);
            },

            ctrlShiftLeft: function () {
                var newPosition = codeEditor._findWordBoundary('left');
                if (codeEditor.state.lineIndex !== newPosition.lineIndex || codeEditor.state.position !== newPosition.position)
                    codeEditor._setCursor(newPosition.lineIndex, newPosition.position, true, true);
            },

            ctrlShiftRight: function () {
                var newPosition = codeEditor._findWordBoundary('right');
                if (codeEditor.state.lineIndex !== newPosition.lineIndex || codeEditor.state.position !== newPosition.position)
                    codeEditor._setCursor(newPosition.lineIndex, newPosition.position, true, true);
            },

            ctrlShiftUp: function () {
                codeEditor._moveLines(-1);
            },

            ctrlShiftDown: function () {
                codeEditor._moveLines(1);
            },

            ctrlX: function () {
                codeEditor._onCutCodeEditor();
            },

            ctrlC: function () {
                if (codeEditor._hasSelection(codeEditor.state)) {
                    codeEditor.buffer = codeEditor._getSelectionTextAsArray();
                    codeEditor._onCopyCodeEditorText();
                }
            },

            ctrlV: function () {
                codeEditor._onPasteCodeEditor();
            },

            ctrlA: function () {
                var lastLine = codeEditor.$lines.length - 1,
                    lastPosition = codeEditor._getLineLength(lastLine);
                codeEditor._setCursor(0, 0);
                codeEditor._setCursor(lastLine, lastPosition, true, true);
            },

            ctrlZ: function () {
                if (codeEditor.stateStack[codeEditor.stateIndex - 1]) {
                    codeEditor._undoState(codeEditor.stateIndex);
                    codeEditor.stateIndex = codeEditor.stateIndex - 1;
                }
            },
            ctrlShiftZ: function () {
                if (codeEditor.stateStack[codeEditor.stateIndex + 1]) {
                    codeEditor._redoState(codeEditor.stateIndex);
                    codeEditor.stateIndex = codeEditor.stateIndex + 1;
                }
            }
        };
    };

    CodeEditor.prototype._createSelectionObject = function () {
        var codeEditor = this,
            selection = null;
        if (window.getSelection) {
            selection = window.getSelection();
            selection.removeAllRanges();
        }
        else {
            selection = document.selection;
            selection.clear();
        }

        codeEditor.$textarea[0].value = Util.isMozilla ? codeEditor._getSelectionTextAsArray().join('\n') : '12';

        codeEditor.$textarea[0].select();
    };

    CodeEditor.prototype._onCopyCodeEditorText = function (lineIndex) {
        var codeEditor = this;
        var selection;
        if (window.getSelection) {
            selection = window.getSelection();
            selection.removeAllRanges();
        }
        else {
            selection = document.selection;
            selection.clear();
        }
        codeEditor.holdFocus = true;
        var $newTextarea = codeEditor._createTextareaForCopyCodeEditorText(lineIndex);
        try {
            $newTextarea[0].select();
        }
        catch (e) {
            $newTextarea[0].select();
        }
        NativeMethods.setTimeout.call(window, function () {
            $newTextarea.remove();
            //NOTE: we should set holdFocus flag to true only after codeEditor textarea focused
            // otherwise codeEditor may lost focus (especially in IE)
            NativeMethods.focus.call(codeEditor.$textarea[0]);
            NativeMethods.setTimeout.call(window, function () {
                codeEditor.holdFocus = false;
            }, 0);
        }, 0);
    };

    CodeEditor.prototype._onPasteCodeEditor = function () {
        var codeEditor = this;
        if (codeEditor.inPaste)
            return;
        codeEditor.inPaste = true;
        NativeMethods.setTimeout.call(window, function () {
            var textareaValue = codeEditor.$textarea[0].value;
            if (textareaValue) {
                codeEditor.buffer = [];
                var pasteText = textareaValue.replace(/\r\n/g, '\n'),
                    pasteLines = pasteText.split('\n');
                $.each(pasteLines, function (index, line) {
                    codeEditor.buffer.push(line);
                });
                codeEditor.$textarea[0].value = '';
            }
            codeEditor._insertText(codeEditor.buffer);
            codeEditor.inPaste = false;
        }, 0);
    };

    CodeEditor.prototype._insertText = function (lines) {
        if (lines.length) {
            var text = this._getLineText(this.state.lineIndex);
            if (this._hasSelection(this.state)) {
                this._removeSelectionContent();
                text = this._getLineText(this.state.lineIndex);
            }
            if (lines.length === 1) {
                this._setLineText(this.state.lineIndex, text.substring(0, this.state.position) + lines[0] + text.substring(this.state.position, text.length));
                this._setCursor(this.state.lineIndex, this.state.position + lines[0].length);
            }
            else {
                this._setLineText(this.state.lineIndex, text.substring(0, this.state.position) + lines[0]);
                var endText = text.substring(this.state.position);

                this._addLines(this.state.lineIndex + 1, lines.slice(1, lines.length - 1));
                this.state.lineIndex += lines.length - 1;
                this._addLine(this.state.lineIndex, lines[lines.length - 1] + endText);
                this._setCursor(this.state.lineIndex, lines[lines.length - 1].length);
            }
            this._textChanged();
        }
    };

    CodeEditor.prototype._onCutCodeEditor = function () {
        var codeEditor = this;
        if (codeEditor._hasSelection(codeEditor.state)) {
            codeEditor.buffer = codeEditor._getSelectionTextAsArray();
            codeEditor._onCopyCodeEditorText();
            codeEditor._removeSelectionContent();
        }
        else {
            codeEditor.buffer[0] = codeEditor._getLineText(codeEditor.state.lineIndex);
            codeEditor.buffer[1] = '';
            codeEditor._onCopyCodeEditorText(codeEditor.state.lineIndex);
            if (codeEditor.state.lineIndex !== codeEditor.$lines.length - 1) {
                codeEditor._removeLine(codeEditor.state.lineIndex);

                var curPosition = codeEditor._getLineLength(codeEditor.state.lineIndex) > codeEditor.state.position ?
                    codeEditor.state.position :
                    codeEditor._getLineLength(codeEditor.state.lineIndex);

                codeEditor._setCursor(codeEditor.state.lineIndex, curPosition, false);
            }
            else if (codeEditor._getLineText(codeEditor.state.lineIndex).length !== 0) {
                codeEditor._setLineText(codeEditor.state.lineIndex, '');
                codeEditor._setCursor(codeEditor.state.lineIndex, 0);
            }
            else
                return;
        }
        codeEditor._textChanged();
    };

//utils
    CodeEditor.prototype._getLetterWidth = function (lineIndex) {     //TODO: without lineIndex (or cache value)
        var $span = this.$lines[lineIndex].children('span'),
            textLength = $span.text().length,
            currentWidth = textLength ? ($span.width() / textLength) : 0;

        if (!this.letterWidth)
            this.letterWidth = this._calculateLetterWidth();

        return currentWidth || this.letterWidth;
    };

    CodeEditor.prototype._calculateLineHeight = function () {
        var $line = this.$lines[0],
            $testSpan = $('<span>')
                .text('a')
                .css('fontFamily', $line.css('fontFamily'))
                .css('fontSize', $line.css('fontSize'))
                .css('position', 'absolute')
                .appendTo(ShadowUI.getRoot()),
            height = $testSpan.height();
        $testSpan.remove();
        return height;
    };

    CodeEditor.prototype._calculateLetterWidth = function () {
        var $testSpan = $('<span></span>')
            .text('1234567890')
            .css('position', 'absolute')
            .appendTo(this.$editor);

        ShadowUI.addClass($testSpan, LETTER_WIDTH_FINDER);

        var width = $testSpan.width() / $testSpan.text().length;

        $testSpan.remove();
        return width;
    };

    CodeEditor.prototype._getLineHeight = function () {
        var codeEditor = this;
        if (!codeEditor.lineHeight) {
            codeEditor.lineHeight = codeEditor._calculateLineHeight();
        }
        return codeEditor.lineHeight;
    };

    CodeEditor.prototype._removeLineTabIndent = function (lineIndex) {
        var lineText = this._getLineText(lineIndex),
            tabSpaceCount = 4,
            lineSpaceCount = this._getLineSpaceCount(lineIndex),
            removingSpaceCount = lineSpaceCount >= tabSpaceCount ?
                tabSpaceCount :
                lineSpaceCount;
        if (!removingSpaceCount)
            return false;
        this._setLineText(lineIndex, lineText.substring(removingSpaceCount, lineText.length));
        return true;
    };

    CodeEditor.prototype._getLineSpaceCount = function (lineIndex) {
        var lineText = this._getLineText(lineIndex),
            lineLength = lineText.length,
            lineSpaceCount = 0;
        for (var i = 0; i < lineLength; i++)
            if (/\s/.test(lineText[i]))
                lineSpaceCount++;
            else break;
        return lineSpaceCount;
    };

    CodeEditor.prototype._getCodeEditorRelativePosition = function (x, y) {
        var codeEditor = this,
            position = {
                lineIndex: 0,
                position: 0,
                cursorLineIndex: 0,
                cursorPosition: 0
            };

        position.lineIndex = Math.floor(y / codeEditor._getLineHeight());

        position.cursorLineIndex = position.lineIndex >= codeEditor.$lines.length ?
            codeEditor.$lines.length - 1 :
            position.lineIndex;

        var letterWidth = position.lineIndex >= codeEditor.$lines.length ?
            this.letterWidth : codeEditor._getLetterWidth(position.lineIndex);
        position.position = Math.round(x / letterWidth);

        position.cursorPosition = position.position > codeEditor._getLineLength(position.cursorLineIndex) ?
            codeEditor._getLineLength(position.cursorLineIndex) :
            position.position;
        return position;
    };

    CodeEditor.prototype._findWordBoundary = function (direction) {
        var offset = (direction === 'left') ? -1 : 1;
        var nextPosition = {
            lineIndex: this.state.lineIndex,
            position: this.state.position
        };

        if ((offset === -1 && nextPosition.position === 0) ||
            (offset === 1 && nextPosition.position === this._getLineLength(nextPosition.lineIndex))) {
            if (this.$lines[nextPosition.lineIndex + offset]) {
                nextPosition.lineIndex = nextPosition.lineIndex + offset;
                nextPosition.position = (offset === -1) ? this._getLineLength(nextPosition.lineIndex) : 0;
            }
        }
        else
            nextPosition.position = (offset === -1) ? this._findCurrentWordPosition(nextPosition.lineIndex, nextPosition.position) : this._findNextWordPosition(nextPosition.lineIndex, nextPosition.position);

        return nextPosition;
    };

    CodeEditor.prototype._findCurrentWordPosition = function (lineIndex, position) {
        var nextPosRegExp = /(\b\S)|$|(\s[\W]+)/g,
            lineText = this._getLineText(lineIndex),
            text = lineText.substring(0, position),
            lastRes = '',
            lastIndex = '',
            res = '';

        do {
            lastRes = res;
            lastIndex = nextPosRegExp.lastIndex;
            res = nextPosRegExp.exec(text)[0];
            if (!/^\s+$/.test(res))
                res = res.replace(/^\s+/, '');
        }
        while (res);
        return lastIndex - lastRes.length;
    };

    CodeEditor.prototype._findCurrentWordEndPosition = function (lineIndex, position) {
        var nextPosRegExp = /(\S\b)|$|(\s[\W]+)/g,
            lineText = this._getLineText(lineIndex),
            text = lineText.substring(position, lineText.length);

        nextPosRegExp.exec(text)[0].replace(/^\s+/, '');

        position = position + (nextPosRegExp.lastIndex);

        return position > this._getLineLength(lineIndex) ?
            this._getLineLength(lineIndex) :
            position;
    };

    CodeEditor.prototype._findNextWordPosition = function (lineIndex, position) {
        var nextPosRegExp = /(\b\S)|$|(\s[\W]+)/g,
            lineText = this._getLineText(lineIndex),
            text = lineText.substring(position, lineText.length),
            res = nextPosRegExp.exec(text)[0].replace(/^\s+/, '');

        if (nextPosRegExp.lastIndex < 2)
            res = nextPosRegExp.exec(text)[0].replace(/^\s+/, '');

        position = position + (nextPosRegExp.lastIndex - res.length);

        return position > this._getLineLength(lineIndex) ?
            this._getLineLength(lineIndex) :
            position;
    };

    CodeEditor.prototype._findLongerLineIndex = function () {
        var codeEditor = this;
        if (codeEditor.$lines.length) {
            var index = 0,
                temp = codeEditor._getLineLength(0);
            $.each(codeEditor.$lines, function (i) {
                if (codeEditor._getLineLength(i) > temp) {
                    index = i;
                    temp = codeEditor._getLineLength(i);
                }
            });
            return index;
        }
        return 0;
    };

    CodeEditor.prototype._getEditorOffset = function () {
        return Util.getOffsetPosition(this.$scrollLayout[0]);
    };

    CodeEditor.prototype._updateEditorSize = function () {
        var longestLine = this._findLongerLineIndex(),
            scrollLayoutBorders = parseInt(this.$scrollLayout.css('borderLeftWidth')) + parseInt(this.$scrollLayout.css('borderRightWidth')),
            editLayoutWidth = Math.max(this.editorWidth - scrollLayoutBorders,
                this._getLineLength(longestLine) * this._getLetterWidth(longestLine) + this.$cursor.width());
        if (this.expanded) {
            var fullWidth = Math.min(editLayoutWidth, this._getLetterWidth(longestLine) * EDITOR_MAX_EXPANDED_SYMBOLS_COUNT + this.$cursor.width()) + scrollLayoutBorders;
            this.$editLayout.css('width', editLayoutWidth);
            if (this.expandDirection === 'left') {
                this.$scrollLayout.css('marginLeft', this.editorWidth - fullWidth);
                this.$scrollLayout.css('position', 'absolute');
            }
            this.$scrollLayout.width(fullWidth);
            this.$scrollLayout.css('overflow-x', fullWidth < editLayoutWidth + scrollLayoutBorders ? 'scroll' : 'hidden');

        }
        else {
            var scrollLayoutWidth = this.$scrollLayout[0].clientWidth || this.$scrollLayout.width();
            if (editLayoutWidth > scrollLayoutWidth) {
                this.$editLayout.css('width', editLayoutWidth);
                this.$scrollLayout.css('overflow-x', this.floatingWidth ? 'hidden' : 'scroll');
            }
            else {
                this.$editLayout.css('width', scrollLayoutWidth);
                this.$scrollLayout.css('overflow-x', 'hidden');
            }
        }

        var linesHeight = this.$lines.length * this._getLineHeight(),
            scrollHeight = this._getHorizontalScrollHeight();
        if (this.fixedHeight) {
            if (linesHeight + scrollHeight > this.editorHeight) {
                this.$editLayout.css('height', linesHeight);
                this.$scrollLayout.css('overflowY', 'scroll');
            }
            else {
                this.$editLayout.css('height', this.editorHeight - scrollHeight);
                this.$scrollLayout.css('overflowY', 'hidden');
            }
        }
        else {
            var height = Math.max(this.editorHeight, linesHeight + scrollHeight) + EDITOR_PADDING_BOTTOM;
            this.$editLayout.css('height', '');
            this.$editor.css('height', height);
            this.$scrollLayout.css('height', height);
            this.$editLayout.css('height', this.$scrollLayout.height() - scrollHeight - EDITOR_PADDING_BOTTOM);
        }

        this.$selectionCenter.css('width', this.$editLayout.width() + 'px');
    };

    CodeEditor.prototype._getLineLength = function (lineIndex) {
        return this.$lines[lineIndex] ? this._getLineText(lineIndex).length : 0;
    };

    CodeEditor.prototype._getElementPosition = function (lineIndex, position) {
        var letterWidth = lineIndex >= this.$lines.length ? this.letterWidth : this._getLetterWidth(lineIndex);

        return {
            left: position * letterWidth,
            top: lineIndex * this._getLineHeight()
        };
    };

    CodeEditor.prototype._setElementPosition = function ($element, lineIndex, position) {
        var newPosition = this._getElementPosition(lineIndex, position);

        $element.css('left', Math.floor(newPosition.left));

        if (this.$lines[lineIndex])
            $element.css('top', newPosition.top);
    };

    CodeEditor.prototype._getVerticalScrollWidth = function () {
        var bordersWidth = parseInt(this.$scrollLayout.css('borderLeftWidth').replace('px', '')) +
            parseInt(this.$scrollLayout.css('borderRightWidth').replace('px', ''));
        return this.$scrollLayout[0].offsetWidth - this.$scrollLayout[0].clientWidth - bordersWidth;
    };

    CodeEditor.prototype._getHorizontalScrollHeight = function () {
        var bordersWidth = parseInt(this.$scrollLayout.css('borderTopWidth').replace('px', '')) +
            parseInt(this.$scrollLayout.css('borderBottomWidth').replace('px', ''));
        return this.$scrollLayout[0].offsetHeight - this.$scrollLayout[0].clientHeight - bordersWidth;
    };

    CodeEditor.prototype._hasSelection = function (state) {
        return !(state.selectionRegion.startLine === state.selectionRegion.endLine &&
            state.selectionRegion.startPosition === state.selectionRegion.endPosition);
    };

    CodeEditor.prototype._selectionContains = function (position) {
        return ((position.lineIndex > this.state.selectionRegion.startLine ||
            (position.lineIndex === this.state.selectionRegion.startLine && position.position >= this.state.selectionRegion.startPosition)) &&
            (position.lineIndex < this.state.selectionRegion.endLine ||
                (position.lineIndex === this.state.selectionRegion.endLine && position.position <= this.state.selectionRegion.endPosition)));
    };

    CodeEditor.prototype._createTextareaForCopyCodeEditorText = function (lineIndex) {
        var codeEditor = this,
            newTextarea = document.createElement('textarea'),
            $newTextarea = $(newTextarea);

        $newTextarea.appendTo(ShadowUI.getRoot());

        $newTextarea.css({
            overflow: 'hidden',
            width: '1px',
            height: '1px',
            position: 'absolute',
            top: '-999px',
            left: '-999px'
        });

        var selectionText = typeof lineIndex !== 'undefined' ?
            codeEditor._getLineText(lineIndex) + '\r\n' :
            this._getSelectionTextAsArray().join('\n');
        $newTextarea.text(selectionText);
        return $newTextarea;
    };

    CodeEditor.prototype._checkBracketPairCharCodes = function (charCodeStart, charCodeEnd) {
        return ((charCodeStart === 40 && charCodeEnd === 41) ||
            (charCodeStart === 91 && charCodeEnd === 93) ||
            (charCodeStart === 123 && charCodeEnd === 125));
    };

    CodeEditor.prototype._checkQuotesPairCharCodes = function (charCodeStart, charCodeEnd) {
        return ((charCodeStart === 39 && charCodeEnd === 39) ||
            (charCodeStart === 34 && charCodeEnd === 34));
    };

    CodeEditor.prototype._getEditorText = function () {
        var codeEditor = this,

            lines = $.map(codeEditor.$lines, function (item, index) {
                return codeEditor._getLineText(index);
            });
        return lines.join('\r\n').replace(/\xA0/g, ' ');
    };

    CodeEditor.prototype._setEditorText = function (value) {
        if (this.$lines.length) {
            this._clearSelection(0, 0);
            while (this.$lines.length)
                this._removeLine(0);
        }
        var text = value.replace(/\r\n/g, '\n');
        this._addLines(undefined, text.split('\n'));
    };

// Undo/Redo
    CodeEditor.prototype._saveState = function (isCursor) {
        this.changes.removed.sort(function (i, j) {
            return (i.index - j.index);
        });
        this.changes.added.sort(function (i, j) {
            return (i.index - j.index);
        });
        var lastState = this.stateStack[this.stateIndex],
            newState = {
                lineIndex: this.state.lineIndex,
                position: this.state.position,

                selectionRegion: {
                    startPosition: this.state.selectionRegion.startPosition,
                    startLine: this.state.selectionRegion.startLine,
                    endPosition: this.state.selectionRegion.endPosition,
                    endLine: this.state.selectionRegion.endLine,
                    inverse: this.state.selectionRegion.inverse
                },
                isCursor: isCursor,
                changes: this.changes
            };

        this._clearChanges();

        if (isCursor && !(newState.changes.added.length || newState.changes.modified.length || newState.changes.removed.length) &&
            newState.lineIndex === lastState.lineIndex && newState.position === lastState.position)
            return;

        if (lastState && this.stateIndex !== this.stateStack.length - 1) {
            if (!isCursor)
                while (this.stateIndex < this.stateStack.length - 1) {
                    this.stateStack.pop();
                }
            else
                return;
        }

        if (isCursor && lastState && lastState.isCursor)
            this.stateStack.pop();

        if (this.stateStack.length === STATE_STACK_SIZE)
            this.stateStack.shift();

        this.stateStack.push(newState);
        this.stateIndex = this.stateStack.length - 1;
    };

    CodeEditor.prototype._redoState = function (stateIndex) {
        var codeEditor = this,
            redoState = codeEditor.stateStack[stateIndex + 1];

        for (var i = 0; i < redoState.changes.modified.length; i++)
            codeEditor._setLineText(redoState.changes.modified[i].index, redoState.changes.modified[i].newValue);

        for (var j = 0; j < redoState.changes.added.length; j++)
            codeEditor._addLine(redoState.changes.added[j].index, redoState.changes.added[j].value);

        for (var k = redoState.changes.removed.length - 1; k >= 0; k--)
            codeEditor._removeLine(redoState.changes.removed[k].index);

        codeEditor._applyState(redoState);
        this.eventEmitter.emit(CodeEditor.CHANGE_EVENT, {text: this.getText()});
    };

    CodeEditor.prototype._undoState = function (stateIndex) {
        var codeEditor = this,
            changes = codeEditor.stateStack[stateIndex].changes,
            redoState = codeEditor.stateStack[stateIndex - 1];

        for (var i = 0; i < changes.modified.length; i++)
            codeEditor._setLineText(changes.modified[i].index, changes.modified[i].oldValue);

        for (var j = changes.added.length - 1; j >= 0; j--)
            codeEditor._removeLine(changes.added[j].index);

        for (var k = 0; k < changes.removed.length; k++)
            codeEditor._addLine(changes.removed[k].index, changes.removed[k].value);

        codeEditor._applyState(redoState);
        this.eventEmitter.emit(CodeEditor.CHANGE_EVENT, {text: this.getText()});
    };

    CodeEditor.prototype._applyState = function (state) {
        var codeEditor = this;
        this._clearChanges();
        codeEditor._updateEditorSize();
        if (codeEditor._hasSelection(state)) {
            var startLineIndex = state.selectionRegion.inverse ?
                    state.selectionRegion.endLine : state.selectionRegion.startLine,
                startPosition = state.selectionRegion.inverse ?
                    state.selectionRegion.endPosition : state.selectionRegion.startPosition;

            codeEditor._setCursor(startLineIndex, startPosition);
            codeEditor.state.selectionRegion.inverse = state.inverse;
            codeEditor._setCursor(state.lineIndex, state.position, false, true);
        }
        else
            codeEditor._setCursor(state.lineIndex, state.position);

        codeEditor._updateScroll();
        codeEditor._highlight();
    };

    CodeEditor.prototype._clearChanges = function () {
        this.changes = {
            modified: [],
            added: [],
            removed: []
        };
    };

    CodeEditor.prototype._getStateIndex = function (changesArray, lineIndex) {
        var changeIndex;
        $.each(changesArray, function (index, value) {
            if (value.index === lineIndex) {
                changeIndex = index;
                return false;
            }
        });
        return changeIndex;
    };

//API
    CodeEditor.prototype.enableEdit = function (value) {
        if (this.allowEdit === value)
            return;
        this.allowEdit = value;
        if (value) {
            if (!this.keyEventParser)
                this._initEventParser();
            ShadowUI.removeClass(this.$editor, DISABLED_CLASS);
        }
        else {
            ShadowUI.addClass(this.$editor, DISABLED_CLASS);
        }
    };

    CodeEditor.prototype.getText = function () {
        return this._getEditorText();
    };

    CodeEditor.prototype.setText = function (value) {
        var prevValue = this._getEditorText();
        if (prevValue !== value) {
            this._setEditorText(value);
            this._setCursor(this.$lines.length - 1, this._getLineLength(this.$lines.length - 1));
            this._textChanged();
        }
    };

    CodeEditor.prototype.insertText = function (value) {
        this._insertText(value.replace(/\r\n/g, '\n').split('\n'));
    };

    CodeEditor.prototype.addClass = function (className) {
        ShadowUI.addClass(this.$editor, className);
    };

    CodeEditor.prototype.removeClass = function (className) {
        ShadowUI.removeClass(this.$editor, className);
    };

    CodeEditor.prototype.getContainer = function () {
        return this.$scrollLayout;
    };

    CodeEditor.prototype.blur = function () {
        var res = NativeMethods.blur.call(this.$textarea[0]);

        this.eventEmitter.emit(CodeEditor.BLUR_EVENT);

        return res;
    };

    CodeEditor.prototype.focus = function () {
        return NativeMethods.focus.call(this.$textarea[0]);
    };

    CodeEditor.prototype.setCursorToEnd = function () {
        var lastLine = this.$lines.length - 1;

        this._setCursor(lastLine, this._getLineLength(lastLine));
    };

    CodeEditor.prototype.stopSelectionProcess = function () {
        this.mouseRightButtonPressed = false;
        this.mouseLeftButtonPressed = false;

        if (!this._hasSelection(this.state)) {
            this.state.selectionRegion.endLine = this.state.lineIndex;
            this.state.selectionRegion.endPosition = this.state.position;
        }

        if (!Util.isMozilla)
            return false;
    };
});