TestCafeClient.define('UI.RecorderWidgets.DialogProperty', function (require) {
    var Hammerhead = HammerheadClient.get('Hammerhead'),
        $ = Hammerhead.$,
        Util = Hammerhead.Util,
        ShadowUI = Hammerhead.ShadowUI,
        EventSandbox = Hammerhead.EventSandbox,
        NativeMethods = Hammerhead.NativeMethods,

        ValidationMessageFactory = require('UI.ValidationMessageFactory'),
        SharedConst = require('Shared.Const'),
        TextSelection = Hammerhead.TextSelection;

    //Const
    var PROPERTY_CLASS = 'dialog-property',
        MAXIMIZED_PROPERTY_CLASS = 'maximized',
        SMALL_PROPERTY_CLASS = 'small',
        FIELD_CLASS = 'field',
        VALUE_CLASS = 'value',
        READONLY_CLASS = 'readonly_property',
        INPUT_CLASS = 'input',
        DISABLED_CLASS = 'disabled',

        INCORRECT_KEYS_ERROR = 'Incorrect keys',

        CHANGE_PROPERTY_VALUE_EVENT_DELAY = 333,
        MAXIMIZED_INPUT_LEFT_INDENT = 52;

    //options:
    //isKeysProperty
    //isNumericProperty
    //isNotNegativeNumericProperty
    //isPositiveNumericProperty
    //isReadOnly
    //maximizeInput

    //DialogProperty
    var DialogProperty = this.exports = function ($container, labelText, value, options) {
        var property = this;

        this.options = options || {};
        this.labelText = labelText;
        this.isDisabled = false;
        this.hasError = false;

        this.$container = $container;
        this.$property = null;
        this.$field = null;
        this.$value = null;
        this.$input = null;

        this.eventEmitter = new Util.EventEmitter();
        this.on = function (ev, listener) {
            property.eventEmitter.on(ev, listener);
        };

        this._createProperty(value);

        if (!this.options.isReadOnly) {
            this._initPropertyInput();
            this._validate();
        }
    };

    //Events
    DialogProperty.VALUE_CHANGED_EVENT = 'valueChangedEvent';

    //Markup
    DialogProperty.prototype._createProperty = function (value) {
        this.$property = $('<div></div>').appendTo(this.$container);
        ShadowUI.addClass(this.$property, PROPERTY_CLASS);

        if (this.options.maximizeInput)
            ShadowUI.addClass(this.$property, MAXIMIZED_PROPERTY_CLASS);

        if (this.options.smallInput)
            ShadowUI.addClass(this.$property, SMALL_PROPERTY_CLASS);

        var inputId = SharedConst.PROPERTY_PREFIX + this.labelText + Math.random().toString().replace('0.', '-');
        this._createLabel(inputId);

        if (!this.options.isReadOnly)
            this._createTextInput(inputId, value);
        else {
            ShadowUI.addClass(this.$property, READONLY_CLASS);
            this._createTextView(value);
        }
    };

    DialogProperty.prototype._createLabel = function (inputId) {
        this.$field = $('<label></label>')
            .appendTo(this.$property)
            .text(this.labelText === 'caretPos' ? 'caretPos:' : this.labelText);
        ShadowUI.addClass(this.$field, FIELD_CLASS);
        this.$field.attr('for', inputId);
    };

    DialogProperty.prototype._createTextView = function (value) {
        if (typeof value === 'string')
            value = [value];

        this.$value = $('<div>').appendTo(this.$property);

        for (var i = 0; i < value.length; i++) {
            $('<span>').text(value[i]).appendTo(this.$value);
            $('<br>').appendTo(this.$value);
        }

        ShadowUI.addClass(this.$value, VALUE_CLASS);
    };

    DialogProperty.prototype._createTextInput = function (inputId, inputValue) {
        this.$value = $('<div></div>').appendTo(this.$property);
        ShadowUI.addClass(this.$value, VALUE_CLASS);

        this.$input = $('<input type="text">')
            .appendTo(this.$value)
            .attr('id', inputId)
            .attr('value', inputValue);
        ShadowUI.addClass(this.$input, INPUT_CLASS);

        if (this.options.maximizeInput)
            this.$input.width(this.$container.width() - MAXIMIZED_INPUT_LEFT_INDENT);
    };

    //Behaviour
    DialogProperty.prototype._initPropertyInput = function () {
        var property = this,
            changeEventDelayTimerId = null;

        this.$input.bind('focus', function (e) {
            if (!property.isDisabled)
                return;

            var $inputs = ShadowUI.select('.' + INPUT_CLASS, property.$container),
                nextFocusableIndex = $.inArray(property.$input[0], $inputs);

            do
                nextFocusableIndex = nextFocusableIndex === $inputs.length - 1 ? 0 : nextFocusableIndex++;
            while
                ($inputs[nextFocusableIndex] !== property.$input[0] && ShadowUI.hasClass($inputs.eq(nextFocusableIndex), DISABLED_CLASS));

            if ($inputs[nextFocusableIndex] !== property.$input[0])
                $inputs.eq(nextFocusableIndex).focus();

            Util.preventDefault(e);
        });

        this.$input.bind('mousedown', function (e) {
            if (property.isDisabled)
                Util.preventDefault(e);
        });

        EventSandbox.addInternalEventListener(window, ['keydown'], function (e) {
            if (e.target !== property.$input[0])
                return;

            if (property.isDisabled) {
                Util.preventDefault(e);
                return false;
            }

            var keyCode = e.keyCode,
                isMinus = keyCode === 189 || keyCode === 109,
                isDigit = (!e.shiftKey && keyCode >= 48 && keyCode <= 57) || (keyCode >= 96 && keyCode <= 105),
                isChar = Util.isCharByKeyCode(keyCode);

            if (isChar) {
                if (property.options.isNumericProperty) {
                    if (!isDigit && !isMinus)
                        Util.preventDefault(e);
                    else if (property.options.isNotNegativeNumericProperty || property.options.isPositiveNumericProperty)
                        if (isMinus)
                            Util.preventDefault(e);
                }
            }
            else
                property._preventInputScrolling(e);
        });

        EventSandbox.addInternalEventListener(window, ['keypress', 'keyup'], function (e) {
            if (e.target === property.$input[0])
                Util.stopPropagation(e);
        });

        this.$input.bind('input', function () {
            if (changeEventDelayTimerId) {
                window.clearTimeout(changeEventDelayTimerId);
                changeEventDelayTimerId = null;
            }

            var isValid = property._isValid();

            if (isValid) {
                property._onSuccess();
                property._onChange();
            }
            else {
                changeEventDelayTimerId = NativeMethods.setTimeout.call(window, function () {
                    property._onError();
                    property._onChange();
                }, CHANGE_PROPERTY_VALUE_EVENT_DELAY);
            }
        });
    };

    DialogProperty.prototype._validate = function () {
        if (!this._isValid())
            this._onError();
        else
            this._onSuccess();
    };

    DialogProperty.prototype._isValid = function () {
        if (this.isDisabled || (!this.options.isKeysProperty && !this.options.isNumericProperty))
            return true;

        var value = this.$input[0].value;

        if (!value.length)
            return !!this.options.isNotNegativeNumericProperty;

        if (this.options.isKeysProperty)
            if (Util.parseKeysString(value).error)
                return false;

        if (this.options.isNumericProperty)
            if (!(/^(-)?\d+$/.test($.trim(value))))
                return false;

        if (this.options.isNotNegativeNumericProperty || this.options.isPositiveNumericProperty) {
            var valueAsNumber = parseInt($.trim(value));

            if (valueAsNumber < 0)
                return false;

            if (this.options.isPositiveNumericProperty && valueAsNumber === 0)
                return false;
        }

        return true;
    };

    DialogProperty.prototype._onError = function () {
        ValidationMessageFactory.error(this.$input, this.options.isKeysProperty ? INCORRECT_KEYS_ERROR : null);
        this.hasError = true;
    };

    DialogProperty.prototype._onSuccess = function () {
        if (this.hasError) {
            ValidationMessageFactory.success(this.$input);
            this.hasError = false;
        }
    };

    DialogProperty.prototype._onChange = function () {
        var args = {};

        args.value = (this.options.isNumericProperty && this.$input[0].value.length) ? parseInt(this.$input[0].value) :
            this.$input[0].value;

        this.eventEmitter.emit(DialogProperty.VALUE_CHANGED_EVENT, args);
    };

    DialogProperty.prototype._preventInputScrolling = function (ev) {
        if (Util.isWebKit) {
            var startSelection = TextSelection.getSelectionStart(this.$input[0]),
                endSelection = TextSelection.getSelectionEnd(this.$input[0]);

            if (startSelection === endSelection)
                if ((ev.keyCode === Util.KEYS_MAPS.SPECIAL_KEYS.left && startSelection === 0) ||
                    (ev.keyCode === Util.KEYS_MAPS.SPECIAL_KEYS.right && startSelection === this.$input[0].value.length)) {
                    Util.preventDefault(ev);
                }
        }
    };

    DialogProperty.prototype.disable = function () {
        this.isDisabled = true;
        if (this.$input) {
            ShadowUI.addClass(this.$input, DISABLED_CLASS);
            //NOTE: In IE9 setting of 'disabled' attribute leads to the appearance of text decoration (white shadow)
            // that can not be removed. Therefore, we should emulate the setting of this attribute.
            if (!(Util.isIE && Util.browserVersion === 9))
                this.$input.attr('disabled', 'disabled');
        }
    };

    DialogProperty.prototype.enable = function () {
        this.isDisabled = false;
        if (this.$input) {
            ShadowUI.removeClass(this.$input, DISABLED_CLASS);
            if (!(Util.isIE && Util.browserVersion === 9))
                this.$input.removeAttr('disabled');
        }
    };

    DialogProperty.prototype.isValid = function () {
        return !this.hasError;
    };

    DialogProperty.prototype.getLabelText = function () {
        return this.labelText;
    };

    DialogProperty.prototype.getInputElement = function () {
        if (this.$input)
            return this.$input[0];
    };

    DialogProperty.prototype.getEditorText = function () {
        if (this.$input)
            return this.$input[0].value;
    };

    DialogProperty.prototype.setEditorText = function (text) {
        if (this.$input) {
            this.$input[0].value = text;
            this._onChange();
        }
    };
});