var http = require('http'),
    int64 = require('node-int64'),
    crc32 = require('buffer-crc32');

//Utils
//----------------------------------------------------------
function noop() {
    //Do nothing =)
}

//Tracking data packet
//----------------------------------------------------------
//Packet format:
//* Part1
//      -- 8 bytes (int64 LE) - time in utc win filetime
//* Part2
//      -- 1 byte (uint8) - kind of action
//      -- 1 byte (uint8) - product code
//* Part3
//      -- 4 bytes (uint32 LE) - action number, it's a CRC32 of the tracking message
//* Part4
//      -- 2 bytes (int16 LE) - message length
//* Part5
//      -- Up to 2048 bytes - message

function createTrackingDataPacket(logItem) {
    var part1 = toLE(new int64(logItem.date).buffer),
        part2 = new Buffer(2);

    part2.writeUInt8(logItem.actionKind, 0);
    part2.writeUInt8(UXLog.PRODUCT_ID, 1);

    var msg = logItem.msg.substr(0, 2048),
        part3 = new Buffer(4),
        part4 = new Buffer(2),
        part5 = new Buffer(msg, 'binary');

    part3.writeUInt32LE(crc32.unsigned(part5), 0);
    part4.writeInt16LE(msg.length, 0);

    return Buffer.concat([part1, part2, part3, part4, part5]);
}

function toLE(buffer) {
    return [].reverse.call(buffer);
}

//UXLog
//----------------------------------------------------------
var UXLog = module.exports = function (config) {
    this.collectedData = [];
    this.loggingAllowed = false;
    this.version = '';
    this.userID = '';
    this.sendDataInterval = setInterval(this.sendCollectedData.bind(this), UXLog.SEND_DATA_INTERVAL);

    this._getLoggingSettings(config);
};

//Const
UXLog.UXLOG_HOSTNAME = 'uxlog.devexpress.com';
UXLog.UXLOG_PORT = 80;
UXLog.SEND_DATA_INTERVAL = 5 * 60 * 1000;
UXLog.SEND_DATA_REQUEST_TIMEOUT = 3000;

//NOTE: see $/2014.1/Win/DevExpress.Data/Utils/UAlgoConstants.cs in DXVCS for the following constants:
UXLog.PRODUCT_ID = 11;
//NOTE: expose this constants via proto, so we don't need to reference this module to get access to the UXLog
//class. So we can use instance instead(e.g. uxlog.RECORDER_ACTION_KIND)
UXLog.prototype.RECORDER_ACTION_KIND = 100;
UXLog.prototype.CONTROL_PANEL_ACTION_KIND = 101;
UXLog.prototype.FATAL_ERROR_ACTION_KIND = 102;

//Static
UXLog.dateStrToWindowsFileTime = function (dateStr) {
    //NOTE: http://stackoverflow.com/questions/6161776/convert-windows-filetime-to-second-in-unix-linux
    var WINDOWS_TICK = 10000000,
        SEC_TO_UNIX_EPOCH = 11644473600,
        unixSec = (new Date(dateStr)).getTime() / 1000;

    return (unixSec + SEC_TO_UNIX_EPOCH) * WINDOWS_TICK;
};

//NOTE: UXLog should VersionIdMinor-format for product version (e.g. see $/2014.2/Win/DevExpress.Data/AssemblyVersion.cs)
UXLog.toVersionIdMinor = function (version) {
    if (version) {
        var parts = version.split('.');

        //NOTE: append zero to path version (e.g '14.1.1' -> '14.1.01')
        if (parseInt(parts[2], 10) < 10)
            parts[2] = '0' + parts[2];

        return parts.join('');
    }

    return version;
};

//Properties
Object.defineProperties(UXLog.prototype, {
    enabled: {
        get: function () {
            return !!(this.loggingAllowed && this.userID && this.version);
        }
    }
});


//Internals
UXLog.prototype._getSendDataReqOptions = function (contentLength) {
    return {
        hostname: UXLog.UXLOG_HOSTNAME,
        port: UXLog.UXLOG_PORT,
        path: ['/?id=', this.userID, '&version=', this.version].join(''),
        method: 'POST',
        headers: {
            'content-type': 'app/x-www-form-urlencoded',
            'content-length': contentLength
        }
    };
};

UXLog.prototype._getLoggingSettings = function (config) {
    var uxlog = this;

    config.getLoggingAllowedFlag(function (loggingAllowed) {
        uxlog.loggingAllowed = loggingAllowed;
    });

    config.getUserID(function (userID) {
        uxlog.userID = userID.replace(/-/g, '');
    });

    config.getCurrentVersion(function (version) {
        uxlog.version = UXLog.toVersionIdMinor(version);
    });
};

UXLog.prototype._getBinaryTrackingData = function () {
    return Buffer.concat(this.collectedData.map(createTrackingDataPacket));
};

//API
UXLog.prototype.addEntry = function (actionKind, msg) {
    if (this.enabled) {
        var dateStr = (new Date()).toUTCString();

        this.collectedData.push({
            date: UXLog.dateStrToWindowsFileTime(dateStr),
            msg: msg,
            actionKind: actionKind
        });
    }
};

UXLog.prototype.shutDown = function () {
    this.loggingAllowed = false;
    this.collectedData = [];
    clearInterval(this.sendDataInterval);
};

UXLog.prototype.sendCollectedData = function () {
    if (this.collectedData.length) {
        var uxlog = this,
            data = this._getBinaryTrackingData();

        this.collectedData = [];

        var requestTimeout = setTimeout(function () {
            req.abort();
        }, UXLog.SEND_DATA_REQUEST_TIMEOUT);

        var halt = function () {
            uxlog.shutDown();
            clearTimeout(requestTimeout);
        };

        var req = http.request(this._getSendDataReqOptions(data.length), function (res) {
            clearTimeout(requestTimeout);
            res.on('data', noop);
            res.on('end', noop);
            res.on('error', halt);
        });

        req.on('error', halt);
        req.end(data);
    }
};

