﻿using System;
using System.Diagnostics;
using System.Drawing;
using System.Reflection;
using System.Windows.Forms;

namespace TestCafe_Runner {
    static class Program {
        [STAThread]
        static void Main() {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);
            Application.Run(new RunnerForm());
        }
    }

    public class RunnerForm : Form {
        private NotifyIcon trayIcon = null;
        private ContextMenuStrip trayMenu = null;
        private Process nodeProcess = null;
        private bool nodeClosed = false;
        private Assembly assembly = Assembly.GetExecutingAssembly();
        private SplashScreen splash = new SplashScreen();

        private string
           nodeFilePath = "node.exe",
           trayIconToolTipFormat = "Control Panel URL - http://{0}/",
           errorTextFormat = "TestCafe failed to run.\r\n\r\n{0}",
           errorCaption = "TestCafe - Error",
           menuOpenControlPanelItemText = "Open Control Panel",
           menuCopyControlPanelURLItemText = "Copy URL to clipboard",
           menuCopyControlPanelURLBalloonTitle = "TestCafe",
           menuCopyControlPanelURLBalloonText = "Control Panel URL has been copied to clipboard.",
           menuExitItemText = "Exit",
           controlPanelHost = null,
           tooltipMsg = "";

        private void getConfig() {
            ProcessStartInfo startInfo = new ProcessStartInfo();
            startInfo.FileName = nodeFilePath;
            startInfo.Arguments = "-e \"var Config=require('./lib/config'); var config=new Config();" +
                "config._initFromFile(function(){console.log(config.controlPanelHost);});\"";
            startInfo.WorkingDirectory = AppDomain.CurrentDomain.BaseDirectory;
            startInfo.WindowStyle = ProcessWindowStyle.Hidden;
            startInfo.UseShellExecute = false;
            startInfo.RedirectStandardOutput = true;
            startInfo.RedirectStandardError = true;
            startInfo.CreateNoWindow = true;

            Process getCfgProcess = Process.Start(startInfo);
            getCfgProcess.EnableRaisingEvents = true;
            getCfgProcess.Exited += delegate(object o, EventArgs e) {
                controlPanelHost = getCfgProcess.StandardOutput.ReadToEnd();

                if (!string.IsNullOrEmpty(controlPanelHost))
                    controlPanelHost = controlPanelHost.Trim();
                
                RunNode();
            };
 
        }

        private void RunNode() {
            ProcessStartInfo startInfo = new ProcessStartInfo();
            startInfo.FileName = nodeFilePath;
            startInfo.Arguments = "--max-old-space-size=1024 -e \"new (require('./index').TestCafe)(null, true);\"";
            startInfo.WorkingDirectory = AppDomain.CurrentDomain.BaseDirectory;
            startInfo.WindowStyle = ProcessWindowStyle.Hidden;
            startInfo.UseShellExecute = false;
            startInfo.RedirectStandardOutput = true;
            startInfo.RedirectStandardError = true;
            startInfo.CreateNoWindow = true;

            nodeProcess = Process.Start(startInfo);
            nodeProcess.EnableRaisingEvents = true;
            nodeProcess.Exited += node_Exited;

            tooltipMsg = string.Format(trayIconToolTipFormat, controlPanelHost);
            if(tooltipMsg.Length > 63)
                tooltipMsg = tooltipMsg.Substring(0, 60) + "...";

            trayIcon.Text = tooltipMsg;
            splash.ThreadSafeClose();
        }

        private void node_Exited(object sender, EventArgs e) {
            if (nodeProcess.ExitCode != 0 && !nodeClosed) {
                var err = nodeProcess.StandardError.ReadToEnd();

                if (!string.IsNullOrEmpty(err)) {
                    string errorMessage = string.Format(errorTextFormat, err);

                    MessageBox.Show(errorMessage, errorCaption, MessageBoxButtons.OK, MessageBoxIcon.Error,
                        MessageBoxDefaultButton.Button1, MessageBoxOptions.ServiceNotification);
                }
            }

            trayIcon.Visible = false;
            Application.Exit();
        }

        public RunnerForm() {
            splash.Show();

            trayMenu = new ContextMenuStrip();
            trayMenu.Items.Add(menuOpenControlPanelItemText, Bitmap.FromStream(assembly.GetManifestResourceStream("TestCafe.icons.icon-control-panel.png")), OnOpenControlPanel);
            trayMenu.Items.Add(menuCopyControlPanelURLItemText, Bitmap.FromStream(assembly.GetManifestResourceStream("TestCafe.icons.icon-copy-url.png")), OnCopyControlPanelURL);
            trayMenu.Items.Add(menuExitItemText, Bitmap.FromStream(assembly.GetManifestResourceStream("TestCafe.icons.icon-exit.png")), delegate(object sender, EventArgs e) {
                nodeProcess.Kill();
                nodeClosed = true;       
            });

            trayIcon = new NotifyIcon();
            trayIcon.Icon = new Icon(assembly.GetManifestResourceStream("TestCafe.icons.testcafe.ico"));
            trayIcon.DoubleClick += OnOpenControlPanel;
            trayIcon.ContextMenuStrip = trayMenu;
            trayIcon.Visible = true;

            getConfig();
        }

        protected override void OnLoad(EventArgs e) {
            Visible = false;
            ShowInTaskbar = false;

            base.OnLoad(e);
        }

        private void OnOpenControlPanel(object sender, EventArgs e) {
            string url = string.Format("http://{0}/", controlPanelHost);

            try{
                Process.Start(url);
            } catch(Exception) {
                //HACK: Win8 and Chrome issue
                //(see: http://stackoverflow.com/questions/12206368/process-starturl-broken-on-windows-8-chrome-are-there-alternatives)
                Process.Start("explorer.exe", url);
            }
        }

        private void OnCopyControlPanelURL(object sender, EventArgs e) {
            string url = string.Format("http://{0}/", controlPanelHost);

            Clipboard.SetText(url);
            trayIcon.ShowBalloonTip(2000, menuCopyControlPanelURLBalloonTitle, menuCopyControlPanelURLBalloonText, ToolTipIcon.Info);
        }

        protected override void Dispose(bool isDisposing) {
            if (isDisposing)
                trayIcon.Dispose();

            base.Dispose(isDisposing);
        }
    }
}
