var http = require('http'),
    childProcess = require('child_process'),
    Hammerhead = require('../../../hammerhead'),
    analysisUtil = require('./lib/analysis_util.js'),
    mongo = require('./lib/mongo.js'),
    runner = require('./lib/runner.js'),
    parseString = require('xml2js').parseString,
    config = require('./conf.json');

function HealthMonitor() {
    this.db = mongo;
    this.runOnce = runner.runOnce.bind(runner);
    this.runCycle = runner.runCycle.bind(runner);

    this._socketUrlAttemptsCount = this.GET_SOCKET_URL_ATTEMPTS;
}

HealthMonitor.prototype.CHROME_DEBUG_PARAM = '--remote-debugging-port=9222';
HealthMonitor.prototype.CHROME_NO_SANDBOX_PARAM = '--no-sandbox';
HealthMonitor.prototype.BROWSER_PATH_REG_CMD = 'reg query "HKEY_LOCAL_MACHINE\\SOFTWARE\\Clients\\StartMenuInternet\\Google Chrome\\shell\\open\\command"';
HealthMonitor.prototype.CHROME_INFO_ABOUT_TABS_URL = 'http://localhost:9222/json';
HealthMonitor.prototype.GET_SOCKET_URL_ATTEMPTS = 2;
HealthMonitor.prototype.GET_SOCKET_URL_DELAY = 15 * 1000;

HealthMonitor.prototype.start = function(settings, callback) {
    var healthMonitor = this;

    mongo.connect(config.mongodb, function () {
        if (settings.runInspector) {
            healthMonitor._setBuildInfo(settings.buildInfo);
            healthMonitor._runHammerhead(settings.port1, settings.port2);
            healthMonitor._runChrome(settings.startUrl);
            healthMonitor._getSocketUrl(0, function(socketUrl) {
                runner.init(socketUrl, settings.startUrl, callback);
            });
        } else if (callback)
            callback();
    });
};


HealthMonitor.prototype._runChrome = function(url) {
    var healthMonitor = this;

    childProcess.exec(this.BROWSER_PATH_REG_CMD, function(error, stdout) {
        var path = stdout.match(/reg_sz\s+"(.+)"/i)[1];

        if (path)
            childProcess.execFile(path, [healthMonitor.CHROME_DEBUG_PARAM, healthMonitor.CHROME_NO_SANDBOX_PARAM, url]);
    });
};

HealthMonitor.prototype._runHammerhead = function(port1, port2) {
    var hammerhead = new Hammerhead.Hammerhead(port1, port2, 'localhost');

    hammerhead.serviceEvents.for('').listen('taskScriptRequested', function (refererInfo, cookie, callback) {
        callback('if(window === window.top) localStorage.clear();');
    });
    hammerhead.crossDomainServiceEvents.for('').listen('taskScriptRequested', function (refererInfo, cookie, callback) {
        callback('localStorage.clear();');
    });
    hammerhead.start();

    analysisUtil.init(hammerhead);
};

HealthMonitor.prototype._setBuildInfo = function(buildInfo) {
    var healthMonitor = this;

    parseString(buildInfo, {
        normalizeTags: true,
        explicitArray: false
    }, function(err, result) {
        var processedBuildInfo = healthMonitor._processBuildInfo(err, result);

        if (processedBuildInfo)
            mongo.setBuildInfo(processedBuildInfo);
    });
};

HealthMonitor.prototype._processBuildInfo = function(err, result) {
    if (!err && result && result.modificationsinfo && result.modificationsinfo.modification) {
        var buildInfo = {},
            modifications = result.modificationsinfo.modification;

        for (var i = 0, j = modifications.length; i < j; i++) {
            if (!(modifications[i].version in buildInfo)) {
                buildInfo[modifications[i].version] = {
                    version: +modifications[i].version,
                    comment: modifications[i].comment,
                    modifiedTime: new Date(modifications[i].modifiedtime),
                    userName: modifications[i].username
                };
            }
        }

        buildInfo = Object.keys(buildInfo).map(function(key) {
            return buildInfo[key];
        });

        return buildInfo;
    }
};

HealthMonitor.prototype._getSocketUrl = function(position, callback) {
    var chunks = [],
        healthMonitor = this;

    http.get(healthMonitor.CHROME_INFO_ABOUT_TABS_URL, function(res) {
        res.on('data', chunks.push.bind(chunks));
        res.on('end', function() {
            callback(JSON.parse(chunks.join(''))[position].webSocketDebuggerUrl);
        });
    }).on('error', function(err) {
        if (healthMonitor._socketUrlAttemptsCount--)
            setTimeout(healthMonitor._getSocketUrl.bind(healthMonitor, position, callback), this.GET_SOCKET_URL_DELAY);
        else
            throw err;
    });
};


module.exports = new HealthMonitor();
