var analysisUtil = require('./analysis_util');

function Analysis() { }

Analysis.prototype.processHistory = function(data) {
    this._initData(data);
    this._clearResources();
    this._mergeInfo();

    return this._formatHistory();
};

Analysis.prototype.processStatistics = function(timestamp, history) {
    var statisticsTimingProps = this._getStatisticsTimingProps(),
        statistics = {
            timestamp: timestamp,
            sitesWithoutErrors: 0,
            sitesWithErrors: 0,
            timing: { }
        },
        initTimingProp = function(timing, prop) {
            timing[prop] = [];
        },
        calcTimingRatio = function(statisticsTiming, historyTiming, prop) {
            statisticsTiming[prop].push(historyTiming[prop].proxy / historyTiming[prop].origin || 0);
        },
        calcMedian = function(timing, prop) {
            timing[prop] = analysisUtil.calcMedian(timing[prop]);
        };

    statisticsTimingProps.forEach(initTimingProp.bind(null, statistics.timing));

    for (var i = 0; i < history.length; i++) {
        if (!history[i].console.length) {
            statisticsTimingProps.forEach(calcTimingRatio.bind(null, statistics.timing, history[i].timing));
            statistics.sitesWithoutErrors++;
        } else
            statistics.sitesWithErrors++;
    }

    if (statistics.sitesWithoutErrors) {
        statisticsTimingProps.forEach(calcMedian.bind(null, statistics.timing));
        return statistics;
    } else
        return null;
};

Analysis.prototype.mergeHistory = function(historyArray) {
    var mergedHistory = historyArray[0],
        timingArray = new Array(historyArray.length);

    for (var i = 0; i < historyArray.length; i++)
        timingArray[i] = historyArray[i].timing;

    mergedHistory.timing = this._mergeTiming(timingArray);

    return mergedHistory;
};


Analysis.prototype._formatHistory = function() {
    return {
        network: this._getResourcesAbsentOnProxy(),
        console: this._proxyErrors,
        timing: this._timing,
        url: this._url
    };
};

Analysis.prototype._initData = function(data) {
    this._timing = this._createTimingObj(data.originInfo.page, data.proxyInfo.page);
    this._originResources = data.originInfo.network;
    this._proxyResources = data.proxyInfo.network;
    this._originErrors = data.originInfo.console;
    this._proxyErrors = data.proxyInfo.console;
    this._url = data.url;
};

Analysis.prototype._createTimingObj = function(originPage, proxyPage) {
    return {
        load: { origin: originPage.load, proxy: proxyPage.load },
        domLoad: { origin: originPage.domContentLoad, proxy: proxyPage.domContentLoad },
        document: { origin: 0, proxy: 0 },
        script: { origin: 0, proxy: 0 },
        stylesheet: { origin: 0, proxy: 0 },
        other: { origin: 0, proxy: 0 },
        hh: { proxy: 0 }
    };
};

Analysis.prototype._clearResources = function() {
    this._removeServiceResources();
    this._removeResourcesWithDataProtocol(this._originResources);
    this._removeResourcesWithDataProtocol(this._proxyResources);
};

Analysis.prototype._removeServiceResources = function() {
    var index = 0;

    while (index < this._proxyResources.length) {
        if (analysisUtil.isServiceUrl(this._proxyResources[index].url)) {
            this._timing.hh.proxy += this._proxyResources[index].download;
            this._removeResources(this._proxyResources, index);
        } else
            index++;
    }
};

Analysis.prototype._removeResourcesWithDataProtocol = function(resources) {
    var index = 0;

    while (index < resources.length) {
        if (!resources[index].url.indexOf('data:'))
            this._removeResources(resources, index);
        else
            index++;
    }
};

Analysis.prototype._removeResources = function(resources, indexes) {
    indexes = indexes instanceof Array ? indexes : [indexes];

    for (var index = 0; index < indexes.length; index++)
        resources.splice(indexes[index] - index, 1);
};

Analysis.prototype._mergeInfo = function() {
    this._mergeResources();
    this._mergeErrors();
};

Analysis.prototype._mergeErrors = function() {
    var originIndex = 0;

    while (originIndex < this._originErrors.length) {
        var proxyIndex = this._searchInProxyErrors(this._originErrors[originIndex]);

        if (proxyIndex !== -1) {
            this._originErrors.splice(originIndex, 1);
            this._proxyErrors.splice(proxyIndex, 1);
        } else
            originIndex++;
    }
};

Analysis.prototype._searchInProxyErrors = function(originError) {
    for (var index = 0; index < this._proxyErrors.length; index++) {
        if (originError.text === this._proxyErrors[index].text && originError.type === this._proxyErrors[index].type)
            return index;
    }

    return -1;
};

Analysis.prototype._mergeResources = function() {
    var index = 0,
        getProxyResource = function(index) {
            return this[index];
        }.bind(this._proxyResources);

    while (index < this._originResources.length) {
        var originResource = this._originResources[index],
            proxyIndexes = this._searchInProxyResources(originResource),
            proxyResources = proxyIndexes.map(getProxyResource);

        if (proxyIndexes.length > 0) {
            this._addTimingData(originResource, proxyResources);
            this._removeResources(this._originResources, index);
            this._removeResources(this._proxyResources, proxyIndexes);
            this._removeMergingResourcesErrors(originResource, proxyResources);
        } else
            index++;
    }
};

Analysis.prototype._searchInProxyResources = function(originResource) {
    var proxyIndexes = [];

    for (var index = 0; index < this._proxyResources.length; index++) {
        var proxyResource = this._proxyResources[index];

        if (this._equalResources(originResource, proxyResource)) {
            proxyIndexes.push(index);

            if (proxyIndexes.length === 2 || proxyResource.status < 400 ||
                /*originResource.type !== 'image' || */analysisUtil.isProxyUrl(proxyResource.url))
                break;
        }
    }

    return proxyIndexes;
};

Analysis.prototype._equalResources = function(originResource, proxyResource) {
    return (!(originResource.type && proxyResource.type) || originResource.type === proxyResource.type) &&
        this._equalUrls(originResource.url, proxyResource.url);
};

Analysis.prototype._equalUrls = function(originUrl, proxyUrl) {
    return (analysisUtil.isProxyUrl(proxyUrl) ? analysisUtil.getOriginUrl(proxyUrl) : proxyUrl) === originUrl;
};

Analysis.prototype._addTimingData = function(originResource, proxyResources) {
    var type = originResource.type;

    type = !this._timing.hasOwnProperty(type) ? 'other' : type;

    if (proxyResources.length === 1) {
        if (originResource.status < 400 && proxyResources[0].status < 400 && analysisUtil.isProxyUrl(proxyResources[0].url)) {
            this._timing[type].origin += originResource.download;
            this._timing[type].proxy += proxyResources[0].download;
        }
    } else if (proxyResources.length === 2) {
        if (originResource.status < 400 && proxyResources[1].status < 400) {
            this._timing[type].origin += originResource.download;
            this._timing[type].proxy += proxyResources[0].download;
            this._timing[type].proxy += proxyResources[1].download;
        }
    }
};

Analysis.prototype._removeMergingResourcesErrors = function(originResource, proxyResources) {
    if (proxyResources.length === 1) {
        if (originResource.status >= 400) {
            this._removeErrorByResourceId(this._originErrors, originResource.networkRequestId);
            this._removeErrorByResourceId(this._proxyErrors, proxyResources[0].networkRequestId);
        }
    } else if (proxyResources.length === 2) {
        if (originResource.status >= 400) {
            this._removeErrorByResourceId(this._originErrors, originResource.networkRequestId);
            this._removeErrorByResourceId(this._proxyErrors, proxyResources[0].networkRequestId);
            this._removeErrorByResourceId(this._proxyErrors, proxyResources[1].networkRequestId);
        } else {
            this._removeErrorByResourceId(this._proxyErrors, proxyResources[0].networkRequestId);
        }
    }
};

Analysis.prototype._getResourcesAbsentOnProxy = function() {
    var result = [],
        testFunc = function(withError, resource) {
            if (withError ? resource.status >= 400: resource.status < 400)
                this.push(resource);
        };

    this._proxyResources.forEach(testFunc.bind(result, true));
    this._originResources.forEach(testFunc.bind(result, false));

    return result;
};

Analysis.prototype._removeErrorByResourceId = function(errors, networkRequestId) {
    var index = 0;

    while (index < errors.length) {
        if (errors[index].networkRequestId === networkRequestId)
            errors.splice(index, 1);
        else
            index++;
    }
};

Analysis.prototype._getStatisticsTimingProps = function() {
    var statisticsTimingProps = Object.keys(this._createTimingObj({}, {}));

    statisticsTimingProps.splice(statisticsTimingProps.indexOf('hh'), 1);
    statisticsTimingProps.splice(statisticsTimingProps.indexOf('other'), 1);

    return statisticsTimingProps;
};

Analysis.prototype._mergeTiming = function(timingArray) {
    var mergedTiming = timingArray[0];

    Object.keys(mergedTiming).forEach(function(prop) {
        mergedTiming[prop].origin = [mergedTiming[prop].origin];
        mergedTiming[prop].proxy = [mergedTiming[prop].proxy];

        for (var i = 1; i < timingArray.length; i++) {
            mergedTiming[prop].origin.push(timingArray[i][prop].origin);
            mergedTiming[prop].proxy.push(timingArray[i][prop].proxy);
        }

        mergedTiming[prop].origin = analysisUtil.calcMedian(mergedTiming[prop].origin);
        mergedTiming[prop].proxy = analysisUtil.calcMedian(mergedTiming[prop].proxy);
    });

    return mergedTiming;
};


module.exports = new Analysis();
