var util = require('util'),
    EventEmitter = require('events').EventEmitter,
    WebSocketClient = require('websocket').client;

function PageInspector() {
    this._timeoutAfterPageLoad = 0;
    this._timeoutInfinitePageLoad = 0;
    this._timeoutInfinitePageLoadId = null;
    this._initState = 0;
    this._connection = null;
    this._data = null;
    this._isListenMassage = false;
}

util.inherits(PageInspector, EventEmitter);

PageInspector.prototype.OPEN_EVENT = 'open';
PageInspector.prototype.CLOSE_EVENT = 'close';
PageInspector.prototype.PAGE_INSPECTED_EVENT = 'pageInspected';
PageInspector.prototype.NETWORK_INSPECT_ID = 1;
PageInspector.prototype.CONSOLE_INSPECT_ID = 2;
PageInspector.prototype.PAGE_INSPECT_ID = 4;
PageInspector.prototype.CACHE_DISABLED_ID = 8;
PageInspector.prototype.PAGE_NAVIGATE_ID = 16;
PageInspector.prototype.INFINITE_PAGE_LOAD_ERR = 'Infinite page load';


PageInspector.prototype.inspectPage = function(url, callback) {
    this.once(this.PAGE_INSPECTED_EVENT, callback);
    this._resetData();
    this._loadPage(url);

    this._timeoutInfinitePageLoadId = setTimeout(this._onPageInspected.bind(this,
        { text: this.INFINITE_PAGE_LOAD_ERR, type: 'err' }), this._timeoutInfinitePageLoad);
};

PageInspector.prototype.init = function(socketUrl, timeoutAfterPageLoad, timeoutInfinitePageLoad) {
    this._timeoutAfterPageLoad = timeoutAfterPageLoad;
    this._timeoutInfinitePageLoad = timeoutInfinitePageLoad;
    this._initWebSocket(socketUrl);
};


PageInspector.prototype._resetData = function() {
    this._data = {
        page: {},
        network: { requests: 0, finished: 0 },
        console: []
    };
};

PageInspector.prototype._loadPage = function(url) {
    this._connection.send(JSON.stringify({
        id: this.PAGE_NAVIGATE_ID,
        method: 'Page.navigate',
        params: { url: url }
    }));
};

PageInspector.prototype._initWebSocket = function(socketUrl) {
    var webSocketClient = new WebSocketClient();

    webSocketClient.on('connectFailed', this.emit.bind(this, this.OPEN_EVENT));
    webSocketClient.on('connect', this._onSocketConnect.bind(this));
    webSocketClient.connect(socketUrl);
};

PageInspector.prototype._onSocketConnect = function(connection) {
    var onConnectionClose = this.emit.bind(this, this.CLOSE_EVENT);

    connection.on('error', onConnectionClose);
    connection.on('close', onConnectionClose);
    connection.on('message', this._onMessage.bind(this));

    this._connection = connection;
    this._prepareBrowser();
};

PageInspector.prototype._prepareBrowser = function() {
    this._connection.send(JSON.stringify({ id: this.NETWORK_INSPECT_ID, method: 'Network.enable' }));
    this._connection.send(JSON.stringify({ id: this.CONSOLE_INSPECT_ID, method: 'Console.enable' }));
    this._connection.send(JSON.stringify({ id: this.PAGE_INSPECT_ID, method: 'Page.enable' }));
    this._connection.send(JSON.stringify({
        id: this.CACHE_DISABLED_ID,
        method: 'Network.setCacheDisabled',
        params: { cacheDisabled: true }
    }));
};

PageInspector.prototype._updateInitState = function(id) {
    /* jshint -W016 */
    this._initState |= id;

    if (this._initState === (this.NETWORK_INSPECT_ID | this.CONSOLE_INSPECT_ID | this.PAGE_INSPECT_ID | this.CACHE_DISABLED_ID))
        this.emit(this.OPEN_EVENT);
    /* jshint +W016 */
};

PageInspector.prototype._onMessage = function(message) {
    var parsedMsg = JSON.parse(message.utf8Data),
        msgType = parsedMsg.method && parsedMsg.method.substring(0, parsedMsg.method.indexOf('.')),
        msgFuncName = '_on' + msgType + 'DataMsg';

    switch (parsedMsg.id) {
        case this.NETWORK_INSPECT_ID: case this.CONSOLE_INSPECT_ID: case this.PAGE_INSPECT_ID: case this.CACHE_DISABLED_ID:
            this._updateInitState(parsedMsg.id);
            break;
        case this.PAGE_NAVIGATE_ID:
            if (!parsedMsg.result)
                throw parsedMsg.error.message;
            else
                this._data.page.mainFrameId = parsedMsg.result.frameId;
            break;
        default:
            if (this._data && this[msgFuncName])
                this[msgFuncName](parsedMsg);
    }

    // for debug
    //console.log('isListenMassage - ' + (this._isListenMassage ? 'unlock' : 'lock'));
    //console.log(util.inspect(parsedMsg, { showHidden: false, colors: true, depth: 3 }));
};

PageInspector.prototype._onPageInspected = function(err) {
    var data = this._data,
        pageFrameInfo = data.page[data.page.mainFrameId],
        processedNetwork = [],
        notLoadedUrls = [],
        pageTiming = {};

    this._isListenMassage = false;
    clearTimeout(this._timeoutInfinitePageLoadId);

    if (pageFrameInfo) {
        pageTiming.domContentLoad = (data.page.domContentLoad || new Date().getTime()) - pageFrameInfo.startedLoading;
        pageTiming.load = (data.page.load || new Date().getTime()) - pageFrameInfo.startedLoading;
    }

    Object.keys(data.network)
        .filter(function(networkId) {
            return networkId.indexOf(data.id) === 0;
        })
        .sort(function(a, b) {
            var lengthDif = a.length - b.length;

            return lengthDif ? lengthDif :  a.localeCompare(b);
        })
        .forEach(function(networkId) {
            var resource = data.network[networkId];

            if (resource.finished) {
                processedNetwork.push({
                    url: resource.url,
                    type: resource.type && resource.type.toLowerCase(),
                    status: typeof resource.status === 'undefined' ? 500 : resource.status,
                    download: resource.finished - resource.willBeSent,
                    networkRequestId: networkId
                });
            } else
                notLoadedUrls.push(resource.url);
        });

    if (err) {
        err.stack = notLoadedUrls.join('\n');
        data.console.push(err);
    }

    this.emit(this.PAGE_INSPECTED_EVENT, {
        page: pageTiming,
        network: processedNetwork,
        console: data.console
    }, err);
};

PageInspector.prototype._onPageDataMsg = function(dataMsg) {
    if (!dataMsg.params)
        return;

    var frameId = dataMsg.params.frameId || (dataMsg.params.frame && dataMsg.params.frame.id),
        data = this._data;

    switch (dataMsg.method) {
        case 'Page.frameStartedLoading':
            if (!data.page[frameId])
                data.page[frameId] = {};

            data.page[frameId].startedLoading = new Date().getTime();

            if (!data.id) {
                data.id = frameId.split('.')[0];
                this._isListenMassage = true;
            }

            break;
        case 'Page.frameStoppedLoading':
            if (data.page[frameId])
                data.page[frameId].stoppedLoading = new Date().getTime();

            break;
        case 'Page.frameNavigated':
            if (data.page[frameId])
                data.page[frameId].navigated = new Date().getTime();

            break;
        case 'Page.domContentEventFired':
            data.page.domContentLoad = new Date().getTime();

            break;
        case 'Page.loadEventFired':
            data.page.load = new Date().getTime();

            setTimeout(this._onPageInspected.bind(this), this._timeoutAfterPageLoad);

            break;
    }
};

PageInspector.prototype._onNetworkDataMsg = function(dataMsg) {
    var requestId = dataMsg.params.requestId,
        data = this._data;

    switch (dataMsg.method) {
        case 'Network.requestWillBeSent':
            if (requestId.indexOf(data.id) === 0) {
                if (!data.network[requestId]) {
                    if (this._isListenMassage) {
                        data.network[requestId] = {};
                        data.network.requests++;
                    } else
                        break;
                }

                data.network[requestId].url = dataMsg.params.request.url;
                data.network[requestId].willBeSent = new Date().getTime();
            }

            break;
        case 'Network.responseReceived':
            if (data.network[requestId]) {
                data.network[requestId].type = dataMsg.params.type;
                data.network[requestId].status = dataMsg.params.response.status;
                data.network[requestId].received = new Date().getTime();
            }

            break;
        case 'Network.loadingFinished':
            if (data.network[requestId]) {
                data.network[requestId].finished = new Date().getTime();
                data.network.finished++;
            }

            break;
        case 'Network.loadingFailed':
            if (data.network[requestId]) {
                data.network[requestId].finished = new Date().getTime();
                data.network.finished++;
            }

            break;
    }
};

PageInspector.prototype._onConsoleDataMsg = function(dataMsg) {
    if (this._isListenMassage && dataMsg.method === 'Console.messageAdded' && dataMsg.params.message.level === 'error') {
        this._data.console.push({
            text: dataMsg.params.message.text,
            type: dataMsg.params.message.type,
            stack: this._formatStackTrace(dataMsg.params.message.stackTrace),
            networkRequestId: dataMsg.params.message.networkRequestId
        });
    }
};

PageInspector.prototype._formatStackTrace = function(stackTrace) {
    if (stackTrace) {
        var result = [];

        for (var i = 0, j = stackTrace.length; i < j; i++) {
            result.push('<b>');
            result.push(stackTrace[i].functionName || '(anonymous function)');
            result.push('</b>\t\t');
            result.push(stackTrace[i].url);
            result.push(':');
            result.push(stackTrace[i].lineNumber);
            result.push(':');
            result.push(stackTrace[i].columnNumber);
            result.push('\n');
        }

        return result.join('');
    } else
        return null;
};


module.exports = new PageInspector();
