/* global console: true */

var pageInspector = require('./page_inspector.js'),
    analysis = require('./analysis.js'),
    analysisUtil = require('./analysis_util.js'),
    mongo = require('./mongo.js'),
    async = require('async'),
    config = require('./../conf.json');

function Runner() {
    this._timestamp = null;
    this._pageAfterTest = null;
    this._history = null;
    this._firstRunHistory = null;
    this._testingUrl = null;
}

Runner.prototype.ADDITIONAL_REPEAT_COUNT = 5;
Runner.prototype.REPEAT_COUNT = config.tests.repeatCount || 3;
Runner.prototype.LIMIT_ERROR_PERCENT = 0.1;
Runner.prototype.TOLERANCE_LOAD_PERCENT = 0.2;

Runner.prototype.init = function(socketUrl, pageAfterTest, callback) {
    this._pageAfterTest = pageAfterTest;

    pageInspector.on(pageInspector.OPEN_EVENT, function(err) {
        if (err)
            throw err;

        if(callback)
            callback();
    });

    pageInspector.on(pageInspector.CLOSE_EVENT, function(err) {
        console.warn('Page inspector are closed' + (err ? ' with error ' + err.toString() : '') + '.');
    });

    pageInspector.init(socketUrl, config.tests.timeoutAfterLoad, config.tests.pageMaxLoadTime);
};

Runner.prototype.runOnce = function(callback) {
    var runner = this;

    runner._timestamp = new Date().getTime();
    runner._history = [];
    runner._firstRunHistory = [];

    mongo.getEnableUrlList(function(urlList) {
        runner._testPageList(urlList, function() {
            runner._saveStatistics(function() {
                runner._isSucceed(function(success) {
                    if (runner._pageAfterTest)
                        pageInspector.inspectPage(runner._pageAfterTest, function() { });

                    callback(success);
                });
            });
        });
    });
};

Runner.prototype.runCycle = function(delay) {
    this.runOnce(setTimeout.bind(null, this.runCycle.bind(this, delay), delay));
};


Runner.prototype._saveStatistics = function(callback) {
    var runner = this;

    var statistics = analysis.processStatistics(runner._timestamp, runner._history),
        statisticsForFirstRun = analysis.processStatistics(runner._timestamp, runner._firstRunHistory),
        finishTest = function() {
            if (callback)
                callback();
        };

    if (statistics) {
        if (statisticsForFirstRun)
            statistics.timingForFirstRun = statisticsForFirstRun.timing;

        mongo.insertStatistics(statistics, finishTest);
    } else
        finishTest();
};

Runner.prototype._testPageList = function(urlList, callback) {
    var runner = this,
        testPageFunc = runner._testPage.bind(runner);

    async.eachSeries(urlList, function(urlObj, callback) {
        runner._testingUrl = urlObj.url;

        async.mapSeries([
            urlObj.url,
            analysisUtil.getProxyUrl(urlObj.url)
        ], testPageFunc, function(err, data) {
            runner._saveHistory(data[0], data[1], callback);
        });
    }, callback);
};

Runner.prototype._saveHistory = function(originData, proxyData, callback) {
    var historyArray = new Array(Math.min(originData.length, proxyData.length)),
        firstRun, history;

    for (var i = 0; i < historyArray.length; i++) {
        historyArray[i] = analysis.processHistory({
            url: this._testingUrl,
            originInfo: originData[i],
            proxyInfo: proxyData[i]
        });
    }

    firstRun = historyArray.shift();
    history = analysis.mergeHistory(historyArray);

    this._firstRunHistory.push(firstRun);
    this._history.push(history);
    mongo.insertInDataByTimestamp(history, this._timestamp, callback);
};

Runner.prototype._testPage = function(url, callback) {
    var runner = this,
        succeedCount = 0,
        failedCount = 0,
        attemptCount = this.ADDITIONAL_REPEAT_COUNT + this.REPEAT_COUNT,
        cacheAttempt = null,
        error = null,
        inspectedResult = [];

    pageInspector.inspectPage(url, function(obj, err) {
        attemptCount--;

        if (cacheAttempt) {
            if (!err && !obj.console.length) {
                inspectedResult.push(obj);
                succeedCount++;
            } else {
                error = obj;
                failedCount++;

                if (failedCount === runner.REPEAT_COUNT && succeedCount === 0)
                    return callback(null, [cacheAttempt, obj]);
            }
        } else
            cacheAttempt = obj;

        if (succeedCount < runner.REPEAT_COUNT && attemptCount)
            pageInspector.inspectPage(url, arguments.callee);
        else if(succeedCount < runner.REPEAT_COUNT)
            callback(null, [cacheAttempt, error]);
        else {
            inspectedResult.unshift(cacheAttempt);
            callback(null, inspectedResult);
        }
    });
};

Runner.prototype._isSucceed = function(callback) {
    var runner = this;

    mongo.getStatistics(0, function(docs) {
        if (docs) {
            var currentLoad = Infinity,
                minLoad = Infinity;

            for (var i = 0; i < docs.length; i++) {
                var statistics = docs[i];

                if (minLoad > statistics.timing.load)
                    minLoad = statistics.timing.load;

                if (runner._timestamp === statistics.timestamp) {
                    currentLoad = statistics.timing.load;

                    if (statistics.sitesWithErrors / (statistics.sitesWithoutErrors + statistics.sitesWithErrors) > runner.LIMIT_ERROR_PERCENT)
                        return callback(false);
                }
            }

            callback((minLoad * (1 + runner.TOLERANCE_LOAD_PERCENT)) >= currentLoad);
        } else
            callback(false);
    });
};


module.exports = new Runner();
