var async = require('async'),
    fs = require('fs'),
    util = require('util'),
    errMsgBuilder = require('../../../../lib/err_msg_builder'),
    VirtualFS = require('../../../../lib/vfs'),
    Api = require('../../../../lib/server/api').Api,
    TestCafe = require('../../../../lib/server/test_cafe').TestCafe,
    WorkerPool = require('../../../../lib/test_runner/worker_pool'),
    testUtils = require('../../test_utils'),
    ERR = require('../../../../lib/server/server_errs');

var testCafe = null,
    watchdog = null;

var getStartupConfig = function (testsPath) {
    return{
        'controlPanelPort': 1837,
        'servicePort': 1838,
        'servicePort2': 1839,
        'testsDir': testsPath ? fs.realpathSync(['test/nodeunit/core/data/api/', testsPath].join('')) : fs.realpathSync('test/nodeunit/core/data/api/'),
        'browsers': {
        }
    };
};

exports['Workers list'] = {
    setUp: function (done) {
        testCafe = new TestCafe(getStartupConfig());
        watchdog = new testUtils.Watchdog();
        done();
    },

    tearDown: function (done) {
        testCafe.close();
        watchdog.shrink();

        process.nextTick(function () {
            done();
        });
    },

    'Get worker list': function (t) {
        var workerName = 'newWorker',
            expectedJs = ['newWorker'];

        testCafe.testRunner.workerPool.connectWorker(workerName, null);
        var workers = testCafe.listConnectedWorkers();

        t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(workers));
        t.done();
    },

    'Get empty worker list': function (t) {
        var workers = testCafe.listConnectedWorkers();

        t.strictEqual(workers.length, 0);
        t.done();
    }
};

exports['Browsers list'] = {
    setUp: function (done) {
        testCafe = new TestCafe(getStartupConfig());
        watchdog = new testUtils.Watchdog();

        done();
    },

    tearDown: function (done) {
        testCafe.close();
        watchdog.shrink();

        process.nextTick(function () {
            done();
        });
    },

    'Get browsers list': function (t) {
        var expectedJs = [];

        var result = testCafe.listAvailableBrowsers();
        t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(result));

        t.done();
    }
};

exports['List directory'] = {
    setUp: function (done) {
        testCafe = new TestCafe(getStartupConfig('/ls/'));
        watchdog = new testUtils.Watchdog();
        done();
    },

    tearDown: function (done) {
        testCafe.close();
        watchdog.shrink();

        process.nextTick(function () {
            done();
        });
    },

    'Directory list': function (t) {
        var lsDirPath = [],
            expectedJs = {
                buildErrs: null,
                fixtures: [
                    {
                        uid: '',
                        name: 'test',
                        fileName: 'fixture.test.js',
                        page: 'localhost:1337/',
                        username: null,
                        password: null,
                        tests: []
                    }
                ],
                dirs: [
                    {
                        name: 'listDir1',
                        path: 'listDir1'
                    }
                ]
            };

        testCafe.listDirectory(lsDirPath, function (err, pathInfo) {
            t.ok(/[0-9a-z-]{36}/.test(pathInfo.fixtures[0].uid));
            pathInfo.fixtures[0].uid = '';

            t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(pathInfo));

            t.done();
        });
    },

    'Directory not exists': function (t) {
        var lsDirPath = ['notExistingPath'],
            expectedJs = null;

        testCafe.listDirectory(lsDirPath, function (err, pathInfo) {
            t.ok(err, VirtualFS.PATH_NOT_EXISTS);
            t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(pathInfo));

            t.done();
        });
    },

    'Directory not specified': function (t) {
        var expectedJs = {
            buildErrs: null,
            fixtures: [
                {
                    uid: '',
                    name: 'test',
                    fileName: 'fixture.test.js',
                    page: 'localhost:1337/',
                    username: null,
                    password: null,
                    tests: []
                }
            ],
            dirs: [
                {
                    name: 'listDir1',
                    path: 'listDir1'
                }
            ]
        };

        testCafe.listDirectory(function (err, pathInfo) {
            t.ok(/[0-9a-z-]{36}/.test(pathInfo.fixtures[0].uid));
            pathInfo.fixtures[0].uid = '';

            t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(pathInfo));

            t.done();
        });
    },

    'Directory is string': function (t) {
        var lsDirPath = 'listDir1',
            expectedJs = {
                buildErrs: null,
                fixtures: [
                    {
                        uid: '',
                        name: 'test',
                        fileName: 'fixture.test.js',
                        page: 'localhost:1337/',
                        username: null,
                        password: null,
                        tests: []
                    }
                ],
                dirs: []
            };

        testCafe.listDirectory(lsDirPath, function (err, pathInfo) {
            t.ok(/[0-9a-z-]{36}/.test(pathInfo.fixtures[0].uid));
            pathInfo.fixtures[0].uid = '';

            t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(pathInfo));

            t.done();
        });
    },

    'Directory is slash': function (t) {
        var lsDirPath = '/',
            expectedJs = {
                buildErrs: null,
                fixtures: [
                    {
                        uid: '',
                        name: 'test',
                        fileName: 'fixture.test.js',
                        page: 'localhost:1337/',
                        username: null,
                        password: null,
                        tests: []
                    }
                ],
                dirs: [
                    {
                        name: 'listDir1',
                        path: 'listDir1'
                    }
                ]
            };

        testCafe.listDirectory(lsDirPath, function (err, pathInfo) {
            t.ok(/[0-9a-z-]{36}/.test(pathInfo.fixtures[0].uid));
            pathInfo.fixtures[0].uid = '';

            t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(pathInfo));

            t.done();
        });
    }
};

exports['Run test'] = {
    setUp: function (done) {
        testCafe = new TestCafe(getStartupConfig('runTest'));
        watchdog = new testUtils.Watchdog();
        done();
    },

    tearDown: function (done) {
        testCafe.close();
        watchdog.shrink();

        process.nextTick(function () {
            done();
        });
    },

    'Run test': function (t) {
        var firstWorkerName = 'testRun1',
            secondWorkerName = 'testRun2',
            runningWorkers = [firstWorkerName, secondWorkerName],
            options = {
                source: 'testDir',
                sourceType: 'dir',
                workers: runningWorkers,
                browsers: []
            };

        testCafe.testRunner.workerPool.connectWorker(firstWorkerName, null);
        testCafe.testRunner.workerPool.connectWorker(secondWorkerName, null);

        testCafe.runTests(options, function (errs, taskUid, workers) {
            t.ok(taskUid);
            t.ok(!errs);
            t.strictEqual(JSON.stringify(workers), JSON.stringify(runningWorkers));

            t.done();
        });
    },

    'Run test without source arguments (Root directory)': function (t) {
        var firstWorkerName = 'testRun1',
            secondWorkerName = 'testRun2',
            runningWorkers = [firstWorkerName, secondWorkerName],
            options = {
                workers: runningWorkers,
                browsers: []
            };

        testCafe.testRunner.workerPool.connectWorker(firstWorkerName, null);
        testCafe.testRunner.workerPool.connectWorker(secondWorkerName, null);

        testCafe.runTests(options, function (errs, taskUid, workers) {
            t.ok(taskUid);
            t.ok(!errs);
            t.strictEqual(JSON.stringify(workers), JSON.stringify(runningWorkers));

            t.done();
        });
    },

    'Run test without sourceType arguments (Directory)': function (t) {
        var firstWorkerName = 'testRun1',
            secondWorkerName = 'testRun2',
            runningWorkers = [firstWorkerName, secondWorkerName],
            options = {
                source: 'testDir',
                workers: runningWorkers,
                browsers: []
            };

        testCafe.testRunner.workerPool.connectWorker(firstWorkerName, null);
        testCafe.testRunner.workerPool.connectWorker(secondWorkerName, null);

        testCafe.runTests(options, function (errs, taskUid, workers) {
            t.ok(taskUid);
            t.ok(!errs);
            t.strictEqual(JSON.stringify(workers), JSON.stringify(runningWorkers));

            t.done();
        });
    },

    'Run test without sourceType arguments (Fixture)': function (t) {
        var firstWorkerName = 'testRun1',
            secondWorkerName = 'testRun2',
            runningWorkers = [firstWorkerName, secondWorkerName],
            options = {
                source: 'testDir/test',
                workers: runningWorkers,
                browsers: []
            };

        testCafe.testRunner.workerPool.connectWorker(firstWorkerName, null);
        testCafe.testRunner.workerPool.connectWorker(secondWorkerName, null);

        testCafe.runTests(options, function (errs, taskUid, workers) {
            t.ok(taskUid);
            t.ok(!errs);
            t.strictEqual(JSON.stringify(workers), JSON.stringify(runningWorkers));

            t.done();
        });
    },

    'Run test without sourceType arguments (Single test)': function (t) {
        var firstWorkerName = 'testRun1',
            secondWorkerName = 'testRun2',
            runningWorkers = [firstWorkerName, secondWorkerName],
            options = {
                source: 'testDir/test/Testing localhost',
                workers: runningWorkers,
                browsers: []
            };

        testCafe.testRunner.workerPool.connectWorker(firstWorkerName, null);
        testCafe.testRunner.workerPool.connectWorker(secondWorkerName, null);

        testCafe.runTests(options, function (errs, taskUid, workers) {
            t.ok(taskUid);
            t.ok(!errs);
            t.strictEqual(JSON.stringify(workers), JSON.stringify(runningWorkers));

            t.done();
        });
    },

    'Run test with unknown source type': function(t) {
        var workerName = 'testRun',
            source = 'testDir',
            runningWorkers = [workerName],
            options = {
                source: source,
                sourceType: 'UNKNOWN',
                workers: runningWorkers,
                browsers: []
            },
            expectedJs = [errMsgBuilder.build({code: ERR.API_UNKNOWN_FIXTURE_TYPE, type: options.sourceType})];

        testCafe.testRunner.workerPool.connectWorker(workerName, null);

        testCafe.runTests(options, function (errs, taskUid, workers) {
            t.ok(!taskUid);
            t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(errs));
            t.ok(!workers);

            t.done();
        });
    },

    'Run task on nonexistent worker': function (t) {
        var firstWorkerName = 'testRun1',
            secondWorkerName = 'testRun2',
            workerWhoNotExists = 'workerWhoNotExists',
            source = 'testDir',
            runningWorkers = [firstWorkerName, secondWorkerName, workerWhoNotExists],
            options = {
                source: source,
                sourceType: 'dir',
                workers: runningWorkers,
                browsers: []
            },
            expectedJs =
                [
                    {
                        code: ERR.API_TASK_WORKER_DOESNT_EXIST,
                        taskName: util.format(Api.DIRECTORY_TEST_TEXT, source),
                        workerName: workerWhoNotExists
                    }
                ];

        testCafe.testRunner.workerPool.connectWorker(firstWorkerName, null);
        testCafe.testRunner.workerPool.connectWorker(secondWorkerName, null);

        testCafe.runTests(options, function (errs, taskUid, workers) {
            t.ok(taskUid);
            t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(errs));
            t.strictEqual(JSON.stringify(workers), JSON.stringify([firstWorkerName, secondWorkerName]));

            t.done();
        });
    },

    'Run nonexistent test': function (t) {
        var firstWorkerName = 'testRun1',
            secondWorkerName = 'testRun2',
            runningWorkers = [firstWorkerName, secondWorkerName],
            options = {
                source: 'testWhoNotExist',
                sourceType: 'dir',
                workers: runningWorkers,
                browsers: []
            },
            expectedJs =
                [
                    errMsgBuilder.build({code: ERR.API_NO_TESTS_TO_RUN})
                ];

        testCafe.testRunner.workerPool.connectWorker(firstWorkerName, null);
        testCafe.testRunner.workerPool.connectWorker(secondWorkerName, null);

        testCafe.runTests(options, function (errs, taskUid, workers) {
            t.ok(!taskUid);
            t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(errs));
            t.ok(!workers);

            t.done();
        });
    },

    'Run target does not containt any tests': function (t) {
        var firstWorkerName = 'testRun1',
            secondWorkerName = 'testRun2',
            runningWorkers = [firstWorkerName, secondWorkerName],
            options = {
                source: 'noTests',
                sourceType: 'dir',
                workers: runningWorkers,
                browsers: []
            },
            expectedJs =
                [
                    errMsgBuilder.build({code: ERR.API_NO_TESTS_TO_RUN})
                ];

        testCafe.testRunner.workerPool.connectWorker(firstWorkerName, null);
        testCafe.testRunner.workerPool.connectWorker(secondWorkerName, null);

        testCafe.runTests(options, function (errs, taskUid, workers) {
            t.ok(!taskUid);
            t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(errs));
            t.ok(!workers);
            t.done();
        });
    },

    'Run fixture by fixture path': function(t) {
        var workerName = 'testRun',
            options = {
                source: 'testDir/test',
                sourceType: 'fixture',
                workers: [workerName],
                browsers: []
            };

        testCafe.testRunner.workerPool.connectWorker(workerName, null);

        testCafe.runTests(options, function (errs, taskUid, workers) {
            t.ok(taskUid);
            t.ok(!errs);
            t.strictEqual(JSON.stringify(workers), JSON.stringify([workerName]));

            options.source = 'Fixture in root directory';

            testCafe.runTests(options, function (errs, taskUid, workers) {
                t.ok(taskUid);
                t.ok(!errs);
                t.strictEqual(JSON.stringify(workers), JSON.stringify([workerName]));

                t.done();
            });
        });
    },

    'Run group tests': function(t) {
        var workerName = 'testRun',
            runningWorkers = [workerName],
            options = {
                sourceType: 'group',
                workers: runningWorkers,
                browsers: []
            };

        testCafe.testRunner.workerPool.connectWorker(workerName, null);

        testCafe.vfs.getSuite(function(vfsSuite) {
            options.source = Object.keys(vfsSuite.tests);

            testCafe.runTests(options, function (errs, taskUid, workers) {
                t.ok(taskUid);
                t.ok(!errs);
                t.strictEqual(JSON.stringify(workers), JSON.stringify(runningWorkers));

                t.done();
            });
        });
    },

    'Run group tests - different test targets': function(t) {
        var workerName = 'testRun',
            runningWorkers = [workerName],
            directoryTarget = 'testDir',
            fixtureTarget = 'testDir/test',
            testTarget = 'testDir/test/Testing localhost',
            options = {
                sourceType: 'group',
                source: [directoryTarget, fixtureTarget, testTarget],
                groupName: 'Custom suite',
                workers: runningWorkers,
                browsers: []
            };

        testCafe.testRunner.workerPool.connectWorker(workerName, null);

        testCafe.runTests(options, function (errs, taskUid, workers) {
            t.ok(taskUid);
            t.ok(!errs);
            t.strictEqual(JSON.stringify(workers), JSON.stringify(runningWorkers));

            t.done();
        });
    },

    'Run group tests - one item doesn\'t exists': function(t) {
        var workerName = 'testRun',
            runningWorkers = [workerName],
            options = {
                sourceType: 'group',
                workers: runningWorkers,
                browsers: []
            },
            notExistingTarget = 'testValue',
            expectedJs =  [errMsgBuilder.build({code: ERR.API_TESTS_TARGET_NOT_EXIST, source: notExistingTarget})];

        testCafe.testRunner.workerPool.connectWorker(workerName, null);

        testCafe.vfs.getSuite(function(vfsSuite) {
            options.source = Object.keys(vfsSuite.tests).concat(notExistingTarget);

            testCafe.runTests(options, function (errs, taskUid, workers) {
                t.ok(!taskUid);
                t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(errs));
                t.ok(!workers);

                t.done();
            });
        });
    },

    'Run group tests - sourceType missed': function(t) {
        var workerName = 'testRun',
            runningWorkers = [workerName],
            options = {
                workers: runningWorkers,
                browsers: []
            };

        testCafe.testRunner.workerPool.connectWorker(workerName, null);

        testCafe.vfs.getSuite(function(vfsSuite) {
            options.source = Object.keys(vfsSuite.tests);

            testCafe.runTests(options, function (errs, taskUid, workers) {
                t.ok(taskUid);
                t.ok(!errs);
                t.strictEqual(JSON.stringify(workers), JSON.stringify(runningWorkers));

                t.done();
            });
        });
    },

    'Detect source type': function(t) {
        var suite = null,
            dirPath = 'testDir',
            fixturePath = 'testDir/test',
            testPath = 'testDir/test/Testing localhost';

        async.series({
            'Get suite': function(done) {
                testCafe.vfs.getSuite(function(vfsSuite) {
                    suite = vfsSuite;
                    done();
                });
            },

            'Directory type': function(done) {
                testCafe._tryDetectSourceType(dirPath, suite, function(sourceType) {
                    t.strictEqual(Api.DIR_TEST_SOURCE, sourceType);
                    done();
                });
            },

            'Fixture type': function(done) {
                testCafe._tryDetectSourceType(fixturePath, suite, function (sourceType) {
                    t.strictEqual(Api.FIXTURE_TEST_SOURCE, sourceType);
                    done();
                });
            },

            'Single test type': function(done) {
                testCafe._tryDetectSourceType(testPath, suite, function(sourceType) {
                    t.strictEqual(Api.SINGLE_TEST_SOURCE, sourceType);
                    done();
                });
            },

            'Fixture uid': function(done) {
                var fixtureUid = testCafe.vfs.getFixtureUidByName('testDir/test', suite);

                testCafe._tryDetectSourceType(fixtureUid, suite, function(sourceType) {
                    t.strictEqual(Api.FIXTURE_TEST_SOURCE, sourceType);
                    done();
                });
            },

            'Single test uid': function(done) {
                var testUid = testCafe.vfs.getTestUidByName('testDir/test/Testing localhost', suite);

                testCafe._tryDetectSourceType(testUid, suite, function(sourceType) {
                    t.strictEqual(Api.SINGLE_TEST_SOURCE, sourceType);
                    done();
                });
            },

            'Group tests': function(done) {
                var testUids = [1, 2, 3, 4, 5];

                testCafe._tryDetectSourceType(testUids, suite, function(sourceType) {
                    t.strictEqual(Api.GROUP_TEST_SOURCE, sourceType);
                    done();
                });
            }
        }, function() {
            t.done();
        });
    }
};

exports['Events'] = {
    setUp: function (done) {
        testCafe = new TestCafe(getStartupConfig());
        watchdog = new testUtils.Watchdog();
        done();
    },

    tearDown: function (done) {
        testCafe.close();
        watchdog.shrink();

        process.nextTick(function () {
            done();
        });
    },

    'workerAdded': function (t) {
        t.expect(1);

        var workerName = 'testAdded';

        testCafe.on('workerAdded', function (name) {
            t.strictEqual(workerName, name);

            t.done();
        });

        testCafe.testRunner.workerPool.connectWorker(workerName, null);
    },

    'workerDiconnected': function (t) {
        t.expect(1);

        var tmp = WorkerPool.WORKER_HEARTBEAT_TIMEOUT,
            workerName = 'disconnectedWorker';

        WorkerPool.WORKER_HEARTBEAT_TIMEOUT = 0;

        testCafe.testRunner.workerPool.on('workerDisconnectedError', function () {
            t.done();
        });

        testCafe.on('workerDisconnected', function (name) {
            //NOTE: this event should be fired before workerDisconnectedError event
            t.strictEqual(workerName, name);

            WorkerPool.WORKER_HEARTBEAT_TIMEOUT = tmp;
        });

        testCafe.testRunner.workerPool.connectWorker(workerName, null);
    },

    'taskUpdated': function (t) {
        t.expect(1);

        var workerName = 'testRun1',
            dirName = 'testDir',
            options = {
                source: 'runTest/' + dirName,
                sourceType: 'dir',
                workers: [workerName],
                browsers: []
            },
            expectedJs = {
                uid: '',
                name: util.format(Api.DIRECTORY_TEST_TEXT, dirName),
                status: 'pending',
                startedAt: null,
                completedAt: null,
                time: 0,
                workerNames: ['testRun1'],
                browserVersions:{'testRun1':'0.0.0'},
                testCount: 1,
                failed: 0,
                passed: 0,
                testErrReports: {},
                passedTests: {}
            };

        testCafe.testRunner.workerPool.connectWorker(workerName, null);

        testCafe.on('taskUpdated', function (res) {
            process.nextTick(function () {
                t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(res));

                t.done();
            });
        });

        testCafe.runTests(options, function (errs, taskUid, workers) {
            expectedJs.uid = taskUid;
            testCafe.testRunner.reporter.taskCompleted(taskUid);
        });
    },

    'taskComplete': function (t) {
        var workerName = 'testRun1',
            dirName = 'testDir',
            options = {
                source: 'runTest/' + dirName,
                sourceType: 'dir',
                workers: [workerName],
                browsers: []
            },
            expectedJs = {
                uid: '',
                name: util.format(Api.DIRECTORY_TEST_TEXT, dirName),
                status: 'succeeded',
                startedAt: null,
                completedAt: null,
                time: 0,
                workerNames: ['testRun1'],
                browserVersions:{'testRun1':'0.0.0'},
                testCount: 1,
                failed: 0,
                passed: 0,
                testErrReports: {},
                passedTests: {}
            };

        testCafe.testRunner.workerPool.connectWorker(workerName, null);

        testCafe.on('taskComplete', function (res) {
            process.nextTick(function () {
                t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(res));

                t.done();
            });
        });

        testCafe.runTests(options, function (errs, taskUid, workers) {
            expectedJs.uid = taskUid;
            testCafe.testRunner.reporter.taskCompleted(taskUid);
        });
    }
};

exports['Api utils'] = {
    setUp: function (done) {
        testCafe = new TestCafe(getStartupConfig());
        watchdog = new testUtils.Watchdog();
        done();
    },

    tearDown: function (done) {
        testCafe.close();
        watchdog.shrink();

        process.nextTick(function () {
            done();
        });
    },

    'Cut private fields': function(t) {
        var obj = {
                a: 1,
                _b: 2,
                c: {
                    d: 1,
                    _e: 2,
                    f: {
                        g: 1,
                        _h: 2
                    }
                }
            },
            expected = {
                a: 1,
                c: {
                    d: 1,
                    f: {
                        g: 1
                    }
                }
            };

        t.strictEqual(JSON.stringify(expected), JSON.stringify(testCafe._clonePublicFields(obj)));
        t.done();
    }
};