var Cli = require('../../../../lib/cli/cli'),
    ERR = require('../../../../lib/server/server_errs'),
    errMsgBuilder = require('../../../../lib/err_msg_builder'),
    testUtils = require('../../test_utils'),

    fs = require('fs'),
    execFile = require('child_process').execFile,
    path = require('path'),
    tmp = require('tmp');

var helpFileContent = JSON.parse(fs.readFileSync('./lib/cli/help.json', 'UTF8')),
    testCafeDir = path.join(__dirname, '../../../../'),
    runnerPath = path.join(testCafeDir, './lib/cli/index'),
    configJSON = path.join(__dirname, '../data/cli/test_config.json');

var cli = null,
    stdout = '',
    stderr = '';

var setCliConfig = function (cli) {
    cli.cfg = {
        hostname: 'localhost',
        controlPanelPort: 1837,
        servicePort: 1838,
        testsDir: 'C:\\tests',
        browsers: {}
    };
};

exports['Cli'] = {
    setUp: function (done) {
        cli = new Cli();

        cli._logSync = function (out) {
            stdout = out;
        };

        cli._fatalError = function (err) {
            stderr = err;
        };

        done();
    },
    'Get next command': function (t) {
        var cmdArr = [
                [
                    '/runTests',
                    'source="ASPxperience"',
                    '/r'
                ],
                [
                    '/runTests'
                ],
                [
                    '/runTests',
                    'source="ASPxperience/ASPxCallback"'
                ],
                [
                    'test',
                    'fakeParam=1'
                ],
                [
                    '/unknownOptions',
                    'fakeParam=2'
                ]
            ],
            expectedArr = [
                {
                    type: '/runTests',
                    params: {
                        source: 'ASPxperience'
                    }
                },
                {
                    type: '/runTests',
                    params: {
                    }
                },
                {
                    type: '/runTests',
                    params: {
                        source: 'ASPxperience/ASPxCallback'
                    }
                },
                {
                    type: 'test',
                    params: {}
                },
                {
                    type: '/unknownOptions',
                    params: {
                        fakeParam: "2"
                    }
                }
            ];

        for (var i = 0, length = cmdArr.length; i < length; i++) {
            cli.cmd = cmdArr[i];

            var command = cli._getNextCommand();

            t.ok(testUtils.compareCode(JSON.stringify(expectedArr[i]), JSON.stringify(command)));
        }

        t.done();
    },

    'Help task': function (t) {
        var expectedHelp = '';

        Object.keys(helpFileContent).forEach(function (key) {
            expectedHelp = expectedHelp ? helpFileContent[key] : [expectedHelp, helpFileContent[key]].join('\n');
        });

        var expectedHelpByOption = helpFileContent['runTests'];

        cli._helpTask({}, function (data1) {
            t.strictEqual(expectedHelp, data1);

            cli._helpTask({'/runTests': ''}, function (data2) {
                t.strictEqual(expectedHelpByOption, data2);

                cli._helpTask({'/unknownOptions': ''}, function () {
                    t.strictEqual(JSON.stringify(stderr), JSON.stringify({code: ERR.CLI_HELP_UNKNOWN_OPTION, option: '/unknownOptions'}));
                    t.done();
                });
            });
        });
    },

    'Config task - get/set': function (t) {
        var expectedOut = 'TestCafe config:\n\thostname=localhost\n\tservicePort=1838\n\ttestsDir=C:\\tests\n',
            newHostname = 'newHostname',
            newTestsDir = 'newTestsDir',
            newControlPanelPort = 1500,
            newServicePort = 1501;

        setCliConfig(cli);

        cli._configTask({}, function () {
            t.strictEqual(stdout, expectedOut);

            cli._configTask({path: configJSON}, function () {
                var expectedConfig = JSON.parse(fs.readFileSync(configJSON, 'UTF8'));

                t.strictEqual(cli.cfg.hostname, expectedConfig.hostname);
                t.strictEqual(cli.cfg.testsDir, expectedConfig.testsDir);
                t.strictEqual(cli.cfg.servicePort, expectedConfig.servicePort);
                t.strictEqual(cli.cfg.controlPanelPort, expectedConfig.controlPanelPort);
                t.strictEqual(JSON.stringify(cli.cfg.browsers), JSON.stringify(expectedConfig.browsers));

                cli._configTask({hostname: newHostname, servicePort: newServicePort, controlPanelPort: newControlPanelPort, testsDir: newTestsDir}, function () {
                    t.strictEqual(cli.cfg.hostname, newHostname);
                    t.strictEqual(cli.cfg.servicePort, newServicePort);
                    t.strictEqual(cli.cfg.controlPanelPort, newControlPanelPort);
                    t.strictEqual(cli.cfg.testsDir, newTestsDir);

                    t.strictEqual(JSON.stringify(cli.cfg.browsers), JSON.stringify(expectedConfig.browsers));

                    t.done();
                });
            });
        });
    },

    'List browser': function (t) {
        cli.cfg = {
            hostname: 'localhost',
            controlPanelPort: 1837,
            servicePort: 1838,
            testsDir: 'C:\\tests',
            browsers: {
                browserOne: {
                    path: 'somePath'
                },
                browserTwo: {
                    path: 'somePath'
                }
            }
        };

        cli._listBrowsersTask({}, function () {
            t.strictEqual(stdout, 'Browsers:\nbrowserOne\nbrowserTwo');

            t.done();
        });
    },

    'Add browser': function (t) {
        var browserName = 'newBrowser',
            browserPath = 'testPath';

        setCliConfig(cli);

        cli._addBrowserTask({name: browserName, path: browserPath}, function () {
            t.strictEqual(cli.cfg.browsers[browserName].path, browserPath);

            delete cli.cfg.browsers[browserName];

            cli._addBrowserTask({name: browserName}, function () {
                t.ok(!cli.cfg.browsers.hasOwnProperty(browserName));
                t.strictEqual(stderr.code, ERR.CLI_BROWSER_PARAM_NOT_VALID);

                cli._addBrowserTask({path: browserPath}, function () {
                    t.strictEqual(stderr.code, ERR.CLI_BROWSER_PARAM_NOT_VALID);

                    t.done();
                });
            });
        });
    }/*,

    'Logging output': function (t) {
        tmp.dir(function (tmpErr, basePath) {
            var logFile = path.join(basePath, 'log.txt'),
            //NOTE: don't childProcess.fork() here because it will hang debugger. Use execFile with current process
            //execPath instead.
                subProcess = execFile(process.execPath, [runnerPath, '/log', 'path="' + logFile + '"', '/browser'], {
                    cwd: testCafeDir,
                    encoding: 'UTF8',
                    silent: true
                });

            subProcess.on('error', function (err) {
                t.ok(false, 'Uncaught exception: ' + err);
            });

            subProcess.on('exit', function () {
                t.strictEqual(fs.readFileSync(logFile, 'UTF8'), errMsgBuilder.build({code: ERR.CLI_BROWSER_PARAM_NOT_VALID}));

                t.done();
            });
        });
    }*/
};