var Config = require('../../../../lib/config'),
    ERR = require('../../../../lib/server/server_errs'),
    async = require('async'),
    fs = require('fs'),
    exec = require('child_process').exec,
    os = require('os'),
    path = require('path'),
    testUtils = require('../../test_utils'),
    tmp = require('tmp'),
    http = require('http');

Config.TESTCAFE_DIR = path.join(__dirname, '../data/config/');

var WIN_PLATFORM = /^win/.test(process.platform),
    WIN64_PLATFORM = WIN_PLATFORM && process.arch === 'x64',
    LICENSE_DATE_REG_KEY = 'HKEY_CLASSES_ROOT\\Licenses\\0378852D-D597-4A32-B6D9-680A16A3CDA6',
    USER_ID_REG_KEY = 'HKEY_CLASSES_ROOT\\Licenses\\5FCB5B35-D11E-4374-AD58-19A608D8228C',
    TRIAL_SEND_INFO_REG_KEY = WIN64_PLATFORM ?
        'HKEY_LOCAL_MACHINE\\SOFTWARE\\Wow6432Node\\DevExpress' :
        'HKEY_LOCAL_MACHINE\\SOFTWARE\\DevExpress';

var cfgPath = Config.CFG_FILE_PATH = path.join(Config.TESTCAFE_DIR, 'testcafe_config.json'),
    testsDir = path.join(__dirname, '../data/config/testsDir'),
    newTestsDir = path.join(__dirname, '../data/config/newTestsDir'),
    reportsDir = path.join(__dirname, '../data/config/reports'),
    defaultTestsDir = path.join(__dirname, '../../../../my_tests');

var testCfg = {
    controlPanelPort: 1777,
    servicePort: 1778,
    servicePort2: 1779,
    hostname: os.hostname(),
    testsDir: testsDir,
    reportsPath: path.join(Config.USER_DIR, 'reports'),
    browsers: {
        firefox: {
            path: path.join(__dirname, '../data/config/fakeBrowser.txt')
        }
    },
    recorder: {
        showSteps: false
    }
};

var testHttpServer1 = null,
    testHttpServer2 = null,
    basePath = null;

function checkTestCfg(t, cfg) {
    t.strictEqual(cfg.controlPanelPort, 1777);
    t.strictEqual(cfg.servicePort, 1778);
    t.strictEqual(cfg.servicePort2, 1779);
    t.strictEqual(cfg.hostname, os.hostname());
    t.strictEqual(cfg.testsDir, testsDir);
    t.strictEqual(cfg.reportsPath, path.join(Config.USER_DIR, 'reports'));
    t.ok(cfg.browsers.firefox);
    t.strictEqual(cfg.browsers.firefox.path, path.join(__dirname, '../data/config/fakeBrowser.txt'));
    t.strictEqual(cfg.controlPanelHost, [os.hostname(), ':1777'].join(''));
}

function checkDefaultConfig(t, cfg) {
    t.ok(cfg.servicePort && cfg.servicePort2 && cfg.controlPanelPort && cfg.testsDir && cfg.hostname &&
        (cfg.controlPanelHost === cfg.hostname + ':' + cfg.controlPanelPort));

    for (var browser in cfg.browsers) {
        t.ok((cfg.browsers[browser].cmd || cfg.browsers[browser].cmd === '') && cfg.browsers[browser].path &&
            (cfg.browsers[browser].icon || cfg.browsers[browser].icon === ''));

        break;
    }
}

exports['Config'] = {
    setUp: function (done) {
        fs.writeFileSync(cfgPath, '');

        done();
    },

    tearDown: function (done) {
        fs.writeFileSync(cfgPath, '');

        if (testHttpServer1) {
            testHttpServer1.close();
            testHttpServer1 = null;
        }

        if (testHttpServer2) {
            testHttpServer2.close();
            testHttpServer2 = null;
        }

        if (basePath) {
            testUtils.rmdirRecursive(basePath);
            basePath = null;
        }

        done();
    },

    'Load from config file': function (t) {
        t.expect(10);

        fs.writeFileSync(cfgPath, JSON.stringify(testCfg));

        var config = new Config();

        config._initFromFile(function (err) {
            checkTestCfg(t, config);

            t.ok(!err);
            t.done();
        });
    },

    'Load from config file: properties auto detection': function (t) {
        t.expect(3);

        var config = new Config();

        config._initFromFile(function (err) {
            checkDefaultConfig(t, config);

            t.ok(!err);
            t.done();
        });
    },

    'Load from config file: invalid config': function (t) {
        t.expect(5);

        fs.writeFileSync(cfgPath, JSON.stringify({
            servicePort: "value"
        }));

        var config = new Config();

        config._initFromFile(function (errs) {
            checkDefaultConfig(t, config);

            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, ERR.CONFIG_PORT_VALUE_IS_NOT_NUMBER);
            t.strictEqual(errs[0].property, 'servicePort');
            t.done();
        });
    },

    'Load from config file: file does not exist': function (t) {
        t.expect(4);

        fs.unlinkSync(cfgPath);

        var config = new Config();

        config._initFromFile(function (errs) {
            checkDefaultConfig(t, config);

            t.ok(!errs);
            t.ok(!fs.existsSync(path));

            fs.appendFileSync(cfgPath, '');
            t.done();
        });
    },

    'Load from config file: partial loading/saving': function (t) {
        t.expect(10);

        fs.writeFileSync(cfgPath, JSON.stringify({
            testsDir: testsDir,
            hostname: 123,
            servicePort: '1234'
        }));

        var config = new Config();

        config._initFromFile(function (errs) {
            config.setTestsDir(newTestsDir, function (err) {
                checkDefaultConfig(t, config);

                t.ok(!err);
                t.strictEqual(errs.length, 2);
                t.strictEqual(errs[0].code, ERR.CONFIG_PORT_VALUE_IS_NOT_NUMBER);
                t.strictEqual(errs[0].property, 'servicePort');
                t.strictEqual(errs[1].code, ERR.CONFIG_PROPERTY_IS_NOT_STRING);
                t.strictEqual(errs[1].property, 'hostname');
                t.ok(config.servicePort !== '1234' && config.testsDir === newTestsDir && config.hostname !== 123);

                var readedCfg = JSON.parse(fs.readFileSync(cfgPath));
                delete readedCfg.browsers;

                t.deepEqual(readedCfg, {
                    testsDir: newTestsDir,
                    hostname: 123,
                    servicePort: '1234'
                });

                t.done();
            });
        });
    },

    'Load from config file: setTestsDir meth': function (t) {
        t.expect(12);

        fs.writeFileSync(cfgPath, JSON.stringify(testCfg));

        var config = new Config();

        config._initFromFile(function (initErr) {
            checkTestCfg(t, config);

            t.ok(!initErr);
            config.setTestsDir(newTestsDir, function (setErr) {
                t.ok(!setErr);

                config._getConfig(function () {
                    var fileObj = JSON.parse(fs.readFileSync(cfgPath));

                    fileObj.controlPanelHost = fileObj.hostname + ':' + fileObj.controlPanelPort;
                    fileObj.initedFromObj = false;
                    fileObj.reportsPath = path.join(Config.USER_DIR, 'reports');

                    t.deepEqual(fileObj, config);
                    t.done();
                });
            });
        });
    },

    'Load from config file: config file has been created': function (t) {
        t.expect(2);

        fs.unlinkSync(cfgPath);

        var config = new Config();

        config._initFromFile(function () {
            config.setTestsDir(newTestsDir, function () {
                t.ok(fs.existsSync(cfgPath));

                var readedCfg = JSON.parse(fs.readFileSync(cfgPath));
                delete readedCfg.browsers;

                t.deepEqual(readedCfg, {
                    testsDir: newTestsDir
                });

                t.done();
            });
        });
    },

    'Load from config file: config not exists -  auto saving browsers': function (t) {
        t.expect(5);

        fs.unlinkSync(cfgPath);

        var config = new Config();

        config._initFromFile(function (errs) {
            var readedCfg = JSON.parse(fs.readFileSync(cfgPath));

            t.ok(!errs);
            t.ok(Object.keys(readedCfg.browsers).length > 0);
            t.ok(!readedCfg.controlPanelPort);
            t.ok(!readedCfg.servicePort);
            t.ok(!readedCfg.hostname);
            t.done();
        });
    },

    'Load from config file: tests dir not exists': function (t) {
        t.expect(2);

        var newTestsCfg = {
            testsDir: 'not_exists'
        };

        fs.writeFileSync(cfgPath, JSON.stringify(newTestsCfg));

        var config = new Config();

        if (/^win/.test(process.platform))
            defaultTestsDir = path.join(process.env['USERPROFILE'], 'Documents', 'TestCafe', 'Tests');

        config._initFromFile(function (errs) {
            t.ok(!errs);
            t.strictEqual(config.testsDir, defaultTestsDir);

            t.done();
        });
    },

    'Load from config file: auto detecting busy port': function (t) {
        t.expect(3);

        fs.writeFileSync(cfgPath, JSON.stringify(testCfg));

        var config = new Config();

        testHttpServer1 = http.createServer().listen(testCfg.servicePort);
        testHttpServer2 = http.createServer().listen(testCfg.controlPanelPort);

        config._initFromFile(function (errs) {
            t.ok(!errs);
            t.notStrictEqual(config.servicePort, testCfg.servicePort);
            t.notStrictEqual(config.controlPanelPort, testCfg.controlPanelPort);

            t.done();
        });
    },

    'Load from arguments': function (t) {
        t.expect(10);

        var config = new Config();

        config._initFromObj(testCfg, function (errs) {
            checkTestCfg(t, config);
            t.ok(!errs);
            t.done();
        });
    },

    'Load from arguments: invalid config': function (t) {
        t.expect(7);

        var config = new Config();

        config._initFromObj({
            servicePort: 'value',
            controlPanelPort: 1777,
            servicePort2: 1779,
            testsDir: 'testsDir'
        }, function (errs) {
            t.strictEqual(config.controlPanelPort, 1777);
            t.strictEqual(config.servicePort, "value");
            t.strictEqual(config.servicePort2, 1779);
            t.strictEqual(config.testsDir, 'testsDir');

            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, ERR.CONFIG_PORT_VALUE_IS_NOT_NUMBER);
            t.strictEqual(errs[0].property, 'servicePort');
            t.done();
        });
    },

    'Load from arguments: partial loading/saving denied': function (t) {
        t.expect(7);

        var config = new Config();

        config._initFromObj({}, function (errs) {
            t.strictEqual(errs.length, 3);
            t.strictEqual(errs[0].code, ERR.CONFIG_PROPERTY_IS_REQUIRED);
            t.strictEqual(errs[0].property, 'controlPanelPort');
            t.strictEqual(errs[1].code, ERR.CONFIG_PROPERTY_IS_REQUIRED);
            t.strictEqual(errs[1].property, 'servicePort');
            t.strictEqual(errs[2].code, ERR.CONFIG_PROPERTY_IS_REQUIRED);
            t.strictEqual(errs[2].property, 'testsDir');
            t.done();
        });
    },

    'Load from arguments: setTestsDir meth': function (t) {
        t.expect(12);

        var config = new Config();

        config._initFromObj(testCfg, function (errs) {
            checkTestCfg(t, config);

            t.ok(!errs);
            config.setTestsDir(newTestsDir, function (setErr) {
                t.ok(!setErr);
                config._getConfig(function () {
                    t.strictEqual(config.testsDir, newTestsDir);
                    t.done();
                });
            });
        });
    },

    'Load from arguments: config file should not created': function (t) {
        t.expect(1);

        fs.unlinkSync(cfgPath);

        var config = new Config();

        config._initFromObj(testCfg, function () {
            t.ok(!fs.existsSync(cfgPath));

            t.done();
        });
    },

    'Validation': function (t) {
        t.expect(46);

        var testData = [
            {
                cfg: { servicePort: '1234', servicePort2: 1235, controlPanelPort: 1236, testsDir: 'testDir', browsers: {} },
                errCode: ERR.CONFIG_PORT_VALUE_IS_NOT_NUMBER,
                property: 'servicePort'
            },
            {
                cfg: { servicePort: -1, servicePort2: 1235, controlPanelPort: 1236, testsDir: 'testDir', browsers: {} },
                errCode: ERR.CONFIG_PORT_VALUE_IS_OUT_OF_RANGE,
                property: 'servicePort'
            },
            {
                cfg: { servicePort: 65536, servicePort2: 1235, controlPanelPort: 1236, testsDir: 'testDir', browsers: {} },
                errCode: ERR.CONFIG_PORT_VALUE_IS_OUT_OF_RANGE,
                property: 'servicePort'
            },
            {
                cfg: { servicePort: 1234, servicePort2: '1235', controlPanelPort: 1236, testsDir: 'testDir', browsers: {} },
                errCode: ERR.CONFIG_PORT_VALUE_IS_NOT_NUMBER,
                property: 'servicePort2'
            },
            {
                cfg: { servicePort: 1234, servicePort2: -1, controlPanelPort: 1236, testsDir: 'testDir', browsers: {} },
                errCode: ERR.CONFIG_PORT_VALUE_IS_OUT_OF_RANGE,
                property: 'servicePort2'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 65536, controlPanelPort: 1236, testsDir: 'testDir', browsers: {} },
                errCode: ERR.CONFIG_PORT_VALUE_IS_OUT_OF_RANGE,
                property: 'servicePort2'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: '1234', testsDir: 'testDir', browsers: {} },
                errCode: ERR.CONFIG_PORT_VALUE_IS_NOT_NUMBER,
                property: 'controlPanelPort'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: -1, testsDir: 'testDir', browsers: {} },
                errCode: ERR.CONFIG_PORT_VALUE_IS_OUT_OF_RANGE,
                property: 'controlPanelPort'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: 65536, testsDir: 'testDir', browsers: {} },
                errCode: ERR.CONFIG_PORT_VALUE_IS_OUT_OF_RANGE,
                property: 'controlPanelPort'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: 1236, testsDir: '', browsers: {} },
                errCode: ERR.CONFIG_PROPERTY_CAN_NOT_BE_EMPTY,
                property: 'testsDir'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: 1236, testsDir: 0, browsers: {} },
                errCode: ERR.CONFIG_PROPERTY_IS_NOT_STRING,
                property: 'testsDir'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: 1236, testsDir: 'testsDir', hostname: 1, browsers: {} },
                errCode: ERR.CONFIG_PROPERTY_IS_NOT_STRING,
                property: 'hostname'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: 1236, testsDir: 'testsDir', hostname: '', browsers: {} },
                errCode: ERR.CONFIG_PROPERTY_CAN_NOT_BE_EMPTY,
                property: 'hostname'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: 1236, testsDir: 'testsDir', browsers: '' },
                errCode: ERR.CONFIG_INVALID_PROPERTY_FORMAT,
                property: 'browsers'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: 1236, testsDir: 'testsDir', browsers: { br1: '' } },
                errCode: ERR.CONFIG_INVALID_PROPERTY_FORMAT,
                property: 'browsers'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: 1236, testsDir: 'testsDir', browsers: { br1: { path: 'path', cmd: 1 } } },
                errCode: ERR.CONFIG_PROPERTY_IS_NOT_STRING,
                property: 'cmd'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: 1236, testsDir: 'testsDir', browsers: { br1: { path: 'path', icon: 1 } } },
                errCode: ERR.CONFIG_PROPERTY_IS_NOT_STRING,
                property: 'icon'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: 1236, testsDir: 'testsDir', browsers: { br1: { } } },
                errCode: ERR.CONFIG_PROPERTY_IS_REQUIRED,
                property: 'path'
            },
            {
                cfg: { servicePort: 1234, servicePort2: 1235, controlPanelPort: 1236, testsDir: 'testsDir', browsers: { br1: { path: 1 } } },
                errCode: ERR.CONFIG_PROPERTY_IS_NOT_STRING,
                property: 'path'
            },
            {
                cfg: {  servicePort: 1234, servicePort2: 1235, controlPanelPort: 1236, testsDir: 'testsDir', browsers: { br1: { path: '' } } },
                errCode: ERR.CONFIG_PROPERTY_CAN_NOT_BE_EMPTY,
                property: 'path'
            },
            {
                strongValidation: true,
                cfg: { servicePort2: 1234, controlPanelPort: 1235, testsDir: 'testsDir' },
                errCode: ERR.CONFIG_PROPERTY_IS_REQUIRED,
                property: 'servicePort'
            },
            {
                strongValidation: true,
                cfg: {  servicePort: 1234, servicePort2: 1235, testsDir: 'testsDir' },
                errCode: ERR.CONFIG_PROPERTY_IS_REQUIRED,
                property: 'controlPanelPort'
            },
            {
                strongValidation: true,
                cfg: {  servicePort: 1234, servicePort2: 1235, controlPanelPort: 1236 },
                errCode: ERR.CONFIG_PROPERTY_IS_REQUIRED,
                property: 'testsDir'
            }
        ];

        var checkValidationErr = function (index) {
            if (index >= testData.length)
                t.done();
            else {
                var config = new Config();

                config._initFromObj(testData[index].cfg, function (errs) {
                    t.strictEqual(errs[0].code, testData[index].errCode);
                    t.strictEqual(errs[0].property, testData[index].property);
                    checkValidationErr(index + 1);
                });
            }
        };

        checkValidationErr(0);
    },

    'Get hostname': function (t) {
        var config = new Config();

        async.series({
            'localhost': function (done) {
                config._getQualifiedHostname('localhost', function (hostname) {
                    t.strictEqual('localhost', hostname);
                    done();
                });
            },
            'non existing host': function (done) {
                config._getQualifiedHostname('localhost-test', function (hostname) {
                    t.strictEqual('127.0.0.1', hostname);
                    done();
                });
            },
            'local ip': function (done) {
                config._getQualifiedHostname('127.0.0.2', function (hostname) {
                    t.strictEqual('127.0.0.2', hostname);
                    done();
                });
            },
            'non valid symbols in hostname': function (done) {
                config._getQualifiedHostname('☀☆☂☻♞☯☭☢€→☎❄♫✂▷✇♎⇧☮♻⌘⌛', function (hostname) {
                    t.strictEqual('127.0.0.1', hostname);
                    done();
                });
            }
        }, function () {
            t.strictEqual('localhost', config._getHostname('localhost'));
            t.strictEqual('localhost-123456.test', config._getHostname('localhost-123456.test'));
            t.strictEqual('127.0.0.2', config._getHostname('127.0.0.2'));
            t.strictEqual('127.0.0.1', config._getHostname('localhost_test'));
            t.strictEqual('127.0.0.1', config._getHostname('☀☆☂☻♞☯☭☢€→☎❄♫✂▷✇♎⇧☮♻⌘⌛'));

            t.done();
        });
    },

    'Regression - check updates': function (t) {
        var config = new Config(),
            values = [
                {
                    versions: ['14.1.1', '14.1.1'],
                    expected: {
                        curVersion: '14.1.1',
                        newVersion: ''
                    }
                },
                {
                    versions: ['14.1.1', '14.1.113164994113'],
                    expected: {
                        curVersion: '14.1.1',
                        newVersion: '14.1.113164994113'
                    }
                },
                {
                    versions: ['14.1.123423534534', '14.1.1'],
                    expected: {
                        curVersion: '14.1.123423534534',
                        newVersion: ''
                    }
                },
                {
                    versions: ['14.1.1', '14.1.2'],
                    expected: {
                        curVersion: '14.1.1',
                        newVersion: '14.1.2'
                    }
                },
                {
                    versions: ['14.1.1', '14.1.223234234'],
                    expected: {
                        curVersion: '14.1.1',
                        newVersion: '14.1.223234234'
                    }
                },
                {
                    versions: ['14.1.1234345345', '14.1.23434534'],
                    expected: {
                        curVersion: '14.1.1234345345',
                        newVersion: '14.1.23434534'
                    }
                },
                {
                    versions: ['14.1.14334534', '14.1.234534534'],
                    expected: {
                        curVersion: '14.1.14334534',
                        newVersion: '14.1.234534534'
                    }
                },
                {
                    versions: ['14.1.2', '14.1.1'],
                    expected: {
                        curVersion: '14.1.2',
                        newVersion: ''
                    }
                },
                {
                    versions: ['14.1.2123123', '14.1.1'],
                    expected: {
                        curVersion: '14.1.2123123',
                        newVersion: ''
                    }
                },
                {
                    versions: ['14.1.2123123', '14.1.12342342342'],
                    expected: {
                        curVersion: '14.1.2123123',
                        newVersion: ''
                    }
                }
            ];

        async.forEachSeries(values, function (data, dataCallback) {
            config.getCurrentVersion = function (callback) {
                callback(data.versions[0]);
            };
            config._getActualVersion = function (callback) {
                callback(data.versions[1]);
            };

            config.checkUpdates(function (curVersion, newVersion) {
                t.strictEqual(curVersion, data.expected.curVersion);
                t.strictEqual(newVersion, data.expected.newVersion);

                dataCallback();
            });
        }, function () {
            t.done();
        });
    },

    'Get user ID (Win)': function (t) {
        var config = new Config(),
            userId = 'aaaaaaaa-1111-aaaa-1111-aaaaaaaaaaaa';

        config._getRegistryKeyValue = function (key, field, callback) {
            t.strictEqual(key, USER_ID_REG_KEY);
            t.strictEqual(field, '');
            callback(userId);
        };

        config.getUserID(function (val) {
            t.strictEqual(userId, val);

            t.done();
        })
    },

    'Check enable send info - trial mode': function (t) {
        var config = new Config(),
            version = testUtils.getCurVersionRegKeySync();

        var modifyRegKeyGetter = function (licenseValue, trialSendInfoValue) {
            config._getRegistryKeyValue = function (key, field, callback) {
                if (key === LICENSE_DATE_REG_KEY) {
                    t.strictEqual(field, version);
                    callback(licenseValue);
                }
                else if (key === TRIAL_SEND_INFO_REG_KEY)
                    callback(trialSendInfoValue);
                else {
                    t.ok(false, 'Unknown registry key' + key);
                    callback();
                }
            };
        };

        async.series({
            'Trial: send info not confirm': function (done) {
                modifyRegKeyGetter('', '');

                config.getLoggingAllowedFlag(function (isEnabled) {
                    t.ok(!isEnabled);
                    done();
                });
            },
            'Trial: send info confirm': function (done) {
                modifyRegKeyGetter('', '1');

                config.getLoggingAllowedFlag(function (isEnabled) {
                    t.ok(isEnabled);
                    done();
                });
            },
            'License: send info not confirm': function (done) {
                modifyRegKeyGetter('1234', '');

                config.getLoggingAllowedFlag(function (isEnabled) {
                    t.ok(!isEnabled);
                    done();
                });
            },
            'License: send info confirm': function (done) {
                modifyRegKeyGetter('1234', '1');

                config.getLoggingAllowedFlag(function (isEnabled) {
                    t.ok(!isEnabled);
                    done();
                });
            }
        }, t.done);
    }
};