var fs = require('fs'),
    testUtils = require('../../../test_utils'),
    tmp = require('tmp'),
    path = require('path'),
    async = require('async'),
    Hammerhead = require('./../../../../../hammerhead'),
    FixtureCode = require('../../../../../lib/fixture_code');

var np = testUtils.normalizePath,
    isNumber = testUtils.isNumber,
    watchdog = null;

exports['FixtureCode.Compiler.compile()'] = {
    setUp: function (done) {
        watchdog = new testUtils.Watchdog();
        done();
    },

    tearDown: function (done) {
        watchdog.shrink();
        done();
    },

    'Compile': function (t) {
        var compiler = new FixtureCode.Compiler('test/nodeunit/core/data/fixture_code_compiler/compile/src.test.js', {
            'testModule': [
                'test/nodeunit/core/data/fixture_code_compiler/compile/module/m_req1.js',
                'test/nodeunit/core/data/fixture_code_compiler/compile/module/m_req2.js'
            ]
        });

        compiler.compile(function (errs, out) {
            t.ok(!errs);

            t.strictEqual(out.fixture, 'Test fixture');
            t.strictEqual(out.page, 'http://my.page.url');
            t.strictEqual(out.authCredentials.username, 'myLogin');
            t.strictEqual(out.authCredentials.password, 'myPassword');

            var expectedRequireJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/compile/expected_require_js.js').toString();
            t.ok(testUtils.compareCode(out.requireJs, expectedRequireJs));

            var expectedRemainderJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/compile/expected_remainder_js.js').toString();
            t.ok(testUtils.compareCode(out.remainderJs, expectedRemainderJs));

            var expectedTest1Js = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/compile/test1_expected.js').toString();
            t.ok(testUtils.compareCode(out.testsStepData['My first test'].js, expectedTest1Js));
            t.deepEqual(out.testsStepData['My first test'].names, ['1.Do smthg cool', '2.Stop here']);

            var expectedTest2Js = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/compile/test2_expected.js').toString();
            t.ok(testUtils.compareCode(out.testsStepData['I want more tests!'].js, expectedTest2Js));
            t.deepEqual(out.testsStepData['I want more tests!'].names, [
                '1.Here we go',
                "2.I'm really tired creating stupid names for test steps",
                '3.This is a final step'
            ]);

            t.done();
        });
    },

    'Read file failed': function (t) {
        var filename = './non/existent/file',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.READ_FILE_FAILED);
            t.strictEqual(np(errs[0].filename), filename);
            t.done();
        });
    },

    'Javascript parsing failed': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/parsing_failed.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.JAVASCRIPT_PARSING_FAILED);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(errs[0].parserErr);
            t.done();
        });
    },

    'Auth directive redefinition': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/auth_directive_redefinition.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.AUTH_DIRECTIVE_REDEFINITION);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Fixture redefinition': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/fixture_redefinition.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.FIXTURE_DIRECTIVE_REDEFINITION);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Page directive redefinition': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/page_directive_redefinition.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.PAGE_DIRECTIVE_REDEFINITION);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Auth credentials invalid format': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/auth_credentials_invalid_format.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.INVALID_NETWORK_AUTHENTICATION_CREDENTIALS_FORMAT);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Duplicate require': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/duplicate_require.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.REQUIRED_FILE_ALREADY_INCLUDED);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Misplaced directive expression': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/misplaced_directive_expression.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.MISPLACED_DIRECTIVE);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Misplaced test directive': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/misplaced_test_directive.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.MISPLACED_TEST_DECLARATION);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Duplicate test name': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/duplicate_test_name.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.DUPLICATE_TEST_NAME);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(errs[0].name);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Empty test name': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/empty_test_name.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.EMPTY_TEST_NAME);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Invalid test assignment': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/invalid_test_assignment.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.INVALID_TEST_ASSIGNMENT);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Test step is not a function': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/test_step_is_not_a_function.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.TEST_STEP_IS_NOT_A_FUNCTION_OR_MIXIN);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Test is empty': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/test_is_empty.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.TEST_IS_EMPTY);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Test step is empty': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/test_step_is_empty.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.TEST_STEP_IS_EMPTY);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Action func is not a last entry': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/action_func_is_not_a_last_entry.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.ACTION_FUNC_IS_NOT_A_LAST_ENTRY);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Action func is called in shared code': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/action_func_call_in_shared_code/src.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 3);

            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.ACTION_FUNC_CALL_IN_SHARED_CODE);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));

            t.strictEqual(errs[1].code, FixtureCode.ErrCodes.ACTION_FUNC_CALL_IN_SHARED_CODE);
            t.ok(np(errs[1].filename) !== filename);
            t.ok(isNumber(errs[1].line));

            t.strictEqual(errs[2].code, FixtureCode.ErrCodes.ACTION_FUNC_CALL_IN_SHARED_CODE);
            t.ok(np(errs[2].filename) !== filename);
            t.ok(isNumber(errs[2].line));

            t.done();
        });
    },

    'Fixture is undefined': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/fixture_is_undefined.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.FIXTURE_DIRECTIVE_IS_UNDEFINED);
            t.strictEqual(np(errs[0].filename), filename);
            t.done();
        });
    },

    'Load expression is undefined': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/page_directive_is_undefined.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.PAGE_DIRECTIVE_IS_UNDEFINED);
            t.strictEqual(np(errs[0].filename), filename);
            t.done();
        });
    },

    'Async func call': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/async_func_call/async_func_call.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.ASYNC_FUNC_CALL);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Failed load require': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/failed_load_require.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.FAILED_LOAD_REQUIRE);
            t.strictEqual(np(errs[0].ownerFilename), filename);
            t.ok(errs[0].filename);
            t.done();
        });
    },

    'Wrap dom accessors in a module': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/wrap_dom_accessors/src.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs, out) {
            t.ok(!errs);

            var expectedRequireJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/wrap_dom_accessors/require_expected.js').toString();
            t.ok(testUtils.compareCode(out.requireJs, expectedRequireJs));

            var expectedRemainderJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/wrap_dom_accessors/remainder_js_expected.js').toString();
            t.ok(testUtils.compareCode(out.remainderJs, expectedRemainderJs));

            var expectedTestJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/wrap_dom_accessors/test_expected.js').toString();
            t.ok(testUtils.compareCode(out.testsStepData['My first test'].js, expectedTestJs));

            t.done();
        });
    },

    'Module not found': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/module_not_found.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.MODULE_NOT_FOUND);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(errs[0].moduleName);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Misplaced mixin declaration': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/misplaced_mixin_declaration.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.MISPLACED_MIXIN_DECLARATION);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Empty mixin name': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/empty_mixin_name.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.EMPTY_MIXIN_NAME);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Duplicate mixin name': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/duplicate_mixin_name.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.DUPLICATE_MIXIN_NAME);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(errs[0].name);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Duplicate mixin name in require': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/duplicate_in_require/src.test.js',
            req1Filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/duplicate_in_require/req1.js',
            req2Filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/duplicate_in_require/req2.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 2);

            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.DUPLICATE_MIXIN_NAME_IN_REQUIRE);
            t.strictEqual(errs[0].name, 'Mixin1');
            t.strictEqual(np(errs[0].filename), filename);
            t.strictEqual(np(errs[0].defFilename1), req2Filename);
            t.strictEqual(np(errs[0].defFilename2), req1Filename);

            t.strictEqual(errs[1].code, FixtureCode.ErrCodes.DUPLICATE_MIXIN_NAME_IN_REQUIRE);
            t.strictEqual(errs[1].name, 'Mixin2');
            t.strictEqual(np(errs[1].filename), filename);
            t.strictEqual(np(errs[1].defFilename1), req2Filename);
            t.strictEqual(np(errs[1].defFilename2), filename);

            t.done();
        });
    },

    'Mixin is empty': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/mixin_is_empty.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.MIXIN_IS_EMPTY);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Mixin step is not a function': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/mixin_step_is_not_a_function.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.MIXIN_STEP_IS_NOT_A_FUNCTION);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Mixin step is empty': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/mixin_step_is_empty.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.MIXIN_STEP_IS_EMPTY);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Invalid mixin assignment': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/invalid_mixin_assignment.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.INVALID_MIXIN_ASSIGNMENT);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Async func call in mixin': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/async_func_call.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.ASYNC_FUNC_CALL);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Action func is not a last entry in mixin step': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/action_func_is_not_a_last_entry.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.ACTION_FUNC_IS_NOT_A_LAST_ENTRY);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Mixin inside mixin': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/mixin_inside_mixin.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.MIXIN_USED_IN_MIXIN);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Undefined mixin used': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/mixins/undefined_mixin_used.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.UNDEFINED_MIXIN_USED);
            t.ok(errs[0].mixinName);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(isNumber(errs[0].line));
            t.done();
        });
    },

    'Compile with mixins': function (t) {
        var compiler = new FixtureCode.Compiler('test/nodeunit/core/data/fixture_code_compiler/mixins/compile/src.test.js');

        compiler.compile(function (errs, out) {
            t.ok(!errs);


            var expectedTest1Js = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/mixins/compile/test1_expected.js').toString();
            t.ok(testUtils.compareCode(out.testsStepData['Test1'].js, expectedTest1Js));
            t.deepEqual(out.testsStepData['Test1'].names, [
                '1.Add mixin | a.This awesome mixin step',
                '1.Add mixin | b.Do more awesome stuff',
                '2.Do smthg cool',
                '3.Stop here'
            ]);

            var expectedTest2Js = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/mixins/compile/test2_expected.js').toString();
            t.ok(testUtils.compareCode(out.testsStepData['Test2'].js, expectedTest2Js));
            t.deepEqual(out.testsStepData['Test2'].names, [
                '1.Here we go',
                '2.Do the mix | a.This awesome mixin step',
                '2.Do the mix | b.Do more awesome stuff',
                "3.I'm really tired creating stupid names for test steps",
                '4.This is a final step'
            ]);

            var expectedTest3Js = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/mixins/compile/test3_expected.js').toString();
            t.ok(testUtils.compareCode(out.testsStepData['Test3'].js, expectedTest3Js));
            t.deepEqual(out.testsStepData['Test3'].names, [
                '1.Yoyo',
                '2.Shake it, baby',
                '3.Ready to go | a.Yo dawg',
                '3.Ready to go | b.I heard you like tests',
                '3.Ready to go | c.So we put test in your test'
            ]);

            var expectedRemainderJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/mixins/compile/remainder_expected.js').toString();
            t.ok(testUtils.compareCode(out.remainderJs, expectedRemainderJs));

            var expectedRequireJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/mixins/compile/req_expected.js').toString();
            t.ok(testUtils.compareCode(out.requireJs, expectedRequireJs));

            t.done();
        });
    },

    'Compile - Source index': function (t) {
        var sourceIndex = [],
            compiler = new FixtureCode.Compiler('test/nodeunit/core/data/fixture_code_compiler/source_index/src.test.js', null, null, sourceIndex);

        compiler.compile(function (errs, out) {
            t.ok(!errs);

            //Test1
            var expectedTest1Js = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/source_index/test1_expected.js').toString();
            t.ok(testUtils.compareCode(out.testsStepData['Test1'].js, expectedTest1Js));

            //Test2Case1
            var expectedTest2Case1Js = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/source_index/test2_case1_expected.js').toString();
            t.ok(testUtils.compareCode(out.testsStepData['Test2   \u2192   Case1'].js, expectedTest2Case1Js));

            //Test2Case2
            var expectedTest2Case2Js = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/source_index/test2_case2_expected.js').toString();
            t.ok(testUtils.compareCode(out.testsStepData['Test2   \u2192   Case2'].js, expectedTest2Case2Js));

            var expectedRemainderJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/source_index/remainder_expected.js').toString();
            t.ok(testUtils.compareCode(out.remainderJs, expectedRemainderJs));

            var expectedRequireJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/source_index/require_expected.js').toString();
            t.ok(testUtils.compareCode(out.requireJs, expectedRequireJs));

            var expectedSourceIndex = require('../../data/fixture_code_compiler/source_index/source_index_expected');
            t.deepEqual(sourceIndex, expectedSourceIndex);

            t.done();
        });
    },

    'Compile with inIFrame': function (t) {
        var compiler = new FixtureCode.Compiler('test/nodeunit/core/data/fixture_code_compiler/in_iframe/compile/src.test.js');

        compiler.compile(function (errs, out) {
            t.ok(!errs);

            var expectedTestJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/in_iframe/compile/expected.js').toString();
            t.ok(testUtils.compareCode(out.testsStepData['Test1'].js, expectedTestJs));

            t.done();
        });
    },

    'inIFrame - insufficient arguments': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/in_iframe/insufficient_arguments.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.INIFRAME_FUNCTION_SHOULD_ACCEPT_TWO_ARGS);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(errs[0].line);
            t.done();
        });
    },

    'inIFrame - test step is not a function': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/in_iframe/test_step_is_not_a_function.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.TEST_STEP_IS_NOT_A_FUNCTION_OR_MIXIN);
            t.strictEqual(np(errs[0].filename), filename);
            t.ok(errs[0].line);
            t.done();
        });
    },

    'Compile with test cases': function (t) {
        var compiler = new FixtureCode.Compiler('test/nodeunit/core/data/fixture_code_compiler/test_cases/compile/src.test.js');

        compiler.compile(function (errs, out) {
            t.ok(!errs);
            t.strictEqual(Object.keys(out.testsStepData).length, 5);

            var test1ExpectedStepNames = [
                '-INIT TEST CASE-',
                '1.Here we go',
                '2.I\'m really tired creating stupid names for test steps',
                '3.This is a final step'
            ];

            var test2ExpectedStepNames = [
                '-INIT TEST CASE-',
                '1.Do smthg cool',
                '2.Stop here'
            ];

            //Test1Case1
            var name = 'Test1   \u2192   TestCase1',
                expectedJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/test_cases/compile/test1_testcase1_expected.js').toString();

            t.ok(testUtils.compareCode(out.testsStepData[name].js, expectedJs));
            t.deepEqual(out.testsStepData[name].names, test1ExpectedStepNames);
            t.strictEqual(out.testGroupMap[name], 'Test1');

            //Test1Case2
            name = 'Test1   \u2192   TestCase2';
            expectedJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/test_cases/compile/test1_testcase2_expected.js').toString();

            t.ok(testUtils.compareCode(out.testsStepData[name].js, expectedJs));
            t.deepEqual(out.testsStepData[name].names, test1ExpectedStepNames);
            t.strictEqual(out.testGroupMap[name], 'Test1');

            //Test1Case3
            name = 'Test1   \u2192   Test case at index 2';
            expectedJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/test_cases/compile/test1_testcase3_expected.js').toString();

            t.ok(testUtils.compareCode(out.testsStepData[name].js, expectedJs));
            t.deepEqual(out.testsStepData[name].names, test1ExpectedStepNames);
            t.strictEqual(out.testGroupMap[name], 'Test1');

            //Test2Case1
            name = 'My first test   \u2192   OneMoreCase';
            expectedJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/test_cases/compile/test2_testcase1_expected.js').toString();

            t.ok(testUtils.compareCode(out.testsStepData[name].js, expectedJs));
            t.deepEqual(out.testsStepData[name].names, test2ExpectedStepNames);
            t.strictEqual(out.testGroupMap[name], 'My first test');

            //Test2Case2
            name = 'My first test   \u2192   And more';
            expectedJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_compiler/test_cases/compile/test2_testcase2_expected.js').toString();

            t.ok(testUtils.compareCode(out.testsStepData[name].js, expectedJs));
            t.deepEqual(out.testsStepData[name].names, test2ExpectedStepNames);
            t.strictEqual(out.testGroupMap[name], 'My first test');

            t.done();
        });
    },

    'Failed to read external test case': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/test_cases/failed_to_read_external_test_cases.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.FAILED_TO_READ_EXTERNAL_TEST_CASES);
            t.ok(errs[0].testCasesPath);
            t.strictEqual(np(errs[0].filename), filename);
            t.done();
        });
    },

    'External test cases list is not array': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/test_cases/external_test_cases_list_is_not_array/src.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.TEST_CASES_LIST_IS_NOT_ARRAY);
            t.ok(errs[0].line);
            t.ok(errs[0].filename);
            t.ok(errs[0].filename !== filename);
            t.done();
        });
    },

    'Test cases list is not array': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/test_cases/test_cases_list_is_not_array.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.TEST_CASES_LIST_IS_NOT_ARRAY);
            t.ok(errs[0].line);
            t.strictEqual(np(errs[0].filename), filename);
            t.done();
        });
    },

    'Test cases list is empty': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/test_cases/test_cases_list_is_empty.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.TEST_CASES_LIST_IS_EMPTY);
            t.ok(errs[0].line);
            t.strictEqual(np(errs[0].filename), filename);
            t.done();
        });
    },

    'Test case is not an object': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/test_cases/test_case_is_not_an_object.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.TEST_CASE_IS_NOT_AN_OBJECT);
            t.ok(errs[0].line);
            t.strictEqual(np(errs[0].filename), filename);
            t.done();
        });
    },

    'Test case doesn\'t contain any fields': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/test_cases/test_case_doesnt_contain_any_fields.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.TEST_CASE_DOESNT_CONTAIN_ANY_FIELDS);
            t.ok(errs[0].line);
            t.strictEqual(np(errs[0].filename), filename);
            t.done();
        });
    },

    'Test case name is not a string': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/test_cases/test_case_name_is_not_a_string.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.TEST_CASE_NAME_IS_NOT_A_STRING);
            t.ok(errs[0].line);
            t.strictEqual(np(errs[0].filename), filename);
            t.done();
        });
    },

    'Duplicate test case name': function (t) {
        var filename = 'test/nodeunit/core/data/fixture_code_compiler/test_cases/duplicate_test_case_name.test.js',
            compiler = new FixtureCode.Compiler(filename);

        compiler.compile(function (errs) {
            t.strictEqual(errs.length, 1);
            t.strictEqual(errs[0].code, FixtureCode.ErrCodes.DUPLICATE_TEST_CASE_NAME);
            t.ok(errs[0].line);
            t.strictEqual(errs[0].testCaseName, 'TestCase1');
            t.strictEqual(np(errs[0].filename), filename);
            t.done();
        });
    },


    'T105929 - Mixins - Playback and page markup crashed': function (t) {
        var compiler = new FixtureCode.Compiler('test/nodeunit/core/data/fixture_code_compiler/mixins/T105929/src.test.js');

        compiler.compile(function (errs, out) {
            t.ok(!errs);
            t.strictEqual(out.requireJs, '');

            t.done();
        });
    },

    'T105923 - Error raised when two tests in fixture contain the same mixin (connected via @require directive)': function (t) {
        var requiresDescriptorCache = {},
            compiler1 = new FixtureCode.Compiler('test/nodeunit/core/data/fixture_code_compiler/mixins/T105923/src.test.js', null, requiresDescriptorCache),
            compiler2 = new FixtureCode.Compiler('test/nodeunit/core/data/fixture_code_compiler/mixins/T105923/src.test.js', null, requiresDescriptorCache);

        //NOTE: mixins that was stored in requiresDescriptorCache were merged twice during compilation. This caused duplicate mixin
        //declaration error.
        compiler1.compile(function () {
            compiler2.compile(function (errs) {
                t.ok(!errs);
                t.done();
            });
        });
    },

    'T113201 - Mixin in require with the precending comment causes compilation error': function (t) {
        var compiler = new FixtureCode.Compiler('test/nodeunit/core/data/fixture_code_compiler/mixins/T113201/src.test.js');

        compiler.compile(function (errs, out) {
            t.ok(!errs);
            t.ok(Object.keys(out.testsStepData).length, 1);

            t.done();
        });
    },

    'T215683: TestCafe cannot perform Windows authentication if the password contains a colon': function (t) {
        var compiler = new FixtureCode.Compiler('test/nodeunit/core/data/fixture_code_compiler/T215683.test.js');

        compiler.compile(function (errs, out) {
            t.ok(!errs);
            t.strictEqual(out.authCredentials.username, 'someName');
            t.strictEqual(out.authCredentials.password, 'd7W:8LK:XKS2w#T');

            t.done();
        });
    }
};