var fs = require('fs'),
    path = require('path'),
    tmp = require('tmp'),
    Recorder = require('../../../../lib/server/recorder').Recorder,
    Hammerhead = require('../../../../hammerhead'),
    CLIENT_ERR = require('../../../../hammerhead/shared/client_errs'),
    VirtualFS = require('../../../../lib/vfs'),
    FixtureCode = require('../../../../lib/fixture_code/'),
    cmd = require('../../../../lib/shared/service_msg_cmd'),
    testUtils = require('../../test_utils'),
    ERR = require('../../../../lib/server/server_errs');


var hammerhead = null,
    recorder = null,
    vfs = null,
    watchdog = null,
    basePath = null,
    rmdirRecursive = testUtils.rmdirRecursive;

exports['Start recording'] = {
    setUp: function (done) {
        watchdog = new testUtils.Watchdog();
        hammerhead = new Hammerhead.Hammerhead(testUtils.TEST_CAFE_PROXY_PORT);
        hammerhead.start();

        done();
    },

    tearDown: function (done) {
        watchdog.shrink();
        vfs.removeFSWatchers();
        hammerhead.close();

        done();
    },

    'Start recording': function (t) {
        vfs = new VirtualFS('test/nodeunit/core/data/recorder/start_recording');
        recorder = new Recorder(hammerhead, vfs);

        vfs.getSuite(function (suite) {
            var fixtureUid = Object.keys(suite.fixtures)[0],
                fixture = suite.fixtures[fixtureUid],
                returnUrl = 'www.test.test';

            recorder.startRecording(fixtureUid, returnUrl, function (err, recordingUrl) {
                t.ok(recordingUrl);
                t.ok(!err);

                var recordingUids = Object.keys(recorder.recordings);
                t.strictEqual(recordingUids.length, 1);

                var recording = recorder.recordings[recordingUids[0]];

                t.strictEqual(recording.fixturePath, fixture.path);
                t.strictEqual(recording.fixtureFileName, fixture.fileName);
                t.strictEqual(recording.returnUrl, returnUrl);
                t.ok(recording.uid);
                t.ok(recording.stepsInfo);
                t.strictEqual(recording.playback.run, false);
                t.strictEqual(recording.playback.nextStep, 0);
                t.strictEqual(recording.playback.steps, '[]');

                t.done();
            });
        });
    },

    'Fixture not exists ': function (t) {
        vfs = new VirtualFS('test/nodeunit/core/data/recorder/start_recording');
        recorder = new Recorder(hammerhead, vfs);

        recorder.startRecording('some unknown uid', 'www.test.test', function (err, recordingUrl) {
            t.strictEqual(err.code, ERR.RECORDER_TARGET_FIXTURE_DOESNT_EXISTS);
            t.ok(!recordingUrl);

            t.done();
        });
    },

    'Invalid url': function (t) {
        vfs = new VirtualFS('test/nodeunit/core/data/recorder/start_recording_invalid_url');
        recorder = new Recorder(hammerhead, vfs);

        vfs.getSuite(function (suite) {
            var fixtureUid = Object.keys(suite.fixtures)[0];

            recorder.startRecording(fixtureUid, 'www.test.test', function (err, recordingUrl) {
                t.strictEqual(err.code, CLIENT_ERR.URL_UTIL_PROTOCOL_IS_NOT_SUPPORTED);
                t.ok(!recordingUrl);

                t.done();
            });
        });
    }
};

exports["Complete recording"] = {
    setUp: function (done) {
        watchdog = new testUtils.Watchdog();
        hammerhead = new Hammerhead.Hammerhead(testUtils.TEST_CAFE_PROXY_PORT);
        hammerhead.start();

        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;

            done();
        });
    },

    tearDown: function (done) {
        watchdog.shrink();
        vfs.removeFSWatchers();
        hammerhead.close();
        rmdirRecursive(basePath);

        done();
    },

    'Save test': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            recorder = new Recorder(hammerhead, vfs),
            testName = 'Successfully saving',
            recordingUid = '00000000-0000-0000-0000-000000000000',
            fileContent = "'@fixture test';'@page http://testPage';",
            fixtureFileName = 'fixtureFile.test.js',
            filePath = [],
            fileName = path.join(fs.realpathSync(basePath), fixtureFileName),
            stepsInfo = [
                {
                    name: 'Click element',
                    action: 'click',
                    currentSelectorIndex: 0,
                    selectors: [
                        { id: 'id', selector: "$(\"#titleLink\")"}
                    ],
                    actionArgs: [
                        {
                            options: {
                                ctrl: true,
                                alt: true,
                                shift: true,
                                meta: true,
                                offsetX: -1,
                                offsetY: -1
                            }
                        }
                    ],
                    useOffsets: true
                }
            ],
            expectedTestContent = '{' +
                '"1.Click element": function() {' +
                'act.click("#titleLink", {' +
                'ctrl: true,' +
                'alt: true,' +
                'shift: true,' +
                'meta: true,' +
                'offsetX: -1,' +
                'offsetY: -1' +
                '});' +
                '}' +
                '};',
            expectedJs = fileContent + '"@test"["' + testName + '"]=' + expectedTestContent;


        var fileDesc = fs.openSync(fileName, "w");

        fs.writeSync(fileDesc, fileContent, null, null);
        fs.close(fileDesc);

        recorder.recordings[recordingUid] = {
            uid: recordingUid,
            fixturePath: filePath,
            fixtureFileName: fixtureFileName,
            returnUrl: '',
            stepsInfo: stepsInfo
        };

        recorder.saveTest(recordingUid, testName, function (res) {
            var actualJs = fs.readFileSync(fileName, 'utf8'),
                err = res.err;

            t.ok(!err);
            t.ok(testUtils.compareCode(expectedJs, actualJs));

            fs.unlinkSync(fileName);

            t.done();
        });
    },

    'Save test with empty testName': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            recorder = new Recorder(hammerhead, vfs),
            testName = '',
            recordingUid = '00000000-0000-0000-0000-000000000000',
            fileContent = "'@fixture test';'@page http://testPage';",
            fixtureFileName = 'fixtureFile.test.js',
            filePath = [],
            fileName = path.join(fs.realpathSync(basePath), fixtureFileName),
            stepsInfo = [
                {
                    name: 'Click element',
                    action: 'click',
                    selectors: [
                        { id: 'id', selector: "$(\"#titleLink\")"}
                    ],
                    actionArgs: [
                        {
                            options: {
                                ctrl: false,
                                alt: false,
                                shift: false,
                                meta: false,
                                offsetX: -1,
                                offsetY: -1
                            }
                        }
                    ]
                }
            ];

        var fileDesc = fs.openSync(fileName, "w");

        fs.writeSync(fileDesc, fileContent, null, null);
        fs.close(fileDesc);

        recorder.recordings[recordingUid] = {
            uid: recordingUid,
            fixturePath: filePath,
            fixtureFileName: fixtureFileName,
            returnUrl: '',
            stepsInfo: stepsInfo
        };

        recorder.saveTest(recordingUid, testName, function (res) {
            t.strictEqual(res.err.code, ERR.RECORDER_TEST_NAME_IS_EMPTY);

            fs.unlinkSync(fileName);

            t.done();
        });
    },

    'Save test with empty stepInfo': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            recorder = new Recorder(hammerhead, vfs),
            testName = 'Test with empty steps info',
            recordingUid = '00000000-0000-0000-0000-000000000000',
            fileContent = "'@fixture test';'@page http://testPage';",
            fixtureFileName = 'fixtureFile.test.js',
            filePath = [],
            fileName = path.join(fs.realpathSync(basePath), fixtureFileName),
            stepsInfo = [];

        var fileDesc = fs.openSync(fileName, "w");

        fs.writeSync(fileDesc, fileContent, null, null);
        fs.close(fileDesc);

        recorder.recordings[recordingUid] = {
            uid: recordingUid,
            fixturePath: filePath,
            fixtureFileName: fixtureFileName,
            returnUrl: '',
            stepsInfo: stepsInfo
        };

        recorder.saveTest(recordingUid, testName, function (res) {
            t.strictEqual(res.err.code, ERR.RECORDER_RECORDING_IS_EMPTY);

            fs.unlinkSync(fileName);

            t.done();
        });
    },

    'Save to non existing file': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            recorder = new Recorder(hammerhead, vfs),
            testName = 'Test with empty steps info',
            recordingUid = '00000000-0000-0000-0000-000000000000',
            fixtureFileName = 'fixtureFile.test.js',
            filePath = [],
            stepsInfo = [
                {
                    name: 'Click element',
                    action: 'click',
                    selectors: [
                        { id: 'id', selector: "$(\"#titleLink\")"}
                    ],
                    actionArgs: [
                        {
                            options: {
                                ctrl: false,
                                alt: false,
                                shift: false,
                                meta: false,
                                offsetX: -1,
                                offsetY: -1
                            }
                        }
                    ]
                }
            ];


        recorder.recordings[recordingUid] = {
            uid: recordingUid,
            fixturePath: filePath,
            fixtureFileName: fixtureFileName,
            returnUrl: '',
            stepsInfo: stepsInfo
        };

        recorder.saveTest(recordingUid, testName, function (res) {
            t.strictEqual(res.err.code, FixtureCode.ErrCodes.READ_FILE_FAILED);

            t.done();
        });
    },

    'Save non existing recording': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            recorder = new Recorder(hammerhead, vfs),
            testName = 'Test with empty steps info',
            recordingUid = 'non_existing_value_123';

        recorder.saveTest(recordingUid, testName, function (res) {
            t.strictEqual(res.err.code, ERR.RECORDER_RECORDING_DOESNT_EXISTS);

            t.done();
        });
    },

    'Remove default options from code': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            recorder = new Recorder(hammerhead, vfs),
            testName = 'Without default options',
            recordingUid = '00000000-0000-0000-0000-000000000000',
            fileContent = "'@fixture test';'@page http://testPage';",
            fixtureFileName = 'fixtureFile.test.js',
            filePath = [],
            fileName = path.join(fs.realpathSync(basePath), fixtureFileName),
            stepsInfo = [
                {
                    name: 'Click element',
                    action: 'click',
                    currentSelectorIndex: 0,
                    selectors: [
                        { id: 'id', selector: "$(\"#titleLink\")"}
                    ],
                    actionArgs: [
                        {
                            options: {
                                ctrl: false,
                                alt: false,
                                shift: false,
                                meta: false,
                                offsetX: -1,
                                offsetY: -1
                            }
                        }
                    ],
                    useOffsets: false
                },

                {
                    name: 'Type in element',
                    action: 'type',
                    currentSelectorIndex: 0,
                    selectors: [
                        { id: 'id', selector: "$(\"#titleLink\")"}
                    ],
                    actionArgs: [
                        {
                            text: 'text'
                        },
                        {
                            options: {
                                ctrl: false,
                                alt: false,
                                shift: false,
                                meta: false,
                                offsetX: -1,
                                offsetY: -1,
                                replace: true,
                                caretPos: ''
                            }
                        }
                    ],
                    useOffsets: false
                },

                {
                    name: 'Drag element',
                    action: 'drag',
                    currentSelectorIndex: 0,
                    selectors: [
                        { id: 'id', selector: "$(\"#titleLink\")"}
                    ],
                    actionArgs: [
                        {
                            dragOffsetX: 100
                        },
                        {
                            dragOffsetY: 50
                        },
                        {
                            options: {
                                ctrl: false,
                                alt: false,
                                shift: false,
                                meta: false,
                                offsetX: 0,
                                offsetY: 0
                            }
                        }
                    ],
                    useOffsets: false
                }
            ],

            expectedTestContent = '{' +
                '"1.Click element": function() {' +
                'act.click("#titleLink");' +
                '},' +
                '"2.Type in element": function() {' +
                'act.type("#titleLink", "text", {' +
                'replace: true' +
                '});' +
                '},' +
                '"3.Drag element": function() {' +
                'act.drag("#titleLink", 100, 50);' +
                '}' +
                '};',

            expectedJs = fileContent + '"@test"["' + testName + '"]=' + expectedTestContent;

        var fileDesc = fs.openSync(fileName, "w");

        fs.writeSync(fileDesc, fileContent, null, null);
        fs.close(fileDesc);

        recorder.recordings[recordingUid] = {
            uid: recordingUid,
            fixturePath: filePath,
            fixtureFileName: fixtureFileName,
            returnUrl: '',
            stepsInfo: stepsInfo
        };

        recorder.saveTest(recordingUid, testName, function (res) {
            var actualJs = fs.readFileSync(fileName, 'utf8'),
                err = res.err;

            t.ok(!err);
            t.ok(testUtils.compareCode(expectedJs, actualJs));

            fs.unlinkSync(fileName);

            t.done();
        });
    }
};

exports["Start playback"] = {
    setUp: function (done) {
        watchdog = new testUtils.Watchdog();
        hammerhead = new Hammerhead.Hammerhead(testUtils.TEST_CAFE_PROXY_PORT);
        hammerhead.start();

        done();
    },

    tearDown: function (done) {
        watchdog.shrink();
        vfs.removeFSWatchers();
        hammerhead.close();

        done();
    },

    'Start playback': function (t) {
        var vfs = new VirtualFS('test/nodeunit/core/data/recorder/start_recording');
        recorder = new Recorder(hammerhead, vfs);

        vfs.getSuite(function (suite) {
            var fixtureUid = Object.keys(suite.fixtures)[0],
                fixture = suite.fixtures[fixtureUid],
                returnUrl = 'www.test.test';

            recorder.startRecording(fixtureUid, returnUrl, function (err, recordingUrl) {
                var recordingUids = Object.keys(recorder.recordings),
                    recording = recorder.recordings[recordingUids[0]];

                recording.stepsInfo = [
                    {
                        name: 'Click element',
                        action: 'click',
                        selectors: [
                            { id: 'id', selector: "$(\"#titleLink\")"}
                        ],
                        actionArgs: [
                            {
                                options: {
                                    ctrl: true,
                                    alt: true,
                                    shift: true,
                                    meta: true,
                                    offsetX: -1,
                                    offsetY: -1
                                }
                            }
                        ],
                        useOffsets: true
                    },

                    {
                        name: 'assertion',
                        isAssertion: true,
                        blocks: [
                            {
                                assertions: [
                                    {
                                        operator: 'eq',
                                        arguments: [ 'document.location', 'test' ], //T172267
                                        message: 'check document location'
                                    }
                                ]
                            }
                        ]
                    }
                ];

                recorder.startPlayback(recordingUids[0], function (err, recordingUrl) {
                    t.ok(recordingUrl);
                    t.ok(!err);

                    t.strictEqual(recording.fixturePath, fixture.path, 'fixture path is correct');
                    t.strictEqual(recording.fixtureFileName, fixture.fileName, 'fixture file name is correct');
                    t.strictEqual(recording.returnUrl, returnUrl, 'return url is correct');
                    t.ok(recording.uid, 'recording uid exists');
                    t.ok(recording.stepsInfo.length === 2, 'step info count is correct');
                    t.strictEqual(recording.playback.run, true, 'playback flag exists');
                    t.strictEqual(recording.playback.nextStep, 0, 'next step is correct');
                    t.ok(testUtils.compareCode(recording.playback.steps, Hammerhead.wrapDomAccessors(FixtureCode.CodeGenerator.getTestStepBodies(recording.stepsInfo))), 'steps are correct');

                    t.done();

                });
            });
        });
    }
};

exports["T173757 - TD_14_2 TestCafe is crashed after some test recorded and browser's 'Back' button pressed"] = {
    'Start recording': function (t) {
        var hammerhead = new Hammerhead.Hammerhead(testUtils.TEST_CAFE_PROXY_PORT),
            vfs = new VirtualFS('test/nodeunit/core/data/recorder/start_recording');

        var recorder = new Recorder(hammerhead, vfs),
            err = null,
            res = null;

        try {
            recorder._onTaskScriptRequest({
                jobInfo: { uid: '1234-1234-1234-1234' }
            }, null, function (script) {
                res = script;
            });
        }
        catch (e) {
            err = e;
        }

        t.equal(err, null);
        t.equal(res, '');

        t.done();
    }
};

exports["T172267 - TD_14_2 - Assertion params don't wraps"] = {
    'Start recording': function (t) {
        var hammerhead = new Hammerhead.Hammerhead(testUtils.TEST_CAFE_PROXY_PORT),
            vfs = new VirtualFS('test/nodeunit/core/data/recorder/start_recording');

        var recorder = new Recorder(hammerhead, vfs),
            res = null;

        vfs.getSuite(function (suite) {
            var fixtureUid = Object.keys(suite.fixtures)[0],
                returnUrl = 'www.test.test';

            recorder.startRecording(fixtureUid, returnUrl, function () {
                var stepsInfo = [
                    {
                        name: 'assertion',
                        isAssertion: true,
                        blocks: [
                            {
                                assertions: [
                                    {
                                        operator: 'eq',
                                        arguments: [ 'document.location', 'test' ],
                                        message: 'check document location'
                                    }
                                ]
                            }
                        ]
                    }
                ];

                recorder._onServiceMsg({ jobUid: recorder.recordings[Object.keys(recorder.recordings)].uid, cmd: cmd.STEPS_INFO_SET, stepsInfo: stepsInfo, getStepsScript: true}, function (res) {
                    t.ok(testUtils.compareCode(res, Hammerhead.wrapDomAccessors(FixtureCode.CodeGenerator.getTestStepBodies(stepsInfo))));

                    t.done();
                });
            });
        });
    }
};