var VirtualFS  = require('../../../../../lib/vfs'),
    TestRunner = require('../../../../../lib/test_runner'),
    Config     = require('../../../../../lib/config'),
    Hammerhead = require('../../../../../hammerhead').Hammerhead,
    testUtils  = require('../../../test_utils'),
    cmd        = require('../../../../../lib/shared/service_msg_cmd'),
    ERR        = require('../../../../../lib/server/server_errs');

function objToKeyValPairArr (obj) {
    return Object.keys(obj).map(function (key) {
        return {
            key: key,
            val: obj[key]
        };
    });
}

function getTestUids (targetSuite) {
    var testDescrs = [];

    //NOTE: we sort uids by test name, so we can rely on this order in tests
    Object.keys(targetSuite.tests).forEach(function (testUid) {
        testDescrs.push({
            uid:  testUid,
            name: targetSuite.tests[testUid].name
        });
    });

    return testDescrs.sort(function (d1, d2) {
        if (d1.name === d2.name)
            return 0;

        return d1.name < d2.name ? -1 : 1;
    }).map(function (descr) {
        return descr.uid;
    });
}

var watchdog   = null,
    hammerhead = null,
    testRunner = null,
    vfs        = null;

exports['Tasks and testRuns'] = {
    setUp: function (done) {
        watchdog   = new testUtils.Watchdog();
        hammerhead = new Hammerhead(testUtils.TEST_CAFE_PROXY_PORT);
        hammerhead.start();

        testRunner = new TestRunner(hammerhead, new Config());

        done();
    },

    tearDown: function (done) {
        hammerhead.close();

        if (vfs) {
            vfs.removeFSWatchers();
            vfs = null;
        }

        //NOTE: clear worker heartbeat timeouts to let process terminate immidiately
        testRunner.workerPool.removeAllWorkers();

        watchdog.shrink();

        done();
    },

    'Add task': function (t) {
        vfs = new VirtualFS('test/nodeunit/core/data/test_runner/test_suite');
        t.expect(23);

        vfs.getSuite(function (suite) {
            testRunner.workerPool.connectWorker('worker1', null);
            testRunner.workerPool.connectWorker('worker2', null);

            testRunner.reporter.taskAdded = function (taskUid, name, testCount, workerNames, reportUnstable) {
                t.ok(taskUid);
                t.strictEqual(name, 'Test task');
                t.strictEqual(testCount, 1);
                t.ok(workerNames.indexOf('worker1') > -1);
                t.ok(workerNames.indexOf('worker2') > -1);
            };

            var taskUid = testRunner.addTask('Test task', suite, getTestUids(suite), ['worker1', 'worker2'],
                { quarantineMode: false, reportFormat: null, reportPath: '' });

            var tasksKeyValArr = objToKeyValPairArr(testRunner.tasks),
                task           = tasksKeyValArr[0].val;

            t.strictEqual(tasksKeyValArr.length, 1);
            t.strictEqual(taskUid, tasksKeyValArr[0].key);
            t.strictEqual(task.pendingTestRunCount, 2);

            var testRunsArray = objToKeyValPairArr(testRunner.testRuns);
            t.strictEqual(testRunsArray.length, 2);
            t.strictEqual(testRunsArray[0].val.stepsSharedData.__workerName, 'worker1');
            t.strictEqual(testRunsArray[1].val.stepsSharedData.__workerName, 'worker2');

            var worker1 = testRunner.workerPool.get('worker1');

            t.strictEqual(worker1.testRuns.length, 1);

            var worker1TestRun = testRunner.testRuns[worker1.testRuns[0]];

            taskUid = tasksKeyValArr[0].key;

            t.strictEqual(worker1TestRun.workerName, 'worker1');
            t.strictEqual(worker1TestRun.taskUid, taskUid);
            t.ok(worker1TestRun.test);
            t.ok(worker1TestRun.fixture);
            t.ok(!worker1TestRun.started);

            var worker2 = testRunner.workerPool.get('worker2');

            t.strictEqual(worker2.testRuns.length, 1);

            var worker2TestRun = testRunner.testRuns[worker2.testRuns[0]];

            t.strictEqual(worker2TestRun.workerName, 'worker2');
            t.strictEqual(worker2TestRun.taskUid, taskUid);
            t.ok(worker2TestRun.test);
            t.ok(worker2TestRun.fixture);
            t.ok(!worker2TestRun.started);

            t.done();
        });
    },

    'Source assigned to task': function (t) {
        vfs = new VirtualFS('test/nodeunit/core/data/test_runner/test_suite');

        vfs.getSuite(function (suite) {
            testRunner.workerPool.connectWorker('worker1', null);
            testRunner.workerPool.connectWorker('worker2', null);


            var taskUid1 = testRunner.addTask('Test task', suite, getTestUids(suite), ['worker1', 'worker2'],
                { quarantineMode: false, reportFormat: null, reportPath: '' });

            var sourceIndex1 = testRunner.tasks[taskUid1].sourceIndex;

            t.ok(sourceIndex1.length);

            vfs.requireRebuild = true;

            vfs.getSuite(function (suite) {
                var taskUid2 = testRunner.addTask('Test task2', suite, getTestUids(suite), ['worker1', 'worker2'],
                    { quarantineMode: false, reportFormat: null, reportPath: '' });

                var sourceIndex2 = testRunner.tasks[taskUid2].sourceIndex;

                t.ok(sourceIndex2.length);
                t.strictEqual(sourceIndex1, testRunner.tasks[taskUid1].sourceIndex);
                t.ok(sourceIndex1 !== sourceIndex2);

                t.done();
            });
        });
    },

    'Convert source index to code': function (t) {
        var testRun = {
                errs:    [],
                taskUid: 'taskID'
            },
            task    = {
                sourceIndex: [
                    'source1',
                    'source2'
                ]
            };

        testRunner.tasks['taskID']   = task;
        testRunner.testRuns['jobID'] = testRun;

        var noop = function () {
            //Do nothing =)
        };

        testRunner._onServiceMsg({
            cmd:    cmd.ASSERTION_FAILED,
            jobUid: 'jobID',
            err:    {
                __sourceIndex: 1
            }
        }, noop);

        testRunner._onServiceMsg({
            cmd:    cmd.ASSERTION_FAILED,
            jobUid: 'jobID',
            err:    {}
        }, noop);

        testRunner._onServiceMsg({
            cmd:    cmd.TEST_FAIL,
            jobUid: 'jobID',
            err:    {
                __sourceIndex: 0
            }
        }, noop);

        t.strictEqual(testRun.errs.length, 3);
        t.strictEqual(testRun.errs[0].relatedSourceCode, 'source2');
        t.ok(!testRun.errs[1].relatedSourceCode);
        t.strictEqual(testRun.errs[2].relatedSourceCode, 'source1');

        t.done();
    },

    'Quarantine - success/.../.../.../...': function (t) {
        var resetCalledCount = 0,
            testRun          = {
                errs:       [],
                quarantine: {
                    succeeded: 0,
                    failed:    0
                }
            };

        testRunner._resetTestRun = function () {
            resetCalledCount++;
        };

        //NOTE: test shouldn't be placed to quarantine and we don't have any errors
        t.ok(!testRunner._quarantine(testRun));
        t.strictEqual(resetCalledCount, 0);
        t.strictEqual(testRun.quarantine.succeeded, 1);
        t.done();
    },

    'Quarantine - fail in 5 cases': function (t) {
        var resetCalledCount = 0,
            testRun          = {
                errs:       ['someError'],
                quarantine: {
                    succeeded: 0,
                    failed:    0
                }
            };

        testRunner._resetTestRun = function () {
            resetCalledCount++;
        };

        //NOTE: test added to quarantine
        t.ok(testRunner._quarantine(testRun));
        testRun.errs = ['secondError'];

        //NOTE: we should run challenge 3 more time before exit from quarantine
        t.ok(testRunner._quarantine(testRun));
        testRun.errs = ['thirdError'];

        t.ok(testRunner._quarantine(testRun));
        testRun.errs = ['fourthError'];

        t.ok(testRunner._quarantine(testRun));
        testRun.errs = ['fifthError'];

        //NOTE: we done with quarantine
        t.ok(!testRunner._quarantine(testRun));

        t.strictEqual(testRun.quarantine.succeeded, 0);
        t.strictEqual(testRun.quarantine.failed, 5);
        t.strictEqual(resetCalledCount, 4);

        t.strictEqual(testRun.quarantine.runErrs.testRun1Errs[0], 'someError');
        t.strictEqual(testRun.quarantine.runErrs.testRun2Errs[0], 'secondError');
        t.strictEqual(testRun.quarantine.runErrs.testRun3Errs[0], 'thirdError');
        t.strictEqual(testRun.quarantine.runErrs.testRun4Errs[0], 'fourthError');
        t.strictEqual(testRun.quarantine.runErrs.testRun5Errs[0], 'fifthError');

        t.done();
    },

    'Quarantine - fail/fail/fail/fail/success': function (t) {
        var resetCalledCount = 0,
            testRun          = {
                errs:       ['someError'],
                quarantine: {
                    succeeded: 0,
                    failed:    0
                }
            };

        testRunner._resetTestRun = function () {
            resetCalledCount++;
        };

        //NOTE: test added to quarantine
        t.ok(testRunner._quarantine(testRun));
        testRun.errs = ['secondError'];
        t.ok(testRunner._quarantine(testRun));
        testRun.errs = ['thirdError'];
        t.ok(testRunner._quarantine(testRun));
        testRun.errs = ['fourthError'];

        //NOTE: we should run challenge one more time before exit from quarantine
        t.ok(testRunner._quarantine(testRun));

        testRun.errs = [];

        //NOTE: we done with quarantine
        t.ok(!testRunner._quarantine(testRun));

        t.strictEqual(testRun.quarantine.succeeded, 1);
        t.strictEqual(testRun.quarantine.failed, 4);
        t.strictEqual(resetCalledCount, 4);

        //NOTE: test run should not contain any errors
        t.strictEqual(testRun.errs.length, 0);

        t.done();
    },

    'Restart current test run': function (t) {
        var workerName     = 'testWorker',
            testRunUid     = 'testTestRunUid',
            startedTestRun = {
                workerName:      workerName,
                stepsSharedData: {
                    testData: 'testData'
                },
                halted:          true,
                errs:            [
                    { code: 'SOME_ERR_CODE' }
                ],
                restartCount:    0
            };

        testRunner.workerPool.connectWorker(workerName, null);
        testRunner.testRuns[testRunUid] = startedTestRun;
        testRunner.workerPool.get(workerName).testRuns.push(testRunUid);

        testRunner.restartCurrentTestRun(workerName);

        t.strictEqual(startedTestRun.stepsSharedData.__workerName, workerName);
        t.ok(!startedTestRun.stepsSharedData.testData);
        t.ok(!startedTestRun.halted);
        t.strictEqual(startedTestRun.errs.length, 0);
        t.strictEqual(startedTestRun.restartCount, 1);

        t.done();
    },

    'Max test run restarts reached': function (t) {
        var workerName = 'testWorker',
            testRunUid = 'testTestRunUid',
            testRun    = {
                stepsSharedData: null,
                halted:          false,
                errs:            [],
                restartCount:    0
            };

        testRunner.workerPool.connectWorker(workerName, null);
        testRunner.testRuns[testRunUid] = testRun;
        testRunner.workerPool.get(workerName).testRuns.push(testRunUid);

        for (var i = 0; i < TestRunner.MAX_TEST_RUN_RESTARTS + 1; i++)
            testRunner.restartCurrentTestRun(workerName);

        t.ok(testRun.halted);
        t.strictEqual(testRun.errs[0].code, ERR.TEST_RUNNER_MAX_TEST_RUN_RESTARTS);

        t.done();
    },

    'Regression - disable quarantine if worker has been disconnected': function (t) {
        t.expect(1);

        var workerName      = 'worker',
            isTaskCompleted = false;

        vfs = new VirtualFS('test/nodeunit/core/data/test_runner/test_suite');

        vfs.getSuite(function (suite) {
            testRunner.workerPool.connectWorker(workerName, null);

            var taskUid = testRunner.addTask('Test task', suite, getTestUids(suite), [workerName],
                { quarantineMode: true, reportFormat: null, reportPath: '' });

            testRunner.reporter.taskCompleted = function (uid) {
                if (uid === taskUid)
                    isTaskCompleted = true;
            };

            setTimeout(function () {
                t.ok(isTaskCompleted);
                t.done();
            }, 200);

            testRunner.workerPool._disconnectWorker(workerName);
        });
    },

    'Set maximum testRuns time for test': function (t) {
        var workerNames = ['wrk1', 'wrk2', 'wrk3'];

        vfs = new VirtualFS('test/nodeunit/core/data/test_runner/test_suite');

        vfs.getSuite(function (suite) {
            testRunner.workerPool.connectWorker(workerNames[0], null);
            testRunner.workerPool.connectWorker(workerNames[1], null);
            testRunner.workerPool.connectWorker(workerNames[2], null);

            var taskUid     = testRunner.addTask('Test task', suite, getTestUids(suite), workerNames,
                { quarantineMode: true, reportFormat: null, reportPath: '' }),
                taskReport  = testRunner.reporter.getTaskReportByUid(taskUid),
                testRunUids = Object.keys(testRunner.testRuns);

            testRunner.testRuns[testRunUids[0]].time = testRunner.testRuns[testRunUids[2]].time = 1;
            testRunner.testRuns[testRunUids[1]].time = 2;

            testRunUids.forEach(function (testRunUid) {
                var testRun = testRunner.testRuns[testRunUid];

                testRun.errs = [{
                    err:            {
                        actual:   false,
                        code:     'CLIENT_API_OK_ASSERTION_FAILED',
                        stepName: 'test',
                        stepNum:  1
                    },
                    possibleCauses: []
                }];
                testRunner.reporter._processTestRunErrs(testRun, taskReport, testRun.test.uid);
            });

            var testErrsReports = testRunner.reporter.getTaskReportByUid(taskUid).testErrReports;

            t.strictEqual(testErrsReports[Object.keys(testErrsReports)[0]].time, 2);

            testRunner.workerPool._disconnectWorker(workerNames[0]);
            testRunner.workerPool._disconnectWorker(workerNames[1]);
            testRunner.workerPool._disconnectWorker(workerNames[2]);

            t.done();
        });
    },

    'Proxy error should be proccessed if proxy context doesn\'t contain originResContentInfo (T468247)': function (t) {
        var noop    = function () {

        };
        var errMock = {};

        var proxyCtxMock = {
            jobInfo: {
                uid: 'testTestRunUid'
            },

            res: {
                end:       noop,
                setHeader: noop,
                writeHead: noop
            }
        };

        testRunner._onProxyError(proxyCtxMock, errMock);

        t.done();
    }
};
