var async = require('async'),
    VirtualFS = require('../../../../../lib/vfs'),
    TestRunner = require('../../../../../lib/test_runner'),
    WorkerPool = require('../../../../../lib/test_runner/worker_pool'),
    Config = require('../../../../../lib/config'),
    Hammerhead = require('../../../../../hammerhead').Hammerhead,
    testUtils = require('../../../test_utils'),
    ERR = require('../../../../../lib/server/server_errs');

function getTestUids(targetSuite) {
    var testDescrs = [];

    //NOTE: we sort uids by test name, so we can rely on this order in tests
    Object.keys(targetSuite.tests).forEach(function (testUid) {
        testDescrs.push({
            uid: testUid,
            name: targetSuite.tests[testUid].name
        });
    });

    return  testDescrs.sort(function (d1, d2) {
        if (d1.name === d2.name)
            return 0;

        return d1.name < d2.name ? -1 : 1;
    }).map(function (descr) {
            return descr.uid;
        });
}

function getTestRunByTestName(testRuns, testName) {
    var result = null;

    Object.keys(testRuns).forEach(function (testRunUid) {
        var testRun = testRuns[testRunUid];

        if (testRun.test.name === testName && !result)
            result = testRun;
    });

    return result;
}

function getObjPropCount(obj) {
    var count = 0;

    Object.keys(obj).forEach(function (prop) {
        count++;
    });

    return count;
}

var watchdog = null,
    hammerhead = null,
    testRunner = null,
    vfs = null;

exports['Workers'] = {
    setUp: function (done) {
        watchdog = new testUtils.Watchdog();
        hammerhead = new Hammerhead(testUtils.TEST_CAFE_PROXY_PORT, testUtils.TEST_CAFE_CROSS_DOMAIN_PROXY_PORT, '127.0.0.1');
        testRunner = new TestRunner(hammerhead, new Config());
        hammerhead.start();
        done();
    },

    tearDown: function (done) {
        hammerhead.close();

        if (vfs) {
            vfs.removeFSWatchers();
            vfs = null;
        }


        //NOTE: clear worker heartbeat timeouts to let process terminate immidiately
        testRunner.workerPool.removeAllWorkers();
        watchdog.shrink();

        done();
    },

    'Add worker': function (t) {
        var actualName = 'new_worker_name';

        testRunner.workerPool.workerAddedHandlers[actualName] = function () {
            t.ok(testRunner.workerPool.get(actualName).heartbeatTimeout);
            t.done();
        };

        testRunner.workerPool.connectWorker(' new\n_worker_name', null);
    },

    'Add worker with empty name': function (t) {
        t.expect(1);

        try {
            testRunner.workerPool.connectWorker('  \n  ', null);
        } catch (err) {
            t.strictEqual(err.code, ERR.WORKER_POOL_WORKER_NAME_IS_EMPTY);
        }

        t.done();
    },

    'Add worker with duplicate name': function (t) {
        var name = 'wrkr';

        testRunner.workerPool.connectWorker(name, null);
        t.expect(2);

        try {
            testRunner.workerPool.connectWorker(name, null);
        } catch (err) {
            t.strictEqual(err.code, ERR.WORKER_POOL_WORKER_NAME_IS_ALREADY_USED);
            t.strictEqual(err.name, name);
        }

        t.done();
    },

    'List connected workers': function (t) {
        testRunner.workerPool.connectWorker('wrkr1', null);
        testRunner.workerPool.connectWorker('wrkr2', null);
        testRunner.workerPool.createWorkerInNewBrowserWindow('browser', null);

        var workers = testRunner.workerPool.listConnectedWorkers();

        t.strictEqual(2, workers.length);
        t.ok(workers.indexOf('wrkr1') > -1);
        t.ok(workers.indexOf('wrkr2') > -1);

        t.done();
    },

    'Heartbeat stop': function (t) {
        var savedTimeout = WorkerPool.WORKER_HEARTBEAT_TIMEOUT;
        WorkerPool.WORKER_HEARTBEAT_TIMEOUT = 0;

        var name = 'wrkr';

        testRunner._completeTestRun = function (testRun) {
            t.ok(testRun);
            t.strictEqual(testRun.errs[0].code, ERR.WORKER_POOL_WORKER_DISCONNECTED);
        };

        testRunner.workerPool.connectWorker(name, null);
        testRunner.workerPool.get(name).testRuns.push('testRun1');
        testRunner.workerPool.get(name).testRuns.push('testRun2');
        testRunner.testRuns['testRun1'] = {uid: 'testRun1', errs: []};
        testRunner.testRuns['testRun2'] = {uid: 'testRun2', errs: []};

        t.expect(5);

        setTimeout(function () {
            WorkerPool.WORKER_HEARTBEAT_TIMEOUT = savedTimeout;

            t.ok(!testRunner.workerPool.get(name));

            t.done();
        }, 100);
    },

    'Worker idle heartbeat': function (t) {
        var name = 'wrkr';

        t.expect(5);

        testRunner.workerPool.connectWorker(name, null);

        vfs = new VirtualFS('test/nodeunit/core/data/test_runner/worker_idle_heartbeat');

        vfs.getSuite(function (suite) {
            var startedTaskUid = testRunner.addTask('Test task', suite, getTestUids(suite), [name],
                {quarantineMode: false, reportFormat: null, reportPath: ''});

            testRunner.reporter.taskStarted = function (taskUid) {
                t.strictEqual(taskUid, startedTaskUid);
            };

            var res = testRunner.getWorkerStatus(name);
            t.strictEqual(res.status, 'run');
            t.ok(res.testRunUrl);

            getTestRunByTestName(testRunner.testRuns, 'Test1').halted = true;

            var testRun2 = getTestRunByTestName(testRunner.testRuns, 'Test2');

            res = testRunner.getWorkerStatus(name);
            t.strictEqual(testRun2.errs.length, 1);
            t.strictEqual(res.status, 'idle');

            vfs.removeFSWatchers();

            t.done();
        });
    },

    'Stop test run and task': function (t) {
        t.expect(4);

        vfs = new VirtualFS('test/nodeunit/core/data/test_runner/test_suite');

        vfs.getSuite(function (suite) {
            var workerName = 'wrkr';

            testRunner.workerPool.connectWorker(workerName, null);
            var addedTaskUid = testRunner.addTask('Test task', suite, getTestUids(suite), [workerName],
                {quarantineMode: false, reportFormat: null, reportPath: ''});

            testRunner.reporter.taskCompleted = function (taskUid) {
                t.strictEqual(taskUid, addedTaskUid);
            };

            var worker = testRunner.workerPool.get(workerName),
                testRun = getTestRunByTestName(testRunner.testRuns, 'Test');

            testRunner._completeTestRun(testRun);

            t.strictEqual(worker.testRuns.length, 0);
            t.strictEqual(getObjPropCount(testRunner.testRuns), 0);
            t.strictEqual(getObjPropCount(testRunner.tasks), 0);

            vfs.removeFSWatchers();
            t.done();
        });
    },

    'Get worker status': function (t) {
        var workerName = 'testWorker';

        testRunner.workerPool.connectWorker(workerName, null);
        var idleStatus = testRunner.workerPool.getWorkerStatus(workerName, null).status;
        testRunner.workerPool.workers[workerName].utilized = true;

        var utilizedStatus = testRunner.workerPool.getWorkerStatus(workerName, null).status;

        t.strictEqual(idleStatus, 'idle');
        t.strictEqual(utilizedStatus, 'utilized');
        t.done();
    },

    'Get auto created worker name': function (t) {
        var chromeUserAgent = 'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.22 (KHTML, like Gecko) Chrome/25.0.1364.172 Safari/537.22',
            expectedValue = 'Google Chrome';

        t.strictEqual(testRunner.workerPool.getWorkerNameFromUserAgent(chromeUserAgent), expectedValue);
        t.done();
    },

    'Get workers info': function (t) {
        var workerName = 'testWorker',
            userAgent = 'test',
            expectedJs = [
                { workerName: workerName,
                    workerType: WorkerPool.WORKER_TYPE.CONNECTED,
                    userAgent: userAgent,
                    browserVersion: '0.0.0',
                    taskReport: null,
                    currentTest: null,
                    connected: true}
            ];

        testRunner.workerPool.connectWorker(workerName, userAgent);

        var workersInfo = testRunner.workerPool.getWorkersInfo();

        t.strictEqual(JSON.stringify(expectedJs), JSON.stringify(workersInfo));

        t.done();
    },

    'Restart worker': function(t) {
        t.expect(4);

        WorkerPool.WORKER_RESTART_TIMEOUT = 50;

        var cases = [
            {
                name: 'Init from object - success',
                eventsCount: 3,
                workerType: WorkerPool.WORKER_TYPE.NEW_BROWSER_WINDOW,
                initedFromObj: true
            },
            {
                name: 'Init from object - fail -  other worker type: control panel window',
                eventsCount: 0,
                workerType: WorkerPool.WORKER_TYPE.CONTROL_PANEL_WINDOW,
                initedFromObj: true
            },
            {
                name: 'Init from object - fail -  other worker type: connected',
                eventsCount: 0,
                workerType: WorkerPool.WORKER_TYPE.CONNECTED,
                initedFromObj: true
            },
            {
                name: 'Init from file',
                eventsCount: 0,
                workerType: WorkerPool.WORKER_TYPE.NEW_BROWSER_WINDOW,
                initedFromObj: false
            }
        ];

        async.forEachSeries(cases, function(elm, cb) {
            var eventsCount = 0,
                workerName = 'case: ' + elm.name;

            testRunner.workerPool._addWorker(workerName, elm.workerType, '', '');
            testRunner.workerPool.config.initedFromObj = elm.initedFromObj;
            testRunner.workerPool.get(workerName).browserProc = {};

            testRunner.workerPool.get(workerName).browserProc.kill = function() {
                eventsCount++;
            };

            testRunner.workerPool.removeAllListeners('workerRestart');
            testRunner.workerPool.on('workerRestart', function() {
                eventsCount++;
            });

            testRunner.workerPool.execBrowsersAndCreateWorkers = function() {
                eventsCount++;
            };

            testRunner.workerPool._disconnectWorker(workerName);

            setTimeout(function() {
                t.strictEqual(eventsCount, elm.eventsCount);
                cb();
            }, 200);
        }, function() {
            t.done();
        });
    }
};
