var http = require('http'),
    Config = require('../../../../lib/config'),
    testUtils = require('../../test_utils'),
    UXLog = require('../../../../lib/server/uxlog');

var httpServer = null,
    watchdog = null;

UXLog.prototype._getLoggingSettings = function () {
    this.loggingAllowed = true;
    this.version = '123';
    this.userID = 'TestUser';
};

exports["UXLog"] = {
    setUp: function (done) {
        watchdog = new testUtils.Watchdog();
        done();
    },
    tearDown: function (done) {
        if (httpServer) {
            httpServer.close();
            httpServer = null;
        }

        watchdog.shrink();
        done();
    },

    'Convert date to Windows file time': function (t) {
        t.strictEqual(UXLog.dateStrToWindowsFileTime('1/25/2014 13:25:17 GMT+0400'), 130351155170000000);
        t.strictEqual(UXLog.dateStrToWindowsFileTime('8/23/2015 17:10:01 GMT+0400'), 130848090010000000);
        t.done();
    },

    'Convert version to VersionIdMinor format': function (t) {
        t.strictEqual(UXLog.toVersionIdMinor('14.1.0'), '14100');
        t.strictEqual(UXLog.toVersionIdMinor('14.1.5'), '14105');
        t.strictEqual(UXLog.toVersionIdMinor('14.1.10'), '14110');
        t.strictEqual(UXLog.toVersionIdMinor('13.2.15'), '13215');
        t.done();
    },

    'Add entry': function (t) {
        var uxlog = new UXLog(new Config());

        uxlog.addEntry(2, 'Msg1');
        uxlog.addEntry(1, 'Msg2');

        t.strictEqual(uxlog.collectedData.length, 2);

        t.ok(testUtils.isNumber(uxlog.collectedData[0].date));
        t.strictEqual(uxlog.collectedData[0].actionKind, 2);
        t.strictEqual(uxlog.collectedData[0].msg, 'Msg1');

        t.ok(testUtils.isNumber(uxlog.collectedData[1].date));
        t.strictEqual(uxlog.collectedData[1].actionKind, 1);
        t.strictEqual(uxlog.collectedData[1].msg, 'Msg2');

        t.done();
    },

    'Add entry - disable via logging settings': function (t) {
        var uxlog = new UXLog(new Config());

        uxlog.loggingAllowed = false;
        uxlog.version = '123';
        uxlog.userID = 'TestUser';

        uxlog.addEntry(2, 'Msg1');
        t.strictEqual(uxlog.collectedData.length, 0);

        uxlog.loggingAllowed = true;
        uxlog.version = '';
        uxlog.userID = 'TestUser';

        uxlog.addEntry(2, 'Msg1');
        t.strictEqual(uxlog.collectedData.length, 0);

        uxlog.loggingAllowed = true;
        uxlog.version = '123';
        uxlog.userID = '';

        t.done();
    },

    'Get binary tracking data': function (t) {
        var uxlog = new UXLog(new Config()),
            expectedBytesStr = JSON.stringify([
                128, 60, 248, 91, 175, 25, 207, 1, 1, 11, 133, 17, 74, 13, 11, 0, 104, 101, 108, 108, 111, 32, 119,
                111, 114, 108, 100, 128, 218, 206, 139, 211, 190, 207, 1, 3, 11, 27, 171, 32, 50, 3, 0, 89, 111, 33
            ]);

        uxlog.collectedData.push({
            date: UXLog.dateStrToWindowsFileTime('1/25/2014 13:25:17 GMT+0400'),
            msg: 'hello world',
            actionKind: 1
        });

        uxlog.collectedData.push({
            date: UXLog.dateStrToWindowsFileTime('8/23/2014 17:10:01 GMT+0400'),
            msg: 'Yo!',
            actionKind: 3
        });

        t.strictEqual(JSON.stringify(uxlog._getBinaryTrackingData()), expectedBytesStr);
        t.done();
    },

    'Send collected data - success': function (t) {
        UXLog.UXLOG_HOSTNAME = '127.0.0.1';
        UXLog.UXLOG_PORT = 1700;

        var uxlog = new UXLog(new Config()),
            receivedData = '',
            expectedBytes = [
                128, 60, 248, 91, 175, 25, 207, 1, 1, 11, 133, 17, 74, 13, 11, 0,
                104, 101, 108, 108, 111, 32, 119, 111, 114, 108, 100
            ];

        uxlog.collectedData.push({
            date: UXLog.dateStrToWindowsFileTime('1/25/2014 13:25:17 GMT+0400'),
            msg: 'hello world',
            actionKind: 1
        });

        httpServer = http.createServer(function (req, res) {
            req.setEncoding('binary');

            req.on('data', function (chunk) {
                receivedData += chunk;
            });

            req.on('end', function () {
                var actualBytesStr = JSON.stringify(new Buffer(receivedData, 'binary'));

                res.writeHead(200, {'Content-Type': 'text/plain'});
                res.end('\n');

                t.strictEqual(req.headers['content-type'], 'app/x-www-form-urlencoded');
                t.strictEqual(req.headers['content-length'], expectedBytes.length.toString());
                t.strictEqual(actualBytesStr, JSON.stringify(expectedBytes));
                t.ok(/\?id=TestUser&version=123/.test(req.url));
                t.ok(uxlog.loggingAllowed);

                t.done();
            });
        }).listen(UXLog.UXLOG_PORT, UXLog.UXLOG_HOSTNAME);

        uxlog.sendCollectedData();
    },

    'Send collected data - ECONNREFUSED': function (t) {
        UXLog.UXLOG_HOSTNAME = '127.0.0.1';
        UXLog.UXLOG_PORT = 1701;

        var uxlog = new UXLog(new Config());

        uxlog.collectedData.push({
            date: UXLog.dateStrToWindowsFileTime('1/25/2014 1:25:17 PM'),
            msg: 'hello world',
            actionKind: 1
        });

        uxlog.sendCollectedData();

        uxlog.shutDown = function () {
            //Wait for ECONNREFUSED
            t.done();
        };
    },

    'Send collected data - connection hang up': function (t) {
        UXLog.UXLOG_HOSTNAME = '127.0.0.1';
        UXLog.UXLOG_PORT = 1702;
        UXLog.SEND_DATA_REQUEST_TIMEOUT = 300;

        httpServer = http.createServer(function (req, res) {
            req.on('data', function () {
            });
            req.on('end', function () {
                res.writeHead(200, {'Content-Type': 'text/plain'});
            });
            //NOTE: we don't end response for the emulating the behaviour of blocking by firewall
        }).listen(UXLog.UXLOG_PORT, UXLog.UXLOG_HOSTNAME);

        var uxlog = new UXLog(new Config());

        uxlog.collectedData.push({
            date: UXLog.dateStrToWindowsFileTime('1/25/2014 1:25:17 PM'),
            msg: 'hello world',
            actionKind: 1
        });

        uxlog.sendCollectedData();

        uxlog.shutDown = function () {
            //Wait to ECONNREFUSED
            t.done();
        };
    }
};