var fs = require('fs'),
    path = require('path'),
    tmp = require('tmp'),
    uuid = require('node-uuid'),
    async = require('../../../../../vendor/async'),
    testUtils = require('../../../test_utils'),
    VirtualFS = require('../../../../../lib/vfs'),
    FixtureCode = require('../../../../../lib/fixture_code'),
    ERR = require('../../../../../lib/server/server_errs');

var watchdog = null,
    vfs = null,
    fileDesc1 = null,
    fileDesc2 = null,
    rmdirRecursive = testUtils.rmdirRecursive,
    basePath = null;

var TRIM_REG_EXP = /\n|\r|\s|'|"/g;

function getFixtureCount(fixtures) {
    return Object.keys(fixtures).length;
}

function getFixture(fixturesArr, fixtureName) {
    for (var i = 0; fixturesArr.length; i++) {
        if (fixturesArr[i].name === fixtureName)
            return fixturesArr[i];
    }

    return null;
}

function unwatchTmpDirectory(vfs) {
    Object.keys(vfs.fsWatchers).forEach(function (fsWatcher) {
        vfs.fsWatchers[fsWatcher].close();
    });
}

function hasFixture(fixtures, fixtureName) {
    var result = false;

    Object.keys(fixtures).forEach(function (fixtureUid) {
        if (fixtures[fixtureUid].name === fixtureName)
            result = true;
    });

    return result;
}

function createSuiteFiles(callback) {
    tmp.dir(function (tmpErr, basePath) {
        fs.mkdirSync(path.join(basePath, './1'));
        fs.mkdirSync(path.join(basePath, './2'));

        fs.writeFileSync(path.join(basePath, 'top.test.js'), fs.readFileSync('test/nodeunit/core/data/vfs/get_suite/top.test.js'));
        fs.writeFileSync(path.join(basePath, './1/1.test.js'), fs.readFileSync('test/nodeunit/core/data/vfs/get_suite/1.test.js'));

        //Clean up
        process.on('exit', function () {
            rmdirRecursive(basePath);
        });

        callback(basePath);
    });
}

function fixtureExists(basePath, fixtureName, expectedJs) {
    var fileList = fs.readdirSync(basePath),
        result = false;

    Object.keys(fileList).forEach(function (i) {
        var fileElm = fileList[i];

        if (fileElm.indexOf(fixtureName) === 0) {
            fileElm = path.join(basePath, fileList[i]);
            var actualJs = fs.readFileSync(fileElm).toString();

            if (testUtils.compareCode(expectedJs, actualJs))
                result = true;
        }
    });

    return result;
}

function copyFile(src, dest) {
    fs.writeFileSync(dest, fs.readFileSync(src).toString());
}

function copyDirRecursiveSync(from, to) {
    fs.readdirSync(from).forEach(function (fileName) {
        var filePath = path.join(from, fileName),
            stats = fs.statSync(filePath);

        if (stats.isDirectory()) {
            fs.mkdirSync(path.join(to, fileName));
            copyDirRecursiveSync(filePath, path.join(to, fileName));
        }
        else {
            fs.writeFileSync(path.join(to, fileName), fs.readFileSync(filePath));
        }
    });
}

exports['Navigation'] = {
    setUp: function (done) {
        watchdog = new testUtils.Watchdog();
        done();
    },

    tearDown: function (done) {
        vfs.removeFSWatchers();
        watchdog.shrink();

        if (basePath) {
            rmdirRecursive(basePath);
            basePath = null;
        }

        done();
    },

    'VirtualFS.getSuite': function (t) {
        var savedTimeout = VirtualFS.REBUILD_TIMEOUT;

        createSuiteFiles(function (basePath) {
            var requireRebuild = false,
                prevSourceIndex = void 0;

            vfs = new VirtualFS(basePath);
            VirtualFS.REBUILD_TIMEOUT = 0;

            vfs.on('suiteChanged', function () {
                requireRebuild = true;
            });

            async.series({
                'Get first revision of the suite': function (next) {
                    vfs.getSuite(function (suiteR1) {
                        t.strictEqual(getFixtureCount(suiteR1.fixtures), 2);
                        t.ok(hasFixture(suiteR1.fixtures, 'top'));
                        t.ok(hasFixture(suiteR1.fixtures, '1'));

                        t.ok(suiteR1.sourceIndex.length);
                        t.ok(suiteR1.sourceIndex != prevSourceIndex);
                        prevSourceIndex = suiteR1.sourceIndex;

                        next();
                    });
                },

                'Create new fixture and give OS some time to raise fsWatcher event': function (next) {
                    fs.writeFileSync(path.join(basePath, './2/2.test.js'), fs.readFileSync('test/nodeunit/core/data/vfs/get_suite/2.test.js'));

                    setTimeout(function () {
                        next();
                    }, 200);
                },

                'Get second revision of the suite': function (next) {
                    vfs.getSuite(function (suiteR2) {
                        t.ok(requireRebuild);
                        t.strictEqual(getFixtureCount(suiteR2.fixtures), 3);
                        t.ok(hasFixture(suiteR2.fixtures, 'top'));
                        t.ok(hasFixture(suiteR2.fixtures, '1'));
                        t.ok(hasFixture(suiteR2.fixtures, '2'));

                        t.ok(suiteR2.sourceIndex.length);
                        t.ok(suiteR2.sourceIndex != prevSourceIndex);
                        prevSourceIndex = suiteR2.sourceIndex;

                        requireRebuild = false;

                        next();
                    });
                },

                'Modify fixture and give OS some time to raise fsWatcher event': function (next) {
                    fs.writeFileSync(path.join(basePath, './top.test.js'), fs.readFileSync('test/nodeunit/core/data/vfs/get_suite/top_modified.test.js'));

                    setTimeout(function () {
                        next();
                    }, 50);
                },

                'Get third revision of the suite': function (next) {
                    vfs.getSuite(function (suiteR3) {
                        t.ok(requireRebuild);
                        t.strictEqual(getFixtureCount(suiteR3.fixtures), 3);
                        t.ok(hasFixture(suiteR3.fixtures, 'top_modified'));
                        t.ok(hasFixture(suiteR3.fixtures, '1'));
                        t.ok(hasFixture(suiteR3.fixtures, '2'));

                        t.ok(suiteR3.sourceIndex.length);
                        t.ok(suiteR3.sourceIndex != prevSourceIndex);
                        prevSourceIndex = suiteR3.sourceIndex;

                        requireRebuild = false;

                        next();
                    });
                },

                'Delete fixture and give OS some time to raise fsWatcher event': function (next) {
                    fs.unlink(path.join(basePath, './2/2.test.js'));

                    setTimeout(function () {
                        next();
                    }, 50);
                },

                'Get fourth revision of the suite': function (next) {
                    vfs.getSuite(function (suiteR4) {
                        t.ok(requireRebuild);
                        t.strictEqual(getFixtureCount(suiteR4.fixtures), 2);
                        t.ok(hasFixture(suiteR4.fixtures, 'top_modified'));
                        t.ok(hasFixture(suiteR4.fixtures, '1'));

                        t.ok(suiteR4.sourceIndex.length);
                        t.ok(suiteR4.sourceIndex != prevSourceIndex);

                        VirtualFS.REBUILD_TIMEOUT = savedTimeout;
                        t.done();
                        next();
                    });
                }
            });
        });


    },

    'VirtualFS.getPathInfo': function (t) {
        vfs = new VirtualFS('test/nodeunit/core/data/vfs/navigation');

        async.series([
            function (next) {
                vfs.getPathInfo([], function (err, pathInfo) {
                    t.ok(!pathInfo.errs);

                    t.strictEqual(pathInfo.dirs.length, 2);
                    t.strictEqual(pathInfo.dirs[0].name, '1');
                    t.strictEqual(pathInfo.dirs[0].path, '1');
                    //NOTE: if directory contains test_config.json, it will includes in pathInfo.dirs
                    t.strictEqual(pathInfo.dirs[1].name, '2');
                    t.strictEqual(pathInfo.dirs[1].path, '2');

                    t.strictEqual(pathInfo.fixtures.length, 2);

                    var f1 = getFixture(pathInfo.fixtures, 'f1');
                    t.ok(f1.uid);
                    t.strictEqual(f1.page, './index');
                    t.strictEqual(f1.fileName, 'f1.test.js');
                    t.strictEqual(f1.tests.length, 1);
                    t.ok(f1.tests[0].uid);
                    t.strictEqual(f1.tests[0].name, 'f1_test');

                    var f2 = getFixture(pathInfo.fixtures, 'f2');
                    t.ok(f2.uid);
                    t.strictEqual(f2.page, './index');
                    t.strictEqual(f2.fileName, 'f2.test.js');
                    t.strictEqual(f2.tests.length, 1);
                    t.ok(f2.tests[0].uid);
                    t.strictEqual(f2.tests[0].name, 'f2_test');

                    next();
                });
            },
            function (next) {
                vfs.getPathInfo(['1'], function (err, pathInfo) {
                    t.ok(!pathInfo.errs);

                    t.strictEqual(pathInfo.dirs.length, 2);
                    t.strictEqual(pathInfo.dirs[0].name, '1');
                    t.strictEqual(pathInfo.dirs[0].path, '1/1');

                    t.strictEqual(pathInfo.dirs[1].name, '2');
                    t.strictEqual(pathInfo.dirs[1].path, '1/2');

                    t.strictEqual(pathInfo.fixtures.length, 2);

                    var f1 = getFixture(pathInfo.fixtures, '1/f1');
                    t.ok(f1.uid);
                    t.strictEqual(f1.page, './index');
                    t.strictEqual(f1.fileName, 'f1.test.js');
                    t.strictEqual(f1.tests.length, 1);
                    t.ok(f1.tests[0].uid);
                    t.strictEqual(f1.tests[0].name, '1/f1_test');

                    var f2 = getFixture(pathInfo.fixtures, '1/f2');
                    t.ok(f2.uid);
                    t.strictEqual(f2.page, './index');
                    t.strictEqual(f2.fileName, 'f2.test.js');
                    t.strictEqual(f2.tests.length, 1);
                    t.ok(f2.tests[0].uid);
                    t.strictEqual(f2.tests[0].name, '1/f2_test');

                    next();
                });
            },
            function (next) {
                vfs.getPathInfo([uuid.v4()], function (errs, pathInfo) {
                    t.ok(!pathInfo);

                    t.done();
                    next();
                });

            }
        ]);
    },

    'VirtualFS.getTestUidsByPath': function (t) {
        vfs = new VirtualFS('test/nodeunit/core/data/vfs/navigation');

        vfs.getSuite(function (suite) {
            var testUids = VirtualFS.getTestUidsByPath(suite, ['1']),
                testNames = [];

            testUids.forEach(function (testUid) {
                testNames.push(suite.tests[testUid].name);
            });

            t.strictEqual(testNames.length, 4);
            t.ok(testNames.indexOf('1/1/f1_test') > -1);
            t.ok(testNames.indexOf('1/2/f1_test') > -1);
            t.ok(testNames.indexOf('1/f1_test') > -1);
            t.ok(testNames.indexOf('1/f2_test') > -1);

            t.done();
        });
    },

    'VirtualFS.getSuite Regression - T173576': function(t) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;
            copyDirRecursiveSync('test/nodeunit/core/data/vfs/get_suite', basePath);

            vfs = new VirtualFS(basePath);

            vfs.getSuite(function () {
                vfs.on('projectRemoved', function () {
                    t.done();
                });

                rmdirRecursive(basePath);
                basePath = null;
            });
        });
    },

    'VirtualFS.getSuite Regression - T239085': function(t) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;
            copyDirRecursiveSync('test/nodeunit/core/data/vfs/T239085', basePath);

            vfs = new VirtualFS(basePath);

            vfs.getSuite(function () {
                vfs.on('projectRemoved', function () {
                    t.ok(false, 'the project should not be removed');
                });

                setTimeout(function() {
                    t.done();
                }, 200);

                rmdirRecursive(path.join(basePath, 'dir'));
                basePath = null;
            });
        });
    },

    'VirtualFS.renameDuplicates': function (t) {
        var lv0 = ['fixtureName', 'fixtureName (2)', 'fixtureName (3)'],
            lv1 = ['fixtureName', 'fixtureName (2)', 'fixtureName (3)'];

        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;
            copyDirRecursiveSync('test/nodeunit/core/data/vfs/duplicates', basePath);

            vfs = new VirtualFS(basePath);

            vfs.getSuite(function (suite) {
                Object.keys(suite.fixtures).forEach(function (fixtureUid) {
                    var fixture = suite.fixtures[fixtureUid],
                        name = fixture.name,
                        fixturePath = fixture.path.join('/');

                    if (!fixturePath && lv0.indexOf(name) > -1)
                        lv0.splice(lv0.indexOf(name), 1);
                    else if (lv1.indexOf(name) > -1)
                        lv1.splice(lv1.indexOf(name), 1);
                });

                t.ok(!lv0.length && !lv1.length);
                t.done();
            });
        });
    },

    'VirtualFS.suiteStartBuild Event': function(t) {
        tmp.dir(function (err, tmpDir) {
            var eventCount = 0;

            basePath = tmpDir;
            copyDirRecursiveSync('test/nodeunit/core/data/vfs/get_suite', basePath);

            vfs = new VirtualFS(basePath);

            vfs.on('suiteStartBuild', function () {
                eventCount++;
            });

            vfs.getSuite(function () {
                vfs.getSuite(function () {
                    setTimeout(function () {
                        vfs.removeAllListeners('suiteStartBuild');

                        t.strictEqual(eventCount, 1);
                        t.done();
                    }, 100);
                });
            });
        });
    }
};

exports['Base path validation'] = {
    'Incorrect path separator': function (t) {
        t.expect(1);

        tmp.dir(function (err, tmpDir) {
            var basePath = tmpDir.replace(VirtualFS.WIN_PLATFORM ? /\\/g : /\//g, VirtualFS.WIN_PLATFORM ? '/' : '\\');
            vfs = new VirtualFS(basePath);

            t.ok(true);
            t.done();
        });
    },

    'Path is not a dir': function (t) {
        t.expect(2);

        var basePath = path.join(VirtualFS.WIN_PLATFORM ? './' : '/', __filename);

        try {
            new VirtualFS(basePath);
        } catch (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_BASE_PATH_IS_NOT_DIR);
            t.strictEqual(err.basePath, basePath);
        }

        t.done();
    },

    'Path doesnt exists': function (t) {
        t.expect(2);

        var basePath = path.join(__dirname, uuid.v4());

        try {
            new VirtualFS(basePath);
        } catch (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_BASE_PATH_DOESNT_EXISTS);
            t.strictEqual(err.basePath, basePath);
        }

        t.done();
    }
};

exports['Create Dir'] = {
    setUp: function (done) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;

            done();
        });
    },

    tearDown: function (done) {
        rmdirRecursive(basePath);

        done();
    },

    'Create dir': function (t) {
        t.expect(3);

        var vfs = new VirtualFS(basePath),
            dirPath = [],
            dirName = 'newDir';

        vfs.createDir(dirPath, dirName, function (err) {
            var realPath = fs.realpathSync(path.join(basePath, dirName)),
                fStats = fs.statSync(realPath),
                testConfigStats = fs.statSync(path.join(realPath, 'test_config.json'));

            t.ok(!err);
            t.ok(fStats.isDirectory());
            t.ok(testConfigStats.isFile());

            unwatchTmpDirectory(vfs);

            t.done();
        });
    },

    'Dir already exists': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            dirPath = [],
            dirName = 'test',
            fixtureFilename = 'testing.test.js',
            fixtureContent = '\'@fixture test\'\r\n\'@page test\'';

        fs.mkdirSync(path.join(basePath, dirName));

        var tmpFileDesc = fs.openSync(path.join(basePath, dirName, fixtureFilename), "w");

        fs.writeSync(tmpFileDesc, fixtureContent, null, null);
        fs.close(tmpFileDesc);

        vfs.createDir(dirPath, dirName, function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_DIR_ALREADY_EXISTS);

            unwatchTmpDirectory(vfs);

            t.done();


        });
    },

    'Path is not valid': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            firstDirPath = ['test', '..'],
            secondDirPath = ['test', '.'],
            dirName = 'test';


        vfs.createDir(firstDirPath, dirName, function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_DIR_PATH);

            vfs.createDir(secondDirPath, dirName, function (err) {
                t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_DIR_PATH);

                unwatchTmpDirectory(vfs);

                t.done();
            });
        });
    }
};

exports['Rename Dir'] = {
    setUp: function (done) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;

            done();
        });
    },

    tearDown: function (done) {
        if (fileDesc1)
            fs.closeSync(fileDesc1);

        rmdirRecursive(basePath);

        done();
    },

    'Rename dir': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            dirPath = ['oldName'],
            newName = 'newName',
            fixtureFile = 'fixtureFile.test.js',
            fixtureContent =
                ['\'', '@fixture ', 'test', '\'\r\n',
                    '\'@page ', 'test', '\''].join('');

        var oldPath = path.join(basePath, dirPath[0]),
            newPath = path.join(basePath, newName);

        fs.mkdirSync(oldPath);
        var fileDesc = fs.openSync(path.join(oldPath, fixtureFile), "w");
        fs.writeSync(fileDesc, fixtureContent, null, null);
        fs.closeSync(fileDesc);

        vfs.renameDir(dirPath, newName, function (err) {
            t.ok(!err);

            fs.stat(newPath, function (err) {
                t.ok(!err);

                unwatchTmpDirectory(vfs);

                t.done();
            });
        });
    },

    'Dir not exists': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            dirPath = ['test_rename'],
            newName = 'new_name';

        vfs.renameDir(dirPath, newName, function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_DIR_DOESNT_EXIST);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    },

    'Dir with new Name already exists': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            dirPath = ['test'],
            newName = 'name_exists',
            fixtureFile = 'test.test.js',
            fixtureContent =
                ['\'', '@fixture ', 'test', '\'\r\n',
                    '\'@page ', 'test', '\''].join('');

        fs.mkdirSync(path.join(basePath, dirPath[0]));
        var fileDesc = fs.openSync(path.join(basePath, dirPath[0], fixtureFile), "w");
        fs.writeSync(fileDesc, fixtureContent, null, null);
        fs.closeSync(fileDesc);
        fs.mkdirSync(path.join(basePath, newName));
        fileDesc = fs.openSync(path.join(basePath, newName, fixtureFile), "w");
        fs.writeSync(fileDesc, fixtureContent, null, null);
        fs.closeSync(fileDesc);

        vfs.renameDir(dirPath, newName, function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_DIR_ALREADY_EXISTS);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    },

    'Path is not valid': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            firstDirPath = ['test', '..'],
            secondDirPath = ['test', '.'],
            newName = "new_name";

        vfs.renameDir(firstDirPath, newName, function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_DIR_PATH);

            vfs.renameDir(secondDirPath, newName, function (err) {
                t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_DIR_PATH);

                unwatchTmpDirectory(vfs);

                t.done();
            });
        });
    },

    'New Name is not valid': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            dirPath = ['test'],
            fixtureFile = 'test.test.js',
            fixtureContent = '\'@fixture test\'\r\n\'@page test\'',
            newName = "../.!@#%*$()_{}";

        fs.mkdirSync(path.join(basePath, dirPath[0]));
        var fileDesc = fs.openSync(path.join(basePath, dirPath[0], fixtureFile), "w");
        fs.writeSync(fileDesc, fixtureContent, null, null);
        fs.closeSync(fileDesc);

        vfs.renameDir(dirPath, newName, function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_DIR_PATH);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    },

    'Parent directory doesn\'t contains fixture files': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            dirPath = ['parentDir'],
            nestedDir = 'nestedDir',
            newName = 'newName',
            fixtureFile = 'fixtureFile.test.js',
            fixtureContent =
                ['\'', '@fixture ', 'test', '\'\r\n',
                    '\'@page ', 'test', '\''].join('');

        var oldPath = path.join(basePath, dirPath[0]),
            newPath = path.join(basePath, newName);

        fs.mkdirSync(oldPath);
        fs.mkdirSync(path.join(oldPath, nestedDir));
        var fileDesc = fs.openSync(path.join(oldPath, nestedDir, fixtureFile), "w");
        fs.writeSync(fileDesc, fixtureContent, null, null);
        fs.closeSync(fileDesc);

        vfs.renameDir(dirPath, newName, function (err) {
            t.ok(!err);

            fs.stat(newPath, function (err) {
                t.ok(!err);

                unwatchTmpDirectory(vfs);

                t.done();
            });
        });
    },

    'Nested file is opened.': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            dirPath = ['test'],
            newName = "newName",
            fileName = "fixtureFile.test.js",
            fixtureContent = '\'@fixture test\'\r\n\'@page test\'',
            oldDirPath = path.join(basePath, dirPath[0]);

        fs.mkdirSync(oldDirPath);
        fileDesc1 = fs.openSync(path.join(oldDirPath, fileName), "w");
        fs.write(fileDesc1, fixtureContent, null, null);
        fs.close(fileDesc1);
        fileDesc1 = fs.openSync(path.join(oldDirPath, fileName), "r");

        vfs.renameDir(dirPath, newName, function (err) {
            if (VirtualFS.WIN_PLATFORM)
                t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_RENAME_DIR_FAILED);
            else
                t.ok(!err);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    }
};

exports['Delete Dir'] = {

    setUp: function (done) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;

            done();
        });
    },

    tearDown: function (done) {
        if (fileDesc2)
            fs.closeSync(fileDesc2);

        rmdirRecursive(basePath);

        done();
    },

    'Delete empty dir': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            dirPath = ['emptyDir'];

        fs.mkdirSync(path.join(basePath, dirPath[0]));

        var realPath = fs.realpathSync(path.join(basePath, dirPath[0]));

        vfs.deleteDir(dirPath, function (deleteErr) {
            t.ok(!deleteErr);

            fs.stat(realPath, function (statErr) {
                t.ok(statErr);

                unwatchTmpDirectory(vfs);

                t.done();
            });
        });
    },

    'Delete dir with nested directories and files': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            dirPath = ['dirWithFiles'],
            testFile1 = 'test1.txt',
            testFile2 = 'test2.txt',
            nestedDirPath1 = ['dirWithFiles', 'nestedDir1'],
            nestedDirPath2 = ['dirWithFiles', 'nestedDir2'];

        //make dir and two files
        fs.mkdirSync(path.join(basePath, dirPath[0]));
        var firstFileDesc = fs.openSync(path.join(basePath, dirPath[0], testFile1), 'w');
        fs.closeSync(firstFileDesc);
        var secondFileDesc = fs.openSync(path.join(basePath, dirPath[0], testFile2), 'w');
        fs.closeSync(secondFileDesc);
        //make two nested dir and file in ones.
        fs.mkdirSync(path.join(basePath, nestedDirPath1.join('/')));
        fs.mkdirSync(path.join(basePath, nestedDirPath2.join('/')));
        var threeFileDesc = fs.openSync(path.join(basePath, nestedDirPath2.join('/'), testFile1), 'w');
        fs.closeSync(threeFileDesc);

        var realPath = path.join(basePath, dirPath[0]);

        vfs.deleteDir(dirPath, function (deleteErr) {
            t.ok(!deleteErr);

            fs.stat(realPath, function (statErr) {

                t.ok(statErr);

                unwatchTmpDirectory(vfs);

                t.done();
            });
        });
    },


    'Dir is not exists': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            dirPath = ['dirNotExists'];

        vfs.deleteDir(dirPath, function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_DELETE_DIR_FAILED);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    },

    'Path is not valid': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            firstDirPath = ['test', '..'],
            secondDirPath = ['test', '.'];


        vfs.deleteDir(firstDirPath, function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_DIR_PATH);

            vfs.deleteDir(secondDirPath, function (err) {
                t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_DIR_PATH);

                unwatchTmpDirectory(vfs);

                t.done();
            });
        });
    },

    'Nested file is opened.': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            dirName = ['test'],
            fileName = 'fixtureFile',
            dirPath = path.join(basePath, dirName[0]);

        fs.mkdirSync(dirPath);
        fileDesc2 = fs.openSync(path.join(dirPath, fileName), "w");

        vfs.deleteDir(dirName, function (err) {
            if (VirtualFS.WIN_PLATFORM)
                t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_DELETE_DIR_FAILED);
            else
                t.ok(!err);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    }
};

exports['Edit File'] = {
    setUp: function (done) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;

            done();
        });
    },

    tearDown: function (done) {
        rmdirRecursive(basePath);

        done();
    },

    'Rename file': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            dirPath = [],
            oldFilename = 'oldName.test.js',
            filename = 'newName',
            newName = 'newName';

        var expectedJs = '"@fixture newName";\r\n"@page test";';
        var tmpFileDesc = fs.openSync(path.join(basePath, oldFilename), "w");
        fs.writeSync(tmpFileDesc, expectedJs, null, null);
        fs.closeSync(tmpFileDesc);

        vfs.editFixture(dirPath, oldFilename, filename, newName, '', '', '', function (err) {
            t.ok(fixtureExists(basePath, filename, expectedJs));
            t.ok(!err);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    },

    'File not exists': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            dirPath = [],
            oldFilename = 'test_rename.test.js',
            filename = 'new_name',
            newName = 'new_name.test.js';

        vfs.editFixture(dirPath, oldFilename, filename, newName, '', '', '', function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_FIXTURE_FILE_DOESNT_EXIST);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    },

    'Path is not valid': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            firstDirPath = ['..'],
            secondDirPath = ['.'],
            oldFilename = 'test.test.js',
            filename = 'new_name',
            newName = 'new_name.txt';

        vfs.editFixture(firstDirPath, oldFilename, filename, newName, '', '', '', function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_FIXTURE_PATH);

            vfs.editFixture(secondDirPath, oldFilename, filename, newName, '', '', '', function (err) {
                t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_FIXTURE_PATH);

                unwatchTmpDirectory(vfs);

                t.done();
            });
        });
    },

    'New Name is not valid': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            dirPath = [],
            oldFilename = 'test.test.js',
            filename = 'new_name./../',
            fixtureContent = '\'@fixture test \'\r\n\'@page test\'',
            newName = "newName";

        var tmpfileDesc = fs.openSync(path.join(basePath, oldFilename), "w");
        fs.writeSync(tmpfileDesc, fixtureContent, null, null);
        fs.closeSync(tmpfileDesc);

        vfs.editFixture(dirPath, oldFilename, filename, newName, '', '', '', function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_FIXTURE_PATH);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    },

    'New Name exists': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            dirPath = [],
            oldFilename = 'test.test.js',
            newFileName = 'newName.test.js',
            fixtureContent = '\'@fixture test \'\r\n\'@page test\'',
            newName = "newName";

        var tmpfileDesc = fs.openSync(path.join(basePath, oldFilename), "w");
        fs.writeSync(tmpfileDesc, fixtureContent, null, null);
        fs.closeSync(tmpfileDesc);

        tmpfileDesc = fs.openSync(path.join(basePath, newFileName), "w");
        fs.writeSync(tmpfileDesc, fixtureContent, null, null);
        fs.closeSync(tmpfileDesc);

        vfs.editFixture(dirPath, oldFilename, newFileName, newName, '', '', '', function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_FIXTURE_FILE_ALREADY_EXISTS);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    },

    'File name is the same': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            dirPath = [],
            filename = 'test.test.js',
            fixtureContent = '\'@fixture test \'\r\n\'@page test\'',
            newName = "newName";

        var tmpfileDesc = fs.openSync(path.join(basePath, filename), "w");
        fs.writeSync(tmpfileDesc, fixtureContent, null, null);
        fs.closeSync(tmpfileDesc);

        vfs.editFixture(dirPath, filename, filename, newName, '', '', '', function (err) {
            t.ok(!err);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    }
};

exports['Delete File'] = {
    setUp: function (done) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;

            done();
        });
    },

    tearDown: function (done) {
        rmdirRecursive(basePath);

        done();
    },

    'Delete file': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            filePath = [],
            fixtureFilename = 'test.test.js',
            fixtureContent = '\'@fixture test\'\r\n\'@page test\'';

        var tmpfileDesc = fs.openSync(path.join(basePath, fixtureFilename), "w");
        fs.writeSync(tmpfileDesc, fixtureContent, null, null);
        fs.closeSync(tmpfileDesc);

        var realPath = fs.realpathSync(path.join(basePath, fixtureFilename));

        vfs.deleteFixture(filePath, fixtureFilename, function (deleteErr) {
            t.ok(!deleteErr);

            fStats = fs.stat(realPath, function (statErr) {
                t.ok(statErr);

                unwatchTmpDirectory(vfs);

                t.done();
            });
        });
    },

    'File is not exists': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            filePath = [],
            fixtureFilename = 'fileNotExists.txt';

        vfs.deleteFixture(filePath, fixtureFilename, function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_FIXTURE_FILE_DOESNT_EXIST);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    },

    'Path is not valid': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            firstFilePath = ['..'],
            secondFilePath = ['.'],
            fixtureFilename = 'test.test.js';

        vfs.deleteFixture(firstFilePath, fixtureFilename, function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_FIXTURE_PATH);

            vfs.deleteFixture(secondFilePath, fixtureFilename, function (err) {
                t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_FIXTURE_PATH);

                unwatchTmpDirectory(vfs);

                t.done();
            });
        });
    }
};

exports["Create fixture"] = {
    setUp: function (done) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;

            done();
        });
    },

    tearDown: function (done) {
        rmdirRecursive(basePath);

        done();
    },

    'Create fixture': function (t) {
        t.expect(8);

        var vfs = new VirtualFS(basePath),
            fixtureName1 = 'test1',
            fixtureName2 = 'test2',
            fixtureName3 = 'test3',
            fixtureName4 = 'test4',
            filename1 = 'testFilename1',
            filename2 = 'testFilename2',
            filename3 = 'testFilename3',
            filename4 = 'testFilename4',
            username1 = 'testUsername1',
            password1 = 'testPassword1',
            fixturePage1 = 'testPage1',
            fixturePage2 = '/testPage2',
            fixturePage3 = './testPage3',
            fixturePage4 = '../testPage4',
            controlValue1 = "'@fixture test1';'@page http://testPage1';'@auth testUsername1:testPassword1';",
            controlValue2 = "'@fixture test2';'@page /testPage2';",
            controlValue3 = "'@fixture test3';'@page ./testPage3';",
            controlValue4 = "'@fixture test4';'@page ../testPage4';";

        vfs.createFixture([], fixtureName1, filename1, fixturePage1, username1, password1, function (err) {
            t.ok(!err);
            t.ok(fixtureExists(basePath, filename1, controlValue1));

            vfs.createFixture([], fixtureName2, filename2, fixturePage2, '', '', function (err) {
                t.ok(!err);
                t.ok(fixtureExists(basePath, filename2, controlValue2));

                vfs.createFixture([], fixtureName3, filename3, fixturePage3, '', '', function (err) {
                    t.ok(!err);
                    t.ok(fixtureExists(basePath, filename3, controlValue3));

                    vfs.createFixture([], fixtureName4, filename4, fixturePage4, '', '', function (err) {
                        t.ok(!err);
                        t.ok(fixtureExists(basePath, filename4, controlValue4));

                        unwatchTmpDirectory(vfs);

                        t.done();
                    });
                });
            });
        });
    },

    'Path is not valid': function (t) {
        t.expect(1);

        var vfs = new VirtualFS(basePath),
            fixtureName = 'test',
            filename = 'test/../',
            fixturePage = 'testPage';

        vfs.createFixture([], fixtureName, filename, fixturePage, '', '', function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_INVALID_FIXTURE_PATH);

            unwatchTmpDirectory(vfs);

            t.done();
        });
    }
};

exports['Delete test'] = {
    setUp: function (done) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;

            done();
        });
    },

    tearDown: function (done) {
        rmdirRecursive(basePath);

        done();
    },

    'Delete test': function (t) {
        var vfs = new VirtualFS(basePath),
            testName = 'First test',
            dirPath = [],
            filename = 'src.test.js';

        copyFile('test/nodeunit/core/data/fixture_code_code_generator/delete_test/src.test.js', path.join(basePath, filename));
        var expectedJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_code_generator/delete_test/success.test.js', 'UTF8');
        expectedJs = expectedJs.replace(TRIM_REG_EXP, '');

        vfs.deleteTest(dirPath, filename, testName, function (err) {
            var res = fs.readFileSync(path.join(basePath, filename), 'UTF8');

            res = res.replace(TRIM_REG_EXP, '');

            t.ok(!err);
            t.strictEqual(res, expectedJs);

            t.done();
        });
    },

    'Path is not exists': function (t) {
        var vfs = new VirtualFS(basePath),
            testName = 'First test',
            dirPath = [],
            filename = 'src.test.js';

        vfs.deleteTest(dirPath, filename, testName, function (err) {
            t.strictEqual(err.code, FixtureCode.ErrCodes.READ_FILE_FAILED);
            t.strictEqual(err.filename, path.join(basePath, filename));

            t.done();
        });
    },

    'Test is not found': function (t) {
        var vfs = new VirtualFS(basePath),
            testName = 'I\'m not exists',
            dirPath = [],
            filename = 'src.test.js';

        copyFile('test/nodeunit/core/data/fixture_code_code_generator/delete_test/src.test.js', path.join(basePath, filename));

        vfs.deleteTest(dirPath, filename, testName, function (err) {
            t.strictEqual(err.code, FixtureCode.ErrCodes.TEST_IS_NOT_FOUND);
            t.strictEqual(err.testName, testName);

            t.done();
        });
    }
};

exports['Rename test'] = {
    setUp: function (done) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;

            done();
        });
    },

    tearDown: function (done) {
        rmdirRecursive(basePath);

        done();
    },

    'Test renamed': function (t) {
        var vfs = new VirtualFS(basePath),
            testName = 'Second test',
            newName = 'Test with new" name',
            dirPath = [],
            filename = 'src.test.js';

        copyFile('test/nodeunit/core/data/fixture_code_code_generator/rename_test/src.test.js', path.join(basePath, filename));
        var expectedJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_code_generator/rename_test/success.test.js', 'UTF8');
        expectedJs = expectedJs.replace(TRIM_REG_EXP, '');

        vfs.renameTest(dirPath, filename, testName, newName, function (err) {
            var res = fs.readFileSync(path.join(basePath, filename), 'UTF8');

            res = res.replace(TRIM_REG_EXP, '');

            t.ok(!err);
            t.strictEqual(res, expectedJs);

            t.done();
        });
    },

    'Path is not exists': function (t) {
        var vfs = new VirtualFS(basePath),
            testName = 'First test',
            newName = '',
            dirPath = [],
            filename = 'src.test.js';

        vfs.renameTest(dirPath, filename, testName, newName, function (err) {
            t.strictEqual(err.code, FixtureCode.ErrCodes.READ_FILE_FAILED);
            t.strictEqual(err.filename, path.join(basePath, filename));

            t.done();
        });
    },

    'Test is not found': function (t) {
        var vfs = new VirtualFS(basePath),
            testName = 'I\'m not exists',
            newName = '',
            dirPath = [],
            filename = 'src.test.js';

        copyFile('test/nodeunit/core/data/fixture_code_code_generator/rename_test/src.test.js', path.join(basePath, filename));

        vfs.renameTest(dirPath, filename, testName, newName, function (err) {
            t.strictEqual(err.code, FixtureCode.ErrCodes.TEST_IS_NOT_FOUND);
            t.strictEqual(err.testName, testName);

            t.done();
        });
    },

    'Test already exists': function (t) {
        var vfs = new VirtualFS(basePath),
            testName = 'First test',
            newName = 'Second test',
            dirPath = [],
            filename = 'src.test.js';

        copyFile('test/nodeunit/core/data/fixture_code_code_generator/rename_test/src.test.js', path.join(basePath, filename));
        var expectedJs = fs.readFileSync('test/nodeunit/core/data/fixture_code_code_generator/rename_test/src.test.js', 'UTF8');
        expectedJs = expectedJs.replace(TRIM_REG_EXP, '');

        vfs.renameTest(dirPath, filename, testName, newName, function (err) {
            var res = fs.readFileSync(path.join(basePath, filename), 'UTF8');

            res = res.replace(TRIM_REG_EXP, '');

            t.ok(err.code, ERR.TEST_NAME_CHANGED_TO_ALREADY_USED);
            t.strictEqual(res, expectedJs);

            t.done();
        });
    }
};

exports['Get dirs by path'] = {
    setUp: function (done) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;

            done();
        });
    },

    tearDown: function (done) {
        rmdirRecursive(basePath);

        done();
    },

    'Get dirs': function (t) {
        var vfs = new VirtualFS(basePath),
            fsPath = fs.realpathSync('test/nodeunit/core/data/vfs/get_dirs_by_path'),
            expectedJS = [
                {
                    text: 'test1',
                    id: encodeURIComponent(path.join('get_dirs_by_path', 'test1')),
                    type: "dir",
                    isProjectDir: true
                },
                {
                    text: 'test2',
                    id: encodeURIComponent(path.join('get_dirs_by_path', 'test2')),
                    type: "dir"
                }
            ];

        vfs.getFileStructureByPath(fsPath, false, function (err, dirs) {
            t.ok(!err);

            dirs[0].id = dirs[0].id.substr(dirs[0].id.indexOf('get_dirs_by_path'));
            dirs[1].id = dirs[1].id.substr(dirs[1].id.indexOf('get_dirs_by_path'));

            t.strictEqual(JSON.stringify(expectedJS), JSON.stringify(dirs));

            t.done();
        });
    },

    'Path is not exist': function (t) {
        var vfs = new VirtualFS(basePath),
            fsPath = path.join(fs.realpathSync('test/nodeunit/core/data/vfs/get_dirs_by_path'), 'path_non_exists');

        vfs.getFileStructureByPath(fsPath, false, function (err, dirs) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_DIR_DOESNT_READ);
            t.strictEqual(dirs.length, 0);

            t.done();
        });
    }
};

exports['Get local drives'] = {
    setUp: function (done) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;

            done();
        });
    },

    tearDown: function (done) {
        rmdirRecursive(basePath);

        done();
    },

    'Get local drives': function (t) {
        var vfs = new VirtualFS(basePath);

        vfs.getLocalDrives(function (err, list) {
            t.ok(!err);

            if (VirtualFS.WIN_PLATFORM) {
                var driveList = '';

                for (var i = 0, length = list.length; i < length; i++) {
                    driveList += list[i].text;
                }

                t.ok(/^([A-Z]{1}:\\)*$/.test(driveList));
            } else
                t.strictEqual(list[0].text, '/');

            t.done();
        });
    },

    'Error getting list of local drives': function (t) {

        var tmp = VirtualFS.GET_LOCAL_DRIVES_CMD;

        VirtualFS.GET_LOCAL_DRIVES_CMD = 'errorExecutingCommand';

        vfs.getLocalDrives(function (err, list) {
            if (VirtualFS.WIN_PLATFORM) {
                t.strictEqual(list.length, 0);
                t.strictEqual(err.code, exports.VIRTUAL_FILE_SYSTEM_NODEJS_CANT_GET_DRIVE_LIST);

                VirtualFS.GET_LOCAL_DRIVES_CMD = tmp;
            } else {
                t.ok(!err);
                t.strictEqual(list[0].text, '/');
            }

            t.done();
        });
    }
};

exports['Fixture edit'] = {
    setUp: function (done) {
        tmp.dir(function (err, tmpDir) {
            basePath = tmpDir;

            done();
        });
    },

    tearDown: function (done) {
        rmdirRecursive(basePath);

        done();
    },

    'Read fixture code': function (t) {
        t.expect(2);

        var filename = '1.test.js',
            sourcePath = path.join(__dirname, '../../data/vfs/get_suite/', filename),
            destPath = path.join(basePath, filename);

        fs.writeFileSync(destPath, fs.readFileSync(sourcePath));

        var vfs = new VirtualFS(basePath),
            expectedJS = fs.readFileSync(destPath, 'utf8');

        vfs.getFixtureCode(filename, function (err, data) {
            t.ok(!err);
            t.strictEqual(JSON.stringify(data), JSON.stringify(expectedJS));
            t.done();
        });
    },

    'Read fixture code - fixture not exists': function (t) {
        t.expect(2);

        var vfs = new VirtualFS(basePath),
            filePath = 'nonExistingPath';

        vfs.getFixtureCode(filePath, function (err, data) {
            t.ok(err.code, ERR.VIRTUAL_FILE_SYSTEM_FIXTURE_FILE_DOESNT_EXIST);
            t.ok(!data);
            t.done();
        });
    },

    'Save fixture code': function (t) {
        t.expect(2);

        var sourceDir = path.join(__dirname, '../../data/vfs/get_suite/'),
            filename1 = '1.test.js',
            filename2 = '2.test.js',
            sourcePath = path.join(sourceDir, filename1),
            destPath = path.join(basePath, filename1),
            newCode = fs.readFileSync(path.join(sourceDir, filename2));

        fs.writeFileSync(destPath, fs.readFileSync(sourcePath));

        var vfs = new VirtualFS(basePath);

        vfs.saveFixtureCode(filename1, newCode, function (err) {
            var actualJS = fs.readFileSync(destPath);

            t.ok(!err);
            t.strictEqual(JSON.stringify(actualJS), JSON.stringify(newCode));
            t.done();
        });
    },

    'Save fixture code - fixture not exists': function (t) {
        t.expect(1);

        var sourceDir = path.join(__dirname, '../../data/vfs/get_suite/'),
            filename = 'nonExistingPath';

        var vfs = new VirtualFS(basePath);

        vfs.saveFixtureCode(filename, 'any code value', function (err) {
            t.strictEqual(err.code, ERR.VIRTUAL_FILE_SYSTEM_FIXTURE_FILE_DOESNT_EXIST);
            t.done();
        });
    }
};