var fs = require('fs'),
    zlib = require('zlib'),
    brotli = require('brotli'),
    iconv = require('iconv-lite'),
    injector = require('../../../../hammerhead/lib/injector'),
    testUtils = require('../../test_utils'),
    ERR = require('../../../../hammerhead/lib/server_errs'),
    path = require('path'),
    pageProc = require('../../../../hammerhead/shared/page_processor');

var CHUNK_SIZE = 50;

function makeChunks(buf) {
    if (!Buffer.isBuffer(buf))
        buf = new Buffer(buf);

    var chunks = [],
        start = 0,
        end = Math.min(CHUNK_SIZE, buf.length);

    while (start < buf.length) {
        var chunk = new Buffer(end - start);
        buf.copy(chunk, 0, start, end);
        chunks.push(chunk);
        start = end;
        end = Math.min(end + CHUNK_SIZE, buf.length);
    }

    return chunks;
}

var srcHtml = null,
    watchdog = null,
    cc = testUtils.compareCode;

exports['Compatible meta'] = {
    setUp: function (done) {
        watchdog = new testUtils.Watchdog();

        done();
    },

    tearDown: function (done) {
        watchdog.shrink();
        done();
    },

    'Not exists': function (t) {
        var srcChunks = makeChunks(fs.readFileSync(path.join(__dirname, '../data/injector/compatible_meta/not_exists_src.html')).toString()),
            expectedHtml = fs.readFileSync(path.join(__dirname, '../data/injector/compatible_meta/not_exists_expected.html')).toString(),
            options = { scripts: [ '/testscript' ] };

        injector.injectInPage(srcChunks, '', 'utf8', options, function (err, htmlBuf) {
            t.ok(!err);
            t.ok(testUtils.compareCode(htmlBuf.toString(), expectedHtml));
            t.done();
        });
    },

    'Exists': function (t) {
        var srcChunks = makeChunks(fs.readFileSync(path.join(__dirname, '../data/injector/compatible_meta/exists_src.html')).toString()),
            expectedHtml = fs.readFileSync(path.join(__dirname, '../data/injector/compatible_meta/exists_expected.html')).toString(),
            options = { scripts: [ '/testscript' ] };

        injector.injectInPage(srcChunks, '', 'utf8', options, function (err, htmlBuf) {
            t.ok(!err);
            t.ok(testUtils.compareCode(htmlBuf.toString(), expectedHtml));
            t.done();
        });
    }
};

exports['Encodings'] = {
    setUp: function (done) {
        srcHtml = fs.readFileSync(path.join(__dirname, '../data/injector/encodings.html')).toString();
        watchdog = new testUtils.Watchdog();

        done();
    },

    tearDown: function (done) {
        watchdog.shrink();
        done();
    },

    'Plain text passed - plain text returned': function (t) {
        var srcChunks = makeChunks(srcHtml);

        injector.injectInPage(srcChunks, '', 'utf8', null, function (err, html) {
            t.ok(!err);
            t.ok(Buffer.isBuffer(html));
            t.ok(cc(html.toString(), srcHtml));
            t.done();
        });
    },

    'Gzip passed - gzip returned': function (t) {
        zlib.gzip(srcHtml, function (err, encodedSrcHtml) {
            var srcChunks = makeChunks(encodedSrcHtml);

            injector.injectInPage(srcChunks, 'gzip', 'utf8', null, function (err, encodedHtml) {
                t.ok(!err);
                t.ok(Buffer.isBuffer(encodedHtml));

                zlib.gunzip(encodedHtml, function (err, html) {
                    t.ok(cc(html.toString(), srcHtml));
                    t.done();
                });
            });
        });
    },

    'Deflate passed - deflate returned': function (t) {
        zlib.deflate(srcHtml, function (err, encodedSrcHtml) {
            var srcChunks = makeChunks(encodedSrcHtml);

            injector.injectInPage(srcChunks, 'deflate', 'utf8', null, function (err, encodedHtml) {
                t.ok(!err);
                t.ok(Buffer.isBuffer(encodedHtml));

                zlib.inflate(encodedHtml, function (err, html) {
                    t.ok(cc(html.toString(), srcHtml));
                    t.done();
                });
            });
        });
    },

    'br passed - br returned': function (t) {
        var data           = new Buffer(srcHtml);
        var encodedSrcHtml = brotli.compress(data);
        var srcChunks      = makeChunks(encodedSrcHtml);

        injector.injectInPage(srcChunks, 'br', 'utf8', null, function (err, encodedHtml) {
            t.ok(!err);
            t.ok(Buffer.isBuffer(encodedHtml));

            var html = new Buffer(brotli.decompress(encodedHtml));

            t.ok(cc(html.toString(), srcHtml));
            t.done();
        });
    },

    'Gzip decoding failed': function (t) {
        var srcChunks = makeChunks('Answer to the Ultimate Question of Life, the Universe, and Everything.');

        injector.injectInPage(srcChunks, 'gzip', 'utf8', null, function (err) {
            t.strictEqual(err.code, ERR.INJECTOR_RESOURCE_DECODING_FAILED);
            t.strictEqual(err.encoding, 'gzip');
            t.done();
        });
    },

    'Deflate decoding failed': function (t) {
        var srcChunks = makeChunks('Answer to the Ultimate Question of Life, the Universe, and Everything.');

        injector.injectInPage(srcChunks, 'deflate', 'utf8', null, function (err) {
            t.strictEqual(err.code, ERR.INJECTOR_RESOURCE_DECODING_FAILED);
            t.strictEqual(err.encoding, 'deflate');
            t.done();
        });
    },

    'Charset decoding': function (t) {
        var srcChunks = makeChunks(iconv.encode(srcHtml, 'win1251'));

        injector.injectInPage(srcChunks, '', 'win1251', null, function (err, html) {
            t.ok(!err);
            t.ok(Buffer.isBuffer(html));
            t.ok(cc(iconv.decode(html, 'win1251'), srcHtml));
            t.done();
        });
    },

    'Decoding failed': function (t) {
        var srcChunks = makeChunks(iconv.encode(srcHtml, 'win1251'));

        injector.injectInPage(srcChunks, '', 'test', null, function (err) {
            t.strictEqual(err.code, ERR.INJECTOR_RESOURCE_CHARSET_DECODING_FAILED);
            t.strictEqual(err.charset, 'test');
            t.done();
        });
    }
};

exports['injector.injectInPage'] = {
    setUp: function (done) {
        watchdog = new testUtils.Watchdog(true);
        done();
    },

    tearDown: function (done) {
        watchdog.shrink();
        process.nextTick(function () {
            done();
        });
    },

    'Startup script': function (t) {
        var srcChunks = makeChunks(fs.readFileSync(path.join(__dirname, '../data/injector/inject_startup_script/src.html')).toString()),
            expectedHtml = fs.readFileSync(path.join(__dirname, '../data/injector/inject_startup_script/expected.html')).toString(),
            options = {
                scripts: [
                    '/testscript1',
                    '/testscript2'
                ]
            };

        injector.injectInPage(srcChunks, '', 'utf8', options, function (err, htmlBuf) {
            t.ok(!err);
            t.ok(testUtils.compareCode(htmlBuf.toString(), expectedHtml));
            t.done();
        });
    },

    'Inject UI stylesheet': function (t) {
        var srcChunks = makeChunks(fs.readFileSync(path.join(__dirname, '../data/injector/inject_ui_stylesheet/src.html')).toString()),
            expectedHtml = fs.readFileSync(path.join(__dirname, '../data/injector/inject_ui_stylesheet/expected.html')).toString(),
            options = {
                styleUrl: '/testStyleUrl'
            };

        injector.injectInPage(srcChunks, '', 'utf8', options, function (err, htmlBuf) {
            t.ok(!err);
            t.ok(testUtils.compareCode(htmlBuf.toString(), expectedHtml));
            t.done();
        });
    },

    'Process resources': function (t) {
        var srcChunks = makeChunks(fs.readFileSync(path.join(__dirname, '../data/injector/process_resources/src.html')).toString()),
            expectedHtml = fs.readFileSync(path.join(__dirname, '../data/injector/process_resources/expected.html')).toString(),
            urlReplacement = 'http://i.am.replaced',
            options = {
                urlReplacer: function () {
                    return urlReplacement;
                }};


        injector.injectInPage(srcChunks, '', 'utf8', options, function (err, htmlBuf) {
            t.ok(!err);
            t.ok(testUtils.compareCode(htmlBuf.toString(), expectedHtml));
            t.done();
        });
    }
};

exports['Regression'] = {
    setUp: function (done) {
        watchdog = new testUtils.Watchdog();

        done();
    },

    tearDown: function (done) {
        watchdog.shrink();
        done();
    },

    'B239430 - Google images - TestCafe script does not inject on empty page': function (t) {
        var srcHtml = fs.readFileSync(path.join(__dirname, '../data/injector/B239430/src.html'), 'utf8').toString(),
            expectedHtml = fs.readFileSync(path.join(__dirname, '../data/injector/B239430/expected.html'), 'utf8').toString(),
            srcChunks = makeChunks(srcHtml),
            injectedScript = ['/testscript'];

        injector.injectInPage(srcChunks, '', 'iso-8859-1', { scripts: injectedScript }, function (err, html) {
            t.ok(cc(html.toString(), expectedHtml));
            t.done();
        });
    },

    'T272552: TestCafe recorder does not render elements on the Northstarmlsuatv2.tlcengine.com page': function (t) {
        var srcScriptText = 'var x = 7;',
            expectedScriptText = pageProc.processScript(srcScriptText),
            srcBuffer = iconv.encode(srcScriptText, 'utf16le', { addBOM: true }),
            expectedBuffer = iconv.encode(expectedScriptText, 'utf16le', { addBOM: true });

        injector.injectInScript([srcBuffer], '', 'utf-8', function (err, scriptBuffer) {
            if (scriptBuffer.length === expectedBuffer.length) {
                for (var i = 0, j = scriptBuffer.length; i < j; i++) {
                    if (scriptBuffer[i] !== expectedBuffer[i])
                        break;
                }

                t.ok(i === j);
            } else
                t.ok(false);

            t.done();
        });
    }
};
